#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#CANADA_NAPS_processor.py

#module which handles parsing of CANADA NAPS network data files.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import copy
import crossnetwork_processing_functions
import datetime
import GHOST_standards
import numpy as np
import os
import pandas as pd
import standardisation_mapping_functions
import time
import unit_converter

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def read_data(slurm_job_ID, parameter, data_year):

    '''Convert one data provider original file for a specific year into multiple GHOST standard .npy station specific files'''

    print('{} -- PARSING RAW DATA'.format(data_year))

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = crossnetwork_processing_functions.get_run_dictionaries(slurm_job_ID, 'CANADA_NAPS', parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get data file specific to year
    data_file = crossnetwork_processing_functions.get_data_files(run_details, parameter_details, specific_year=data_year)[0]

    #define dictionary of reported units by NAPS pollutant
    reported_units_dict = {'O3':'ppbv', 'NO':'ppbv', 'NO2':'ppbv', 'SO2':'ppbv', 'CO':'ppmv'}
    #get reported units for species being processed
    reported_units = reported_units_dict[run_details['parameter']]

    #load external metadata into memory

    #------------------------------------------------------------------------------------#
    #READ DATA FILE INTO MEMORY (PER STATION)
    #------------------------------------------------------------------------------------# 

    data_by_station = {}
    station_object  = {'measurements':[], 'start_time':[]}

    with open(data_file, 'r') as f:
        
        for index,line in enumerate(f):

            #check if pollutant code is not correct for species being processed, if not do not process line
            pollutant_code = line[:3]
            if pollutant_code != parameter_details['naps_code']:
                continue 
            
            #get NAPS reference ID code    
            naps_reference = line[3:9] 
            
            #check if have already stored data for NAPS reference ID, if not add empty frame to store it in
            if naps_reference not in list(data_by_station.keys()):
                data_by_station[naps_reference] = copy.deepcopy(station_object)

            #append time (local time) by station
            #convert each start measurement datetime to minutes from reference time 
            year = int(line[9:13])
            month = int(line[13:15])
            day = int(line[15:17])
            start_i = 29
            end_i = 33
            for hour in range(24):
                
                data_by_station[naps_reference]['start_time'].append(crossnetwork_processing_functions.datetime_to_fractional_minutes_from_reference(datetime.datetime(year, month, day, hour, 0)))
                #also append measurements
                data_by_station[naps_reference]['measurements'].append(line[start_i:end_i])
                start_i += 4
                end_i += 4

    #------------------------------------------------------------------------------------#
    #ITERATE THROUGH ALL STATION SPECIFIC DATA IN FILE AND WRITE OUT TO SPECIFIC .NPY FILES
    #------------------------------------------------------------------------------------#

    #create site instances of metadata/data dictionaries
    site_data_orig = GHOST_standards.get_standard_data(parameter_details)
    site_metadata_orig = GHOST_standards.get_standard_metadata(parameter_details)

    #get standard measurement process information for specific species being processed
    site_metadata_orig, measurement_qa_flags, standard_method_abbreviation = standardisation_mapping_functions.map_network_measurement_information_to_standard_measurement_information('', site_metadata_orig, run_details)

    #iterate through measurement QA flags and write standard qa flag code if required
    qa = []
    for standard_flag_name in list(measurement_qa_flags.keys()):
        if measurement_qa_flags[standard_flag_name] == True:
            for qq in range(len(qa)):
                qa[qq].append(GHOST_standards.standard_QA_name_to_QA_code[standard_flag_name])

    #convert measurement units (if necessary)
    if reported_units != site_data[parameter_details['bsc_parameter_name']]['units']
        conv_obj = unit_converter.convert_units(reported_units,site_data[parameter_details['bsc_parameter_name']]['units'],1.0, measured_species = parameter_details['chemical_formula'])    
        #do conversion
        conv_obj.do_conversion()  
        #convert all station specific measurements
        for naps_reference in list(data_by_station.keys()):
            for meas_n, meas in enumerate(data_by_station[naps_reference]['measurements']):
                try:
                    float_meas = float(meas)
                except:
                    float_meas = np.NaN
                #only do conversion if measurement value is above 0, is not infinite 
                if (float_meas > 0.0) & (np.isinf(float_meas) == False):
                    data_by_station[naps_reference]['measurements'][meas_n] = float_meas*conv_obj.conversion_factor
                else:
                    data_by_station[naps_reference]['measurements'][meas_n] = float_meas

    for naps_reference in list(data_by_station.keys()):

        #copy original site instances of metadata/data dictionaries
        site_data =     copy.deepcopy(site_data_orig)
        site_metadata = GHOST_standards.get_standard_metadata(parameter_details_orig)

        #create time variable
        frac_time = data_by_station[naps_reference]['start_time']
        
        #------------------------------------------------------------------------#
        #PARSE METADATA
        #------------------------------------------------------------------------#
        #get station specific metadata from external metadata file

        #GHOST version
        site_metadata['GHOST_version']['value'].append(run_details['ghost_version'])
        
        #save meta_update_stamp, set as first sorted time processed (as minutes relative to reference time)
        site_metadata['meta_update_stamp']['value'].append(sorted(frac_time)[0])

        #write known network specific metadata  
        site_metadata['horizontal_geodetic_datum']['value'].append('WSG84')
        site_metadata['country']['value'].append('CANADA') 
        site_metadata['network']['value'].append('CANADA NAPS')

        #process standard classification metadata information
        native_classifications = {'area_classification':area_classification, 'station_classification':station_classification, 'main_emission_source':main_emission_source, 'land_use':land_use, 'area_classification_rural_subdivision':area_classification_rural_subdivision, 'representative_radius':representative_radius}        
        site_metadata = standardisation_mapping_functions.map_network_classifications_to_standard_classifications(native_classifications, site_metadata, run_details)
        
        #iterate through metadata variables, if any columns have not been appended to add a NaN
        #check number of refs in list
        for column_header in list(site_metadata.keys()):
            n_column = len(site_metadata[column_header]['value'])
            #if column is empty, then append NaN
            if n_column < 1:
                site_metadata[column_header]['value'].append(np.NaN)

        #process non-network specific metadata (station_timezone, data_download_stamp)
        site_metadata = crossnetwork_processing_functions.process_non_network_specific_metadata(site_metadata, data_file)
        
        #set station_reference (provided network ID + standard measurement method abbreviation)
        site_metadata['station_reference']['value'][-1] = '%s_%s'%(naps_reference,standard_method_abbreviation)

        #check if longitude/latitude coordinates are suspect
        new_longitude, new_latitude = crossnetwork_processing_functions.check_geographical_position(site_metadata['longitude']['value'][-1], site_metadata['latitude']['value'][-1])
        site_metadata['longitude']['value'][-1] = new_longitude
        site_metadata['latitude']['value'][-1] = new_latitude

        #------------------------------------------------------------------------#
        #WRITE OUT DATA VARIABLES
        #------------------------------------------------------------------------#
        
        site_data['time']['value'].extend(frac_time)
        site_data[parameter_details['bsc_parameter_name']]['value'].extend(data_by_station[naps_reference]['measurements'])
        site_data['temporal_resolution']['value'].extend([60]*len(frac_time))   
        site_data['reported_lower_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['reported_upper_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['reported_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['derived_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['flag']['value'].extend([[] for i in range(len(frac_time))])
        site_data['qa']['value'].extend(qa)
        site_data['classification']['value'].extend([[] for i in range(len(frac_time))])
        #------------------------------------------------------------------------#

        #standardise metadata/data and write out specific file station data to temporary .npy
        crossnetwork_processing_functions.post_process_file_parse(site_metadata,site_data,run_details,parameter_details)

    #if have written all file data for all unique stations, return valid process code (0)
    print('PC:0,PT:{}'.format(time.time()-run_details['stage_process_start_time']))