#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#EBAS_processor.py

#module which handles parsing of EBAS network data files.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import crossnetwork_processing_functions
import datetime
from dateutil.relativedelta import *
import GHOST_standards
import numpy as np
import pandas as pd
import standardisation_mapping_functions
import sys
import time
import unit_converter

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def read_data(slurm_job_ID, parameter, station_reference):

    '''Convert one data provider original file for a specific station into a GHOST standard .npy station specific file'''

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- PARSING RAW DATA\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = crossnetwork_processing_functions.get_run_dictionaries(slurm_job_ID, 'EBAS', parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get data files specific to station
    data_files = crossnetwork_processing_functions.get_data_files(run_details, parameter_details, specific_station=station_reference)

    #sort files in date order
    data_file_start_dates = [int(data_file.split('/')[-1].split('.')[1]) for data_file in data_files]
    data_files = np.array([x for _,x in sorted(zip(data_file_start_dates,data_files))]) 
   
    #----------------------------------------------------------------------------------
    #set flag to know when have written data for station    
    written_data = False

    #iterate through site files
    for file_n, data_file in enumerate(data_files):

        log_file_str += 'PROCESSING FILE: {}\n'.format(data_file) 
        
        #create site instances of metadata/data dictionaries
        site_data =     GHOST_standards.get_standard_data(parameter_details)
        site_metadata = GHOST_standards.get_standard_metadata(parameter_details)
 
        #Get data level - this is given in both in filename and inside file as metadata field. 
        #Level 2 data is the maximum quality assured data - class all other levels as preliminary data
        #If exists in metadata, write this out preferentially. If not there, write from filename
        #if data level is not given anywhere, assume level 2
        if '.lev' in data_file:
            data_level = data_file.split('.lev')[-1].split('.nas')[0]
            data_level = data_level.split('_')[0]
            if data_level != '':
                data_level_string = str(data_level)              

        #----------------------------------------------------------------------------#
        #READ FILE
        #----------------------------------------------------------------------------#

        site_nasa_ames_obj = crossnetwork_processing_functions.process_nasa_ames(data_file,log_file_str,station_reference,run_details)
        site_nasa_ames_obj.read_header()
        site_nasa_ames_obj.read_data()
    
        #get file reference start date
        start_date_reference = site_nasa_ames_obj.date_start

        #get start/end time arrays
        s_times = site_nasa_ames_obj.structured_data[0]
        e_times = site_nasa_ames_obj.structured_data[1]
    
        #if length of s_times == 0, then do not process file, as zero valid data
        if len(s_times) == 0:
            break

        #create datetime from YYYYMMDD start date file reference 
        dt_start = datetime.datetime(int(start_date_reference[:4]),int(start_date_reference[4:6]),int(start_date_reference[6:]))
        
        #iterate through file start/end times (in units of days from start date reference), and calculate the time in integer minutes from the desired start year
        #create start/end datetimes for each time given by adding to datetime date file reference, round to nearest minute to avoid issues due to floating point estimation
        start_dt = [crossnetwork_processing_functions.round_to_minute(dt_start+datetime.timedelta(days=float(s_time))) for s_time in s_times]
        end_dt = [crossnetwork_processing_functions.round_to_minute(dt_start+datetime.timedelta(days=float(e_time))) for e_time in e_times]

        #convert each start measurement datetime to minutes from reference time 
        #do not need to shift time base as already in UTC         
        frac_time = [crossnetwork_processing_functions.datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

        #get timedeltas between start and end of measurement window
        #get list of valid temporal resolution indices, where temporal resolution of measurement window is >= 1 minute
        temporal_resolutions = []
        valid_temporal_resolution_inds = []
        for temp_ii, (start, end) in enumerate(zip(start_dt,end_dt)):
            dt_window = end - start
            #convert total seconds to total minutes
            dt_window_minutes = int(dt_window.total_seconds()/60.)
            #only append measurement window resolution if >= 1 minute in length
            if dt_window_minutes >= 1:
                temporal_resolutions.append(dt_window_minutes)    
                valid_temporal_resolution_inds.append(temp_ii)

        #cut start time data array to be just those times for measurement windows, where temporal resolution is >= 1 minute
        if len(valid_temporal_resolution_inds) != len(frac_time):
            frac_time = [frac_time[x] for x in valid_temporal_resolution_inds]
        
        #set variable flags for parsing header of data file
        var_ind = -999
        flag = -999
    
        #get desired measured parameter data
        #check how many inds there are matched for relevant parameter
        relevant_var_inds = np.where(np.array(site_nasa_ames_obj.primary_names) == parameter_details['ebas_parameter_name'])[0]
        #typically there will be more than 1 set of units for only gas phase species (mixing ratio and mass density) 
        if len(relevant_var_inds) > 1:
            #if more than one valid ind for wanted parameter, cut for preferential units 
            #each parameter has defined preferential units as a list (first element is preferred), iterate list until have units desired 
            priority_units_list = parameter_details['ebas_priority_units']
            for priority_units in priority_units_list:
                unit_inds = [idx for idx,s in enumerate(site_nasa_ames_obj.primary_units) if priority_units in s]
                #if have got desired units in file, break out of loop
                if len(unit_inds) > 0:
                    break
            #if have not found any desired units, write error log file to inform of this horrible development (caught error)
            if len(unit_inds) == 0:
                log_file_str += 'CANNOT FIND ANY DESIRED UNITS IN FILE'
                crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,2,log_file_str,station_reference)
            
            #next check overlap of desired var and units
            intersect_inds = np.array([val for val in relevant_var_inds if val in unit_inds])
            #if more than 1 valid intersect ind, must be due to reporting different metrics (mean, min, max etc.) 
            #cut to just get arithmetic mean
            if len(intersect_inds) > 1:
            
                #have to check for duplicate column data, as there some instances where there are repeats (e.g. - CH0002R, 2007, for NO)
                #check if variable names, units and extra details are same for all intersects
                #if so, just take first intersect ind preferentially
            
                if (len(np.unique(np.array(site_nasa_ames_obj.primary_names)[intersect_inds])) == 1) & (len(np.unique(np.array(site_nasa_ames_obj.primary_units)[intersect_inds])) == 1) & (len(np.unique(np.array(site_nasa_ames_obj.primary_details)[intersect_inds])) < 2):
                    var_ind =  intersect_inds[0]
                
                #otherwise cut for arithmetic mean
                else:
                    for int_i in intersect_inds:
                        for detail in site_nasa_ames_obj.primary_details[int_i]:
                            if 'arithmetic mean' in detail:
                                var_ind = int_i
                                break       
            else:
                var_ind = intersect_inds[0]
        else:
            var_ind = relevant_var_inds[0]
            
        #get parameter measurements
        file_measurements = np.array(site_nasa_ames_obj.structured_data[var_ind+1])
        
        #get data flags
        #check how many inds there are for flags
        relevant_flag_inds = [idx for idx,s in enumerate(site_nasa_ames_obj.primary_names) if 'numflag' in s]
    
        #if more than 1 flag in file then must be because there multiple different metrics reported (mean, min, max etc.), each with associated flag
        #in this case simply add 1 to determined var indice, as the flag is always reported immediately after 
        if len(relevant_flag_inds) > 1:
            flag_ind = var_ind+1 
        else:
            flag_ind = relevant_flag_inds[0]
    
        #add 1 to allow for starttime field (not one of primary names)
        file_flags = np.array(site_nasa_ames_obj.structured_data[flag_ind+1])
        
        #----------------------------------------------------------------------------#
        #PROCESS METADATA
        #----------------------------------------------------------------------------#
        
        #GHOST version
        site_metadata['GHOST_version']['value'].append(run_details['ghost_version'])
    
        #save meta_update_stamp, set as first sorted time processed (as minutes relative to reference time)
        site_metadata['meta_update_stamp']['value'].append(sorted(frac_time)[0])
   
        #create dictionary that metadata that can be appended to in both details and comments sections 
        #no parameter specific metadata overlaps between details and comments, however this is useful when compling metadata from different fields into 1 standardised meta field.
        concat_meta = {}
   
        #dictionary for storing raw metadata units with quantity name
        units_dict = {}
   
        #process metadata in details
        concat_meta,units_dict = process_metadata(site_nasa_ames_obj.primary_details[var_ind],'details',concat_meta,units_dict,run_details,station_reference)

        #process metadata in comments
        concat_meta,units_dict = process_metadata(site_nasa_ames_obj.comments,'comments',concat_meta,units_dict,run_details,station_reference)

        #merge certain saved metadata fields into 1 field
        #delete merged fields from concat_meta afterwards
        
        #associated_networks
        variables_to_merge = ['wdca_id','gaw_id','airs_id','other_ids']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['associated_networks'] = '{};{};{};{}'.format(concat_meta['wdca_id'],concat_meta['gaw_id'],concat_meta['airs_id'],concat_meta['other_ids'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]
        
        #measurement information
        variables_to_merge = ['instrument_type','instrument_manufacturer','instrument_model','analytical_measurement_technique','analytical_instrument_manufacturer','analytical_instrument_model']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['measurement_information'] = '{},{},{},{},{},{}'.format(concat_meta['instrument_type'], concat_meta['instrument_manufacturer'], concat_meta['instrument_model'],concat_meta['analytical_measurement_technique'], concat_meta['analytical_instrument_manufacturer'], concat_meta['analytical_instrument_model'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]
        
        #inlet information
        variables_to_merge = ['inlet_type','inlet_description','inlet_tube_material','inlet_outer_diameter','inlet_inner_diameter','inlet_length']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['measuring_instrument_inlet_information'] = '{};{};{};{};{};{}'.format(concat_meta['inlet_type'], concat_meta['inlet_description'], concat_meta['inlet_tube_material'], concat_meta['inlet_outer_diameter'], concat_meta['inlet_inner_diameter'], concat_meta['inlet_length'])
        #change " to inches in merged string if there - causes formatting issues further down pipe
        concat_meta['measuring_instrument_inlet_information'] = concat_meta['measuring_instrument_inlet_information'].replace('"','inch')
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]
        
        #sampling details
        variables_to_merge = ['filter_face_velocity','exposed_filter_area','filter_description','medium','filter_prefiring','filter_conditioning','coating','sample_preparation']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['network_sampling_details'] = '{};{};{};{};{};{};{};{}'.format(concat_meta['filter_face_velocity'],concat_meta['exposed_filter_area'],concat_meta['filter_description'],concat_meta['medium'],concat_meta['filter_prefiring'],concat_meta['filter_conditioning'],concat_meta['coating'],concat_meta['sample_preparation'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]

        #uncertainty details
        variables_to_merge = ['measurement_uncertainty_explained','detection_limit_explained','zero_values']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['network_uncertainty_details'] = '{};{};{}'.format(concat_meta['measurement_uncertainty_explained'],concat_meta['detection_limit_explained'],concat_meta['zero_values'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]
        
        #maintenance details
        variables_to_merge = ['maintenance_description','zero_type','zero_interval','blank_correction','artifact_correction_description','charring_correction','humidity_control','humidity_description']
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['network_maintenance_details'] = '{};{};{};{};{};{};{};{}'.format(concat_meta['maintenance_description'],concat_meta['zero_type'],concat_meta['zero_interval'],concat_meta['blank_correction'],concat_meta['artifact_correction_description'],concat_meta['charring_correction'],concat_meta['humidity_control'],concat_meta['humidity_description'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]
        
        #Network QA details
        variables_to_merge = ['qa_date','qa_outcome','qa_bias','qa_variability','qa_document_url']  
        for var in variables_to_merge:
            #if variable not in concat_meta, make it empty string
            if var not in list(concat_meta.keys()):
                concat_meta[var] = ''
        #make merged variable
        concat_meta['network_qa_details'] = '{};{};{};{};{}'.format(concat_meta['qa_date'],concat_meta['qa_outcome'],concat_meta['qa_bias'],concat_meta['qa_variability'],concat_meta['qa_document_url'])
        #delete merged fields
        for var in variables_to_merge:
            del concat_meta[var]

        #iterate through concatenated metadata and remove formatting inconsistencies, and deal with empty strings
        for key in list(concat_meta.keys()):
            #strip white space from ends of string
            concat_meta[key] = concat_meta[key].strip()
            #if have empty commas or semi colons, replace with blank space
            concat_meta[key] = concat_meta[key].replace(',,','')
            concat_meta[key] = concat_meta[key].replace(';;','')
            concat_meta[key] = concat_meta[key].replace('  ',' ')
            #if first or last character is semicolon, colon or comma, cut it off
            if len(concat_meta[key]) > 1:
                if (concat_meta[key][0] == ';') or (concat_meta[key][0] == ':') or (concat_meta[key][0] == ','):
                    concat_meta[key] = concat_meta[key][1:]
                if (concat_meta[key][-1] == ';') or (concat_meta[key][-1] == ':') or (concat_meta[key][-1] == ','):
                    concat_meta[key] = concat_meta[key][:-1]
            #strip white space from ends of string
            concat_meta[key] = concat_meta[key].strip()
            #convert empty string, or solo commas/semicolons/colons to NaN
            if (concat_meta[key] == '') or (concat_meta[key] == ';') or (concat_meta[key] == ',') or (concat_meta[key] == ':'):
                concat_meta[key] = np.NaN

        #process station classifications
        #if both file area_classification and land use are present in file, then process standardised classifications
        if ('area_classification' in concat_meta) & ('land_use' in concat_meta):
            #get standardised classifications
            native_classifications = {'area_classification':concat_meta['area_classification'],'land_use':concat_meta['land_use']}
            site_metadata = standardisation_mapping_functions.map_network_classifications_to_standard_classifications(native_classifications, site_metadata, run_details)
        #delete classification parameters from concat_meta
        if 'area_classification' in concat_meta:
            del concat_meta['area_classification']
        if 'land_use' in concat_meta:
            del concat_meta['land_use']

        #process measurement information
        if pd.isnull(concat_meta['measurement_information']) == False:
            site_metadata, measurement_qa_flags, standard_method_abbreviation = standardisation_mapping_functions.map_network_measurement_information_to_standard_measurement_information(concat_meta['measurement_information'], site_metadata, run_details)
        #delete parameter from concat_meta
        del concat_meta['measurement_information']

        #write rest of saved metadata out
        for column_header in list(concat_meta.keys()):
            n_column = len(site_metadata[column_header]['value'])
            #if length of column is less than number of references, then hasn't been written to yet, therefore append
            if n_column < 1:
                site_metadata[column_header]['value'].append(concat_meta[column_header])
            #otherwise it has been written as NaN already, therefore overwrite
            else:
                site_metadata[column_header]['value'][-1] = concat_meta[column_header]

        #write known network specific metadata  
        site_metadata['horizontal_geodetic_datum']['value'].append('WSG84')
        site_metadata['country']['value'].append(GHOST_standards.country_ISO3166_to_name[station_reference[:2]]) 
        site_metadata['network']['value'].append('EBAS')
                    
        #iterate through metadata variables, if any columns have not been appended to add a NaN
        #check number of refs in list
        for column_header in list(site_metadata.keys()):
            n_column = len(site_metadata[column_header]['value'])
            #if column is empty, then append NaN
            if n_column < 1:
                site_metadata[column_header]['value'].append(np.NaN)

        #ensure that measurement position metadata are floats, otherwise append nan
        position_meta = ['latitude','longitude','altitude','sampling_height','measurement_altitude']
        for pm in position_meta:
            try:
                site_metadata[pm]['value'][-1] = np.float64(site_metadata[pm]['value'][-1])
            except:
                site_metadata[pm]['value'][-1] = np.NaN
                
        #process non-network specific metadata (station_timezone, data_download_stamp)
        site_metadata = crossnetwork_processing_functions.process_non_network_specific_metadata(site_metadata, data_file)
        
        #if do not have data level from metadata, try writing out data level from filename
        #if no level in filename either, assume it to be level 2 (highest level) 
        if pd.isnull(site_metadata['data_level']['value'][-1]) == True:
            try:
                site_metadata['data_level']['value'][-1] = str(float(data_level_string))
            except:
                site_metadata['data_level']['value'][-1] = '2.0'
        else:
            site_metadata['data_level']['value'][-1] = str(float(site_metadata['data_level']['value'][-1]))            
        
        #set station_reference (provided network ID + standard measurement method abbreviation)
        site_metadata['station_reference']['value'][-1] = '{}_{}'.format(station_reference,standard_method_abbreviation)

        #check if longitude/latitude coordinates are suspect
        new_longitude, new_latitude = crossnetwork_processing_functions.check_geographical_position(site_metadata['longitude']['value'][-1], site_metadata['latitude']['value'][-1])
        site_metadata['longitude']['value'][-1] = new_longitude
        site_metadata['latitude']['value'][-1] = new_latitude
                    
        #-----------------------------------------------------
        #convert metadata data with units to standard units
        
        #get measurement units
        units = site_nasa_ames_obj.primary_units[var_ind].strip()
        #convert raw units given to standard name
        conv_obj = unit_converter.convert_units(units,units,1, measured_species = parameter_details['chemical_formula'])
        conv_obj.do_conversion()
        site_metadata['measuring_instrument_reported_units']['value'][-1] = conv_obj.output_standard_units
        
        #process standard temperature and pressure first - potentially need later for converting gas concentrations
        #use EBAS stated standards if not given - also set flag to know have assumed gas volume, if later converting units
        #given as 293.15 K / 1013.25 hPa for O3, & 273.15 K / 1013.25 hPa for all other parameters:
        assume_gas_volume = False
        
        if 'measuring_instrument_volume_standard_temperature' in units_dict:
            
            #if metadata units not equal to standard units, convert!
            if units_dict['measuring_instrument_volume_standard_temperature'] != site_metadata['measuring_instrument_volume_standard_temperature']['units']:
                conv_obj = unit_converter.convert_units(units_dict['measuring_instrument_volume_standard_temperature'],site_metadata['measuring_instrument_volume_standard_temperature']['units'],site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1], measured_species = parameter_details['chemical_formula'])
                conv_obj.do_conversion()
                site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1] = conv_obj.converted_value
            #remove from units dict 
            del units_dict['measuring_instrument_volume_standard_temperature']
        else:
            assume_gas_volume = True
            if run_details['parameter'] == 'O3':
                site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1] = '293.15'
            else:
                site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1] = '273.15'
        
        if 'measuring_instrument_volume_standard_pressure' in units_dict:
        
            #if metadata units not equal to standard units, convert!
            if units_dict['measuring_instrument_volume_standard_pressure'] != site_metadata['measuring_instrument_volume_standard_pressure']['units']:
                conv_obj = unit_converter.convert_units(units_dict['measuring_instrument_volume_standard_pressure'],site_metadata['measuring_instrument_volume_standard_pressure']['units'],site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1], measured_species = parameter_details['chemical_formula'])
                conv_obj.do_conversion()
                site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1] = conv_obj.converted_value
             #remove from units dict 
            del units_dict['measuring_instrument_volume_standard_pressure']
        else:
            assume_gas_volume = True
            site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1] = '1013.25'

        #process rest of meta units
        for meta_name in units_dict:
            #if metadata units not equal to standard units, convert!
            if units_dict[meta_name] != site_metadata[meta_name]['units']:
                
                #some gas concentration values may need converting to mole fraction from mass density (using reported standard temperature and pressure)
                if (meta_name in ['measuring_instrument_reported_lower_limit_of_detection','measuring_instrument_reported_uncertainty']) & ('/m3' in units_dict[meta_name]) & ('mol-1' in site_data[parameter_details['bsc_parameter_name']]['units']):
                    input_units = {'temperature':'K', 'pressure':'hPa', 'molar_mass':'kg mol-1', 'mass_density':units_dict[meta_name]}
                    input_vals = {'temperature':site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1], 'pressure':site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1], 'molar_mass':crossnetwork_processing_functions.get_molecular_mass(parameter_details['chemical_formula']), 'mass_density':site_metadata[meta_name]['value'][-1]}
                    conv_obj  = unit_converter.convert_units(input_units,site_metadata[meta_name]['units'],input_vals,measured_species = parameter_details['chemical_formula'])
                #quantities which just need converting i.e. cm2/s --> m2/min
                else:
                    conv_obj = unit_converter.convert_units(units_dict[meta_name],site_metadata[meta_name]['units'],site_metadata[meta_name]['value'][-1], measured_species = parameter_details['chemical_formula'])
                    
                #do conversion
                conv_obj.do_conversion()
                site_metadata[meta_name]['value'][-1] = conv_obj.converted_value
        
        #----------------------------------------------------------------------------#
        #PROCESS DATA
        #----------------------------------------------------------------------------#

        #flags
        flags = []
        qa = []
    
        #get dictionary necessary for mapping of EBAS file data flags to standard data flag codes
        data_flag_mapping_dictionary = standardisation_mapping_functions.network_data_flag_to_standard_data_flag_mapping['EBAS']

        #iterate through EBAS file data flags and map to standard flag codes
        for file_flag in file_flags:

            #create empty list for standard data flags associated with each measurement
            standard_data_flag_names = []
            
            #convert file flag into integer number, equivalent to format as in standard flag mapping dictionary 
            if '.' in file_flag:
                file_flag = file_flag.split('.')[-1]
            if len(file_flag) > 3:
                file_flag = file_flag[:3]
            
            #get mapped standard data flag specific dictionary 
            mapped_standard_data_flag_dict = data_flag_mapping_dictionary[file_flag]
            #unpack mapped standard data flag key variables
            standard_data_flag_names.append(mapped_standard_data_flag_dict['standard_data_flag_name'])
            GHOST_decreed_validity = mapped_standard_data_flag_dict['GHOST_decreed_validity']
            network_decreed_validity = mapped_standard_data_flag_dict['network_decreed_validity']
            
            #write preliminary data flags if necessary (informs if data file contains preliminary data (lower than level 2))
            if site_metadata['data_level']['value'][-1] != '2.0':
                if 'Preliminary Data' not in standard_data_flag_names:
                    standard_data_flag_names.append('Preliminary Data')
            
            #get standard flag codes associated with each standard data flag name, and append to list to write out
            standard_data_flag_codes = [GHOST_standards.standard_data_flag_name_to_data_flag_code[standard_data_flag_name] for standard_data_flag_name in standard_data_flag_names]
            flags.append(standard_data_flag_codes)
            
            #process qa information associated with flags
            qa_codes = []
            #append qa flag code if standard data flag is GHOST decreed invalid
            if GHOST_decreed_validity == 'I':
                qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
            #append qa flag if standard data flag is network decreed invalid
            if network_decreed_validity == 'I':
                qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
            #append qa codes to list to write out
            qa.append(qa_codes)
            
        #--------------------------------------
        #measurements
        #measurement values may need converting (i.e. mass density to mixing ratio)
        #--------------------------------------
        
        measurements = []

        #if gas volume not provided by data provider and units need converting, this conversion is done using assumed network standard pressure and temperatures - write qa data flag to inform of this assumption
        #this flag below alerts for writing of qa flag later
        write_volume_assumption_flag = False

        #check if reported standard units, match required standard units
        if site_metadata['measuring_instrument_reported_units']['value'][-1] != site_data[parameter_details['bsc_parameter_name']]['units']:
            
            #mass density to mixing ratio?
            if ('m-3' in site_metadata['measuring_instrument_reported_units']['value'][-1]) & ('mol-1' in site_data[parameter_details['bsc_parameter_name']]['units']):

                #set flag to instruct writing of flag that am assuming gas volume in converting units 
                if assume_gas_volume == True:
                    write_volume_assumption_flag = True
                
                #get conversion factor
                input_units = {'temperature':'K', 'pressure':'hPa', 'molar_mass':'kg mol-1', 'mass_density':site_metadata['measuring_instrument_reported_units']['value'][-1]}
                input_vals = {'temperature':site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1], 'pressure':site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1], 'molar_mass':crossnetwork_processing_functions.get_molecular_mass(parameter_details['chemical_formula']), 'mass_density':1.0}
                conv_obj = unit_converter.convert_units(input_units,site_data[parameter_details['bsc_parameter_name']]['units'],input_vals,measured_species = parameter_details['chemical_formula'],conversion_input_quantity = 'mass_density')
                     

            #otherwise simple unit conversion, i.e. pptv to ppbv or mass density to mass density
            else:
                conv_obj = unit_converter.convert_units(site_metadata['measuring_instrument_reported_units']['value'][-1],site_data[parameter_details['bsc_parameter_name']]['units'],1.0, measured_species = parameter_details['chemical_formula'])
        
            #do conversion
            conv_obj.do_conversion()    
            
            #process through each measurement and convert value using derived conversion factor
            for n, meas in enumerate(file_measurements):
                try:
                    float_meas = float(meas)
                except:
                    float_meas = np.NaN
                #only do conversion if measurement value is above 0, is not infinite 
                if (float_meas > 0.0) & (np.isinf(float_meas) == False):
                    measurements.append(float_meas*conv_obj.conversion_factor)
                else:
                    measurements.append(float_meas)
            
        #otherwise do nothing to measured values
        else:
            for n, meas in enumerate(file_measurements):
                try:
                    float_meas = float(meas)
                except:
                    float_meas = np.NaN
                measurements.append(float_meas)
                
        #----------------------------------------------------------------------------#    
        #QA
        #----------------------------------------------------------------------------#
        #write measurement associated qa flags
                           
        #iterate through measurement QA flags and write standard qa flag code if required
        for standard_flag_name in list(measurement_qa_flags.keys()):
            if measurement_qa_flags[standard_flag_name] == True:
                for qq in range(len(qa)):
                    qa[qq].append(GHOST_standards.standard_QA_name_to_QA_code[standard_flag_name])
        
        #write assumed volume of gas qa flags (inform have assumed volume of gas for converting units)
        if write_volume_assumption_flag == True:
            for qq in range(len(qa)):
                qa[qq].append(GHOST_standards.standard_QA_name_to_QA_code['Assumed Gas Volume'])
        
        #------------------------------------------------------
        #cut data variables to only keep those associated with measurement windows of >= 1 minute (time and temporal resolution fields already cut)
        if len(valid_temporal_resolution_inds) != len(measurements):
            flags = [flags[x] for x in valid_temporal_resolution_inds]
            qa = [qa[x] for x in valid_temporal_resolution_inds]        
            measurements = [measurements[x] for x in valid_temporal_resolution_inds]

        #------------------------------------------------------
        #write out site variables
        #No LODs or measurement uncertainties per measurement - append list of NaNs    
        site_data['time']['value'].extend(frac_time)
        site_data[parameter_details['bsc_parameter_name']]['value'].extend(measurements)
        site_data['temporal_resolution']['value'].extend(temporal_resolutions)   
        site_data['reported_lower_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['reported_upper_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['reported_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['derived_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
        site_data['flag']['value'].extend(flags)
        site_data['qa']['value'].extend(qa)
        site_data['classification']['value'].extend([[] for i in range(len(frac_time))])
        #------------------------------------------------------
            
        #standardise metadata/data and write out specific file station data to temporary .npy
        crossnetwork_processing_functions.post_process_file_parse(site_metadata,site_data,run_details,parameter_details)

        #update flag to state that have written data for station
        written_data = True
    
    #if stage has completed successfully, with some valid data, write a valid process log file
    if written_data == True:
        log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
        crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,0,log_file_str,station_reference)
    #else, if no data has been written for station write an error log file to inform station is not continuing in the pipeline
    elif written_data == False:
        log_file_str += 'NO DATA HAS BEEN WRITTEN FOR STATION'
        crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,1,log_file_str,station_reference)

def process_metadata(meta,meta_type,concat_meta,units_dict,run_details,station_reference):

    '''function that takes metadata from header details and comments and sorts it.
       many parameters can appear in both header and comments, it handles these cases. 
    '''    

    line_titles = []
    line_data = []
    
    #get splitter for different section of datafile metadata is reported in
    splitter_dict = {'comments':':','details':'='}
    splitter = splitter_dict[meta_type]
    
    for line in meta: 
        line_split = line.split(splitter)
        if len(line_split) >= 2:
            line_titles.append(line_split[0].strip())
            line_data.append(''.join([element.strip() for element in line_split[1:]]))
    
    #data_level
    if 'Data level' in line_titles:
        data_level = line_data[line_titles.index('Data level')]
        if (data_level != '') & ('data_level' not in concat_meta):
            concat_meta['data_level'] = data_level

    #data_timezone
    if 'Timezone' in line_titles:
        data_timezone = line_data[line_titles.index('Timezone')]
        if (data_timezone != '') & ('timezone' not in concat_meta):
            #if timezone data is reported in is not UTC, write error log file as it should be (caught error)
            if data_timezone == 'UTC':
                concat_meta['data_timezone'] = '0.0'
            else:
                log_file_str += 'DATA NOT IN UTC TIME'
                crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,2,log_file_str,station_reference)

    #latitude
    if 'Station latitude' in line_titles:
        latitude = line_data[line_titles.index('Station latitude')]
        if (latitude != '') & ('latitude' not in concat_meta):
            concat_meta['latitude'] = latitude

    #longitude
    if 'Station longitude' in line_titles:
        longitude = line_data[line_titles.index('Station longitude')]
        if (longitude != '') & ('longitude' not in concat_meta):
            concat_meta['longitude'] = longitude

    #altitude 
    if 'Station altitude' in line_titles:
        altitude = line_data[line_titles.index('Station altitude')]
        if (altitude != '') & ('altitude' not in concat_meta):
            concat_meta['altitude'] = altitude.split(' ')[0]
            units_dict['altitude'] = ''.join(altitude.split(' ')[1:])

    #sampling_height 
    if 'Measurement height' in line_titles:
        sampling_height = line_data[line_titles.index('Measurement height')]
        if (sampling_height != '') & ('sampling_height' not in concat_meta):
            concat_meta['sampling_height'] = sampling_height.split(' ')[0]
            units_dict['sampling_height'] = ''.join(sampling_height.split(' ')[1:])

    #station_name
    if 'Station name' in line_titles:
        station_name = line_data[line_titles.index('Station name')]
        if (station_name != '') & ('station_name' not in concat_meta):
            concat_meta['station_name'] = station_name

    #province
    if 'Station state/province' in line_titles:
        province = line_data[line_titles.index('Station state/province')]
        if (province != '') & ('province' not in concat_meta):
            concat_meta['province'] = province

    #flow_rate
    if ('Flow rate' in line_titles):
        reported_flow_rate = line_data[line_titles.index('Flow rate')]
        if (reported_flow_rate != '') & ('reported_flow_rate' not in concat_meta):
            concat_meta['measuring_instrument_reported_flow_rate'] = reported_flow_rate.split(' ')[0]
            units_dict['measuring_instrument_reported_flow_rate'] = ''.join(reported_flow_rate.split(' ')[1:])

    #reported_lower_limit_of_detection 
    if ('Detection limit' in line_titles):
        reported_lower_limit_of_detection = line_data[line_titles.index('Detection limit')]
        if (reported_lower_limit_of_detection != '') & ('measuring_instrument_reported_lower_limit_of_detection' not in concat_meta):
            concat_meta['measuring_instrument_reported_lower_limit_of_detection'] = reported_lower_limit_of_detection.split(' ')[0]
            units_dict['measuring_instrument_reported_lower_limit_of_detection'] = ''.join(reported_lower_limit_of_detection.split(' ')[1:])

    #reported_uncertainty (given either as a percent or absolute value) 
    if ('Measurement uncertainty' in line_titles):
        reported_uncertainty = line_data[line_titles.index('Measurement uncertainty')]
        if (reported_uncertainty != '') & ('measuring_instrument_reported_uncertainty' not in concat_meta):
            concat_meta['measuring_instrument_reported_uncertainty'] = reported_uncertainty.split(' ')[0]
            uncertainty_unit = ''.join(reported_uncertainty.split(' ')[1:])
            if uncertainty_unit != '%':
                units_dict['measuring_instrument_reported_uncertainty'] = uncertainty_unit
    
    #absorption_cross_section 
    if ('Absorption cross section' in line_titles):
        reported_absorption_cross_section = line_data[line_titles.index('Absorption cross section')]
        if (reported_absorption_cross_section != '') & ('measuring_instrument_reported_absorption_cross_section' not in concat_meta):
            concat_meta['measuring_instrument_reported_absorption_cross_section'] = reported_absorption_cross_section.split(' ')[0]
            units_dict['measuring_instrument_reported_absorption_cross_section'] = ''.join(reported_absorption_cross_section.split(' ')[1:]) 

    #calibration_scale
    if ('Calibration scale' in line_titles):
        calibration_scale = line_data[line_titles.index('Calibration scale')]
        if (calibration_scale != '') & ('measuring_instrument_calibration_scale' not in concat_meta):
            concat_meta['measuring_instrument_calibration_scale'] = calibration_scale
        
    #volume_standard_temperature 
    if ('Volume std. temperature' in line_titles):
        volume_standard_temperature = line_data[line_titles.index('Volume std. temperature')]
        if (volume_standard_temperature != '') & (volume_standard_temperature != 'ambient') & ('measuring_instrument_volume_standard_temperature' not in concat_meta):
            #must have units given for temperature, otherwise it is useless. If no units are given, write volume standard temperature out as NaN
            try:
                concat_meta['measuring_instrument_volume_standard_temperature'] = volume_standard_temperature.split(' ')[0]
                units_dict['measuring_instrument_volume_standard_temperature'] = ''.join(volume_standard_temperature.split(' ')[1:]) 
            except:
                concat_meta['measuring_instrument_volume_standard_temperature'] = np.NaN
    
    #volume_standard_pressure 
    if ('Volume std. pressure' in line_titles):
        volume_standard_pressure = line_data[line_titles.index('Volume std. pressure')]
        if (volume_standard_pressure != '') & (volume_standard_pressure != 'ambient') & ('measuring_instrument_volume_standard_pressure' not in concat_meta):
            #must have units given for pressure, otherwise it is useless. If no units are given, write volume standard pressure out as NaN
            try:
                concat_meta['measuring_instrument_volume_standard_pressure'] = volume_standard_pressure.split(' ')[0]
                units_dict['measuring_instrument_volume_standard_pressure'] = ''.join(volume_standard_pressure.split(' ')[1:]) 
            except:
                concat_meta['measuring_instrument_volume_standard_pressure'] = np.NaN
    
    #associated_networks
    if 'Station WDCA-ID' in line_titles:
        wdca_id = line_data[line_titles.index('Station WDCA-ID')].strip()
        if (wdca_id != '') & ('wdca_id' not in concat_meta):
            concat_meta['wdca_id'] = 'WDCA:'+wdca_id
    if 'Station GAW-ID' in line_titles:
        gaw_id = line_data[line_titles.index('Station GAW-ID')].strip()
        if (gaw_id != '') & ('gaw_id' not in concat_meta):
            concat_meta['gaw_id'] = 'GAW:'+gaw_id
    if 'Station AIRS-ID' in line_titles:
        airs_id = line_data[line_titles.index('Station AIRS-ID')].strip()
        if (airs_id != '') & ('airs_id' not in concat_meta):
            concat_meta['airs_id'] = 'AIRS:'+airs_id
    if 'Station other IDs' in line_titles:
        other_ids = line_data[line_titles.index('Station other IDs')].strip()
        if (other_ids != '') & ('other_ids' not in concat_meta):
            concat_meta['other_ids'] = ''
            other_ids = other_ids.split(',')
            for o_id in other_ids:
                o_id_split = o_id.split('(')
                concat_meta['other_ids'] = concat_meta['other_ids'] + '{}:{};'.format(o_id_split[1][:-1],o_id_split[0])
            #remove end semi-colon
            concat_meta['other_ids'] = concat_meta['other_ids'][:-1] 

    #area_classification/land_use
    if 'Station setting' in line_titles:
        area_classification = line_data[line_titles.index('Station setting')].replace(" ", "").lower()
        if (area_classification != '') & ('area_classification' not in concat_meta):
            concat_meta['area_classification'] = area_classification
    if 'Station land use' in line_titles:  
        land_use = line_data[line_titles.index('Station land use')].replace(" ", "").lower()
        if (land_use != '') & ('land_use' not in concat_meta):
            concat_meta['land_use'] = land_use

    #measurement information
    #instrument_type, instrument_manufacturer, instrument_model, analytical_measurement_technique, analytical_instrument_manufacturer, analytical_instrument_model
    if 'Instrument type' in line_titles:
        instrument_type = line_data[line_titles.index('Instrument type')].lower()
        if (instrument_type != '') & ('instrument_type' not in concat_meta):
            concat_meta['instrument_type'] = instrument_type
    if 'Instrument manufacturer' in line_titles:
        instrument_manufacturer = line_data[line_titles.index('Instrument manufacturer')].lower()
        if (instrument_manufacturer != '') & ('instrument_manufacturer' not in concat_meta):
            concat_meta['instrument_manufacturer'] = instrument_manufacturer
    if 'Instrument model' in line_titles:
        instrument_model = line_data[line_titles.index('Instrument model')].lower()
        if (instrument_model != '') & ('instrument_model' not in concat_meta):
            concat_meta['instrument_model'] = instrument_model
    if 'Analytical measurement technique' in line_titles:
        analytical_measurement_technique = line_data[line_titles.index('Analytical measurement technique')].lower()
        if (analytical_measurement_technique != '') & ('analytical_measurement_technique' not in concat_meta):
            concat_meta['analytical_measurement_technique'] = analytical_measurement_technique
    if 'Analytical instrument manufacturer' in line_titles:
        analytical_instrument_manufacturer = line_data[line_titles.index('Analytical instrument manufacturer')].lower()
        if (analytical_instrument_manufacturer != '') & ('analytical_instrument_manufacturer' not in concat_meta):
            concat_meta['analytical_instrument_manufacturer'] = analytical_instrument_manufacturer
    if 'Analytical instrument model' in line_titles:
        analytical_instrument_model = line_data[line_titles.index('Analytical instrument model')].lower()
        if (analytical_instrument_model != '') & ('analytical_instrument_model' not in concat_meta):
            concat_meta['analytical_instrument_model'] = analytical_instrument_model

    #inlet_information  
    if ('Inlet type' in line_titles):
        inlet_type = line_data[line_titles.index('Inlet type')].lower()
        if (inlet_type != '') & ('inlet_type' not in concat_meta):
            concat_meta['inlet_type'] = 'Inlet type: '+inlet_type
    if ('Inlet description' in line_titles):
        inlet_description = line_data[line_titles.index('Inlet description')].lower()
        if (inlet_description != '') & ('inlet_description' not in concat_meta):
            concat_meta['inlet_description'] = 'Description: '+inlet_description
    if ('Inlet tube material' in line_titles):
        inlet_tube_material = line_data[line_titles.index('Inlet tube material')].lower().strip()
        if (inlet_tube_material != '') & ('inlet_tube_material' not in concat_meta):
            concat_meta['inlet_tube_material'] = 'Tube material: '+inlet_tube_material
    if ('Inlet tube outer diameter' in line_titles):
        inlet_outer_diameter = line_data[line_titles.index('Inlet tube outer diameter')].lower().strip()
        if (inlet_outer_diameter != '') & ('inlet_outer_diameter' not in concat_meta):
            concat_meta['inlet_outer_diameter'] = 'Outer diameter: '+inlet_outer_diameter
    if ('Inlet tube inner diameter' in line_titles):
        inlet_inner_diameter = line_data[line_titles.index('Inlet tube inner diameter')].lower().strip()
        if (inlet_inner_diameter != '') & ('inlet_inner_diameter' not in concat_meta):
            concat_meta['inlet_inner_diameter'] = 'Inner diameter: '+inlet_inner_diameter
    if ('Inlet tube length' in line_titles):
        inlet_length = line_data[line_titles.index('Inlet tube length')].lower().strip()
        if (inlet_length != '') & ('inlet_length' not in concat_meta):
            concat_meta['inlet_length'] = 'Length: '+inlet_length

    #principal_investigator information
    #can be multiple - so take first instance
    if 'Originator' in line_titles:
        principal_investigator_split = line_data[line_titles.index('Originator')].split(',')
        
        principal_investigator_name =  '{} {}'.format(principal_investigator_split[1].strip(),principal_investigator_split[0].strip())
        principal_investigator_name = principal_investigator_name.strip()
        if (principal_investigator_name != '') & ('principal_investigator_name' not in concat_meta):
            concat_meta['principal_investigator_name'] = principal_investigator_name
            
        principal_investigator_email_address = principal_investigator_split[2].strip()
        if (principal_investigator_email_address != '') & ('principal_investigator_email_address' not in concat_meta):
            concat_meta['principal_investigator_email_address'] = principal_investigator_email_address
            
        principal_investigator_institution = principal_investigator_split[3].strip()
        if (principal_investigator_institution != '') & ('principal_investigator_institution' not in concat_meta):
            concat_meta['principal_investigator_institution'] = principal_investigator_institution
        
   #contact information
   #can be multiple - so take first instance
    if 'Submitter' in line_titles:
        contact_split = line_data[line_titles.index('Submitter')].split(',')
        
        contact_name = '{} {}'.format(contact_split[1].strip(),contact_split[0].strip())
        contact_name = contact_name.strip()
        if (contact_name != '') & ('contact_name' not in concat_meta):
            concat_meta['contact_name'] = contact_name
        
        contact_email_address = contact_split[2].strip()
        if (contact_email_address != '') & ('contact_email_address' not in concat_meta):
            concat_meta['contact_email_address'] = contact_email_address
        
        contact_institution = contact_split[3].strip()
        if (contact_institution != '') & ('contact_institution' not in concat_meta):
            concat_meta['contact_institution'] = contact_institution
            
    #uncertainty details
    if ('Measurement uncertainty expl.' in line_titles):
        measurement_uncertainty_explained = line_data[line_titles.index('Measurement uncertainty expl.')].strip()   
        if (measurement_uncertainty_explained != '') & ('measurement_uncertainty_explained' not in concat_meta):
            concat_meta['measurement_uncertainty_explained'] = 'Measurement uncertainty explained: '+measurement_uncertainty_explained

    if ('Detection limit expl.' in line_titles):
        detection_limit_explained = line_data[line_titles.index('Detection limit expl.')].strip() 
        if (detection_limit_explained != '') & ('detection_limit_explained' not in concat_meta):
            concat_meta['detection_limit_explained'] = 'Detection limit explained: '+detection_limit_explained

    if ('Zero/negative values' in line_titles):
        zero_values = line_data[line_titles.index('Zero/negative values')].strip() 
        if (zero_values != '') & ('zero_values' not in concat_meta):
            concat_meta['zero_values'] = 'Zero/negative values: '+zero_values

    #measurement maintenance details
    if ('Maintenance description' in line_titles):
        maintenance_description = line_data[line_titles.index('Maintenance description')].strip() 
        if (maintenance_description != '') & ('maintenance_description' not in concat_meta):
            concat_meta['maintenance_description'] = 'Maintenance description: '+maintenance_description

    if ('Zero/span check type' in line_titles):
        zero_type = line_data[line_titles.index('Zero/span check type')].strip() 
        if (zero_type != '') & ('zero_type' not in concat_meta):
            concat_meta['zero_type'] = 'Zero/span check type: '+zero_type

    if ('Zero/span check interval' in line_titles):
        zero_interval = line_data[line_titles.index('Zero/span check interval')].strip() 
        if (zero_interval != '') & ('zero_interval' not in concat_meta):
            concat_meta['zero_interval'] = 'Zero/span check interval: '+zero_interval

    if ('Blank correction' in line_titles):
        blank_correction = line_data[line_titles.index('Blank correction')].strip()
        if (blank_correction != '') & ('blank_correction' not in concat_meta):
            concat_meta['blank_correction'] = 'Blank correction: '+blank_correction
    
    if ('Artifact correction description' in line_titles):
        artifact_correction_description = line_data[line_titles.index('Artifact correction description')].strip()
        if (artifact_correction_description != '') & ('artifact_correction_description' not in concat_meta):
            concat_meta['artifact_correction_description'] = 'Artifact correction description: '+artifact_correction_description
    
    if ('Charring correction' in line_titles):
        charring_correction = line_data[line_titles.index('Charring correction')].strip()
        if (charring_correction != '') & ('charring_correction' not in concat_meta):
            concat_meta['charring_correction'] = 'Charring correction: '+charring_correction
    
    if ('Humidity/temperature control' in line_titles):
        humidity_control = line_data[line_titles.index('Humidity/temperature control')].strip() 
        if (humidity_control != '') & ('humidity_control' not in concat_meta):
            concat_meta['humidity_control'] = 'Humidity/temp. control: '+humidity_control

    if ('Humidity/temperature control description' in line_titles):
        humidity_description = line_data[line_titles.index('Humidity/temperature control description')].strip() 
        if (humidity_description != '') & ('humidity_description' not in concat_meta):
            concat_meta['humidity_description'] = 'Humidity/temp. control description: '+humidity_description
    
    #sampling details 
    if ('Filter face velocity' in line_titles):
        filter_face_velocity = line_data[line_titles.index('Filter face velocity')].strip() 
        if (filter_face_velocity != '') & ('filter_face_velocity' not in concat_meta):
            concat_meta['filter_face_velocity'] = 'Filter face velocity: '+filter_face_velocity
    
    if ('Exposed filter area' in line_titles):
        exposed_filter_area = line_data[line_titles.index('Exposed filter area')].strip() 
        if (exposed_filter_area != '') & ('exposed_filter_area' not in concat_meta):
            concat_meta['exposed_filter_area'] = 'Exposed filter area: '+exposed_filter_area
    
    if ('Filter description' in line_titles):
        filter_description = line_data[line_titles.index('Filter description')].strip()
        if (filter_description != '') & ('filter_description' not in concat_meta):
            concat_meta['filter_description'] = 'Filter description: '+filter_description
    
    if ('Medium' in line_titles):
        medium = line_data[line_titles.index('Medium')].strip()
        if (medium != '') & ('medium' not in concat_meta):
            concat_meta['medium'] = 'Medium: '+medium
    
    if ('Filter prefiring' in line_titles):
        filter_prefiring = line_data[line_titles.index('Filter prefiring')].strip()
        if (zero_values != '') & ('zero_values' not in concat_meta):
            concat_meta['filter_prefiring'] = 'Filter prefiring: '+filter_prefiring
    
    if ('Filter conditioning' in line_titles):
        filter_conditioning = line_data[line_titles.index('Filter conditioning')].strip()
        if (filter_conditioning != '') & ('filter_conditioning' not in concat_meta):
            concat_meta['filter_conditioning'] = 'Filter conditioning: '+filter_conditioning
    
    if ('Coating/Solution' in line_titles):
        coating = line_data[line_titles.index('Coating/Solution')].strip()
        if (coating != '') & ('coating' not in concat_meta):
            concat_meta['coating'] = 'Coating/Solution: '+coating
    
    if ('Sample preparation' in line_titles):
        sample_preparation = line_data[line_titles.index('Sample preparation')].strip()
        if (sample_preparation != '') & ('sample_preparation' not in concat_meta):
            concat_meta['sample_preparation'] = 'Sample preparation: '+sample_preparation
    
    #network QA details
    if ('QA date' in line_titles):  
        qa_date = line_data[line_titles.index('QA date')].strip()
        if (qa_date != '') & ('qa_date' not in concat_meta):
            concat_meta['qa_date'] = 'QA date: '+qa_date
        
    if ('QA outcome' in line_titles):
        qa_outcome = line_data[line_titles.index('QA outcome')].strip()
        if (qa_outcome != '') & ('qa_outcome' not in concat_meta):
            concat_meta['qa_outcome'] = 'QA outcome: '+qa_outcome
    
    if ('QA bias' in line_titles):
        qa_bias = line_data[line_titles.index('QA bias')].strip()
        if (qa_bias != '') & ('qa_bias' not in concat_meta):
            concat_meta['qa_bias'] = 'QA bias: '+qa_bias
    
    if ('QA variability' in line_titles):
        qa_variability = line_data[line_titles.index('QA variability')].strip()
        if (qa_variability != '') & ('qa_variability' not in concat_meta):
            concat_meta['qa_variability'] = 'QA variability: '+qa_variability
    
    if ('QA document URL' in line_titles):
        qa_document_url = line_data[line_titles.index('QA document URL')].strip()
        if (qa_document_url != '') & ('qa_document_url' not in concat_meta):
            concat_meta['qa_document_url'] = 'QA document URL: '+qa_document_url    
    
    #miscellaneous details
    if ('Comment' in line_titles):
        comment = line_data[line_titles.index('Comment')].strip() 
        if (comment != '') & ('network_miscellaneous_details' not in concat_meta):
            concat_meta['network_miscellaneous_details'] = comment

    return concat_meta,units_dict
