#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#EEA_AQ_eReporting_processor.py

#module which handles parsing of EEA AQ eReporting network data files.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import codecs
import crossnetwork_processing_functions
import datetime
import dateutil
import GHOST_standards
import numpy as np
import os
import pandas as pd
import re
import standardisation_mapping_functions
from standardisation_mapping_functions import EEA_AQ_eReporting_external_meta_map, EEA_AQ_eReporting_AIDE_D_external_meta_map, EEA_AIRBASE_external_meta_map
import sys
import time
import unit_converter

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def read_data(slurm_job_ID, parameter, station_reference):

    '''Convert one data provider original file for a specific station into a GHOST standard .npy station specific file'''

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- PARSING RAW DATA\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = crossnetwork_processing_functions.get_run_dictionaries(slurm_job_ID, 'EEA_AQ_eReporting', parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get data files specific to station
    data_files = crossnetwork_processing_functions.get_data_files(run_details, parameter_details, specific_station=station_reference)
    
    #sort files in date order
    data_file_years = [int(data_file.split('/')[-1].split('_')[3]) for data_file in data_files]
    data_files = np.array([x for _,x in sorted(zip(data_file_years,data_files))]) 
    data_file_years = sorted(data_file_years)    

    #load all external metadata into memory (3 files for EEA_AQ_eReporting) -- joins to specific station will be done later
    external_metadata = np.load(run_details['data_root']+'EEA_AQ_eReporting/metadata/processed/EEA_AQ_eReporting_META.npy', allow_pickle=True)
    external_metadata_AIDE_D = np.load(run_details['data_root']+'EEA_AQ_eReporting/metadata/processed/EEA_AQ_eReporting_AIDE_D_META.npy', allow_pickle=True)
    external_metadata_AIRBASE = np.load(run_details['data_root']+'EEA_AIRBASE/metadata/processed/EEA_AIRBASE_META.npy', allow_pickle=True)
    
    #------------------------------------------------------------------------------------#
    #ITERATE THROUGH ALL YEARLY FILES FOR STATION
    #------------------------------------------------------------------------------------#    
    eoicode_data_extent = {}
    eoicode_metadata_extent = {}
    sampling_process_data_extent = {}
    sampling_process_metadata_extent = {}
    
    #set flag to know when have written data for station    
    written_data = False

    for file_i, data_file in enumerate(data_files):
        valid_file = True
    
        #----------------------------------------------------------------------------#
        #READ FILE
        #----------------------------------------------------------------------------#
        log_file_str += 'PROCESSING FILE: {}\n'.format(data_file) 
    
        #get first 32 bytes of file and determine encoding (UTF-8 or UTF-16)
        bytes = min(32, os.path.getsize(data_file))
        raw = open(data_file, 'rb').read(bytes)
        if raw.startswith(codecs.BOM_UTF16):
            encoding = 'utf-16'
        else:
            encoding = 'utf-8'

        with open(data_file, 'r', encoding=encoding) as f:  

            for index,line in enumerate(f): 
        
                if valid_file == False:
                    break

                #strip newline
                line = line.strip("\r\n")
        
                #get column names
                if index == 0:
                    column_names = line.split(',')
                    file_data = [[] for _ in range(len(column_names))]

                #get Data
                elif index >= 1:
                    line_split = line.split(',')
                    for column_name in column_names:
                        try:
                            file_data[column_names.index(column_name)].append(line_split[column_names.index(column_name)])   
                        except:
                            log_file_str += 'FILE: {} IS NOT FORMATTED CORRECTLY. SKIPPING.\n'.format(data_file)
                            valid_file = False
                            break

        
        if valid_file == True:
        
            #convert file_data to numpy record array
            try:
                file_data = np.rec.fromarrays(file_data, names=column_names)
            #have no data in file trying to read
            except:
                log_file_str += 'NO DATA IN FILE: {}. SKIPPING.\n'.format(data_file)
                continue
                            

            #check species codes inside file are the correct ones for the species processing
            file_species_codes = np.unique(file_data['AirPollutantCode'])
            try:
                if (int(file_species_codes[0].split('/')[-1]) != int(parameter_details['airbase_code'])) or (len(file_species_codes) > 1):
                    log_file_str += 'INCORRECT SPECIES CODE INSIDE FILE FOR SPECIES PROCESSING. SKIPPING.\n'
                    continue
            except:
                log_file_str += 'FILE: {} IS NOT FORMATTED CORRECTED. SKIPPING.\n'.format(data_file)
                continue
               
            #Group data in file by EoICode
            all_grouped_data = []
            unique_eoicodes = np.unique(file_data['AirQualityStationEoICode'])
            for unique_eoicode in unique_eoicodes:

                grouped_data_by_eoicode = file_data[np.where(file_data['AirQualityStationEoICode'] == unique_eoicode)[0]]
                #Modifiy EoICodes, taking data only after after dashes to remove issues with incorrectly formatted codes
                grouped_data_by_eoicode['AirQualityStationEoICode'] = [x.split('-')[-1] for x in grouped_data_by_eoicode['AirQualityStationEoICode']] 

                #Group data in file by SamplingPoint
                unique_sampling_points = np.unique(file_data['SamplingPoint'])
                for unique_sampling_point in unique_sampling_points:
                    all_grouped_data.append(grouped_data_by_eoicode[np.where(grouped_data_by_eoicode['SamplingPoint'] == unique_sampling_point)[0]])

            #iterate through grouped data
            for grouped_data in all_grouped_data:
                
                #create site instances of metadata/data dictionaries
                site_data =     GHOST_standards.get_standard_data(parameter_details)
                site_metadata = GHOST_standards.get_standard_metadata(parameter_details)

                #get eoicode/sampling point of group
                eoicode = grouped_data['AirQualityStationEoICode'][0]
                data_sampling_point = grouped_data['SamplingPoint'][0] 
        
                #create a dictionary with unique units associated with indices in the file for conversion 
                unique_measurement_units = np.unique(grouped_data['UnitOfMeasurement'])
                unit_dict = {}
                #make sure have some valid units, otherwise skip file
                for units in unique_measurement_units:
                    if units == '':
                        continue
                    elif units == 'µg/m3':
                        #log_file_str += "Non-ascii micro sign in reported units - changing {} to ug/m3 for simplicity\n".format(units) 
                        changed_units = 'ug/m3'
                        unit_dict[changed_units] = np.where(grouped_data['UnitOfMeasurement'] == units)[0]
                    else:
                        unit_dict[units] = np.where(grouped_data['UnitOfMeasurement'] == units)[0]
                if unit_dict == {}:
                    log_file_str += 'NO UNITS FOR {}:{} PAIRING. SKIPPING\n'.format(eoicode, data_sampling_point)
                    continue
            
                #get external metadata specific for station & species (data is retrieved from 3 files)
        
                #ready key data/metadata variables that will use for making join between data and metadata
                eeacode = grouped_data['AirQualityStation'][0]
                data_sampling_process = grouped_data['SamplingProcess'][0]
                data_species_code = grouped_data['AirPollutantCode'][0]
                pollutant_AIDE_D_format = parameter_details['airbase_name']
                data_species_code_AIRBASE_format = data_species_code.split('/')[-1]

                #STANDARD EXTERNAL METADATA FILE              
                #first try making data/metadata join with standard external meta file using the "SamplingPoint"
                cut_external_metadata_standard = external_metadata[np.where(external_metadata['SamplingPoint'] == data_sampling_point)[0]]
        
                #if cut_external_metadata_standard is empty or not specific - next try doing join with "SamplingProcess"
                if len(cut_external_metadata_standard) == 0:
                     cut_external_metadata_standard = external_metadata[np.where(external_metadata['SamplingProcess'] == data_sampling_process)[0]]
                elif len(cut_external_metadata_standard) > 1:
                    cut_external_metadata_standard = cut_external_metadata_standard[np.where(cut_external_metadata_standard['SamplingProcess'] == data_sampling_process)[0]]
        
                #if cut_external_metadata_standard is empty or not specific - next try doing join with "AirQualityStationEoICode" and "AirPollutantCode"
                if len(cut_external_metadata_standard) == 0:
                    cut_external_metadata_standard = external_metadata[np.intersect1d(np.where(external_metadata['AirQualityStationEoICode'] == eoicode)[0], np.where(external_metadata['AirPollutantCode'] == data_species_code)[0])]
                elif len(cut_external_metadata_standard) > 1:
                    cut_external_metadata_standard = cut_external_metadata_standard[np.intersect1d(np.where(cut_external_metadata_standard['AirQualityStationEoICode'] == eoicode)[0], np.where(cut_external_metadata_standard['AirPollutantCode'] == data_species_code)[0])]
        
                #if cut_external_metadata_standard is empty or not specific - next try doing join with "AirQualityStation" and "AirPollutantCode"
                if len(cut_external_metadata_standard) == 0:
                    cut_external_metadata_standard = external_metadata[np.intersect1d(np.where(external_metadata['AirQualityStation'] == eeacode)[0], np.where(external_metadata['AirPollutantCode'] == data_species_code)[0])]
                elif len(cut_external_metadata_standard) > 1:
                    cut_external_metadata_standard = cut_external_metadata_standard[np.intersect1d(np.where(cut_external_metadata_standard['AirQualityStation'] == eeacode)[0], np.where(cut_external_metadata_standard['AirPollutantCode'] == data_species_code)[0])]
        
                #check if have specific standard metadata
                if len(cut_external_metadata_standard) != 0:
                    #if not specific, preferentially take first index
                    cut_external_metadata_standard = cut_external_metadata_standard[0]
        
    
                #AIDE D METADATA FILE
                #first try making data/metadata join with AIDE D external metadata file using the "SamplingPointId"
                cut_external_metadata_AIDE_D = external_metadata_AIDE_D[np.where(external_metadata_AIDE_D['SamplingPointId'] == data_sampling_point)[0]]
        
                #if cut_external_metadata_AIDE_D is empty or not specific - next try doing join with "SamplingProcess"
                if len(cut_external_metadata_AIDE_D) == 0:
                    cut_external_metadata_AIDE_D = external_metadata_AIDE_D[np.where(external_metadata_AIDE_D['SamplingProcess'] == data_sampling_process)[0]]
                elif len(cut_external_metadata_AIDE_D) > 1:
                    cut_external_metadata_AIDE_D = cut_external_metadata_AIDE_D[np.where(cut_external_metadata_AIDE_D['SamplingProcess'] == data_sampling_process)[0]]
        
                #if cut_external_metadata_AIDE_D is empty or not specific - next try doing join with "EoICode" and "Pollutant"
                if len(cut_external_metadata_AIDE_D) == 0:
                    cut_external_metadata_AIDE_D = external_metadata_AIDE_D[np.intersect1d(np.where(external_metadata_AIDE_D['EoICode'] == eoicode)[0], np.where(external_metadata_AIDE_D['Pollutant'] == pollutant_AIDE_D_format)[0])]
                elif len(cut_external_metadata_AIDE_D) > 1:
                    cut_external_metadata_AIDE_D = cut_external_metadata_AIDE_D[np.intersect1d(np.where(cut_external_metadata_AIDE_D['EoICode'] == eoicode)[0], np.where(cut_external_metadata_AIDE_D['Pollutant'] == pollutant_AIDE_D_format)[0])]
        
                #check if have specific AIDE D metadata
                if (len(cut_external_metadata_AIDE_D) != 0):
                    #if not specific, preferentially take first index
                    cut_external_metadata_AIDE_D = cut_external_metadata_AIDE_D[0]

        
                #AIRBASE v8 METADATA FILE 
                #first try making data/metadata join with AIRBASE external metadata file using the "station_european_code" and "component_code"
                cut_external_metadata_AIRBASE = external_metadata_AIRBASE[np.intersect1d(np.where(external_metadata_AIRBASE['station_european_code'] == eoicode)[0], np.where(external_metadata_AIRBASE['component_code'] == data_species_code_AIRBASE_format)[0])]
                if (len(cut_external_metadata_AIRBASE) == 1):
                    cut_external_metadata_AIRBASE = cut_external_metadata_AIRBASE[0]
                #if cut_external_metadata_AIRBASE is not specific - next try doing join taking into account measurement dates
                elif (len(cut_external_metadata_AIRBASE) > 1):
                    #select the metadata row with the latest "measurement_end_date"
                    #convert string metadata dates to datetime
                    #*first convert empty strings to a date much later in future (i.e. empty string = no end date)
                    measurement_end_dates = cut_external_metadata_AIRBASE['measurement_end_date']
                    measurement_end_dates[np.where(measurement_end_dates == '')[0]] = '3000-01-01'
                    measurement_end_dates = [datetime.datetime.strptime(date, '%Y-%m-%d').date() for date in measurement_end_dates]
                    #finally get index of latest end date and cut cut_external_metadata_AIRBASE
                    cut_external_metadata_AIRBASE = cut_external_metadata_AIRBASE[np.argmax(measurement_end_dates)]

                #log_file_str += 'standard: {}\n'.format(cut_external_metadata_standard['Altitude'])
                #log_file_str += 'aide d: {}\n'.format(cut_external_metadata_AIDE_D['Altitude'])
                #log_file_str += 'airbase: {}\n'.format(cut_external_metadata_AIRBASE['station_altitude'])
            
                #UNIFY 3 METADATA files information into standardised dict
                cut_unified_metadata = {}
                #first load in standard metadata variables
                if len(cut_external_metadata_standard) > 0:
                    for var in external_metadata.dtype.names:                    
                        if var in list(EEA_AQ_eReporting_external_meta_map.keys()):
                            #get standard_name of variable
                            standard_var = EEA_AQ_eReporting_external_meta_map[var]
                            #get variable value
                            var_value = cut_external_metadata_standard[var]
                            #do not write if empty string, or 'unknown' or '-999'
                            if (var_value != '') & (var_value != 'unknown') & (var_value != '-999'): 
                                cut_unified_metadata[standard_var] = var_value

                #next load in AIDE D metadata variables
                if len(cut_external_metadata_AIDE_D) > 0:
                    for var in external_metadata_AIDE_D.dtype.names:                    
                        if var in list(EEA_AQ_eReporting_AIDE_D_external_meta_map.keys()):
                            #get standard_name of variable
                            standard_var = EEA_AQ_eReporting_AIDE_D_external_meta_map[var]
                            #do not proceed if already have variable written in cut_unified_metadata dict
                            if standard_var not in list(cut_unified_metadata.keys()):
                                #get variable value
                                var_value = cut_external_metadata_AIDE_D[var]
                                #do not write if empty string, or 'unknown' or '-999'
                                if (var_value != '') & (var_value != 'unknown') & (var_value != '-999'): 
                                    cut_unified_metadata[standard_var] = var_value
                    
                #next load in AIRBASE metadata variables
                if len(cut_external_metadata_AIRBASE) > 0:
                    for var in external_metadata_AIRBASE.dtype.names:                    
                        if var in list(EEA_AIRBASE_external_meta_map.keys()):
                            #get standard_name of variable
                            standard_var = EEA_AIRBASE_external_meta_map[var]
                            #do not proceed if already have variable written in cut_unified_metadata dict
                            if standard_var not in list(cut_unified_metadata.keys()):
                                #get variable value
                                var_value = cut_external_metadata_AIRBASE[var]
                                #do not write if empty string, or 'unknown' or '-999'
                                if (var_value != '') & (var_value != 'unknown') & (var_value != '-999'): 
                                    cut_unified_metadata[standard_var] = var_value

                #Check have some valid metadata
                if len(cut_unified_metadata) == 0:
                    log_file_str += 'NO VALID EXTERNAL METADATA FOUND FOR {}:{} PAIRING. SKIPPING\n'.format(eoicode, data_sampling_process)
                    continue
    
                #------------------------------------------------------------------------#
                #PARSE TIME/DATA
                #------------------------------------------------------------------------#
          
                #parse start datetime - "DatetimeBegin"  (Defines the start time: yyyy-mm-dd hh:mm:ss Z, of the measurement -- includes timezone)
                #first parse into datetime (taking into account timezone information), convert to UTC datetime, then round to nearest minute
                start_dt = [crossnetwork_processing_functions.round_to_minute(dateutil.parser.parse(date).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in grouped_data['DatetimeBegin']]

                #parse end datetime - "DatetimeEnd" (Defines the end time: yyyy-mm-dd hh:mm:ss Z, of the measurement -- includes timezone)
                #first parse into datetime (taking into account timezone information), convert to UTC datetime, then round to nearest minute
                end_dt = [crossnetwork_processing_functions.round_to_minute(dateutil.parser.parse(date).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in grouped_data['DatetimeEnd']]

                #convert each start measurement datetime to minutes from reference time
                frac_time = [crossnetwork_processing_functions.datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

                #if length of time array is 0, skip to next file
                if len(frac_time) == 0:
                    log_file_str += 'NO DATA IN FILE: {}. SKIPPING.\n'.format(data_file)
                    continue

                #get timedeltas between start and end of measurement window
                temporal_resolutions = []
                for start, end in zip(start_dt,end_dt):
                    dt_window = end - start
                    #convert hours to minutes
                    dt_window_minutes = int(dt_window.total_seconds()/60.)
                    temporal_resolutions.append(dt_window_minutes)  
                    
                #parse data flags
                flags = []
                qa = []
    
                #get dictionary necessary for mapping of EEA AQ eReporting file data flags to standard data flag codes
                data_flag_mapping_dictionary = standardisation_mapping_functions.network_data_flag_to_standard_data_flag_mapping['EEA_AQ_eReporting']                

                #get flags from data file - Validity
                data_flags = grouped_data['Validity']

                #get verification information - informs on data verification
                verification_data = grouped_data['Verification']

                #iterate through EEA AQ eReporting file data flags and map to standard flag codes
                for data_flag, verification_flag in zip(data_flags,verification_data):

                    #create empty list for standard data flags associated with each measurement
                    standard_data_flag_names = []
                
                    #get mapped standard data flag specific dictionary 
                    mapped_standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                    #unpack mapped standard data flag key variables
                    standard_data_flag_names.append(mapped_standard_data_flag_dict['standard_data_flag_name'])
                    GHOST_decreed_validity = mapped_standard_data_flag_dict['GHOST_decreed_validity']
                    network_decreed_validity = mapped_standard_data_flag_dict['network_decreed_validity']
                
                    #write preliminary data flags if necessary (if "Verification" variable informs that data is preliminary)
                    #1 = Verified, 2 = Preliminary verified, 3 = Not verified
                    if (verification_flag == '2') or (verification_flag == '3'):
                        if 'Preliminary Data' not in standard_data_flag_names:
                            standard_data_flag_names.append('Preliminary Data')

                    #get standard flag codes associated with each standard data flag name, and append to list to write out
                    standard_data_flag_codes = [GHOST_standards.standard_data_flag_name_to_data_flag_code[standard_data_flag_name] for standard_data_flag_name in standard_data_flag_names]
                    flags.append(standard_data_flag_codes)
        
                    #process qa information associated with flags
                    qa_codes = []
                    #append qa flag code if standard data flag is GHOST decreed invalid
                    if GHOST_decreed_validity == 'I':
                        qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
                    #append qa flag if standard data flag is network decreed invalid
                    if network_decreed_validity == 'I':
                        qa_codes.append(GHOST_standards.standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
                    #append qa codes to list to write out
                    qa.append(qa_codes)
        
                #parse unique reported measurement unit
                #can be instances where have multiple units in a file, in this case write raw units as modal occurring unit (that is not blank)
                #all raw units will be taken into account when converting concentrations however
                unit_n = -1
                for unit_name in list(unit_dict.keys()):
                    if len(unit_dict[unit_name]) > unit_n:
                        reported_units = unit_name   
                        unit_n = len(unit_dict[unit_name])        
                #convert raw reported units given to standard name
                conv_obj = unit_converter.convert_units(reported_units, reported_units, 1, measured_species = parameter_details['chemical_formula'])
                conv_obj.do_conversion()
                site_metadata['measuring_instrument_reported_units']['value'].append(conv_obj.output_standard_units)

                #parse measurements
                measurements = []
                #measurement values may need converting (i.e. mass density to mixing ratio)
        
                #gas volume is not provided by EEA AQ eReporting and if units need converting, this conversion is done using assumed network standard pressure and temperatures - write qa data flag to inform of this assumption
                #measurement temperature and pressure values are never provided by EEA AQ eReporting so are always assumed to be the network standard pressure and temperatures
                #these are defined as for 293 K temperature, and 1013 hPa for pressure here: https://eur-lex.europa.eu/legal-content/EN/TXT/PDF/?uri=CELEX:32008L0050&from=en and here: https://eur-lex.europa.eu/legal-content/EN/TXT/PDF/?uri=CELEX:32004L0107&from=EN 
                site_metadata['measuring_instrument_volume_standard_temperature']['value'].append('293.0')
                site_metadata['measuring_instrument_volume_standard_pressure']['value'].append('1013.0')
        
                #this flag below alerts for writing of qa flag later
                write_volume_assumption_flag = False
        
                #check if reported standard units, match required standard units
                #take care to deal with multiple units in file
                order_n = []
                for reported_units in list(unit_dict.keys()):

                    #convert raw reported units given to standard name
                    conv_obj = unit_converter.convert_units(reported_units, reported_units, 1, measured_species = parameter_details['chemical_formula'])
                    conv_obj.do_conversion()
                    standard_reported_units = conv_obj.output_standard_units
        
                    #cut concentration data to be that just associated with units
                    cut_concentrations = grouped_data['Concentration'][unit_dict[reported_units]]
            
                    if standard_reported_units != site_data[parameter_details['bsc_parameter_name']]['units']:
                        #mass density to mixing ratio?
                        if ('m-3' in standard_reported_units) & ('mol-1' in site_data[parameter_details['bsc_parameter_name']]['units']):
            
                            #set flag to instruct writing of flag that am assuming gas volume in converting units 
                            write_volume_assumption_flag = True
                
                            #get conversion factor
                            input_units = {'temperature':'K', 'pressure':'hPa', 'molar_mass':'kg mol-1', 'mass_density':standard_reported_units}
                            input_vals = {'temperature':site_metadata['measuring_instrument_volume_standard_temperature']['value'][-1], 'pressure':site_metadata['measuring_instrument_volume_standard_pressure']['value'][-1], 'molar_mass':crossnetwork_processing_functions.get_molecular_mass(parameter_details['chemical_formula']), 'mass_density':1.0}
                            conv_obj = unit_converter.convert_units(input_units,site_data[parameter_details['bsc_parameter_name']]['units'],input_vals,measured_species = parameter_details['chemical_formula'],conversion_input_quantity = 'mass_density')
            
                        #otherwise simple unit conversion, i.e. pptv to ppbv or mass density to mass density
                        else:
                            conv_obj = unit_converter.convert_units(standard_reported_units,site_data[parameter_details['bsc_parameter_name']]['units'],1.0, measured_species = parameter_details['chemical_formula'])
    
                        #do conversion
                        conv_obj.do_conversion()
        
                        #process through each measurement and convert value using derived conversion factor
                        for n, meas in enumerate(cut_concentrations):
                            actual_n = unit_dict[reported_units][n]
                            order_n.append(actual_n)
                            #only do conversion if measurement value is above 0 and not infinite
                            try:
                                float_meas = float(meas)
                            except:
                                float_meas = np.NaN
                            if (float_meas > 0.0) & (np.isinf(float_meas) == False):
                                measurements.append(float_meas*conv_obj.conversion_factor)
                            else:
                                measurements.append(float_meas)
        
                    #otherwise do nothing to measured values
                    else:
                        for n, meas in enumerate(cut_concentrations):
                            order_n.append(unit_dict[reported_units][n])
                            try:
                                float_meas = float(meas)
                            except:
                                float_meas = np.NaN
                            measurements.append(float_meas)
        
                #resort measurements by original order
                measurements = [x for _,x in sorted(zip(order_n,measurements))]
        
                #filter frac_time, temporal_resolutions, flags and qa by order_n (in case there are some measurements associated unknown units in file
                order_n = sorted(order_n)
                frac_time = [frac_time[x] for x in order_n]
                temporal_resolutions = [temporal_resolutions[x] for x in order_n]
                flags = [flags[x] for x in order_n]
                qa = [qa[x] for x in order_n]

                #keep only measurement windows that are >= 1 minute in temporal resolution, as well as all other associated variables across other fields
                valid_temporal_resolution_inds = [ii for ii, res in enumerate(temporal_resolutions) if res >= 1]
                if len(valid_temporal_resolution_inds) != len(frac_time):
                    frac_time = [frac_time[x] for x in valid_temporal_resolution_inds]
                    temporal_resolutions = [temporal_resolutions[x] for x in valid_temporal_resolution_inds]
                    flags = [flags[x] for x in valid_temporal_resolution_inds]
                    qa = [qa[x] for x in valid_temporal_resolution_inds]
                    measurements = [measurements[x] for x in valid_temporal_resolution_inds]
                
                #------------------------------------------------------------------------#
                #PARSE METADATA
                #------------------------------------------------------------------------#

                #GHOST version
                site_metadata['GHOST_version']['value'].append(run_details['ghost_version'])
        
                #save meta_update_stamp, set as first sorted time processed (as minutes relative to reference time)
                site_metadata['meta_update_stamp']['value'].append(sorted(frac_time)[0])
        
                #write known metadata  
                site_metadata['network']['value'].append('EEA_AQ_eReporting')
                site_metadata['data_level']['value'].append('E1a')

                #define list of variable names that require no processing that can be simply added to site_metadata
                metadata_to_add = ['latitude','longitude','altitude','sampling_height','distance_to_kerb','distance_to_building',
                                  'distance_to_junction','distance_to_source','street_type','street_width','daytime_traffic_speed',
                                  'daily_passing_vehicles','station_name','city','representative_radius','principal_investigator_institution',
                                  'principal_investigator_email_address','contact_institution','contact_email_address']

                #iterate through all metadata variable names and put values into standard dictionary
                for variable_name in metadata_to_add:
                    #check variable is in unified metadata dict
                    if variable_name in list(cut_unified_metadata.keys()):
                        site_metadata[variable_name]['value'].append(cut_unified_metadata[variable_name].replace('"',''))

                #parse associated networks 
                #first add AIRBASE with EoICode
                associated_network_string = 'AIRBASE:{}'.format(eoicode)
                #second add EBAS, if have 'emep_station' variable and is 'yes'
                if 'emep_station' in list(cut_unified_metadata.keys()):
                    if cut_unified_metadata['emep_station'] == 'yes':
                        #add EBAS with EoICode
                        ebas_string = 'EBAS:{}'.format(eoicode)
                        associated_network_string = '{};{}'.format(associated_network_string, ebas_string)
                #finally add local network with local code
                if ('local_network_code' in list(cut_unified_metadata.keys())) & ('local_reference' in list(cut_unified_metadata.keys())):
                    local_string = '{}:{}'.format(cut_unified_metadata['local_network_code'],cut_unified_metadata['local_reference'])
                    associated_network_string = '{};{}'.format(associated_network_string, local_string)
                site_metadata['associated_networks']['value'].append(associated_network_string)
    
                #parse horizontal geodetic datum - take from parsed spatial reference system information (certain spatial reference systems use specific datums - i.e. 'EPSG:4979' uses 'WSG84' datum)
                if 'spatial_reference_system' in list(cut_unified_metadata.keys()):
                    if cut_unified_metadata['spatial_reference_system'] == 'EPSG:4979':
                       site_metadata['horizontal_geodetic_datum']['value'].append('WSG84') 
                    #if spatial reference system is unknown then write handled error logfile (2)
                    else:
                        log_file_str += 'UNKNOWN SPATIAL REFERENCE SYSTEM:{}'.format(cut_unified_metadata['spatial_reference_system'])
                        crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,2,log_file_str,station_reference)
                
                #parse population (need to check to remove non-numeric characters)
                if 'population' in list(cut_unified_metadata.keys()):
                    population = cut_unified_metadata['population']
                    population = re.sub("[^0-9]", "", population)
                    site_metadata['population']['value'].append(population) 
        
                #parse 'principal_investigator_name' (concatenate title, first name and last name, if given)
                principal_investigator_name = ''
                if 'principal_investigator_title' in list(cut_unified_metadata.keys()):
                    principal_investigator_name = principal_investigator_name + cut_unified_metadata['principal_investigator_title']
                if 'principal_investigator_first_name' in list(cut_unified_metadata.keys()):
                    if principal_investigator_name == '':
                        principal_investigator_name = principal_investigator_name + cut_unified_metadata['principal_investigator_first_name']
                    else:
                        principal_investigator_name = '{} {}'.format(principal_investigator_name, cut_unified_metadata['principal_investigator_first_name'])
                if 'principal_investigator_last_name' in list(cut_unified_metadata.keys()):
                    if principal_investigator_name == '':
                        principal_investigator_name = principal_investigator_name + cut_unified_metadata['principal_investigator_last_name']
                    else:
                        principal_investigator_name = '{} {}'.format(principal_investigator_name, cut_unified_metadata['principal_investigator_last_name'])
                if principal_investigator_name != '':
                    site_metadata['principal_investigator_name']['value'].append(principal_investigator_name) 
         
                #parse 'contact_name' (concatenate title, first name and last name, if given)
                contact_name = ''
                if 'contact_title' in list(cut_unified_metadata.keys()):
                    contact_name = contact_name + cut_unified_metadata['contact_title']
                if 'contact_first_name' in list(cut_unified_metadata.keys()):
                    if contact_name == '':
                        contact_name = contact_name + cut_unified_metadata['contact_first_name']
                    else:
                        contact_name = '{} {}'.format(contact_name, cut_unified_metadata['contact_first_name'])
                if 'contact_last_name' in list(cut_unified_metadata.keys()):
                    if contact_name == '':
                        contact_name = contact_name + cut_unified_metadata['contact_last_name']
                    else:
                        contact_name = '{} {}'.format(contact_name, cut_unified_metadata['contact_last_name'])
                if contact_name != '':
                    site_metadata['contact_name']['value'].append(contact_name) 
        
                #parse data timezone (in hours from UTC)
                data_timezone = (dateutil.parser.parse(grouped_data['DatetimeBegin'][0]).utcoffset().seconds)/60.0/60.0
                site_metadata['data_timezone']['value'].append(data_timezone) 
        
                #parse miscellaneous details
                #deal with miscellaneous sampling details
                if 'sample_location_miscellaneous' in list(cut_unified_metadata.keys()):
                    miscellaneous_sampling = 'Sampling location: {}'.format(cut_unified_metadata['sample_location_miscellaneous'])
                    site_metadata['network_sampling_details']['value'].append(miscellaneous_sampling) 
        
                #deal with miscellaneous sampling inlet information
                if 'length_sampling_line' in list(cut_unified_metadata.keys()):
                    inlet_info = 'Length: {}'.format(cut_unified_metadata['length_sampling_line'])
                    site_metadata['measuring_instrument_inlet_information']['value'].append(inlet_info)
                        
                #deal with miscellaneous maintenance details
                miscellaneous_maintenance = ''
                if 'calibration_method' in list(cut_unified_metadata.keys()):
                    string_to_add = 'Calibration methodology: {}'.format(cut_unified_metadata['calibration_method'])
                    miscellaneous_maintenance = miscellaneous_maintenance + string_to_add
                if 'calibration_frequency' in list(cut_unified_metadata.keys()):
                    string_to_add = 'Calibration frequency: {}'.format(cut_unified_metadata['calibration_frequency'])
                    if miscellaneous_maintenance == '':
                        miscellaneous_maintenance = miscellaneous_maintenance + string_to_add
                    else:
                        miscellaneous_maintenance = '{};{}'.format(miscellaneous_maintenance, string_to_add)            
                if miscellaneous_maintenance != '':
                    site_metadata['network_maintenance_details']['value'].append(miscellaneous_maintenance)
        
                #deal with other miscellaneous details
                if 'station_description' in list(cut_unified_metadata.keys()):
                    station_description = 'Station description: {}'.format(cut_unified_metadata['station_description'])
                    site_metadata['network_miscellaneous_details']['value'].append(station_description)
        
                #parse country code
                country_code = grouped_data['Countrycode'][0]
                site_metadata['country']['value'].append(GHOST_standards.country_ISO3166_to_name[country_code]) 

                #parse measurement information (from external metadata) - use all possible variables of: 'sampling_equipment', 'measurement_equipment', 'analytical_technique', 'measurement_method', 'measurement_type', 'measurement_automatic'
                if 'sampling_equipment' in list(cut_unified_metadata.keys()):
                    sampling_equipment = cut_unified_metadata['sampling_equipment']
                else:
                    sampling_equipment = ''
                if 'measurement_equipment' in list(cut_unified_metadata.keys()):
                    #remove 'unnessary long string from variable
                    measurement_equipment = cut_unified_metadata['measurement_equipment'].split('measurementequipment/')[-1]
                else:
                    measurement_equipment = ''
                if 'analytical_technique' in list(cut_unified_metadata.keys()):
                    analytical_technique = cut_unified_metadata['analytical_technique']
                else:
                    analytical_technique = ''
                if 'measurement_method' in list(cut_unified_metadata.keys()):
                    #remove 'unnessary long string from variable
                    measurement_method = cut_unified_metadata['measurement_method'].split('measurementmethod/')[-1]
                else:
                    measurement_method = ''
                if 'measurement_type' in list(cut_unified_metadata.keys()):
                    measurement_type = cut_unified_metadata['measurement_type']
                else:
                    measurement_type = ''
                #only parse 'measurement_automatic' if don't have 'measurement_type'
                if ('measurement_automatic' in list(cut_unified_metadata.keys())) & (measurement_type == ''):
                    measurement_automatic = cut_unified_metadata['measurement_automatic']
                    if measurement_automatic == 'yes':
                        measurement_type = 'automatic'
                    elif measurement_automatic == 'no':
                        measurement_type = 'not-automatic'
                else:
                    measurement_automatic = ''
                #concatenate measurement information together
                measurement_information = '{} {} {} {} {}'.format(sampling_equipment, measurement_equipment, analytical_technique, measurement_method, measurement_type)
                #strip leading/trailing white spaces, and change multiple white spaces to be singular
                measurement_information = measurement_information.strip()
                measurement_information = ' '.join(measurement_information.split())
                site_metadata, measurement_qa_flags, standard_method_abbreviation = standardisation_mapping_functions.map_network_measurement_information_to_standard_measurement_information(measurement_information, site_metadata, run_details)

                #parse station classifications (from external metadata) - use all possible variables of: 'area_classification', 'station_classification', 'main_emission_source', 'land_use', 'area_classification_rural_subdivision', 'representative_radius'
                if 'area_classification' in list(cut_unified_metadata.keys()):
                    area_classification = cut_unified_metadata['area_classification']
                else:
                    area_classification = ''
                if 'station_classification' in list(cut_unified_metadata.keys()):
                    station_classification = cut_unified_metadata['station_classification']
                else:
                    station_classification = ''
                if 'main_emission_source' in list(cut_unified_metadata.keys()):
                    main_emission_source = cut_unified_metadata['main_emission_source']
                else:
                    main_emission_source = ''
                if 'land_use' in list(cut_unified_metadata.keys()):
                    land_use = cut_unified_metadata['land_use']
                else:
                    land_use = ''
                if 'area_classification_rural_subdivision' in list(cut_unified_metadata.keys()):
                    area_classification_rural_subdivision = cut_unified_metadata['area_classification_rural_subdivision']
                else:
                    area_classification_rural_subdivision = ''   
                if 'representative_radius' in list(cut_unified_metadata.keys()):
                    representative_radius = cut_unified_metadata['representative_radius']
                else:
                    representative_radius = ''   
                native_classifications = {'area_classification':area_classification, 'station_classification':station_classification, 'main_emission_source':main_emission_source, 'land_use':land_use, 'area_classification_rural_subdivision':area_classification_rural_subdivision, 'representative_radius':representative_radius}
                site_metadata = standardisation_mapping_functions.map_network_classifications_to_standard_classifications(native_classifications, site_metadata, run_details)
        
                #iterate through metadata variables, if any columns have not been appended to add a NaN
                #check number of refs in list
                for column_header in list(site_metadata.keys()):
                    n_column = len(site_metadata[column_header]['value'])
                    #if column is empty, then append NaN
                    if n_column < 1:
                        site_metadata[column_header]['value'].append(np.NaN)
                
                #ensure that measurement position metadata are floats, otherwise append nan
                position_meta = ['latitude','longitude','altitude','sampling_height','measurement_altitude']
                for pm in position_meta:
                    try:
                        site_metadata[pm]['value'][-1] = np.float64(site_metadata[pm]['value'][-1])
                    except:
                        site_metadata[pm]['value'][-1] = np.NaN

                #process non-network specific metadata (station_timezone, data_download_stamp)
                site_metadata = crossnetwork_processing_functions.process_non_network_specific_metadata(site_metadata, data_file)
                
                #set station_reference (provided network ID + standard measurement method abbreviation)
                site_metadata['station_reference']['value'][-1] = '{}_{}'.format(eoicode,standard_method_abbreviation)

                #check if longitude/latitude coordinates are suspect
                new_longitude, new_latitude = crossnetwork_processing_functions.check_geographical_position(site_metadata['longitude']['value'][-1], site_metadata['latitude']['value'][-1])
                site_metadata['longitude']['value'][-1] = new_longitude
                site_metadata['latitude']['value'][-1] = new_latitude

                #----------------------------------------------------------------------------#    
                #QA
                #----------------------------------------------------------------------------#
                #write measurement associated qa flags

                #iterate through measurement QA flags and write standard qa flag code if required
                for standard_flag_name in list(measurement_qa_flags.keys()):
                    if measurement_qa_flags[standard_flag_name] == True:
                        for qq in range(len(qa)):
                            qa[qq].append(GHOST_standards.standard_QA_name_to_QA_code[standard_flag_name])
        
                #write assumed volume of gas qa flags (inform have assumed volume of gas for converting units)
                if write_volume_assumption_flag == True:
                    for qq in range(len(qa)):
                        qa[qq].append(GHOST_standards.standard_QA_name_to_QA_code['Assumed Gas Volume'])
                
                #----------------------------------------------------------------------------#
                #WRITE OUT DATA
                #----------------------------------------------------------------------------#

                #No LODs or measurement uncertainties per measurement - append list of NaNs    
                site_data['time']['value'].extend(frac_time)
                site_data[parameter_details['bsc_parameter_name']]['value'].extend(measurements)
                site_data['temporal_resolution']['value'].extend(temporal_resolutions)   
                site_data['reported_lower_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
                site_data['reported_upper_limit_of_detection_per_measurement']['value'].extend([np.NaN]*len(frac_time))
                site_data['reported_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
                site_data['derived_uncertainty_per_measurement']['value'].extend([np.NaN]*len(frac_time))
                site_data['flag']['value'].extend(flags)
                site_data['qa']['value'].extend(qa)
                site_data['classification']['value'].extend([[] for i in range(len(frac_time))])

                #standardise metadata/data and write out specific file station data to temporary .npy
                crossnetwork_processing_functions.post_process_file_parse(site_metadata,site_data,run_details,parameter_details)

                #update flag to state that have written data for station
                written_data = True

        #if stage has completed successfully, with some valid data, write a valid process log file
    if written_data == True:
        log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
        crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,0,log_file_str,station_reference)
    #else, if no data has been written for station write an error log file to inform station is not continuing in the pipeline
    elif written_data == False:
        log_file_str += 'NO DATA HAS BEEN WRITTEN FOR STATION'
        crossnetwork_processing_functions.create_GHOST_stage_logfile('S1',run_details,1,log_file_str,station_reference)