#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#GHOST_pipeline.py

#Program that handles the submission and management of GHOST pipeline jobs

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import copy
from crossnetwork_processing_functions import check_GHOST_output_directories, check_GHOST_stage_temporary_directories, check_GHOST_stage_arguments_logs_submit, clean_native_directory, clean_output_resolution_directories, get_run_dictionaries
import GHOST_pipeline_functions
import sys
import time

###--------------------------------------------------------------------------------------------------###
#Get root slurm job ID, and active network/parameter to process --> passed from job submittal script
###--------------------------------------------------------------------------------------------------###
slurm_job_ID, network, parameter = sys.argv[1], sys.argv[2], sys.argv[3]

###--------------------------------------------------------------------------------------------------###
#Get dictionaries holding run specific variables, parameter specific variables and network specific variables
###--------------------------------------------------------------------------------------------------###
run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)

###--------------------------------------------------------------------------------------------------###
#Check/create GHOST output native/output resolution directories
###--------------------------------------------------------------------------------------------------###
check_GHOST_output_directories(run_details, parameter_details)

###--------------------------------------------------------------------------------------------------###
#Check/create/clean temporary data directories of relevant defined GHOST stages
###--------------------------------------------------------------------------------------------------###
check_GHOST_stage_temporary_directories(run_details, parameter_details)

###--------------------------------------------------------------------------------------------------###
#Check/create/clean argument/log/submit directories of relevant defined GHOST stages
###--------------------------------------------------------------------------------------------------###
check_GHOST_stage_arguments_logs_submit(run_details)

###---------------------------------------------------------------------------------------
#SUMMARY OF GHOST PIPELINE STAGES:
###---------------------------------------------------------------------------------------

#PARSING OF NETWORK DATA FILES PARSING TO STATION SPECIFIC .npz FILES
#  S1 = Raw network files parse
#  S2 = Concatenate processed data files 
#  S3 = Split stations by significantly changing locational information
#  S4 = Parse globally gridded metadata
#  S5 = Quality control
#  S6 = Process classifications

#TEMPORALLY GRIDDING CONCATENATION OF STATIONS TO netCDF
#  T1 = Temporally grid by station reference
#  T2 = Group temporally gridded station data in monthly netCDFs
    
###--------------------------------------------------------------------------------------------------###
#Start timer
###--------------------------------------------------------------------------------------------------###


#start pipeline logfile
pipeline_logfile = open('{}/run/pipeline_logs/{}_{}.out'.format(run_details['process_root'],network, parameter),"w")
pipeline_logfile.write('**************************************************************************************************************\n')        
pipeline_logfile.write('BEGINNING PROCESSING OF STAGES: {} FOR {}:{}\n'.format(run_details['stages_to_process'],network,parameter))
pipeline_logfile.write('**************************************************************************************************************\n')


#initialise timing variables 
pipeline_start_time = time.time()
total_stage_process_time = 0
total_stage_queue_time = 0

###--------------------------------------------------------------------------------------------------###
#STAGE S1 : 1. Process all raw network files for a species
#           2. Outputs processed files to /native/parameter/temporary_S1 directory as .npz files grouped by defined stations (network_station_reference:measurement_method) 
#           3. If a file contains measurements of varying station reference, different measuring methods, or different measurement instrumentation, all these different instances will be separated out into unique files    
###--------------------------------------------------------------------------------------------------###

if 'S1' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S1(pipeline_start_time, slurm_job_ID, run_details, parameter_details, network_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE S2 : 1. Concatenate /native/component/temporary_S1/*.npz files by station (handling duplicate data reporting)
#           2. Read manually compiled metadata. If any metadata fields are empty through entire time record attempt to set fields through time record by incorporating manually compiled metadata.   
#           3. Handle gaps in selected metadata fields through time record. 
###--------------------------------------------------------------------------------------------------###

if 'S2' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S2(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE S3 : 1. Split /native/parameter/temporary_S2/*.npz station files by significantly changing locational information (creating new stations where necessary)
###--------------------------------------------------------------------------------------------------###

if 'S3' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S3(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE S4 : 1. Process globally gridded metadata by station
###--------------------------------------------------------------------------------------------------###

if 'S4' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S4(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE S5 : 1. Do Quality Control checks per station 
###--------------------------------------------------------------------------------------------------###
    
if 'S5' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S5(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE S6 : 1. Clean native output directory for specific network/parameter
#           2. Do Classification checks per station
#           3. Move /native/parameter/temporary_P6/*.npz files per station to /native/component/ directory as files now finalised
#           4. Some variables that are no longer needed are dropped 
###--------------------------------------------------------------------------------------------------###
  
if 'S6' in run_details['stages_to_process']:
    clean_native_directory(run_details, parameter_details)
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.S6(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

###--------------------------------------------------------------------------------------------------###
#STAGE T1 : 1. Temporally grid station specific data to desired temporal resolution by year/month
#           2. Outputs station specific netCDF files to respective /temporal_resolution/parameter/temporary/year/month directories 
###--------------------------------------------------------------------------------------------------###
        
if 'T1' in run_details['stages_to_process']:
    stage_process_time, stage_queue_time, valid_temporal_resolution_codes = GHOST_pipeline_functions.T1(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile)  
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time            

###--------------------------------------------------------------------------------------------------###
#STAGE T2 : 1. Clean all defined output resolution directories
#           2. Merge temporary station specific files to 1 netCDF by year/month 
#           3. Outputs netCDF (.nc) files by year/month to /temporal_resolution/parameter/ directory
#           4. Clean all temporary directories
###--------------------------------------------------------------------------------------------------###

if 'T2' in run_details['stages_to_process']:
    #if valid_temporal_resolution codes variable does not exist, assume all defined temporal resolutions to process are valid 
    if 'valid_temporal_resolution_codes' not in locals():
         valid_temporal_resolution_codes = copy.deepcopy(run_details['temporal_resolutions'])
    clean_output_resolution_directories(run_details, parameter_details)
    stage_process_time, stage_queue_time = GHOST_pipeline_functions.T2(pipeline_start_time, slurm_job_ID, run_details, parameter_details, valid_temporal_resolution_codes, pipeline_logfile) 
    total_stage_process_time += stage_process_time
    total_stage_queue_time += stage_queue_time

    #clean all temporary GHOST stage data directories
    check_GHOST_stage_temporary_directories(run_details, parameter_details)

###--------------------------------------------------------------------------------------------------###
#FINALISE GHOST PIPELINE :  1. Stop timer       
###--------------------------------------------------------------------------------------------------###

pipeline_end_time = time.time()
total_pipeline_time = (pipeline_end_time - pipeline_start_time)/60.
total_stage_process_time = total_stage_process_time
total_stage_queue_time = total_stage_queue_time
pipeline_overhead_time = total_pipeline_time - (total_stage_process_time + total_stage_queue_time)

pipeline_logfile.write('\n') 
pipeline_logfile.write('PROCESSING OF ALL DESIRED STAGES FOR {}:{} COMPLETE IN {:.2f} MINUTES\n({:.2f} MINUTES STAGE PROCESSING, {:.2f} MINUTES IN STAGE QUEUING, {:.2f} MINUTES IN PIPELINE OVERHEADS)'.format(network,parameter,total_pipeline_time,total_stage_process_time,total_stage_queue_time,pipeline_overhead_time))
pipeline_logfile.close() 
###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###
