
#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#GHOST_pipeline_functions.py

#Program that defines the functions which execute the separate GHOST pipeline stage jobs 

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

from crossnetwork_processing_functions import get_data_files, get_station_references_from_filenames, create_greasy_arguments_file, submit_greasy_job, get_unique_processed_references, get_relevant_yearmonths
import os
import sys
import time
import random

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S1(pipeline_start_time, slurm_job_ID, run_details, parameter_details, network_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S1

       In this stage:
         -- All raw network files for a species are parsed
         -- Processed files are output to /native/parameter/temporary_P1 directory as .npz files grouped by defined stations (network_station_reference:measurement_method)
         -- If a file contains measurements of varying station reference, different measuring methods, or different measurement instrumentation, all these different instances will be separated out into unique files
    '''

    #get all relevant data files for specific parameter, per network
    data_files = get_data_files(run_details, parameter_details)

    #check you have some data to process
    if len(data_files) > 0:

        #set up parallelisation to process data files, dependant on how observations are reported
        #1. If file storage type is all stations per year, parallelise by year
        #2. If file storage is station per year/multiyear, parallelise by station

        #if file storage type is "all_per_year", get all different data years to parallelise by 
        if network_details['file_storage_type'] == 'all_per_year':
            param_to_parallelise_by = range(int(run_details['start_year']),int(run_details['end_year']))

        #otherwise, get all station references to parallelise by
        else:
            #get all unique station references from data filenames
            param_to_parallelise_by = get_station_references_from_filenames(data_files, run_details) 

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, param_to_parallelise_by, GHOST_stage='S1')

        #submit greasy job/s to process all years, or station data in parallel
        job_name = 'S1_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S1',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero data files then stage outcome == process code: 2 (i.e. no data)
    else:
        pipeline_logfile.write('S1 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()
###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S2(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S2

       In this stage:
        -- Concatenate /native/parameter/temporary_P1/*.npz files by station (handling duplicate data reporting)
        -- Read manually compiled metadata. If any metadata fields are empty through entire time record attempt to set fields through time record by incorporating manually compiled metadata.   
        -- Handle gaps in selected metadata fields through time record. 
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference
    processed_temporary_references = get_unique_processed_references('S2', run_details, parameter_details)

    #check you have some data to process
    if len(processed_temporary_references) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, processed_temporary_references, GHOST_stage='S2')

        #submit greasy job/s to process station data in parallel
        job_name = 'S2_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S2',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('S2 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S3(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S3

       In this stage:
        -- Split /native/component/P2_temporary/*.npz station files by significantly changing locational information (creating new stations where necessary)
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference
    processed_temporary_references = get_unique_processed_references('S3', run_details, parameter_details)

    #check you have some data to process
    if len(processed_temporary_references) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, processed_temporary_references, GHOST_stage='S3')

        #submit greasy job/s to process station data in parallel
        job_name = 'S3_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S3',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('S3 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S4(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S4

       In this stage: 
        -- Process globally gridded metadata by station
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference
    processed_temporary_references = get_unique_processed_references('S4', run_details, parameter_details)

    #check you have some data to process
    if len(processed_temporary_references) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, processed_temporary_references, GHOST_stage='S4')

        #submit greasy job/s to process station data in parallel
        job_name = 'S4_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S4',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('S4 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S5(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S5
     
       In this stage:
        -- Do Quality Control checks per station 
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference
    processed_temporary_references = get_unique_processed_references('S5', run_details, parameter_details)

    #check you have some data to process
    if len(processed_temporary_references) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, processed_temporary_references, GHOST_stage='S5')

        #submit greasy job/s to process station data in parallel
        job_name = 'S5_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S5',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('S5 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def S6(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: S6

       In this stage:
        -- Do Classification checks per station
        -- Move /native/parameter/temporary_P6/*.npz files per station to /native/component/ directory as files now finalised
        -- Some variables that are no longer needed are dropped 
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference
    processed_temporary_references = get_unique_processed_references('S6', run_details, parameter_details)

    #check you have some data to process
    if len(processed_temporary_references) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, processed_temporary_references, GHOST_stage='S6')

        #submit greasy job/s to process station data in parallel
        job_name = 'S6_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='S6',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('S6 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def T1(pipeline_start_time, slurm_job_ID, run_details, parameter_details, pipeline_logfile):

    '''GHOST PIPELINE STAGE: T1
      
       In this stage:
        -- Temporally grid station specific data to desired temporal resolution by year/month
        -- Outputs station specific netCDF files to respective /temporal_resolution/parameter/temporary/year/month directories 
    '''

    #get all unique station references from output numpy files in previous stage temporary directory 
    #the stage will be parallelised by station reference/temporal resolution
    processed_final_references = get_unique_processed_references('T1', run_details, parameter_details)

    #processed_final_references = processed_final_references[:5]
    processed_final_references = [processed_final_references[0]]

    #check you have some data to process
    if len(processed_final_references) > 0:

        #create arguments list of each unique station reference and each unique temporal resolution code
        arguments = []
        for ref in processed_final_references:
            for temporal_resolution_code in run_details['temporal_resolutions']:
                arguments.append([ref, temporal_resolution_code])

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, arguments, GHOST_stage='T1')

        #submit greasy job/s to process station data in parallel
        job_name = 'T1_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time, valid_temporal_resolution_codes = submit_greasy_job('stage',job_name,run_details,GHOST_stage='T1',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time + valid temporal resolution codes
        return stage_process_time, stage_queue_time, valid_temporal_resolution_codes

    #if have zero station references then stage outcome == process code: 1 (i.e. no data)
    else:
        pipeline_logfile.write('T1 NOT PROCESSED AS HAVE NO DATA FOR PARAMETER.\nTERMINATING PIPELINE HERE')
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def T2(pipeline_start_time, slurm_job_ID, run_details, parameter_details, valid_temporal_resolution_codes, pipeline_logfile):

    '''GHOST PIPELINE STAGE: T2
      
       In this stage:
        -- Merge temporary station specific files to 1 netCDF by year/month 
        -- Outputs netCDF (.nc) files by year/month to /temporal_resolution/parameter/ directory
    '''

    import numpy as np

    #create list to store all relevant yearmonths + valid temporal resolution codes
    arguments = []

    #iterate through valid temporal resolutions
    for temporal_resolution_code in valid_temporal_resolution_codes:

        #get list of yearmonth strings for which in those time periods exist valid processed station data, of the specific temporal resolution
        yearmonth_strings = get_relevant_yearmonths(run_details, parameter_details, temporal_resolution_code)      

        #iterate through yearmonths
        for yearmonth_ii, yearmonth_str in enumerate(yearmonth_strings):
            #append yearmonth_str and temporal resolution code  
            arguments.append([yearmonth_str, temporal_resolution_code])

    #check you have some data to process
    if len(arguments) > 0:

        #create greasy arguments file/s to store all different configurations of jobs to process for GHOST stage
        create_greasy_arguments_file('stage', run_details, arguments, GHOST_stage='T2')

        #submit greasy job/s to process yearmonth data in parallel
        job_name = 'T2_{}_{}'.format(run_details['network'],run_details['parameter'])
        stage_process_time, stage_queue_time = submit_greasy_job('stage',job_name,run_details,GHOST_stage='T2',pipeline_start_time=pipeline_start_time,pipeline_logfile=pipeline_logfile)

        #return stage process time + stage queue time
        return stage_process_time, stage_queue_time

    #if have zero yearmonths with data then cannot process stage
    else:
        pipeline_logfile.write('T2 NOT PROCESSED AS HAVE NO VALID STATION DATA OF ANY DEFINED TEMPORAL RESOLUTION TO WRITE TO ANY MONTHLY NETCDF IN DESIRED TEMPORAL RANGE.\nTERMINATING PIPELINE HERE'.format(temporal_resolution_code))
        pipeline_logfile.close()
        sys.exit()

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###
