#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#globally_gridded_metadata_functions.py

#define functions which handle parsing of globally gridded metadata fields

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import bisect
import copy
from crossnetwork_processing_functions import weighted_mode, get_unique_longitude_latitude_pairs
import geopy
from geopy import distance
from GHOST_standards import standard_QA_name_to_QA_code
from netCDF4 import Dataset
import numpy as np
import numpy.ma as ma
import pandas as pd
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def get_globally_gridded_metadata(run_details, metadata, data, log_file_str):
        
    '''define serial list of functions that will be run to process each globally gridded dataset''' 
    
    #define list to append all names of metadata fields which are updated
    metadata_fields_to_update = []

    #get unique longitude and latitude pairs across metadata, with associated metadata/data indices of each pair
    unique_pair_dictionary = get_unique_longitude_latitude_pairs(metadata, data)

    #get UMBC Anthrome classifications
    metadata, log_file_str, metadata_fields_to_update = get_UMBC_Anthrome(run_details, metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update)

    #get ETOPO1 altitude
    metadata, data, log_file_str, metadata_fields_to_update = get_ETOPO1(run_details, metadata, data, log_file_str, unique_pair_dictionary, metadata_fields_to_update)

    #get GSFC coastline proximity
    metadata, log_file_str, metadata_fields_to_update = get_GSFC_coastline_proximity(run_details, metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update)

    #get WMO region
    metadata, log_file_str, metadata_fields_to_update = get_WMO_region(metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update)

    return metadata, data, log_file_str, metadata_fields_to_update

###--------------------------------------------------------------------------------------------------###
#DEFINE FUNCTIONS WHICH LOAD A SPECIFIC SET OF GLOBALLY GRIDDED METADATA AND RETURN VALUES FOR ALL 
#LONGITUDE/LATITUDE PAIRINGS BY STATION
###--------------------------------------------------------------------------------------------------###

def get_UMBC_Anthrome(run_details, metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update):

    '''parse globally gridded UMBC anthrome classifications'''

    log_file_str += 'GETTING UMBC ANTHROME METADATA\n'

    #define all distinct Anthrome classifications by code:string
    anthrome_definitions = {
    11: 'Urban',
    12: 'Mixed settlements',
    21: 'Rice villages',
    22: 'Irrigated villages',
    23: 'Rainfed villages',
    24: 'Pastoral villages',
    31: 'Residential irrigated croplands',
    32: 'Residential rainfed croplands',
    33: 'Populated croplands',
    34: 'Remote croplands',
    41: 'Residential rangelands',
    42: 'Populated rangelands',
    43: 'Remote rangelands',
    51: 'Residential woodlands',
    52: 'Populated woodlands',
    53: 'Remote woodlands',
    54: 'Inhabited treeless and barren lands',
    61: 'Wild woodlands',
    62: 'Wild treeless and barren lands',
    99: 'Ocean'
    }

    #load UMBC Anthrome grid
    UMBC_Anthrome_root = Dataset(run_details['data_root']+'globally_gridded_metadata/ANTHROME/anthro2_a2000.nc')

    #get grid centre longitudes
    grid_centre_lons = UMBC_Anthrome_root.variables['lon'][:]
    #get grid centre latitudes
    grid_centre_lats = UMBC_Anthrome_root.variables['lat'][:]
    #need to reverse latitude array (given as 90:-90)
    grid_centre_lats = grid_centre_lats[::-1]
    #read grid of UMBC classifications (indexed by: latitude, longitude)
    UMBC_Anthrome_classes = UMBC_Anthrome_root.variables['anthro2_a2000'][:]
    #need to reverse classification grid due to reversed latitudes (90:-90)
    UMBC_Anthrome_classes = UMBC_Anthrome_classes[::-1,:]

    #calculate grid edges
    grid_edge_lons = np.linspace(-180,180,len(grid_centre_lons)+1)
    grid_edge_lats = np.linspace(-90,90,len(grid_centre_lats)+1)

    #set relevant metadata fields for function
    relevant_metadata_fields = ['UMBC_anthrome_classification', 'UMBC_modal_anthrome_classification_5km', 'UMBC_modal_anthrome_classification_25km']
    metadata_fields_to_update.extend(relevant_metadata_fields)

    #define radiuses of metadata classifications that will be calculated
    radius_var_dict = {'UMBC_modal_anthrome_classification_5km':5.0, 'UMBC_modal_anthrome_classification_25km':25.0}

    #iterate through all unique longitude/latitude pairs
    for unique_pair_ii, unique_pair in unique_pair_dictionary.items():

        #get current unique longitude/latitude
        current_longitude = unique_pair['longitude']
        current_latitude = unique_pair['latitude']

        #get metadata indices associated with current unique longitude/latitude pair
        relevant_metadata_indices = unique_pair['metadata_indices']

        #get gridcell indices that longitude/latitude coordinates are contained in  
        lon_i, lat_i = get_coordinate_gridcell_indices(current_longitude,current_latitude,grid_centre_lons,grid_centre_lats)

        #get gridcell specific 'UMBC_anthrome_classification' code 
        gridcell_UMBC_Anthrome_class_code = UMBC_Anthrome_classes[lat_i,lon_i]

        #if gridcell code is masked, this is because it is the ocean. Force code to be 99.
        if ma.is_masked(gridcell_UMBC_Anthrome_class_code) == True:
            gridcell_UMBC_Anthrome_class_code = 99

        #a correction for costal sites is made:
        # -- if native anthrome class is ocean (i.e. code 99), then the minimum code of the surrounding gridcells is used instead (i.e. most urban)
        # -- if the site is truly an ocean site, all the surrounding gridcells will be ocean also, and therefore the class will be maintained as ocean
        if gridcell_UMBC_Anthrome_class_code == 99:

            #get neighbouring gridcell anthrome class codes (all grid cells which share a boundary with the gridcell, including diagonally)
            neighbouring_gridcell_lon_indices, neighbouring_gridcell_lat_indices =  get_neighbouring_gridcell_indices(lon_i,lat_i,len(grid_centre_lons),len(grid_centre_lats))
            neighbouring_UMBC_Anthrome_class_codes = []
            for neighbour_lon_i, neighbour_lat_i in zip(neighbouring_gridcell_lon_indices, neighbouring_gridcell_lat_indices):
                neighbouring_UMBC_Anthrome_class_code = UMBC_Anthrome_classes[neighbour_lat_i,neighbour_lon_i]
                #if gridcell code is masked, this is because it is the ocean. Force code to be 99.
                if ma.is_masked(neighbouring_UMBC_Anthrome_class_code) == True:
                    neighbouring_UMBC_Anthrome_class_code = 99
                neighbouring_UMBC_Anthrome_class_codes.append(neighbouring_UMBC_Anthrome_class_code)  

            #get minumum neighbouring_UMBC Anthrome class code - use this to overwrite original gridcell_UMBC_Anthrome_class_code
            gridcell_UMBC_Anthrome_class_code = np.min(neighbouring_UMBC_Anthrome_class_codes) 

        #set gridcell classification as string associated with code 
        metadata['UMBC_anthrome_classification'][relevant_metadata_indices] = anthrome_definitions[gridcell_UMBC_Anthrome_class_code]

        #this is the largest fraction of classifications from the combination of the gridcells fractions which intersect in a radius of 5km/25km around the station location
        for vv, radius_var in enumerate(list(radius_var_dict.keys())):

            #Get coordinate points 5km/25km in distance from initial coordinates, around 360 degrees (calculated geodesically)
            perimeter_points, perimeter_extreme_points = calculate_destination_points_perimeter(current_longitude,current_latitude,radius_var_dict[radius_var])

            #get relevant indices of model which intersect with perimeter points
            relevant_grid_indices = get_relevant_gridbox_indices(perimeter_points,perimeter_extreme_points,grid_centre_lons,grid_edge_lons,grid_centre_lats,grid_edge_lats)
            relevant_grid_class_codes = []
            relevant_grid_weights = []
            for box_indices in relevant_grid_indices:
                relevant_gridbox_class_code = UMBC_Anthrome_classes[box_indices[1],box_indices[0]]
                #modify gridcell class code to be 99 if masked (in ocean)
                if ma.is_masked(relevant_gridbox_class_code) == True:
                    relevant_gridbox_class_code = 99
                relevant_grid_class_codes.append(relevant_gridbox_class_code)
                relevant_grid_weights.append(box_indices[2])

            #get weighted mode of classifications from all gridcells intersecting with perimeter, using the fraction of gridcells inside perimeter as weights 
            #if there is a tie, the lowest code is kept preferentially (i.e. most urban)
            modal_value = weighted_mode(relevant_grid_class_codes, relevant_grid_weights)

            #set 'UMBC_modal_anthrome_classification_5km' and 'UMBC_modal_anthrome_classification_25km'
            metadata[radius_var][relevant_metadata_indices] = anthrome_definitions[modal_value]

    #return metadata 
    return metadata, log_file_str, metadata_fields_to_update
            
###------------------------------------------------------------------------------------###
        
def get_ETOPO1(run_details, metadata, data, log_file_str, unique_pair_dictionary, metadata_fields_to_update):

    '''parse globally gridded ETOPO1 altitudes'''

    log_file_str += 'GETTING ETOPO1 METADATA\n'

    #load ETOPO1 altitude grid
    ETOPO1_root = Dataset(run_details['data_root']+'globally_gridded_metadata/ETOPO1/original_files/ETOPO1_Ice_c_gmt4.grd')
    
    #get grid centre longitudes
    grid_centre_lons = ETOPO1_root['x'][:]
    #get grid centre latitudes
    grid_centre_lats = ETOPO1_root['y'][:]
    #read grid of altitudes (indexed by: latitude, longitude)
    ETOPO1_altitudes = ETOPO1_root['z'][:]

    #calculate grid edges
    grid_edge_lons = np.linspace(-180,180,len(grid_centre_lons)+1)
    grid_edge_lats = np.linspace(-90,90,len(grid_centre_lats)+1)

    #set relevant metadata fields for function
    relevant_metadata_fields = ['ETOPO1_altitude', 'ETOPO1_max_altitude_difference_5km', 'altitude', 'measurement_altitude']
    metadata_fields_to_update.extend(relevant_metadata_fields)

    #iterate through all unique longitude/latitude pairs
    for unique_pair_ii, unique_pair in unique_pair_dictionary.items():

        #get current unique longitude/latitude
        current_longitude = unique_pair['longitude']
        current_latitude = unique_pair['latitude']

        #get metadata indices associated with current unique longitude/latitude pair
        relevant_metadata_indices = unique_pair['metadata_indices']

        #get gridcell indices that longitude/latitude coordinates are contained in  
        lon_i, lat_i = get_coordinate_gridcell_indices(current_longitude,current_latitude,grid_centre_lons,grid_centre_lats)

        #get gridcell specific 'ETOPO1_altitude' 
        gridcell_ETOPO1_altitude = ETOPO1_altitudes[lat_i,lon_i]

        #a correction for coastal sites is made:
        # -- if the derived altitude is <= -5 m, the maximum altitude of the neighbouring grid cells will be used instead. 
        # -- if all neighbouring grid cells have altitudes <= -5 m, the original value will be retained.
        if gridcell_ETOPO1_altitude <= -5.0:
        
            #get neighbouring gridcell altitudes (all grid cells which share a boundary with the gridcell, including diagonally)
            neighbouring_gridcell_lon_indices, neighbouring_gridcell_lat_indices =  get_neighbouring_gridcell_indices(lon_i,lat_i,len(grid_centre_lons),len(grid_centre_lats))
            neighbouring_ETOPO1_altitudes = []
            for neighbour_lon_i, neighbour_lat_i in zip(neighbouring_gridcell_lon_indices, neighbouring_gridcell_lat_indices):
                neighbouring_ETOPO1_altitudes.append(ETOPO1_altitudes[neighbour_lat_i,neighbour_lon_i])                 

            #take maximum neighbouring ETOPO1 altitude
            max_neighbouring_ETOPO1_altitude = np.max(neighbouring_ETOPO1_altitudes)                 

            #if all neighbouring grid cells have altitudes <= -5 m, maintain original value
            if max_neighbouring_ETOPO1_altitude > -5.0:
                gridcell_ETOPO1_altitude = copy.deepcopy(max_neighbouring_ETOPO1_altitude)
        
        #set 'ETOPO1_altitude'
        metadata['ETOPO1_altitude'][relevant_metadata_indices] = gridcell_ETOPO1_altitude

        #calculate 'ETOPO1_max_altitude_difference_5km'
        #This is the altitude difference between the previously calcluated ETOPO1_altitude, and the minimum ETOP1 altitude (in metres) from all gridcells where more or equal to 50% of the cell intersects with a geodesic perimeter of 5 km around the station location.      
        
        #Get coordinate points 5km in distance from initial coordinates, around 360 degrees (calculated geodesically)
        perimeter_points, perimeter_extreme_points = calculate_destination_points_perimeter(current_longitude,current_latitude,5.0)

        #get relevant indices of model which intersect with perimeter points
        relevant_grid_indices = get_relevant_gridbox_indices(perimeter_points,perimeter_extreme_points,grid_centre_lons,grid_edge_lons,grid_centre_lats,grid_edge_lats)
        relevant_grid_altitudes = []
        for box_indices in relevant_grid_indices:
            #if gridcell intersection with perimeter is >= than 50%, then gridcell is relevant
            if box_indices[2] >= 0.5:
                relevant_grid_altitudes.append(ETOPO1_altitudes[box_indices[1],box_indices[0]])

        #calculate difference between minumum altitude in relevant gridcells and ETOPO1_altitude
        gridcell_ETOPO1_max_altitude_difference_5km = gridcell_ETOPO1_altitude - np.min(relevant_grid_altitudes)

        #set 'ETOPO1_max_altitude_difference_5km'
        metadata['ETOPO1_max_altitude_difference_5km'][relevant_metadata_indices] = gridcell_ETOPO1_max_altitude_difference_5km

        #if current altitudes == np.NaN, this is because there was no valid altitude metadata through time record. Overwrite with current ETOPO1 altitude estimation.
        #also if have no current measurement altitudes, set these. If sampling height is available = ETOPO1 altitude + sampling height; or if not solely ETOP01 altitude 
        #set appropriate QA flags to inform of changes
        for meta_ii in relevant_metadata_indices:

            current_altitude =             metadata['altitude'][meta_ii]
            current_sampling_height =      metadata['sampling_height'][meta_ii]
            current_measurement_altitude = metadata['measurement_altitude'][meta_ii]

            if pd.isnull(current_altitude) == True:
                metadata['altitude'][meta_ii] = gridcell_ETOPO1_altitude

                #get associated data indices of current meta index
                start_data_ind = bisect.bisect_left(data['time'], metadata['meta_update_stamp'][meta_ii])
                if (meta_ii + 1) == len(metadata['meta_update_stamp']):
                    end_data_ind = len(data['time'])
                else: 
                    end_data_ind = bisect.bisect_left(data['time'], metadata['meta_update_stamp'][meta_ii+1])
                relevant_data_indices = range(start_data_ind, end_data_ind)

                #set QA code to inform have estimated altitude with ETOPO1 altitude
                qa_flag_code = standard_QA_name_to_QA_code['No Altitude Metadata - Used ETOPO1 to Estimate Altitude']
                for inv in relevant_data_indices:
                    data['qa'][inv].append(qa_flag_code)

                #do not have valid measurement altitude?
                if pd.isnull(current_measurement_altitude) == True:
                    #have sampling height?
                    if pd.isnull(current_sampling_height) == False:
                        metadata['measurement_altitude'][meta_ii] = gridcell_ETOPO1_altitude + current_sampling_height
                        qa_flag_code = standard_QA_name_to_QA_code['No Measurement Altitude Metadata - Used ETOPO1 + Sampling Height to Estimate Altitude']
                    else:
                        metadata['measurement_altitude'][meta_ii] = gridcell_ETOPO1_altitude
                        qa_flag_code = standard_QA_name_to_QA_code['No Measurement Altitude Metadata - Used ETOPO1 to Estimate Altitude']
                    #set QA code to inform have estimated measurement altitude
                    for inv in relevant_data_indices:
                        data['qa'][inv].append(qa_flag_code)

    #return metadata/data
    return metadata, data, log_file_str, metadata_fields_to_update

###------------------------------------------------------------------------------------###

def get_GSFC_coastline_proximity(run_details, metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update):
        
    '''parse globally gridded GSFC coastline proximities'''

    log_file_str += 'GETTING GSFC COASTLINE PROXIMITY METADATA\n'

    #load GSFC coastline proximity grid (indexed by: latitude, longitude)
    GSFC_coastline_proximities = np.load(run_details['data_root']+'globally_gridded_metadata/GMT_COAST_DISTANCE/GMT_intermediate_coast_distance_01d.npy')
    #need to reverse grid due to reversed latitudes (90:-90)
    GSFC_coastline_proximities = GSFC_coastline_proximities[::-1,:]

    #get N lons/lats
    n_lons = GSFC_coastline_proximities.shape[1]
    n_lats = GSFC_coastline_proximities.shape[0]

    #calculate grid centres/edges
    grid_edge_lons = np.linspace(-180,180,n_lons+1)
    grid_edge_lats = np.linspace(-90,90,n_lats+1)
    grid_centre_lons = np.empty(len(grid_edge_lons)-1)
    grid_centre_lats = np.empty(len(grid_edge_lats)-1)
    for ii in range(len(grid_centre_lons)):
        left_edge = grid_edge_lons[ii]
        right_edge = grid_edge_lons[ii+1]
        grid_centre_lons[ii] = (left_edge+right_edge)/2.
    for ii in range(len(grid_centre_lats)):
        bottom_edge = grid_edge_lats[ii]
        top_edge = grid_edge_lats[ii+1]
        grid_centre_lats[ii] = (bottom_edge+top_edge)/2.

    #set relevant metadata fields for function
    relevant_metadata_fields = ['GSFC_coastline_proximity']
    metadata_fields_to_update.extend(relevant_metadata_fields)

    #iterate through all unique longitude/latitude pairs
    for unique_pair_ii, unique_pair in unique_pair_dictionary.items():

        #get current unique longitude/latitude
        current_longitude = unique_pair['longitude']
        current_latitude = unique_pair['latitude']

        #get metadata indices associated with current unique longitude/latitude pair
        relevant_metadata_indices = unique_pair['metadata_indices']

        #get gridcell indices that longitude/latitude coordinates are contained in  
        lon_i, lat_i = get_coordinate_gridcell_indices(current_longitude,current_latitude,grid_centre_lons,grid_centre_lats)

        #set gridcell specific 'GSFC_coastline_proximity' 
        metadata['GSFC_coastline_proximity'][relevant_metadata_indices] = GSFC_coastline_proximities[lat_i,lon_i]

    #return metadata 
    return metadata, log_file_str, metadata_fields_to_update

###------------------------------------------------------------------------------------###  

def get_WMO_region(metadata, log_file_str, unique_pair_dictionary, metadata_fields_to_update):

    '''determine which WMO region a lon/lat pairing is contained within (done using Shapely)'''

    log_file_str += 'GETTING WMO REGION METADATA\n'

    #define polygon boundaries for each WMO region
    #Region 1: Africa
    region1 = [[59.81270260311114, 13.140473047666644], [60.0, 13.0], [60.0, -5.0], [80.0, -5.0], [80.0, -50.0], [70.0, -50.0], [70.0, -60.0], [-20.0, -60.0], [-20.0, 5.0], [-35.0, 5.0], [-35.0, 36.0], [-2.0, 36.0], [10.0, 39.0], [13.0, 34.0], [31.960749852491567, 34.0], [32.71072653394139, 33.11565345693667], [33.49379554291177, 32.19184002943465], [34.27685121714466, 31.26802660193263], [34.29031264711197, 31.2312847333785], [34.267578125, 31.216541290283203], [34.287498474121094, 31.162776947021484], [34.40138244628906, 30.85944366455078], [34.48537111624103, 30.698888646891767], [34.504390893218066, 30.6469757262499], [34.55888366699219, 30.48611068725586], [34.54332733154297, 30.43471908569336], [34.54527282714844, 30.406944274902344], [34.60577636046175, 30.370252412321644], [34.873386555914244, 29.63983234187301], [34.87388610839844, 29.63083267211914], [34.86638641357422, 29.606664657592773], [34.874721999526784, 29.55238983177807], [34.536446900113035, 27.98189026696113], [35.043470949393395, 27.078260998379324], [35.55258274950386, 26.174631729797515], [36.06396288751517, 25.271002461215705], [36.58277128272936, 24.36737319263389], [37.10395913027517, 23.46374392405208], [37.625057649157604, 22.560114655470265], [38.14614029606225, 21.656485386888452], [38.66728002427425, 20.75285611830664], [39.18796605637755, 19.849226849724822], [39.70666260601094, 18.945597581143012], [40.22690153381618, 18.041968312561202], [40.74879978550911, 17.138339043979393], [41.272474407000246, 16.234709775397583], [41.7980426363959, 15.331080506815768], [42.3260551813884, 14.427451238233958], [42.85622942954716, 13.523821969652145], [43.38837612843018, 12.620192701070335], [44.879992378778105, 11.709576608246914], [45.87209710027676, 11.899618227618758], [46.86473847458583, 12.089659846990601], [47.85792634980978, 12.279701466362447], [48.851670592417655, 12.469743085734292], [49.845981087577265, 12.659784705106135], [50.85579496535577, 12.708513325457886], [51.865728435952114, 12.757241945809643], [52.87578215653159, 12.805970566161399], [53.885956784553755, 12.854699186513155], [54.89625297777378, 12.903427806864908], [55.90667139424395, 12.952156427216664], [56.91721269231514, 13.000885047568422], [57.927877530638256, 13.049613667920179], [58.93866656816571, 13.098342288271933], [59.81270260311114, 13.140473047666644], [52.34138488769531, 12.144721984863281], [52.38249969482422, 12.148887634277344], [52.39457702636719, 12.153055191040039], [52.39527130126953, 12.196527481079102], [52.362220764160156, 12.19972038269043], [52.291107177734375, 12.201944351196289], [52.27888488769531, 12.200553894042969], [52.234161376953125, 12.20111083984375], [52.220550537109375, 12.202775955200195], [52.20832824707031, 12.205554962158203], [52.19916534423828, 12.210832595825195], [52.18694305419922, 12.219720840454102], [52.17333221435547, 12.233610153198242], [52.164161682128906, 12.238609313964844], [52.13902282714844, 12.246527671813965], [52.09221649169922, 12.240276336669922], [52.08555221557617, 12.223540306091309], [52.103607177734375, 12.201944351196289], [52.11277770996094, 12.196109771728516], [52.129024505615234, 12.187359809875488], [52.17888641357422, 12.171943664550781], [52.268882751464844, 12.147777557373047], [52.34138488769531, 12.144721984863281], [54.22083282470703, 12.650554656982422], [54.18138885498047, 12.674165725708008], [54.146385192871094, 12.692221641540527], [54.12666320800781, 12.698888778686523], [54.0985107421875, 12.703158378601074], [54.0655517578125, 12.685555458068848], [54.00360870361328, 12.658609390258789], [53.821800231933594, 12.605484962463379], [53.766868591308594, 12.622220993041992], [53.71360778808594, 12.662221908569336], [53.6506233215332, 12.703680038452148], [53.499717712402344, 12.7172212600708], [53.40437316894531, 12.654998779296875], [53.40791320800781, 12.621387481689453], [53.40638732910156, 12.601387023925781], [53.393123626708984, 12.566596031188965], [53.375274658203125, 12.553333282470703], [53.33082962036133, 12.545970916748047], [53.37360382080078, 12.491388320922852], [53.57110595703125, 12.34638786315918], [53.59583282470703, 12.339166641235352], [53.6361083984375, 12.328887939453125], [53.755271911621094, 12.308609962463379], [53.79277038574219, 12.312776565551758], [54.01471710205078, 12.344165802001953], [54.1240234375, 12.348193168640137], [54.14430236816406, 12.35291576385498], [54.179443359375, 12.374443054199219], [54.222496032714844, 12.402498245239258], [54.413055419921875, 12.478333473205566], [54.47346878051758, 12.548748970031738], [54.44777297973633, 12.576249122619629], [54.25666427612305, 12.645971298217773], [54.22083282470703, 12.650554656982422], [31.987123713826193, 34.0], [32.0, 34.0], [32.09382805497271, 33.929628958770465], [31.987123713826193, 34.0]]
    #Region 2: Asia
    region2 = [[52.268882751464844, 12.147777557373047], [52.17888641357422, 12.171943664550781], [52.129024505615234, 12.187359809875488], [52.11277770996094, 12.196109771728516], [52.103607177734375, 12.201944351196289], [52.08555221557617, 12.223540306091309], [52.09221649169922, 12.240276336669922], [52.13902282714844, 12.246527671813965], [52.164161682128906, 12.238609313964844], [52.17333221435547, 12.233610153198242], [52.18694305419922, 12.219720840454102], [52.19916534423828, 12.210832595825195], [52.20832824707031, 12.205554962158203], [52.220550537109375, 12.202775955200195], [52.234161376953125, 12.20111083984375], [52.27888488769531, 12.200553894042969], [52.291107177734375, 12.201944351196289], [52.362220764160156, 12.19972038269043], [52.39527130126953, 12.196527481079102], [52.39457702636719, 12.153055191040039], [52.38249969482422, 12.148887634277344], [52.34138488769531, 12.144721984863281], [52.268882751464844, 12.147777557373047], [54.25666427612305, 12.645971298217773], [54.44777297973633, 12.576249122619629], [54.47346878051758, 12.548748970031738], [54.413055419921875, 12.478333473205566], [54.222496032714844, 12.402498245239258], [54.179443359375, 12.374443054199219], [54.14430236816406, 12.35291576385498], [54.1240234375, 12.348193168640137], [54.01471710205078, 12.344165802001953], [53.79277038574219, 12.312776565551758], [53.755271911621094, 12.308609962463379], [53.6361083984375, 12.328887939453125], [53.59583282470703, 12.339166641235352], [53.57110595703125, 12.34638786315918], [53.37360382080078, 12.491388320922852], [53.33082962036133, 12.545970916748047], [53.375274658203125, 12.553333282470703], [53.393123626708984, 12.566596031188965], [53.40638732910156, 12.601387023925781], [53.40791320800781, 12.621387481689453], [53.40437316894531, 12.654998779296875], [53.499717712402344, 12.7172212600708], [53.6506233215332, 12.703680038452148], [53.71360778808594, 12.662221908569336], [53.766868591308594, 12.622220993041992], [53.821800231933594, 12.605484962463379], [54.00360870361328, 12.658609390258789], [54.0655517578125, 12.685555458068848], [54.0985107421875, 12.703158378601074], [54.12666320800781, 12.698888778686523], [54.146385192871094, 12.692221641540527], [54.18138885498047, 12.674165725708008], [54.22083282470703, 12.650554656982422], [54.25666427612305, 12.645971298217773], [34.61209480513503, 30.35300673602901], [34.617800660577906, 30.35649364768854], [34.72334289550781, 30.089344024658203], [34.75360870361328, 29.991107940673828], [34.84777069091797, 29.740833282470703], [34.87277603149414, 29.65083122253418], [34.873386555914244, 29.63983234187301], [34.61209480513503, 30.35300673602901], [-172.509033203125, 70.23301696777344], [-169.0, 68.0], [-169.0, 65.5], [-180.0, 58.9523811340332], [-180.0, 75.0], [-172.509033203125, 70.23301696777344], [44.81637479260081, 39.645822862874155], [44.88859, 39.60582], [44.95458, 39.43235], [45.06304, 39.37305], [45.13055, 39.28555], [45.14353, 39.21686], [45.29839, 39.19762], [45.4211, 39.06082], [45.43348, 39.00319], [46.17825, 38.84115], [46.35109, 38.90998], [46.54038, 38.87559], [46.75895, 39.08596], [46.8747, 39.16749], [46.9529, 39.15637], [47.04346, 39.2379], [47.16457, 39.33528], [47.28694, 39.37443], [47.39166, 39.47054], [47.5425, 39.51527], [47.76985, 39.63374], [47.98471, 39.71558], [48.35788, 39.38991], [48.12388, 39.27818], [48.13332, 39.21096], [48.29388, 39.11249], [48.3254, 39.03221], [48.22471, 38.97137], [48.01207, 38.89707], [48.02464, 38.83339], [48.2469, 38.7154], [48.33166, 38.60284], [48.43277, 38.62457], [48.6629, 38.38915], [48.78374, 38.44554], [48.78374, 38.44554], [48.84117, 38.45048], [48.88828, 38.44241], [50.01484779661026, 38.43357894317056], [50.0, 42.45003578090057], [49.996748393599674, 42.90146458820869], [49.99861324203279, 45.09030899283844], [50.000651855064525, 46.63776117157768], [50.0, 69.08428955078125], [50.00000151773305, 69.0909004211618], [50.00006510586252, 69.3074858303206], [50.0, 69.31358017455813], [50.0, 69.324898990654], [50.0, 69.32868627512715], [50.0, 69.43576313614068], [40.0, 80.0], [40.0, 90.0], [172.5346221923828, 90.0], [180.0, 90.0], [180.0, 75.0], [180.0, 58.9523811340332], [173.41497802734375, 55.03273010253906], [170.0, 53.0], [176.0055389404297, 49.997230529785156], [180.0, 48.0], [180.0, 32.032], [180.0, 30.0], [180.0, 23.5], [125.0, 23.5], [110.0, 10.0], [101.740899624126, 6.53962165061756], [101.7451171875, 6.532638072967529], [101.76193237304688, 6.507222175598145], [101.78678131103516, 6.477082252502441], [101.8124771118164, 6.454166412353516], [102.01844787597656, 6.285138130187988], [102.04773712158203, 6.265832901000977], [102.09523010253906, 6.236138343811035], [102.09053039550781, 6.154165267944336], [102.01109313964844, 6.038331985473633], [101.98392486572266, 6.01367998123169], [101.95260620117188, 5.975554466247559], [101.9437255859375, 5.956526756286621], [101.94080352783203, 5.927498817443848], [101.94371032714844, 5.903332710266113], [101.94405364990234, 5.867081642150879], [101.82608795166016, 5.740970611572266], [101.80072021484375, 5.745971202850342], [101.7718505859375, 5.777499198913574], [101.74635314941406, 5.78916597366333], [101.69954681396484, 5.760971069335938], [101.69232177734375, 5.764305114746094], [101.65857696533203, 5.82173490524292], [101.65940856933594, 5.850624084472656], [101.60649108886719, 5.905276775360107], [101.57101440429688, 5.915207386016846], [101.50604248046875, 5.880571365356445], [101.45218658447266, 5.863054275512695], [101.31189727783203, 5.807221412658691], [101.2594223022461, 5.714582443237305], [101.22322845458984, 5.670970916748047], [101.18968963623047, 5.651110172271729], [101.16497802734375, 5.638332366943359], [101.14232635498047, 5.633471012115479], [101.05745697021484, 5.729166030883789], [100.99465942382812, 5.779623031616211], [100.99114990234375, 5.794304370880127], [101.04225158691406, 5.901804924011231], [101.06900024414062, 5.9145827293396], [101.09566497802734, 5.923957347869873], [101.11920166015625, 5.99027681350708], [101.10635375976562, 6.050276756286621], [101.09913635253906, 6.118887901306152], [101.11426544189453, 6.182463645935059], [101.11204528808594, 6.250693798065186], [101.00969696044922, 6.269443511962891], [100.95206451416016, 6.240831851959229], [100.85323333740234, 6.243887901306152], [100.84691619873047, 6.284998893737793], [100.84786987304688, 6.296963691711426], [100.85441589355469, 6.32249927520752], [100.82816314697266, 6.427776336669922], [100.7513656616211, 6.502777099609375], [100.7169189453125, 6.479999542236328], [100.68135070800781, 6.453332901000977], [100.6566390991211, 6.448678970336914], [100.43689727783203, 6.515555381774902], [100.41321836112911, 6.525404688778491], [100.38748, 6.53611], [100.3744, 6.54389], [100.33518, 6.58611], [100.33066, 6.62042], [100.30025, 6.69979], [100.26414, 6.70667], [100.21498, 6.71111], [100.20385, 6.70937], [100.19407, 6.69743], [100.17454, 6.62236], [100.18413, 6.5975], [100.19275, 6.57528], [100.18105658015898, 6.525053494529113], [100.16844177246094, 6.470832347869873], [100.14192199707031, 6.436665534973145], [100.12711334228516, 6.424946784973145], [100.1221923828125, 6.448887825012207], [100.09340366369773, 6.524920900783176], [100.0777475545591, 6.524897217576258], [100.05843353271484, 6.511631011962891], [100.04409210871654, 6.52484630653056], [99.69197550268896, 6.524313655003549], [99.66804504394531, 6.496387481689453], [99.65262603759766, 6.494790554046631], [99.63441629035232, 6.524226584408694], [90.05563792866475, 6.509736637516332], [90.0, 6.5], [90.0, -5.0], [60.0, -5.0], [60.0, 9.949312476527963], [60.0, 13.0], [59.81270260311114, 13.140473047666644], [58.93866656816571, 13.098342288271933], [57.927877530638256, 13.049613667920179], [56.91721269231514, 13.000885047568422], [55.90667139424395, 12.952156427216664], [54.89625297777378, 12.903427806864908], [53.885956784553755, 12.854699186513155], [52.87578215653159, 12.805970566161399], [51.865728435952114, 12.757241945809643], [50.85579496535577, 12.708513325457886], [49.845981087577265, 12.659784705106135], [48.851670592417655, 12.469743085734292], [47.85792634980978, 12.279701466362447], [46.86473847458583, 12.089659846990601], [45.87209710027676, 11.899618227618758], [44.879992378778105, 11.709576608246914], [43.38837612843018, 12.620192701070335], [42.85622942954716, 13.523821969652145], [42.3260551813884, 14.427451238233958], [41.7980426363959, 15.331080506815768], [41.272474407000246, 16.234709775397583], [40.74879978550911, 17.138339043979393], [40.22690153381618, 18.041968312561202], [39.70666260601094, 18.945597581143012], [39.18796605637755, 19.849226849724822], [38.66728002427425, 20.75285611830664], [38.14614029606225, 21.656485386888452], [37.625057649157604, 22.560114655470265], [37.10395913027517, 23.46374392405208], [36.58277128272936, 24.36737319263389], [36.06396288751517, 25.271002461215705], [35.55258274950386, 26.174631729797515], [35.043470949393395, 27.078260998379324], [34.536446900113035, 27.98189026696113], [34.874721999526784, 29.55238983177807], [34.877777099609375, 29.53249740600586], [34.90380096435547, 29.486705780029297], [34.93804931640625, 29.523330688476562], [34.962562561035156, 29.55146026611328], [34.96582794189453, 29.549163818359375], [34.97998046875, 29.545753479003906], [35.00367736816406, 29.5281925201416], [34.97846984863281, 29.477497100830078], [34.966941833496094, 29.44888687133789], [34.96041488647461, 29.37742805480957], [34.961387634277344, 29.36083221435547], [35.187217712402344, 29.326942443847656], [35.541107177734375, 29.272499084472656], [36.06999969482422, 29.188888549804688], [36.272727966308594, 29.336528778076172], [36.50054931640625, 29.502777099609375], [36.74360656738281, 29.86471939086914], [37.03443908691406, 29.918052673339844], [37.41009521484375, 29.985580444335938], [37.502777099609375, 30.002220153808594], [37.66749572753906, 30.336387634277344], [38.00138854980469, 30.504165649414062], [37.56108856201172, 30.951087951660156], [37.48999786376953, 31.023052215576172], [37.29499053955078, 31.21722412109375], [37.005271911621094, 31.50555419921875], [37.20220184326172, 31.55615997314453], [37.47193908691406, 31.624996185302734], [37.709495544433594, 31.68466567993164], [37.96388244628906, 31.748332977294922], [37.986663818359375, 31.754051208496094], [38.48832702636719, 31.879165649414062], [39.00499725341797, 32.00555419921875], [39.07221984863281, 32.057777404785156], [39.19674301147461, 32.15494155883789], [39.301109313964844, 32.23638153076172], [39.2599983215332, 32.35554122924805], [39.04365158081055, 32.30405044555664], [38.986019134521484, 32.4776611328125], [39.08600997924805, 32.50151062011719], [38.794700622558594, 33.377593994140625], [38.810272216796875, 33.38221740722656], [38.961387634277344, 33.45555114746094], [39.3861083984375, 33.65888214111328], [39.579139709472656, 33.75], [39.79804992675781, 33.85333251953125], [40.0, 33.947654724121094], [40.16554260253906, 34.024436950683594], [40.199432373046875, 34.04082489013672], [40.663604736328125, 34.260276794433594], [41.003875732421875, 34.41943359375], [41.030540466308594, 34.473609924316406], [41.09943389892578, 34.60859680175781], [41.224849700927734, 34.78290557861328], [41.223052978515625, 34.91693878173828], [41.2147216796875, 35.06915283203125], [41.212493896484375, 35.11277770996094], [41.21124267578125, 35.1885986328125], [41.21240997314453, 35.20062255859375], [41.21527099609375, 35.224708557128906], [41.22248840332031, 35.25360870361328], [41.240962982177734, 35.3065185546875], [41.254432678222656, 35.34332275390625], [41.26221466064453, 35.367774963378906], [41.269161224365234, 35.39874267578125], [41.27082061767578, 35.443328857421875], [41.272491455078125, 35.46832275390625], [41.27777099609375, 35.49554443359375], [41.32249450683594, 35.550559997558594], [41.3840217590332, 35.63096618652344], [41.37833023071289, 35.83694076538086], [41.35693359375, 35.876380920410156], [41.27110290527344, 36.02777099609375], [41.25498962402344, 36.05499267578125], [41.29027557373047, 36.355552673339844], [41.367210388183594, 36.47277069091797], [41.403045654296875, 36.525550842285156], [41.83527374267578, 36.59888458251953], [42.367286682128906, 37.05588912963867], [42.37409973144531, 37.06316375732422], [42.37833023071289, 37.07206344604492], [42.355613708496094, 37.10692596435547], [42.37693786621094, 37.11749267578125], [42.53166198730469, 37.147491455078125], [42.579505920410156, 37.1536750793457], [42.588043212890625, 37.17568588256836], [42.724708557128906, 37.35194396972656], [42.76573181152344, 37.3712158203125], [42.786800384521484, 37.38367462158203], [42.822837829589844, 37.36717987060547], [42.894439697265625, 37.335540771484375], [42.924163818359375, 37.32776641845703], [42.96061706542969, 37.32353210449219], [43.00276565551758, 37.33915710449219], [43.15693664550781, 37.373878479003906], [43.489990234375, 37.251243591308594], [43.623046875, 37.22999572753906], [43.816375732421875, 37.22193145751953], [43.91297912597656, 37.22658920288086], [43.93693542480469, 37.25193786621094], [43.9558219909668, 37.279640197753906], [44.00901412963867, 37.31776809692383], [44.04804992675781, 37.32221984863281], [44.07294464111328, 37.316375732421875], [44.096099853515625, 37.31721496582031], [44.11888122558594, 37.3156852722168], [44.230552673339844, 37.276241302490234], [44.26214599609375, 37.23651885986328], [44.26832580566406, 37.19499969482422], [44.262596130371094, 37.16346740722656], [44.23026657104492, 37.15401840209961], [44.19596481323242, 37.096378326416016], [44.257076263427734, 36.985687255859375], [44.31916046142578, 36.97123718261719], [44.341102600097656, 37.00665283203125], [44.349708557128906, 37.03832244873047], [44.436378479003906, 37.06304168701172], [44.47027587890625, 37.08082580566406], [44.49055480957031, 37.091651916503906], [44.55054473876953, 37.132633209228516], [44.567630767822266, 37.14971160888672], [44.582767486572266, 37.163185119628906], [44.642215728759766, 37.18846130371094], [44.66767433250223, 37.185013833273], [44.85003082442806, 37.11554469349173], [44.85003082442806, 37.20309818114609], [44.80803768299948, 37.22653621357134], [44.8205451965332, 37.26874542236328], [44.818180084228516, 37.29741668701172], [44.801658630371094, 37.32166290283203], [44.648597717285156, 37.427772521972656], [44.61280059814453, 37.43422317504883], [44.58885192871094, 37.443092346191406], [44.584434509277344, 37.61388397216797], [44.598876953125, 37.690818786621094], [44.617767333984375, 37.71797561645508], [44.5767936706543, 37.76679611206055], [44.293052673339844, 37.873321533203125], [44.223968505859375, 37.89915084838867], [44.23888397216797, 37.96027374267578], [44.32331848144531, 38.09721374511719], [44.399017333984375, 38.211299896240234], [44.47322082519531, 38.316749572753906], [44.48251724243164, 38.34130096435547], [44.43672561645508, 38.39480972290039], [44.318145751953125, 38.38205337524414], [44.305259704589844, 38.400535583496094], [44.304161071777344, 38.44943618774414], [44.311767578125, 38.46992492675781], [44.32152557373047, 38.509368896484375], [44.31721496582031, 38.61304473876953], [44.306514739990234, 38.647281646728516], [44.26179504394531, 38.72429275512695], [44.30318832397461, 38.80998611450195], [44.30026626586914, 38.842628479003906], [44.28366470336914, 38.862491607666016], [44.25054931640625, 38.86554718017578], [44.20943069458008, 38.893463134765625], [44.17332458496094, 38.968040466308594], [44.16241455078125, 39.00519943237305], [44.19803237915039, 39.05667495727539], [44.21644973754883, 39.13651657104492], [44.07471466064453, 39.336097717285156], [44.060546875, 39.35248565673828], [44.03495407104492, 39.377445220947266], [44.036380767822266, 39.385475158691406], [44.07686233520508, 39.41165542602539], [44.200958251953125, 39.41693878173828], [44.230918884277344, 39.40845489501953], [44.30452346801758, 39.387290954589844], [44.34027099609375, 39.395545959472656], [44.401100158691406, 39.4165153503418], [44.42582321166992, 39.43644332885742], [44.47789001464844, 39.640655517578125], [44.47108459472656, 39.69886779785156], [44.608463287353516, 39.77915573120117], [44.736717224121094, 39.70180130004883], [44.81304168701172, 39.63081359863281], [44.816375732421875, 39.64582061767578], [44.81637479260081, 39.645822862874155]]
    #Region 3: South America
    region3 = [[-81.71305847167969, 12.490276336669922], [-81.72013854980469, 12.496317863464355], [-81.72014617919922, 12.545276641845703], [-81.71389770507812, 12.564998626708984], [-81.70472717285156, 12.580275535583496], [-81.69236755371094, 12.590276718139648], [-81.68132019042969, 12.585762977600098], [-81.69097900390625, 12.532220840454102], [-81.7025146484375, 12.507221221923828], [-81.71028137207031, 12.490833282470703], [-81.71305847167969, 12.490276336669922], [-64.05555725097656, 10.857221603393555], [-64.05862426757812, 10.864999771118164], [-64.13639831542969, 10.944999694824219], [-64.15084838867188, 10.952220916748047], [-64.17403411865234, 10.960415840148926], [-64.19014739990234, 10.946943283081055], [-64.20112609863281, 10.936943054199219], [-64.2279281616211, 10.931249618530273], [-64.35792541503906, 10.954581260681152], [-64.40570068359375, 10.970137596130371], [-64.37834167480469, 11.0569429397583], [-64.21833801269531, 11.088054656982422], [-64.20181274414062, 11.087777137756348], [-64.19277954101562, 11.082915306091309], [-64.19098663330078, 11.072638511657715], [-64.19743347167969, 11.055901527404785], [-64.18806457519531, 11.040276527404785], [-64.17584228515625, 11.031665802001953], [-64.1058349609375, 10.995277404785156], [-64.0836181640625, 10.988887786865234], [-64.05445098876953, 10.985692977905273], [-64.04278564453125, 10.987637519836426], [-64.02249908447266, 10.999998092651367], [-63.98750305175781, 11.07611083984375], [-63.951393127441406, 11.115554809570312], [-63.88493347167969, 11.175622940063477], [-63.84486770629883, 11.127082824707031], [-63.804725646972656, 11.021387100219727], [-63.81500244140625, 10.97805404663086], [-63.89000701904297, 10.904443740844727], [-63.97236251831055, 10.892915725708008], [-64.05555725097656, 10.857221603393555], [-65.28111267089844, 10.880277633666992], [-65.29972839355469, 10.883054733276367], [-65.39334106445312, 10.906944274902344], [-65.41570281982422, 10.916248321533203], [-65.41612243652344, 10.927221298217773], [-65.38667297363281, 10.9566650390625], [-65.3739013671875, 10.96444320678711], [-65.36445617675781, 10.969165802001953], [-65.30445098876953, 10.976109504699707], [-65.212158203125, 10.9541654586792], [-65.20083618164062, 10.911664962768555], [-65.19959259033203, 10.898332595825195], [-65.21000671386719, 10.891387939453125], [-65.26861572265625, 10.881109237670898], [-65.28111267089844, 10.880277633666992], [-62.38009193870275, 9.894418905915902], [-62.249128285596655, 9.930797698445373], [-62.25048828125, 9.967012405395508], [-62.225438860949566, 9.937378094180676], [-62.20647226064378, 9.942646594265616], [-62.23173904418945, 9.964720726013184], [-62.2092399597168, 10.011039733886719], [-62.17889404296875, 10.014721870422363], [-62.0755615234375, 9.986387252807617], [-62.06483960144875, 9.981988999597569], [-62.0, 10.0], [-35.0, 10.0], [-35.0, 5.0], [-20.0, 5.0], [-20.0, -60.0], [-120.0, -60.0], [-120.0, 5.0], [-80.0, 5.0], [-77.88972473144531, 7.228888511657715], [-77.80750274658203, 7.47868013381958], [-77.77570343017578, 7.47541618347168], [-77.74861907958984, 7.484305381774902], [-77.73222351074219, 7.505902290344238], [-77.72097778320312, 7.536457538604736], [-77.72937774658203, 7.568888187408447], [-77.75459289550781, 7.612221717834473], [-77.75917053222656, 7.6333327293396], [-77.75944519042969, 7.667222023010254], [-77.75889587402344, 7.693055152893066], [-77.74403381347656, 7.719999313354492], [-77.66389465332031, 7.679443836212158], [-77.62417602539062, 7.60333251953125], [-77.61029052734375, 7.562499046325684], [-77.60014343261719, 7.539235591888428], [-77.57743835449219, 7.526180267333984], [-77.33097839355469, 7.701805114746094], [-77.330078125, 7.726176261901856], [-77.37000274658203, 7.778471946716309], [-77.3114013671875, 7.886943817138672], [-77.29521179199219, 7.904860496520996], [-77.21556091308594, 7.937221527099609], [-77.19833374023438, 7.999444007873535], [-77.21542358398438, 8.087915420532227], [-77.243896484375, 8.14527702331543], [-77.27361297607422, 8.198610305786133], [-77.29611206054688, 8.215831756591797], [-77.34889221191406, 8.267776489257812], [-77.36250305175781, 8.284998893737793], [-77.36805725097656, 8.337221145629883], [-77.36862182617188, 8.364166259765625], [-77.37500762939453, 8.39861011505127], [-77.40576934814453, 8.45131778717041], [-77.4291000366211, 8.472498893737793], [-77.46858215332031, 8.471699714660645], [-77.47576904296875, 8.521110534667969], [-77.45223236083984, 8.556943893432617], [-77.43861389160156, 8.566665649414062], [-77.3719482421875, 8.646110534667969], [-77.36666870117188, 8.674999237060547], [-77.34695434570312, 8.662221908569336], [-77.30029296875, 8.578332901000977], [-77.28466033935547, 8.516526222229004], [-77.24097442626953, 8.464720726013184], [-77.22000122070312, 8.45111083984375], [-77.19944763183594, 8.443609237670898], [-77.16389465332031, 8.434165954589844], [-77.134521484375, 8.410207748413086], [-77.118896484375, 8.386388778686523], [-77.08168029785156, 8.32611083984375], [-76.96861267089844, 8.238332748413086], [-76.95556640625, 8.189998626708984], [-76.94862365722656, 8.15916633605957], [-76.88778686523438, 8.121665954589844], [-76.83857727050781, 8.133211135864258], [-76.8341064453125, 8.129096031188965], [-76.83250427246094, 8.086387634277344], [-76.83667755126953, 8.026456832885742], [-76.87055969238281, 8.034651756286621], [-76.8929214477539, 8.04666519165039], [-76.916748046875, 8.031526565551758], [-76.92750549316406, 7.985554695129395], [-76.92388916015625, 7.936040878295898], [-76.8719482421875, 7.909443855285645], [-76.81695556640625, 7.904999732971191], [-76.75791931152344, 7.919166088104248], [-76.74445343017578, 7.935832977294922], [-76.73194885253906, 8.039165496826172], [-76.73896026611328, 8.09062385559082], [-76.75334167480469, 8.197498321533203], [-76.76528930664062, 8.340693473815918], [-76.7738265991211, 8.411249160766602], [-76.82181549072266, 8.482220649719238], [-76.87667083740234, 8.521388053894043], [-76.9281997680664, 8.53722095489502], [-76.9283447265625, 8.56833267211914], [-76.88472747802734, 8.622220993041992], [-76.81806945800781, 8.633888244628906], [-76.67389678955078, 8.676595687866211], [-76.43243408203125, 8.876596450805664], [-76.4267578125, 8.906387329101562], [-76.41889953613281, 8.906665802001953], [-76.39889526367188, 8.910554885864258], [-76.37361145019531, 8.91611099243164], [-76.31751251220703, 8.938610076904297], [-76.26417541503906, 8.99527645111084], [-76.18055725097656, 9.191944122314453], [-76.17264556884766, 9.235832214355469], [-76.11056518554688, 9.315553665161133], [-76.08973693847656, 9.335832595825195], [-76.07139587402344, 9.34555435180664], [-75.91862487792969, 9.422222137451172], [-75.81604766845703, 9.438610076904297], [-75.7252197265625, 9.421241760253906], [-75.68791961669922, 9.41138744354248], [-75.65667724609375, 9.425832748413086], [-75.63431549072266, 9.448193550109863], [-75.62001037597656, 9.472221374511719], [-75.59806823730469, 9.52277660369873], [-75.59111785888672, 9.546943664550781], [-75.58555603027344, 9.58305549621582], [-75.58528137207031, 9.606388092041016], [-75.59140014648438, 9.634721755981445], [-75.60139465332031, 9.66555404663086], [-75.6268081665039, 9.698887825012207], [-75.66084289550781, 9.760276794433594], [-75.63528442382812, 9.82472038269043], [-75.59611511230469, 9.957498550415039], [-75.59028625488281, 9.979442596435547], [-75.58014678955078, 10.045692443847656], [-75.58000183105469, 10.071943283081055], [-75.58251953125, 10.08863353729248], [-75.57167053222656, 10.134443283081055], [-75.5340347290039, 10.18235969543457], [-75.53167724609375, 10.236944198608398], [-75.54707336425781, 10.41789436340332], [-75.48529052734375, 10.446664810180664], [-75.4866714477539, 10.490971565246582], [-75.50403594970703, 10.550138473510742], [-75.47611999511719, 10.592777252197266], [-75.40888977050781, 10.658609390258789], [-75.28807067871094, 10.724443435668945], [-75.26139831542969, 10.72138786315918], [-75.25184631347656, 10.70645523071289], [-75.22660827636719, 10.729165077209473], [-75.23562622070312, 10.762012481689453], [-75.25083923339844, 10.746109008789062], [-75.26819610595703, 10.75305461883545], [-75.2641372680664, 10.798992156982422], [-75.24402618408203, 10.800321578979492], [-75.2227783203125, 10.80555534362793], [-75.0394515991211, 10.905832290649414], [-75.03176879882812, 10.936005592346191], [-75.02305603027344, 10.966526985168457], [-74.92111206054688, 11.068887710571289], [-74.86805725097656, 11.119720458984375], [-74.86080932617188, 11.125486373901367], [-74.83458709716797, 11.105554580688477], [-74.5050048828125, 10.993331909179688], [-74.47611999511719, 10.987220764160156], [-74.45083618164062, 10.983888626098633], [-74.42361450195312, 10.983055114746094], [-74.37973022460938, 10.983055114746094], [-74.3416748046875, 10.986665725708008], [-74.31597900390625, 10.991804122924805], [-74.29222869873047, 10.99902629852295], [-74.30487060546875, 10.983471870422363], [-74.32722473144531, 10.974860191345215], [-74.354736328125, 10.971109390258789], [-74.41902923583984, 10.965831756591797], [-74.46528625488281, 10.971318244934082], [-74.59174346923828, 10.87763786315918], [-74.60750579833984, 10.823609352111816], [-74.59330749511719, 10.783297538757324], [-74.51918029785156, 10.812776565551758], [-74.46528625488281, 10.765971183776855], [-74.45777893066406, 10.742499351501465], [-74.39500427246094, 10.742152214050293], [-74.37541961669922, 10.755414962768555], [-74.35826873779297, 10.783401489257812], [-74.33222961425781, 10.863609313964844], [-74.29055786132812, 10.967498779296875], [-74.26361846923828, 11.016942977905273], [-74.22987365722656, 11.118610382080078], [-74.2388916015625, 11.19666576385498], [-74.23695373535156, 11.23499870300293], [-74.21556091308594, 11.2691650390625], [-74.18890380859375, 11.309720993041992], [-74.15501403808594, 11.331388473510742], [-74.0372314453125, 11.356388092041016], [-74.01278686523438, 11.355276107788086], [-73.96958923339844, 11.33902645111084], [-73.95125579833984, 11.319165229797363], [-73.86695861816406, 11.284721374511719], [-73.79084014892578, 11.265693664550781], [-73.67250061035156, 11.265275955200195], [-73.63139343261719, 11.271248817443848], [-73.5947265625, 11.273054122924805], [-73.55723571777344, 11.273887634277344], [-73.53724670410156, 11.271329879760742], [-73.50889587402344, 11.269721984863281], [-73.38431549072266, 11.273887634277344], [-73.34500122070312, 11.279443740844727], [-73.31695556640625, 11.285831451416016], [-73.28445434570312, 11.295555114746094], [-73.26750183105469, 11.306388854980469], [-73.20626068115234, 11.34902572631836], [-73.19584655761719, 11.376526832580566], [-73.14306640625, 11.424444198608398], [-73.03472900390625, 11.502777099609375], [-72.76876068115234, 11.687776565551758], [-72.73090362548828, 11.706805229187012], [-72.51764678955078, 11.78749942779541], [-72.48736572265625, 11.784095764160156], [-72.45750427246094, 11.78499984741211], [-72.41625213623047, 11.79777717590332], [-72.27972412109375, 11.875831604003906], [-72.25862121582031, 11.889165878295898], [-72.23271179199219, 11.915207862854004], [-72.13871765136719, 12.106388092041016], [-72.13890075683594, 12.126943588256836], [-72.14583587646484, 12.207289695739746], [-72.16885375976562, 12.221561431884766], [-72.15792083740234, 12.241108894348145], [-72.13862609863281, 12.249443054199219], [-72.02862548828125, 12.257776260375977], [-71.96475982666016, 12.251977920532227], [-71.98251342773438, 12.225692749023438], [-72.01181030273438, 12.192638397216797], [-71.98604583740234, 12.157081604003906], [-71.96305847167969, 12.153610229492188], [-71.93778991699219, 12.162776947021484], [-71.86424255371094, 12.210762977600098], [-71.80972290039062, 12.313331604003906], [-71.69292449951172, 12.366388320922852], [-71.63139343261719, 12.421595573425293], [-71.65501403808594, 12.438888549804688], [-71.70438385009766, 12.434477806091309], [-71.73653411865234, 12.414165496826172], [-71.72431182861328, 12.440971374511719], [-71.69084167480469, 12.459304809570312], [-71.66153717041016, 12.463888168334961], [-71.56361389160156, 12.453609466552734], [-71.52862548828125, 12.446109771728516], [-71.27570343017578, 12.346110343933105], [-71.24305725097656, 12.321664810180664], [-71.22084045410156, 12.302082061767578], [-71.1148681640625, 12.098609924316406], [-71.11083984375, 12.075554847717285], [-71.11418151855469, 12.048887252807617], [-71.13125610351562, 12.017359733581543], [-71.17362213134766, 11.984443664550781], [-71.23194885253906, 11.951387405395508], [-71.28250122070312, 11.918609619140625], [-71.3194580078125, 11.865276336669922], [-71.32472229003906, 11.85305404663086], [-71.34112548828125, 11.799165725708008], [-71.37861633300781, 11.753332138061523], [-71.44584655761719, 11.723888397216797], [-71.48056030273438, 11.710277557373047], [-71.5050048828125, 11.707221984863281], [-71.54139709472656, 11.703887939453125], [-71.63833618164062, 11.690832138061523], [-71.71223449707031, 11.674999237060547], [-71.76583862304688, 11.662360191345215], [-71.81639099121094, 11.647499084472656], [-71.95417022705078, 11.594720840454102], [-71.96916961669922, 11.546248435974121], [-71.96640014648438, 11.506387710571289], [-71.96084594726562, 11.46444320678711], [-71.95264434814453, 11.421804428100586], [-71.94486999511719, 11.395554542541504], [-71.93278503417969, 11.363054275512695], [-71.90611267089844, 11.300832748413086], [-71.86695861816406, 11.229442596435547], [-71.83778381347656, 11.182777404785156], [-71.7952880859375, 11.128610610961914], [-71.7408447265625, 11.03416633605957], [-71.76361846923828, 11.013749122619629], [-71.69111633300781, 10.834165573120117], [-71.66181182861328, 10.777498245239258], [-71.61750793457031, 10.741943359375], [-71.59695434570312, 10.727776527404785], [-71.57792663574219, 10.716110229492188], [-71.58168029785156, 10.674720764160156], [-71.6300048828125, 10.476387023925781], [-71.64889526367188, 10.442499160766602], [-71.67778015136719, 10.423332214355469], [-71.75569915771484, 10.369721412658691], [-71.77320098876953, 10.349305152893066], [-71.79444885253906, 10.321805000305176], [-71.82514953613281, 10.255138397216797], [-71.84112548828125, 10.218887329101562], [-71.9112548828125, 10.124443054199219], [-72.12535095214844, 9.818193435668945], [-72.07792663574219, 9.735554695129395], [-72.04653930664062, 9.70777702331543], [-72.00806427001953, 9.663471221923828], [-71.97917175292969, 9.623332977294922], [-71.96583557128906, 9.602499008178711], [-71.9444580078125, 9.543054580688477], [-71.91084289550781, 9.493610382080078], [-71.73320007324219, 9.375693321228027], [-71.72361755371094, 9.344999313354492], [-71.7319564819336, 9.295970916748047], [-71.751953125, 9.22249984741211], [-71.75945281982422, 9.118331909179688], [-71.73417663574219, 9.100555419921875], [-71.7148666381836, 9.077916145324707], [-71.68917083740234, 9.063401222229004], [-71.62389373779297, 9.043054580688477], [-71.59834289550781, 9.039999008178711], [-71.55278015136719, 9.04083251953125], [-71.49723052978516, 9.050415992736816], [-71.31361389160156, 9.110553741455078], [-71.29055786132812, 9.121665954589844], [-71.24139404296875, 9.155415534973145], [-71.22508239746094, 9.174200057983398], [-71.21417236328125, 9.208887100219727], [-71.20834350585938, 9.219720840454102], [-71.16542053222656, 9.273748397827148], [-71.14723205566406, 9.28680419921875], [-71.11778259277344, 9.293333053588867], [-71.07486724853516, 9.312499046325684], [-71.05597686767578, 9.338749885559082], [-71.05555725097656, 9.354442596435547], [-71.07112121582031, 9.390276908874512], [-71.08917236328125, 9.536596298217773], [-71.08306884765625, 9.566665649414062], [-71.06111145019531, 9.61722183227539], [-71.0533447265625, 9.703609466552734], [-71.07340240478516, 9.851110458374023], [-71.09834289550781, 9.894790649414062], [-71.14778137207031, 9.95083236694336], [-71.26945495605469, 10.150137901306152], [-71.29264831542969, 10.175415992736816], [-71.31695556640625, 10.192777633666992], [-71.38555908203125, 10.28277587890625], [-71.43667602539062, 10.369443893432617], [-71.4586181640625, 10.459999084472656], [-71.52195739746094, 10.533332824707031], [-71.54570770263672, 10.568331718444824], [-71.52639770507812, 10.726943969726562], [-71.44667053222656, 10.795555114746094], [-71.45167541503906, 10.916387557983398], [-71.49285888671875, 10.961040496826172], [-71.43473052978516, 10.979581832885742], [-71.41195678710938, 10.98416519165039], [-71.38417053222656, 10.983332633972168], [-71.35389709472656, 10.978333473205566], [-71.28111267089844, 10.989166259765625], [-71.2711181640625, 10.99305534362793], [-71.17807006835938, 11.032220840454102], [-71.07389831542969, 11.079999923706055], [-71.03153228759766, 11.103193283081055], [-71.01445007324219, 11.11722183227539], [-70.9890365600586, 11.133332252502441], [-70.93389892578125, 11.164721488952637], [-70.8900146484375, 11.18610954284668], [-70.8558349609375, 11.200277328491211], [-70.8255615234375, 11.211387634277344], [-70.73695373535156, 11.229999542236328], [-70.65501403808594, 11.238609313964844], [-70.62139892578125, 11.236248970031738], [-70.54681396484375, 11.241943359375], [-70.50847625732422, 11.248193740844727], [-70.49209594726562, 11.26291561126709], [-70.44667053222656, 11.289999008178711], [-70.42140197753906, 11.292221069335938], [-70.32333374023438, 11.3316650390625], [-70.23918151855469, 11.35305404663086], [-70.14250183105469, 11.418333053588867], [-70.09556579589844, 11.429998397827148], [-70.03486633300781, 11.441317558288574], [-70.02389526367188, 11.491388320922852], [-70.04779052734375, 11.517776489257812], [-69.9876480102539, 11.514720916748047], [-69.89771270751953, 11.436179161071777], [-69.87403106689453, 11.426526069641113], [-69.836669921875, 11.424582481384277], [-69.8013916015625, 11.427221298217773], [-69.74240112304688, 11.499199867248535], [-69.77611541748047, 11.609442710876465], [-69.78778076171875, 11.635276794433594], [-69.81653594970703, 11.690971374511719], [-69.85736846923828, 11.689582824707031], [-70.00473022460938, 11.650276184082031], [-70.18000793457031, 11.603610038757324], [-70.21403503417969, 11.60916519165039], [-70.23597717285156, 11.628888130187988], [-70.22653198242188, 11.655068397521973], [-70.243896484375, 11.775276184082031], [-70.29403686523438, 11.861665725708008], [-70.29376220703125, 11.899443626403809], [-70.28666687011719, 11.920276641845703], [-70.25806427001953, 11.989165306091309], [-70.20639038085938, 12.080275535583496], [-70.1875, 12.107776641845703], [-70.06806945800781, 12.17361068725586], [-70.01431274414062, 12.197498321533203], [-69.93472290039062, 12.169721603393555], [-69.90570068359375, 12.139999389648438], [-69.85945129394531, 12.071943283081055], [-69.84056091308594, 12.032776832580566], [-69.8237533569336, 11.988055229187012], [-69.81663513183594, 11.916109085083008], [-69.81889343261719, 11.881109237670898], [-69.81687927246094, 11.850971221923828], [-69.80416870117188, 11.794166564941406], [-69.77084350585938, 11.696109771728516], [-69.75778198242188, 11.661388397216797], [-69.71916961669922, 11.570276260375977], [-69.70611572265625, 11.548887252807617], [-69.67472839355469, 11.511110305786133], [-69.6318130493164, 11.46763801574707], [-69.60848236083984, 11.459165573120117], [-69.57952117919922, 11.464096069335938], [-69.5555648803711, 11.486943244934082], [-69.52890014648438, 11.503470420837402], [-69.5069580078125, 11.506942749023438], [-69.44473266601562, 11.498055458068848], [-69.4127197265625, 11.487568855285645], [-69.36021423339844, 11.493331909179688], [-69.32757568359375, 11.512638092041016], [-69.27487182617188, 11.533957481384277], [-69.1059799194336, 11.486387252807617], [-68.95389556884766, 11.451664924621582], [-68.84368133544922, 11.44708251953125], [-68.66043090820312, 11.349859237670898], [-68.60195922851562, 11.29055404663086], [-68.55347442626953, 11.271804809570312], [-68.53195190429688, 11.261388778686523], [-68.4183349609375, 11.179998397827148], [-68.39862060546875, 11.15805435180664], [-68.33334350585938, 11.044998168945312], [-68.24250793457031, 10.874374389648438], [-68.251953125, 10.856666564941406], [-68.3050765991211, 10.854999542236328], [-68.32528686523438, 10.843610763549805], [-68.33264923095703, 10.815971374511719], [-68.32806396484375, 10.767499923706055], [-68.28056335449219, 10.643888473510742], [-68.25361633300781, 10.588888168334961], [-68.18597412109375, 10.514998435974121], [-68.16514587402344, 10.49888801574707], [-68.11424255371094, 10.484929084777832], [-68.0773696899414, 10.49305534362793], [-68.00222778320312, 10.491109848022461], [-67.89306640625, 10.474443435668945], [-67.86798095703125, 10.46487045288086], [-67.83195495605469, 10.482776641845703], [-67.79611206054688, 10.491943359375], [-67.5433349609375, 10.53277587890625], [-67.46694946289062, 10.538055419921875], [-67.40139770507812, 10.538887023925781], [-67.39224243164062, 10.538785934448242], [-67.28028869628906, 10.54666519165039], [-67.10806274414062, 10.579166412353516], [-67.00222778320312, 10.61027717590332], [-66.50973510742188, 10.627777099609375], [-66.47056579589844, 10.629165649414062], [-66.4183349609375, 10.625555038452148], [-66.385009765625, 10.616109848022461], [-66.3429946899414, 10.604721069335938], [-66.32362365722656, 10.612777709960938], [-66.31028747558594, 10.626018524169922], [-66.28308868408203, 10.644651412963867], [-66.22987365722656, 10.640554428100586], [-66.20973205566406, 10.632499694824219], [-66.08168029785156, 10.576665878295898], [-66.11834716796875, 10.52694320678711], [-66.1197280883789, 10.499999046325684], [-66.09834289550781, 10.472498893737793], [-66.0684814453125, 10.443193435668945], [-65.9586181640625, 10.356666564941406], [-65.93573760986328, 10.298297882080078], [-65.814453125, 10.22833251953125], [-65.78667449951172, 10.220415115356445], [-65.76528930664062, 10.216943740844727], [-65.72531127929688, 10.21921157836914], [-65.54806518554688, 10.178054809570312], [-65.42668151855469, 10.139999389648438], [-65.276123046875, 10.115554809570312], [-65.08139038085938, 10.060555458068848], [-64.82501220703125, 10.082221984863281], [-64.7880630493164, 10.08944320678711], [-64.76445007324219, 10.097221374511719], [-64.7327880859375, 10.113471031188965], [-64.71473693847656, 10.17972183227539], [-64.6259765625, 10.246665000915527], [-64.57972717285156, 10.258609771728516], [-64.53460693359375, 10.249940872192383], [-64.48251342773438, 10.238054275512695], [-64.37854766845703, 10.301109313964844], [-64.39584350585938, 10.337221145629883], [-64.38473510742188, 10.368888854980469], [-64.36805725097656, 10.384164810180664], [-64.2025146484375, 10.44999885559082], [-64.18205261230469, 10.456549644470215], [-64.0925064086914, 10.456526756286621], [-64.07084655761719, 10.446388244628906], [-64.05001068115234, 10.442220687866211], [-63.87611389160156, 10.440277099609375], [-63.8094482421875, 10.442220687866211], [-63.7933349609375, 10.446664810180664], [-63.71556091308594, 10.471942901611328], [-63.69750213623047, 10.485553741455078], [-63.77777862548828, 10.525276184082031], [-63.83889389038086, 10.551666259765625], [-63.966392517089844, 10.578471183776855], [-64.148681640625, 10.569998741149902], [-64.20098114013672, 10.550276756286621], [-64.2158432006836, 10.534027099609375], [-64.2356948852539, 10.514373779296875], [-64.25889587402344, 10.534720420837402], [-64.29959106445312, 10.626526832580566], [-64.26458740234375, 10.657776832580566], [-64.2366714477539, 10.638887405395508], [-64.14681243896484, 10.617915153503418], [-63.97416687011719, 10.629720687866211], [-63.84278106689453, 10.645832061767578], [-63.69194793701172, 10.645554542541504], [-63.66667175292969, 10.641666412353516], [-63.534446716308594, 10.627082824707031], [-63.505558013916016, 10.639998435974121], [-63.34444808959961, 10.672637939453125], [-63.28611755371094, 10.670831680297852], [-63.25305938720703, 10.678333282470703], [-63.23027801513672, 10.688331604003906], [-63.17222595214844, 10.719720840454102], [-62.99375534057617, 10.716387748718262], [-62.972782135009766, 10.705415725708008], [-62.95347595214844, 10.699860572814941], [-62.90666961669922, 10.695833206176758], [-62.86958694458008, 10.706388473510742], [-62.79138946533203, 10.731109619140625], [-62.74833679199219, 10.74222183227539], [-62.69750213623047, 10.74777603149414], [-62.538612365722656, 10.735553741455078], [-62.470001220703125, 10.726943969726562], [-62.358062744140625, 10.708054542541504], [-62.301116943359375, 10.700277328491211], [-62.1925048828125, 10.69416618347168], [-62.13187789916992, 10.701318740844727], [-61.98389434814453, 10.72805404663086], [-61.87958908081055, 10.728331565856934], [-61.883338928222656, 10.694721221923828], [-61.921112060546875, 10.661943435668945], [-61.95305633544922, 10.648887634277344], [-62.035560607910156, 10.632499694824219], [-62.08555603027344, 10.627222061157227], [-62.114585876464844, 10.628193855285645], [-62.13555908203125, 10.634721755981445], [-62.162506103515625, 10.636665344238281], [-62.249168395996094, 10.627012252807617], [-62.2761116027832, 10.600415229797363], [-62.28722381591797, 10.583610534667969], [-62.3013916015625, 10.555276870727539], [-62.331111907958984, 10.531664848327637], [-62.44972229003906, 10.533332824707031], [-62.527225494384766, 10.54041576385498], [-62.58000183105469, 10.555276870727539], [-62.64972686767578, 10.568193435668945], [-62.671112060546875, 10.56805419921875], [-62.830833435058594, 10.544443130493164], [-62.912296295166016, 10.528749465942383], [-63.00423812866211, 10.452984809875488], [-62.977088928222656, 10.464165687561035], [-62.93472671508789, 10.490553855895996], [-62.91118240356445, 10.515624046325684], [-62.87250518798828, 10.524443626403809], [-62.8386116027832, 10.511665344238281], [-62.83555603027344, 10.448331832885742], [-62.83097457885742, 10.428193092346191], [-62.837711334228516, 10.39729118347168], [-62.87180709838867, 10.389860153198242], [-62.88916778564453, 10.403888702392578], [-62.91167068481445, 10.416526794433594], [-62.93430709838867, 10.418471336364746], [-62.954444885253906, 10.41472053527832], [-62.980560302734375, 10.406944274902344], [-63.003684997558594, 10.3941650390625], [-62.998409271240234, 10.271596908569336], [-62.93416976928711, 10.279304504394531], [-62.927642822265625, 10.302915573120117], [-62.94291687011719, 10.32673454284668], [-62.95472717285156, 10.374998092651367], [-62.9466667175293, 10.400900840759277], [-62.9072265625, 10.391729354858398], [-62.88473892211914, 10.37439250946045], [-62.857093811035156, 10.38235855102539], [-62.822418212890625, 10.385638236999512], [-62.79008865356445, 10.401333808898926], [-62.736392974853516, 10.373817443847656], [-62.68556213378906, 10.31027603149414], [-62.66638946533203, 10.280553817749023], [-62.63063049316406, 10.10715103149414], [-62.6693115234375, 10.07499885559082], [-62.70916748046875, 10.065832138061523], [-62.8013916015625, 10.055276870727539], [-62.82917022705078, 10.053609848022461], [-62.89167022705078, 10.0816650390625], [-62.917503356933594, 10.094999313354492], [-62.956947326660156, 10.104304313659668], [-63.00861358642578, 10.101664543151855], [-63.01679992675781, 10.09869384765625], [-63.01597595214844, 10.095693588256836], [-62.983856201171875, 10.06892204284668], [-62.94972229003906, 10.077220916748047], [-62.910560607910156, 10.076944351196289], [-62.87139129638672, 10.046388626098633], [-62.80528259277344, 10.008609771728516], [-62.68250274658203, 10.051666259765625], [-62.66278076171875, 10.059165954589844], [-62.64556121826172, 10.069442749023438], [-62.61576461791992, 10.093193054199219], [-62.60541915893555, 10.12722110748291], [-62.62333679199219, 10.188610076904297], [-62.60944747924805, 10.223957061767578], [-62.5775032043457, 10.225138664245605], [-62.5352783203125, 10.202220916748047], [-62.501319885253906, 10.174235343933105], [-62.4888916015625, 10.153055191040039], [-62.47028350830078, 10.091110229492188], [-62.42694854736328, 9.979166030883789], [-62.38009193870275, 9.894418905915902]]
    #Region 4: North America, Central America and the Caribbean
    region4 = [[180.0, 58.9523811340332], [180.0, 48.0], [176.0055389404297, 49.997230529785156], [170.0, 53.0], [173.41497802734375, 55.03273010253906], [180.0, 58.9523811340332], [-35.0, 10.0], [-62.0, 10.0], [-62.06483960144875, 9.981988999597569], [-62.0755615234375, 9.986387252807617], [-62.17889404296875, 10.014721870422363], [-62.2092399597168, 10.011039733886719], [-62.23173904418945, 9.964720726013184], [-62.20647226064378, 9.942646594265616], [-62.225438860949566, 9.937378094180676], [-62.25048828125, 9.967012405395508], [-62.249128285596655, 9.930797698445373], [-62.38009193870275, 9.894418905915902], [-62.42694854736328, 9.979166030883789], [-62.47028350830078, 10.091110229492188], [-62.4888916015625, 10.153055191040039], [-62.501319885253906, 10.174235343933105], [-62.5352783203125, 10.202220916748047], [-62.5775032043457, 10.225138664245605], [-62.60944747924805, 10.223957061767578], [-62.62333679199219, 10.188610076904297], [-62.60541915893555, 10.12722110748291], [-62.61576461791992, 10.093193054199219], [-62.64556121826172, 10.069442749023438], [-62.66278076171875, 10.059165954589844], [-62.68250274658203, 10.051666259765625], [-62.80528259277344, 10.008609771728516], [-62.87139129638672, 10.046388626098633], [-62.910560607910156, 10.076944351196289], [-62.94972229003906, 10.077220916748047], [-62.983856201171875, 10.06892204284668], [-63.01597595214844, 10.095693588256836], [-63.01679992675781, 10.09869384765625], [-63.00861358642578, 10.101664543151855], [-62.956947326660156, 10.104304313659668], [-62.917503356933594, 10.094999313354492], [-62.89167022705078, 10.0816650390625], [-62.82917022705078, 10.053609848022461], [-62.8013916015625, 10.055276870727539], [-62.70916748046875, 10.065832138061523], [-62.6693115234375, 10.07499885559082], [-62.63063049316406, 10.10715103149414], [-62.66638946533203, 10.280553817749023], [-62.68556213378906, 10.31027603149414], [-62.736392974853516, 10.373817443847656], [-62.79008865356445, 10.401333808898926], [-62.822418212890625, 10.385638236999512], [-62.857093811035156, 10.38235855102539], [-62.88473892211914, 10.37439250946045], [-62.9072265625, 10.391729354858398], [-62.9466667175293, 10.400900840759277], [-62.95472717285156, 10.374998092651367], [-62.94291687011719, 10.32673454284668], [-62.927642822265625, 10.302915573120117], [-62.93416976928711, 10.279304504394531], [-62.998409271240234, 10.271596908569336], [-63.003684997558594, 10.3941650390625], [-62.980560302734375, 10.406944274902344], [-62.954444885253906, 10.41472053527832], [-62.93430709838867, 10.418471336364746], [-62.91167068481445, 10.416526794433594], [-62.88916778564453, 10.403888702392578], [-62.87180709838867, 10.389860153198242], [-62.837711334228516, 10.39729118347168], [-62.83097457885742, 10.428193092346191], [-62.83555603027344, 10.448331832885742], [-62.8386116027832, 10.511665344238281], [-62.87250518798828, 10.524443626403809], [-62.91118240356445, 10.515624046325684], [-62.93472671508789, 10.490553855895996], [-62.977088928222656, 10.464165687561035], [-63.00423812866211, 10.452984809875488], [-62.912296295166016, 10.528749465942383], [-62.830833435058594, 10.544443130493164], [-62.671112060546875, 10.56805419921875], [-62.64972686767578, 10.568193435668945], [-62.58000183105469, 10.555276870727539], [-62.527225494384766, 10.54041576385498], [-62.44972229003906, 10.533332824707031], [-62.331111907958984, 10.531664848327637], [-62.3013916015625, 10.555276870727539], [-62.28722381591797, 10.583610534667969], [-62.2761116027832, 10.600415229797363], [-62.249168395996094, 10.627012252807617], [-62.162506103515625, 10.636665344238281], [-62.13555908203125, 10.634721755981445], [-62.114585876464844, 10.628193855285645], [-62.08555603027344, 10.627222061157227], [-62.035560607910156, 10.632499694824219], [-61.95305633544922, 10.648887634277344], [-61.921112060546875, 10.661943435668945], [-61.883338928222656, 10.694721221923828], [-61.87958908081055, 10.728331565856934], [-61.98389434814453, 10.72805404663086], [-62.13187789916992, 10.701318740844727], [-62.1925048828125, 10.69416618347168], [-62.301116943359375, 10.700277328491211], [-62.358062744140625, 10.708054542541504], [-62.470001220703125, 10.726943969726562], [-62.538612365722656, 10.735553741455078], [-62.69750213623047, 10.74777603149414], [-62.74833679199219, 10.74222183227539], [-62.79138946533203, 10.731109619140625], [-62.86958694458008, 10.706388473510742], [-62.90666961669922, 10.695833206176758], [-62.95347595214844, 10.699860572814941], [-62.972782135009766, 10.705415725708008], [-62.99375534057617, 10.716387748718262], [-63.17222595214844, 10.719720840454102], [-63.23027801513672, 10.688331604003906], [-63.25305938720703, 10.678333282470703], [-63.28611755371094, 10.670831680297852], [-63.34444808959961, 10.672637939453125], [-63.505558013916016, 10.639998435974121], [-63.534446716308594, 10.627082824707031], [-63.66667175292969, 10.641666412353516], [-63.69194793701172, 10.645554542541504], [-63.84278106689453, 10.645832061767578], [-63.97416687011719, 10.629720687866211], [-64.14681243896484, 10.617915153503418], [-64.2366714477539, 10.638887405395508], [-64.26458740234375, 10.657776832580566], [-64.29959106445312, 10.626526832580566], [-64.25889587402344, 10.534720420837402], [-64.2356948852539, 10.514373779296875], [-64.2158432006836, 10.534027099609375], [-64.20098114013672, 10.550276756286621], [-64.148681640625, 10.569998741149902], [-63.966392517089844, 10.578471183776855], [-63.83889389038086, 10.551666259765625], [-63.77777862548828, 10.525276184082031], [-63.69750213623047, 10.485553741455078], [-63.71556091308594, 10.471942901611328], [-63.7933349609375, 10.446664810180664], [-63.8094482421875, 10.442220687866211], [-63.87611389160156, 10.440277099609375], [-64.05001068115234, 10.442220687866211], [-64.07084655761719, 10.446388244628906], [-64.0925064086914, 10.456526756286621], [-64.18205261230469, 10.456549644470215], [-64.2025146484375, 10.44999885559082], [-64.36805725097656, 10.384164810180664], [-64.38473510742188, 10.368888854980469], [-64.39584350585938, 10.337221145629883], [-64.37854766845703, 10.301109313964844], [-64.48251342773438, 10.238054275512695], [-64.53460693359375, 10.249940872192383], [-64.57972717285156, 10.258609771728516], [-64.6259765625, 10.246665000915527], [-64.71473693847656, 10.17972183227539], [-64.7327880859375, 10.113471031188965], [-64.76445007324219, 10.097221374511719], [-64.7880630493164, 10.08944320678711], [-64.82501220703125, 10.082221984863281], [-65.08139038085938, 10.060555458068848], [-65.276123046875, 10.115554809570312], [-65.42668151855469, 10.139999389648438], [-65.54806518554688, 10.178054809570312], [-65.72531127929688, 10.21921157836914], [-65.76528930664062, 10.216943740844727], [-65.78667449951172, 10.220415115356445], [-65.814453125, 10.22833251953125], [-65.93573760986328, 10.298297882080078], [-65.9586181640625, 10.356666564941406], [-66.0684814453125, 10.443193435668945], [-66.09834289550781, 10.472498893737793], [-66.1197280883789, 10.499999046325684], [-66.11834716796875, 10.52694320678711], [-66.08168029785156, 10.576665878295898], [-66.20973205566406, 10.632499694824219], [-66.22987365722656, 10.640554428100586], [-66.28308868408203, 10.644651412963867], [-66.31028747558594, 10.626018524169922], [-66.32362365722656, 10.612777709960938], [-66.3429946899414, 10.604721069335938], [-66.385009765625, 10.616109848022461], [-66.4183349609375, 10.625555038452148], [-66.47056579589844, 10.629165649414062], [-66.50973510742188, 10.627777099609375], [-67.00222778320312, 10.61027717590332], [-67.10806274414062, 10.579166412353516], [-67.28028869628906, 10.54666519165039], [-67.39224243164062, 10.538785934448242], [-67.40139770507812, 10.538887023925781], [-67.46694946289062, 10.538055419921875], [-67.5433349609375, 10.53277587890625], [-67.79611206054688, 10.491943359375], [-67.83195495605469, 10.482776641845703], [-67.86798095703125, 10.46487045288086], [-67.89306640625, 10.474443435668945], [-68.00222778320312, 10.491109848022461], [-68.0773696899414, 10.49305534362793], [-68.11424255371094, 10.484929084777832], [-68.16514587402344, 10.49888801574707], [-68.18597412109375, 10.514998435974121], [-68.25361633300781, 10.588888168334961], [-68.28056335449219, 10.643888473510742], [-68.32806396484375, 10.767499923706055], [-68.33264923095703, 10.815971374511719], [-68.32528686523438, 10.843610763549805], [-68.3050765991211, 10.854999542236328], [-68.251953125, 10.856666564941406], [-68.24250793457031, 10.874374389648438], [-68.33334350585938, 11.044998168945312], [-68.39862060546875, 11.15805435180664], [-68.4183349609375, 11.179998397827148], [-68.53195190429688, 11.261388778686523], [-68.55347442626953, 11.271804809570312], [-68.60195922851562, 11.29055404663086], [-68.66043090820312, 11.349859237670898], [-68.84368133544922, 11.44708251953125], [-68.95389556884766, 11.451664924621582], [-69.1059799194336, 11.486387252807617], [-69.27487182617188, 11.533957481384277], [-69.32757568359375, 11.512638092041016], [-69.36021423339844, 11.493331909179688], [-69.4127197265625, 11.487568855285645], [-69.44473266601562, 11.498055458068848], [-69.5069580078125, 11.506942749023438], [-69.52890014648438, 11.503470420837402], [-69.5555648803711, 11.486943244934082], [-69.57952117919922, 11.464096069335938], [-69.60848236083984, 11.459165573120117], [-69.6318130493164, 11.46763801574707], [-69.67472839355469, 11.511110305786133], [-69.70611572265625, 11.548887252807617], [-69.71916961669922, 11.570276260375977], [-69.75778198242188, 11.661388397216797], [-69.77084350585938, 11.696109771728516], [-69.80416870117188, 11.794166564941406], [-69.81687927246094, 11.850971221923828], [-69.81889343261719, 11.881109237670898], [-69.81663513183594, 11.916109085083008], [-69.8237533569336, 11.988055229187012], [-69.84056091308594, 12.032776832580566], [-69.85945129394531, 12.071943283081055], [-69.90570068359375, 12.139999389648438], [-69.93472290039062, 12.169721603393555], [-70.01431274414062, 12.197498321533203], [-70.06806945800781, 12.17361068725586], [-70.1875, 12.107776641845703], [-70.20639038085938, 12.080275535583496], [-70.25806427001953, 11.989165306091309], [-70.28666687011719, 11.920276641845703], [-70.29376220703125, 11.899443626403809], [-70.29403686523438, 11.861665725708008], [-70.243896484375, 11.775276184082031], [-70.22653198242188, 11.655068397521973], [-70.23597717285156, 11.628888130187988], [-70.21403503417969, 11.60916519165039], [-70.18000793457031, 11.603610038757324], [-70.00473022460938, 11.650276184082031], [-69.85736846923828, 11.689582824707031], [-69.81653594970703, 11.690971374511719], [-69.78778076171875, 11.635276794433594], [-69.77611541748047, 11.609442710876465], [-69.74240112304688, 11.499199867248535], [-69.8013916015625, 11.427221298217773], [-69.836669921875, 11.424582481384277], [-69.87403106689453, 11.426526069641113], [-69.89771270751953, 11.436179161071777], [-69.9876480102539, 11.514720916748047], [-70.04779052734375, 11.517776489257812], [-70.02389526367188, 11.491388320922852], [-70.03486633300781, 11.441317558288574], [-70.09556579589844, 11.429998397827148], [-70.14250183105469, 11.418333053588867], [-70.23918151855469, 11.35305404663086], [-70.32333374023438, 11.3316650390625], [-70.42140197753906, 11.292221069335938], [-70.44667053222656, 11.289999008178711], [-70.49209594726562, 11.26291561126709], [-70.50847625732422, 11.248193740844727], [-70.54681396484375, 11.241943359375], [-70.62139892578125, 11.236248970031738], [-70.65501403808594, 11.238609313964844], [-70.73695373535156, 11.229999542236328], [-70.8255615234375, 11.211387634277344], [-70.8558349609375, 11.200277328491211], [-70.8900146484375, 11.18610954284668], [-70.93389892578125, 11.164721488952637], [-70.9890365600586, 11.133332252502441], [-71.01445007324219, 11.11722183227539], [-71.03153228759766, 11.103193283081055], [-71.07389831542969, 11.079999923706055], [-71.17807006835938, 11.032220840454102], [-71.2711181640625, 10.99305534362793], [-71.28111267089844, 10.989166259765625], [-71.35389709472656, 10.978333473205566], [-71.38417053222656, 10.983332633972168], [-71.41195678710938, 10.98416519165039], [-71.43473052978516, 10.979581832885742], [-71.49285888671875, 10.961040496826172], [-71.45167541503906, 10.916387557983398], [-71.44667053222656, 10.795555114746094], [-71.52639770507812, 10.726943969726562], [-71.54570770263672, 10.568331718444824], [-71.52195739746094, 10.533332824707031], [-71.4586181640625, 10.459999084472656], [-71.43667602539062, 10.369443893432617], [-71.38555908203125, 10.28277587890625], [-71.31695556640625, 10.192777633666992], [-71.29264831542969, 10.175415992736816], [-71.26945495605469, 10.150137901306152], [-71.14778137207031, 9.95083236694336], [-71.09834289550781, 9.894790649414062], [-71.07340240478516, 9.851110458374023], [-71.0533447265625, 9.703609466552734], [-71.06111145019531, 9.61722183227539], [-71.08306884765625, 9.566665649414062], [-71.08917236328125, 9.536596298217773], [-71.07112121582031, 9.390276908874512], [-71.05555725097656, 9.354442596435547], [-71.05597686767578, 9.338749885559082], [-71.07486724853516, 9.312499046325684], [-71.11778259277344, 9.293333053588867], [-71.14723205566406, 9.28680419921875], [-71.16542053222656, 9.273748397827148], [-71.20834350585938, 9.219720840454102], [-71.21417236328125, 9.208887100219727], [-71.22508239746094, 9.174200057983398], [-71.24139404296875, 9.155415534973145], [-71.29055786132812, 9.121665954589844], [-71.31361389160156, 9.110553741455078], [-71.49723052978516, 9.050415992736816], [-71.55278015136719, 9.04083251953125], [-71.59834289550781, 9.039999008178711], [-71.62389373779297, 9.043054580688477], [-71.68917083740234, 9.063401222229004], [-71.7148666381836, 9.077916145324707], [-71.73417663574219, 9.100555419921875], [-71.75945281982422, 9.118331909179688], [-71.751953125, 9.22249984741211], [-71.7319564819336, 9.295970916748047], [-71.72361755371094, 9.344999313354492], [-71.73320007324219, 9.375693321228027], [-71.91084289550781, 9.493610382080078], [-71.9444580078125, 9.543054580688477], [-71.96583557128906, 9.602499008178711], [-71.97917175292969, 9.623332977294922], [-72.00806427001953, 9.663471221923828], [-72.04653930664062, 9.70777702331543], [-72.07792663574219, 9.735554695129395], [-72.12535095214844, 9.818193435668945], [-71.9112548828125, 10.124443054199219], [-71.84112548828125, 10.218887329101562], [-71.82514953613281, 10.255138397216797], [-71.79444885253906, 10.321805000305176], [-71.77320098876953, 10.349305152893066], [-71.75569915771484, 10.369721412658691], [-71.67778015136719, 10.423332214355469], [-71.64889526367188, 10.442499160766602], [-71.6300048828125, 10.476387023925781], [-71.58168029785156, 10.674720764160156], [-71.57792663574219, 10.716110229492188], [-71.59695434570312, 10.727776527404785], [-71.61750793457031, 10.741943359375], [-71.66181182861328, 10.777498245239258], [-71.69111633300781, 10.834165573120117], [-71.76361846923828, 11.013749122619629], [-71.7408447265625, 11.03416633605957], [-71.7952880859375, 11.128610610961914], [-71.83778381347656, 11.182777404785156], [-71.86695861816406, 11.229442596435547], [-71.90611267089844, 11.300832748413086], [-71.93278503417969, 11.363054275512695], [-71.94486999511719, 11.395554542541504], [-71.95264434814453, 11.421804428100586], [-71.96084594726562, 11.46444320678711], [-71.96640014648438, 11.506387710571289], [-71.96916961669922, 11.546248435974121], [-71.95417022705078, 11.594720840454102], [-71.81639099121094, 11.647499084472656], [-71.76583862304688, 11.662360191345215], [-71.71223449707031, 11.674999237060547], [-71.63833618164062, 11.690832138061523], [-71.54139709472656, 11.703887939453125], [-71.5050048828125, 11.707221984863281], [-71.48056030273438, 11.710277557373047], [-71.44584655761719, 11.723888397216797], [-71.37861633300781, 11.753332138061523], [-71.34112548828125, 11.799165725708008], [-71.32472229003906, 11.85305404663086], [-71.3194580078125, 11.865276336669922], [-71.28250122070312, 11.918609619140625], [-71.23194885253906, 11.951387405395508], [-71.17362213134766, 11.984443664550781], [-71.13125610351562, 12.017359733581543], [-71.11418151855469, 12.048887252807617], [-71.11083984375, 12.075554847717285], [-71.1148681640625, 12.098609924316406], [-71.22084045410156, 12.302082061767578], [-71.24305725097656, 12.321664810180664], [-71.27570343017578, 12.346110343933105], [-71.52862548828125, 12.446109771728516], [-71.56361389160156, 12.453609466552734], [-71.66153717041016, 12.463888168334961], [-71.69084167480469, 12.459304809570312], [-71.72431182861328, 12.440971374511719], [-71.73653411865234, 12.414165496826172], [-71.70438385009766, 12.434477806091309], [-71.65501403808594, 12.438888549804688], [-71.63139343261719, 12.421595573425293], [-71.69292449951172, 12.366388320922852], [-71.80972290039062, 12.313331604003906], [-71.86424255371094, 12.210762977600098], [-71.93778991699219, 12.162776947021484], [-71.96305847167969, 12.153610229492188], [-71.98604583740234, 12.157081604003906], [-72.01181030273438, 12.192638397216797], [-71.98251342773438, 12.225692749023438], [-71.96475982666016, 12.251977920532227], [-72.02862548828125, 12.257776260375977], [-72.13862609863281, 12.249443054199219], [-72.15792083740234, 12.241108894348145], [-72.16885375976562, 12.221561431884766], [-72.14583587646484, 12.207289695739746], [-72.13890075683594, 12.126943588256836], [-72.13871765136719, 12.106388092041016], [-72.23271179199219, 11.915207862854004], [-72.25862121582031, 11.889165878295898], [-72.27972412109375, 11.875831604003906], [-72.41625213623047, 11.79777717590332], [-72.45750427246094, 11.78499984741211], [-72.48736572265625, 11.784095764160156], [-72.51764678955078, 11.78749942779541], [-72.73090362548828, 11.706805229187012], [-72.76876068115234, 11.687776565551758], [-73.03472900390625, 11.502777099609375], [-73.14306640625, 11.424444198608398], [-73.19584655761719, 11.376526832580566], [-73.20626068115234, 11.34902572631836], [-73.26750183105469, 11.306388854980469], [-73.28445434570312, 11.295555114746094], [-73.31695556640625, 11.285831451416016], [-73.34500122070312, 11.279443740844727], [-73.38431549072266, 11.273887634277344], [-73.50889587402344, 11.269721984863281], [-73.53724670410156, 11.271329879760742], [-73.55723571777344, 11.273887634277344], [-73.5947265625, 11.273054122924805], [-73.63139343261719, 11.271248817443848], [-73.67250061035156, 11.265275955200195], [-73.79084014892578, 11.265693664550781], [-73.86695861816406, 11.284721374511719], [-73.95125579833984, 11.319165229797363], [-73.96958923339844, 11.33902645111084], [-74.01278686523438, 11.355276107788086], [-74.0372314453125, 11.356388092041016], [-74.15501403808594, 11.331388473510742], [-74.18890380859375, 11.309720993041992], [-74.21556091308594, 11.2691650390625], [-74.23695373535156, 11.23499870300293], [-74.2388916015625, 11.19666576385498], [-74.22987365722656, 11.118610382080078], [-74.26361846923828, 11.016942977905273], [-74.29055786132812, 10.967498779296875], [-74.33222961425781, 10.863609313964844], [-74.35826873779297, 10.783401489257812], [-74.37541961669922, 10.755414962768555], [-74.39500427246094, 10.742152214050293], [-74.45777893066406, 10.742499351501465], [-74.46528625488281, 10.765971183776855], [-74.51918029785156, 10.812776565551758], [-74.59330749511719, 10.783297538757324], [-74.60750579833984, 10.823609352111816], [-74.59174346923828, 10.87763786315918], [-74.46528625488281, 10.971318244934082], [-74.41902923583984, 10.965831756591797], [-74.354736328125, 10.971109390258789], [-74.32722473144531, 10.974860191345215], [-74.30487060546875, 10.983471870422363], [-74.29222869873047, 10.99902629852295], [-74.31597900390625, 10.991804122924805], [-74.3416748046875, 10.986665725708008], [-74.37973022460938, 10.983055114746094], [-74.42361450195312, 10.983055114746094], [-74.45083618164062, 10.983888626098633], [-74.47611999511719, 10.987220764160156], [-74.5050048828125, 10.993331909179688], [-74.83458709716797, 11.105554580688477], [-74.86080932617188, 11.125486373901367], [-74.86805725097656, 11.119720458984375], [-74.92111206054688, 11.068887710571289], [-75.02305603027344, 10.966526985168457], [-75.03176879882812, 10.936005592346191], [-75.0394515991211, 10.905832290649414], [-75.2227783203125, 10.80555534362793], [-75.24402618408203, 10.800321578979492], [-75.2641372680664, 10.798992156982422], [-75.26819610595703, 10.75305461883545], [-75.25083923339844, 10.746109008789062], [-75.23562622070312, 10.762012481689453], [-75.22660827636719, 10.729165077209473], [-75.25184631347656, 10.70645523071289], [-75.26139831542969, 10.72138786315918], [-75.28807067871094, 10.724443435668945], [-75.40888977050781, 10.658609390258789], [-75.47611999511719, 10.592777252197266], [-75.50403594970703, 10.550138473510742], [-75.4866714477539, 10.490971565246582], [-75.48529052734375, 10.446664810180664], [-75.54707336425781, 10.41789436340332], [-75.53167724609375, 10.236944198608398], [-75.5340347290039, 10.18235969543457], [-75.57167053222656, 10.134443283081055], [-75.58251953125, 10.08863353729248], [-75.58000183105469, 10.071943283081055], [-75.58014678955078, 10.045692443847656], [-75.59028625488281, 9.979442596435547], [-75.59611511230469, 9.957498550415039], [-75.63528442382812, 9.82472038269043], [-75.66084289550781, 9.760276794433594], [-75.6268081665039, 9.698887825012207], [-75.60139465332031, 9.66555404663086], [-75.59140014648438, 9.634721755981445], [-75.58528137207031, 9.606388092041016], [-75.58555603027344, 9.58305549621582], [-75.59111785888672, 9.546943664550781], [-75.59806823730469, 9.52277660369873], [-75.62001037597656, 9.472221374511719], [-75.63431549072266, 9.448193550109863], [-75.65667724609375, 9.425832748413086], [-75.68791961669922, 9.41138744354248], [-75.7252197265625, 9.421241760253906], [-75.81604766845703, 9.438610076904297], [-75.91862487792969, 9.422222137451172], [-76.07139587402344, 9.34555435180664], [-76.08973693847656, 9.335832595825195], [-76.11056518554688, 9.315553665161133], [-76.17264556884766, 9.235832214355469], [-76.18055725097656, 9.191944122314453], [-76.26417541503906, 8.99527645111084], [-76.31751251220703, 8.938610076904297], [-76.37361145019531, 8.91611099243164], [-76.39889526367188, 8.910554885864258], [-76.41889953613281, 8.906665802001953], [-76.4267578125, 8.906387329101562], [-76.43243408203125, 8.876596450805664], [-76.67389678955078, 8.676595687866211], [-76.81806945800781, 8.633888244628906], [-76.88472747802734, 8.622220993041992], [-76.9283447265625, 8.56833267211914], [-76.9281997680664, 8.53722095489502], [-76.87667083740234, 8.521388053894043], [-76.82181549072266, 8.482220649719238], [-76.7738265991211, 8.411249160766602], [-76.76528930664062, 8.340693473815918], [-76.75334167480469, 8.197498321533203], [-76.73896026611328, 8.09062385559082], [-76.73194885253906, 8.039165496826172], [-76.74445343017578, 7.935832977294922], [-76.75791931152344, 7.919166088104248], [-76.81695556640625, 7.904999732971191], [-76.8719482421875, 7.909443855285645], [-76.92388916015625, 7.936040878295898], [-76.92750549316406, 7.985554695129395], [-76.916748046875, 8.031526565551758], [-76.8929214477539, 8.04666519165039], [-76.87055969238281, 8.034651756286621], [-76.83667755126953, 8.026456832885742], [-76.83250427246094, 8.086387634277344], [-76.8341064453125, 8.129096031188965], [-76.83857727050781, 8.133211135864258], [-76.88778686523438, 8.121665954589844], [-76.94862365722656, 8.15916633605957], [-76.95556640625, 8.189998626708984], [-76.96861267089844, 8.238332748413086], [-77.08168029785156, 8.32611083984375], [-77.118896484375, 8.386388778686523], [-77.134521484375, 8.410207748413086], [-77.16389465332031, 8.434165954589844], [-77.19944763183594, 8.443609237670898], [-77.22000122070312, 8.45111083984375], [-77.24097442626953, 8.464720726013184], [-77.28466033935547, 8.516526222229004], [-77.30029296875, 8.578332901000977], [-77.34695434570312, 8.662221908569336], [-77.36666870117188, 8.674999237060547], [-77.3719482421875, 8.646110534667969], [-77.43861389160156, 8.566665649414062], [-77.45223236083984, 8.556943893432617], [-77.47576904296875, 8.521110534667969], [-77.46858215332031, 8.471699714660645], [-77.4291000366211, 8.472498893737793], [-77.40576934814453, 8.45131778717041], [-77.37500762939453, 8.39861011505127], [-77.36862182617188, 8.364166259765625], [-77.36805725097656, 8.337221145629883], [-77.36250305175781, 8.284998893737793], [-77.34889221191406, 8.267776489257812], [-77.29611206054688, 8.215831756591797], [-77.27361297607422, 8.198610305786133], [-77.243896484375, 8.14527702331543], [-77.21542358398438, 8.087915420532227], [-77.19833374023438, 7.999444007873535], [-77.21556091308594, 7.937221527099609], [-77.29521179199219, 7.904860496520996], [-77.3114013671875, 7.886943817138672], [-77.37000274658203, 7.778471946716309], [-77.330078125, 7.726176261901856], [-77.33097839355469, 7.701805114746094], [-77.57743835449219, 7.526180267333984], [-77.60014343261719, 7.539235591888428], [-77.61029052734375, 7.562499046325684], [-77.62417602539062, 7.60333251953125], [-77.66389465332031, 7.679443836212158], [-77.74403381347656, 7.719999313354492], [-77.75889587402344, 7.693055152893066], [-77.75944519042969, 7.667222023010254], [-77.75917053222656, 7.6333327293396], [-77.75459289550781, 7.612221717834473], [-77.72937774658203, 7.568888187408447], [-77.72097778320312, 7.536457538604736], [-77.73222351074219, 7.505902290344238], [-77.74861907958984, 7.484305381774902], [-77.77570343017578, 7.47541618347168], [-77.80750274658203, 7.47868013381958], [-77.88972473144531, 7.228888511657715], [-80.0, 5.0], [-120.0, 5.0], [-140.0, 30.0], [-180.0, 30.0], [-180.0, 58.9523811340332], [-169.0, 65.5], [-169.0, 68.0], [-172.509033203125, 70.23301696777344], [-180.0, 75.0], [-180.0, 90.0], [-56.496, 90.0], [-60.721565634288304, 82.15099200763784], [-61.595555466305846, 81.90326626967301], [-63.888, 81.312], [-67.2338423993058, 80.72219420274243], [-67.7834877005752, 80.57859654224569], [-68.76948205787909, 80.09388076087414], [-73.72044302328452, 78.85842383762537], [-73.392, 74.976], [-66.352, 72.864], [-59.664, 68.64], [-55.0, 59.0], [-35.0, 59.0], [-35.0, 36.0], [-35.024, 23.584], [-35.0, 23.583915789473686], [-35.0, 10.0], [-81.71305847167969, 12.490276336669922], [-81.71028137207031, 12.490833282470703], [-81.7025146484375, 12.507221221923828], [-81.69097900390625, 12.532220840454102], [-81.68132019042969, 12.585762977600098], [-81.69236755371094, 12.590276718139648], [-81.70472717285156, 12.580275535583496], [-81.71389770507812, 12.564998626708984], [-81.72014617919922, 12.545276641845703], [-81.72013854980469, 12.496317863464355], [-81.71305847167969, 12.490276336669922], [-64.05555725097656, 10.857221603393555], [-63.97236251831055, 10.892915725708008], [-63.89000701904297, 10.904443740844727], [-63.81500244140625, 10.97805404663086], [-63.804725646972656, 11.021387100219727], [-63.84486770629883, 11.127082824707031], [-63.88493347167969, 11.175622940063477], [-63.951393127441406, 11.115554809570312], [-63.98750305175781, 11.07611083984375], [-64.02249908447266, 10.999998092651367], [-64.04278564453125, 10.987637519836426], [-64.05445098876953, 10.985692977905273], [-64.0836181640625, 10.988887786865234], [-64.1058349609375, 10.995277404785156], [-64.17584228515625, 11.031665802001953], [-64.18806457519531, 11.040276527404785], [-64.19743347167969, 11.055901527404785], [-64.19098663330078, 11.072638511657715], [-64.19277954101562, 11.082915306091309], [-64.20181274414062, 11.087777137756348], [-64.21833801269531, 11.088054656982422], [-64.37834167480469, 11.0569429397583], [-64.40570068359375, 10.970137596130371], [-64.35792541503906, 10.954581260681152], [-64.2279281616211, 10.931249618530273], [-64.20112609863281, 10.936943054199219], [-64.19014739990234, 10.946943283081055], [-64.17403411865234, 10.960415840148926], [-64.15084838867188, 10.952220916748047], [-64.13639831542969, 10.944999694824219], [-64.05862426757812, 10.864999771118164], [-64.05555725097656, 10.857221603393555], [-65.28111267089844, 10.880277633666992], [-65.26861572265625, 10.881109237670898], [-65.21000671386719, 10.891387939453125], [-65.19959259033203, 10.898332595825195], [-65.20083618164062, 10.911664962768555], [-65.212158203125, 10.9541654586792], [-65.30445098876953, 10.976109504699707], [-65.36445617675781, 10.969165802001953], [-65.3739013671875, 10.96444320678711], [-65.38667297363281, 10.9566650390625], [-65.41612243652344, 10.927221298217773], [-65.41570281982422, 10.916248321533203], [-65.39334106445312, 10.906944274902344], [-65.29972839355469, 10.883054733276367], [-65.28111267089844, 10.880277633666992]]
    #Region 5: South-West Pacific
    region5 = [[-180.0, 30.0], [-140.0, 30.0], [-120.0, 5.0], [-120.0, -60.0], [-180.0, -60.0], [-180.0, 30.0], [180.0, 23.5], [180.0, -60.0], [70.0, -60.0], [70.0, -50.0], [80.0, -50.0], [80.0, -5.0], [90.0, -5.0], [90.0, 6.5], [90.05563792866475, 6.509736637516332], [99.63441629035232, 6.524226584408694], [99.65262603759766, 6.494790554046631], [99.66804504394531, 6.496387481689453], [99.69197550268896, 6.524313655003549], [100.04409210871654, 6.52484630653056], [100.05843353271484, 6.511631011962891], [100.0777475545591, 6.524897217576258], [100.09340366369773, 6.524920900783176], [100.1221923828125, 6.448887825012207], [100.12711334228516, 6.424946784973145], [100.14192199707031, 6.436665534973145], [100.16844177246094, 6.470832347869873], [100.18105658015898, 6.525053494529113], [100.19275, 6.57528], [100.18413, 6.5975], [100.17454, 6.62236], [100.19407, 6.69743], [100.20385, 6.70937], [100.21498, 6.71111], [100.26414, 6.70667], [100.30025, 6.69979], [100.33066, 6.62042], [100.33518, 6.58611], [100.3744, 6.54389], [100.38748, 6.53611], [100.41321836112911, 6.525404688778491], [100.43689727783203, 6.515555381774902], [100.6566390991211, 6.448678970336914], [100.68135070800781, 6.453332901000977], [100.7169189453125, 6.479999542236328], [100.7513656616211, 6.502777099609375], [100.82816314697266, 6.427776336669922], [100.85441589355469, 6.32249927520752], [100.84786987304688, 6.296963691711426], [100.84691619873047, 6.284998893737793], [100.85323333740234, 6.243887901306152], [100.95206451416016, 6.240831851959229], [101.00969696044922, 6.269443511962891], [101.11204528808594, 6.250693798065186], [101.11426544189453, 6.182463645935059], [101.09913635253906, 6.118887901306152], [101.10635375976562, 6.050276756286621], [101.11920166015625, 5.99027681350708], [101.09566497802734, 5.923957347869873], [101.06900024414062, 5.9145827293396], [101.04225158691406, 5.901804924011231], [100.99114990234375, 5.794304370880127], [100.99465942382812, 5.779623031616211], [101.05745697021484, 5.729166030883789], [101.14232635498047, 5.633471012115479], [101.16497802734375, 5.638332366943359], [101.18968963623047, 5.651110172271729], [101.22322845458984, 5.670970916748047], [101.2594223022461, 5.714582443237305], [101.31189727783203, 5.807221412658691], [101.45218658447266, 5.863054275512695], [101.50604248046875, 5.880571365356445], [101.57101440429688, 5.915207386016846], [101.60649108886719, 5.905276775360107], [101.65940856933594, 5.850624084472656], [101.65857696533203, 5.82173490524292], [101.69232177734375, 5.764305114746094], [101.69954681396484, 5.760971069335938], [101.74635314941406, 5.78916597366333], [101.7718505859375, 5.777499198913574], [101.80072021484375, 5.745971202850342], [101.82608795166016, 5.740970611572266], [101.94405364990234, 5.867081642150879], [101.94371032714844, 5.903332710266113], [101.94080352783203, 5.927498817443848], [101.9437255859375, 5.956526756286621], [101.95260620117188, 5.975554466247559], [101.98392486572266, 6.01367998123169], [102.01109313964844, 6.038331985473633], [102.09053039550781, 6.154165267944336], [102.09523010253906, 6.236138343811035], [102.04773712158203, 6.265832901000977], [102.01844787597656, 6.285138130187988], [101.8124771118164, 6.454166412353516], [101.78678131103516, 6.477082252502441], [101.76193237304688, 6.507222175598145], [101.7451171875, 6.532638072967529], [101.740899624126, 6.53962165061756], [110.0, 10.0], [125.0, 23.5], [180.0, 23.5]]
    #Region 6: Europe
    region6 = [[-35.024, 23.584], [-35.0, 36.0], [-35.0, 23.583915789473686], [-35.024, 23.584], [44.81637479260081, 39.645822862874155], [44.816375732421875, 39.64582061767578], [44.81304168701172, 39.63081359863281], [44.736717224121094, 39.70180130004883], [44.608463287353516, 39.77915573120117], [44.47108459472656, 39.69886779785156], [44.47789001464844, 39.640655517578125], [44.42582321166992, 39.43644332885742], [44.401100158691406, 39.4165153503418], [44.34027099609375, 39.395545959472656], [44.30452346801758, 39.387290954589844], [44.230918884277344, 39.40845489501953], [44.200958251953125, 39.41693878173828], [44.07686233520508, 39.41165542602539], [44.036380767822266, 39.385475158691406], [44.03495407104492, 39.377445220947266], [44.060546875, 39.35248565673828], [44.07471466064453, 39.336097717285156], [44.21644973754883, 39.13651657104492], [44.19803237915039, 39.05667495727539], [44.16241455078125, 39.00519943237305], [44.17332458496094, 38.968040466308594], [44.20943069458008, 38.893463134765625], [44.25054931640625, 38.86554718017578], [44.28366470336914, 38.862491607666016], [44.30026626586914, 38.842628479003906], [44.30318832397461, 38.80998611450195], [44.26179504394531, 38.72429275512695], [44.306514739990234, 38.647281646728516], [44.31721496582031, 38.61304473876953], [44.32152557373047, 38.509368896484375], [44.311767578125, 38.46992492675781], [44.304161071777344, 38.44943618774414], [44.305259704589844, 38.400535583496094], [44.318145751953125, 38.38205337524414], [44.43672561645508, 38.39480972290039], [44.48251724243164, 38.34130096435547], [44.47322082519531, 38.316749572753906], [44.399017333984375, 38.211299896240234], [44.32331848144531, 38.09721374511719], [44.23888397216797, 37.96027374267578], [44.223968505859375, 37.89915084838867], [44.293052673339844, 37.873321533203125], [44.5767936706543, 37.76679611206055], [44.617767333984375, 37.71797561645508], [44.598876953125, 37.690818786621094], [44.584434509277344, 37.61388397216797], [44.58885192871094, 37.443092346191406], [44.61280059814453, 37.43422317504883], [44.648597717285156, 37.427772521972656], [44.801658630371094, 37.32166290283203], [44.818180084228516, 37.29741668701172], [44.8205451965332, 37.26874542236328], [44.80803768299948, 37.22653621357134], [44.85003082442806, 37.20309818114609], [44.85003082442806, 37.11554469349173], [44.66767433250223, 37.185013833273], [44.642215728759766, 37.18846130371094], [44.582767486572266, 37.163185119628906], [44.567630767822266, 37.14971160888672], [44.55054473876953, 37.132633209228516], [44.49055480957031, 37.091651916503906], [44.47027587890625, 37.08082580566406], [44.436378479003906, 37.06304168701172], [44.349708557128906, 37.03832244873047], [44.341102600097656, 37.00665283203125], [44.31916046142578, 36.97123718261719], [44.257076263427734, 36.985687255859375], [44.19596481323242, 37.096378326416016], [44.23026657104492, 37.15401840209961], [44.262596130371094, 37.16346740722656], [44.26832580566406, 37.19499969482422], [44.26214599609375, 37.23651885986328], [44.230552673339844, 37.276241302490234], [44.11888122558594, 37.3156852722168], [44.096099853515625, 37.31721496582031], [44.07294464111328, 37.316375732421875], [44.04804992675781, 37.32221984863281], [44.00901412963867, 37.31776809692383], [43.9558219909668, 37.279640197753906], [43.93693542480469, 37.25193786621094], [43.91297912597656, 37.22658920288086], [43.816375732421875, 37.22193145751953], [43.623046875, 37.22999572753906], [43.489990234375, 37.251243591308594], [43.15693664550781, 37.373878479003906], [43.00276565551758, 37.33915710449219], [42.96061706542969, 37.32353210449219], [42.924163818359375, 37.32776641845703], [42.894439697265625, 37.335540771484375], [42.822837829589844, 37.36717987060547], [42.786800384521484, 37.38367462158203], [42.76573181152344, 37.3712158203125], [42.724708557128906, 37.35194396972656], [42.588043212890625, 37.17568588256836], [42.579505920410156, 37.1536750793457], [42.53166198730469, 37.147491455078125], [42.37693786621094, 37.11749267578125], [42.355613708496094, 37.10692596435547], [42.37833023071289, 37.07206344604492], [42.37409973144531, 37.06316375732422], [42.367286682128906, 37.05588912963867], [41.83527374267578, 36.59888458251953], [41.403045654296875, 36.525550842285156], [41.367210388183594, 36.47277069091797], [41.29027557373047, 36.355552673339844], [41.25498962402344, 36.05499267578125], [41.27110290527344, 36.02777099609375], [41.35693359375, 35.876380920410156], [41.37833023071289, 35.83694076538086], [41.3840217590332, 35.63096618652344], [41.32249450683594, 35.550559997558594], [41.27777099609375, 35.49554443359375], [41.272491455078125, 35.46832275390625], [41.27082061767578, 35.443328857421875], [41.269161224365234, 35.39874267578125], [41.26221466064453, 35.367774963378906], [41.254432678222656, 35.34332275390625], [41.240962982177734, 35.3065185546875], [41.22248840332031, 35.25360870361328], [41.21527099609375, 35.224708557128906], [41.21240997314453, 35.20062255859375], [41.21124267578125, 35.1885986328125], [41.212493896484375, 35.11277770996094], [41.2147216796875, 35.06915283203125], [41.223052978515625, 34.91693878173828], [41.224849700927734, 34.78290557861328], [41.09943389892578, 34.60859680175781], [41.030540466308594, 34.473609924316406], [41.003875732421875, 34.41943359375], [40.663604736328125, 34.260276794433594], [40.199432373046875, 34.04082489013672], [40.16554260253906, 34.024436950683594], [40.0, 33.947654724121094], [39.79804992675781, 33.85333251953125], [39.579139709472656, 33.75], [39.3861083984375, 33.65888214111328], [38.961387634277344, 33.45555114746094], [38.810272216796875, 33.38221740722656], [38.794700622558594, 33.377593994140625], [39.08600997924805, 32.50151062011719], [38.986019134521484, 32.4776611328125], [39.04365158081055, 32.30405044555664], [39.2599983215332, 32.35554122924805], [39.301109313964844, 32.23638153076172], [39.19674301147461, 32.15494155883789], [39.07221984863281, 32.057777404785156], [39.00499725341797, 32.00555419921875], [38.48832702636719, 31.879165649414062], [37.986663818359375, 31.754051208496094], [37.96388244628906, 31.748332977294922], [37.709495544433594, 31.68466567993164], [37.47193908691406, 31.624996185302734], [37.20220184326172, 31.55615997314453], [37.005271911621094, 31.50555419921875], [37.29499053955078, 31.21722412109375], [37.48999786376953, 31.023052215576172], [37.56108856201172, 30.951087951660156], [38.00138854980469, 30.504165649414062], [37.66749572753906, 30.336387634277344], [37.502777099609375, 30.002220153808594], [37.41009521484375, 29.985580444335938], [37.03443908691406, 29.918052673339844], [36.74360656738281, 29.86471939086914], [36.50054931640625, 29.502777099609375], [36.272727966308594, 29.336528778076172], [36.06999969482422, 29.188888549804688], [35.541107177734375, 29.272499084472656], [35.187217712402344, 29.326942443847656], [34.961387634277344, 29.36083221435547], [34.96041488647461, 29.37742805480957], [34.966941833496094, 29.44888687133789], [34.97846984863281, 29.477497100830078], [35.00367736816406, 29.5281925201416], [34.97998046875, 29.545753479003906], [34.96582794189453, 29.549163818359375], [34.962562561035156, 29.55146026611328], [34.93804931640625, 29.523330688476562], [34.90380096435547, 29.486705780029297], [34.877777099609375, 29.53249740600586], [34.874721999526784, 29.55238983177807], [34.86638641357422, 29.606664657592773], [34.87388610839844, 29.63083267211914], [34.873386555914244, 29.63983234187301], [34.87277603149414, 29.65083122253418], [34.84777069091797, 29.740833282470703], [34.75360870361328, 29.991107940673828], [34.72334289550781, 30.089344024658203], [34.617800660577906, 30.35649364768854], [34.61209480513503, 30.35300673602901], [34.60577636046175, 30.370252412321644], [34.54527282714844, 30.406944274902344], [34.54332733154297, 30.43471908569336], [34.55888366699219, 30.48611068725586], [34.504390893218066, 30.6469757262499], [34.48537111624103, 30.698888646891767], [34.40138244628906, 30.85944366455078], [34.287498474121094, 31.162776947021484], [34.267578125, 31.216541290283203], [34.29031264711197, 31.2312847333785], [34.27685121714466, 31.26802660193263], [33.49379554291177, 32.19184002943465], [32.71072653394139, 33.11565345693667], [31.960749852491567, 34.0], [13.0, 34.0], [10.0, 39.0], [-2.0, 36.0], [-35.0, 36.0], [-35.0, 59.0], [-55.0, 59.0], [-59.664, 68.64], [-66.352, 72.864], [-73.392, 74.976], [-73.72044302328452, 78.85842383762537], [-68.76948205787909, 80.09388076087414], [-67.7834877005752, 80.57859654224569], [-67.2338423993058, 80.72219420274243], [-63.888, 81.312], [-61.595555466305846, 81.90326626967301], [-60.721565634288304, 82.15099200763784], [-56.496, 90.0], [40.0, 90.0], [40.0, 80.0], [50.0, 69.43576313614068], [50.00000151773305, 69.0909004211618], [50.0, 69.08428955078125], [50.000651855064525, 46.62786887918644], [50.0, 42.44704194167163], [50.01484779661026, 38.43357894317056], [48.88828, 38.44241], [48.84117, 38.45048], [48.78374, 38.44554], [48.6629, 38.38915], [48.43277, 38.62457], [48.33166, 38.60284], [48.2469, 38.7154], [48.02464, 38.83339], [48.01207, 38.89707], [48.22471, 38.97137], [48.3254, 39.03221], [48.29388, 39.11249], [48.13332, 39.21096], [48.12388, 39.27818], [48.35788, 39.38991], [47.98471, 39.71558], [47.76985, 39.63374], [47.5425, 39.51527], [47.39166, 39.47054], [47.28694, 39.37443], [47.16457, 39.33528], [47.04346, 39.2379], [46.9529, 39.15637], [46.8747, 39.16749], [46.75895, 39.08596], [46.54038, 38.87559], [46.35109, 38.90998], [46.17825, 38.84115], [45.43348, 39.00319], [45.4211, 39.06082], [45.29839, 39.19762], [45.14353, 39.21686], [45.13055, 39.28555], [45.06304, 39.37305], [44.95458, 39.43235], [44.88859, 39.60582], [44.81637479260081, 39.645822862874155], [32.09382805497271, 33.929628958770465], [32.0, 34.0], [31.987123713826193, 34.0], [32.09382805497271, 33.929628958770465]]
    #Region 7: Antarctic
    region7 = [[-180,-60],[-180,-90],[180,-90],[180,-60]]

    all_regions = [region1,region2,region3,region4,region5,region6,region7]

    #dictionary to return region by string from defined code
    code_to_string_dict = {1:'Africa', 2:'Asia', 3:'South America', 4:'North America, Central America and the Caribbean', 5:'South-West Pacific', 6:'Europe', 7:'Antarctic'}

    #set relevant metadata fields for function
    relevant_metadata_fields = ['WMO_region']
    metadata_fields_to_update.extend(relevant_metadata_fields)

    #iterate through all unique longitude/latitude pairs
    for unique_pair_ii, unique_pair in unique_pair_dictionary.items():

        #get current unique longitude/latitude
        current_longitude = unique_pair['longitude']
        current_latitude = unique_pair['latitude']

        #get metadata indices associated with current unique longitude/latitude pair
        relevant_metadata_indices = unique_pair['metadata_indices']

        #set default WMO_region as NaN
        WMO_region = np.NaN

        #create Shapely Point from longitude and latitude pairing
        point = Point(current_longitude, current_latitude)

        #iterate through each region to determine which region the lon/lat pairing is contained within 
        for i, region in enumerate(all_regions):
            #construct regional polygon from list of boundary points contained in regional lists
            region_polygon = Polygon(region)
            #check if lon/lat pairing is inside constructed regional polygon
            if region_polygon.contains(point) == True:
                #get which number WMO region the lon/lat pairing is contained within.
                WMO_region_code = i+1
                #get string for WMO region code
                WMO_region = code_to_string_dict[WMO_region_code]
                break

        #check if have determined WMO region (or not...)
        if pd.isnull(WMO_region) == True:
            #return Nan as WMO region - something has gone wrong as the lon/lat pairing is either located exactly on a boundary edge, or not on the earth..
            print('Could not determine WMO region\n', current_longitude, current_latitude)

        #set WMO region
        metadata['WMO_region'][relevant_metadata_indices] = WMO_region

    #return metadata 
    return metadata, log_file_str, metadata_fields_to_update 

###------------------------------------------------------------------------------------###
###------------------------------------------------------------------------------------###    

def get_coordinate_gridcell_indices(longitude,latitude,grid_centre_lons,grid_centre_lats):

    '''based on a specific longitude/latitude coordinate, get the gridcell indices that this point lies in.'''
    
    #get index of gridcell centre longitudes/latitudes closest to respective longitude or latitude coordinate
    lon_i = (np.abs(grid_centre_lons - longitude)).argmin() 
    lat_i = (np.abs(grid_centre_lats - latitude)).argmin() 

    return lon_i,lat_i 

###------------------------------------------------------------------------------------### 

def get_neighbouring_gridcell_indices(lon_i,lat_i,n_lons,n_lats):

    '''get the longitude/latitude indices of all grid cells which surround a defined grid cell (i.e. share a boundary, including diagonally)'''
    
    neigbouring_lon_indices = []
    neigbouring_lat_indices = []

    neighbour_lon_inds = np.arange(lon_i-1,lon_i+1.5,dtype=np.uint32)
    neighbour_lat_inds = np.arange(lat_i-1,lat_i+1.5,dtype=np.uint32)

    for neighbour_lat_ind in neighbour_lat_inds:
        for neighbour_lon_ind in neighbour_lon_inds:
            #do not append indices for centre gridcell
            if (neighbour_lon_ind == lon_i) & (neighbour_lat_ind == lat_i):
                continue

            #make sure lat indices are valid (handling poles convergence)
            if (neighbour_lat_ind < 0) or (neighbour_lat_ind >= n_lats):
                continue           
     
            #handle longitude wrap around
            if neighbour_lon_ind < 0:
                neighbour_lon_ind = int((n_lons - np.abs(neighbour_lon_ind)))

            if neighbour_lon_ind >= n_lons:
                neighbour_lon_ind = int((neighbour_lon_ind - n_lons))

            neigbouring_lon_indices.append(neighbour_lon_ind)
            neigbouring_lat_indices.append(neighbour_lat_ind)
        
    return neigbouring_lon_indices, neigbouring_lat_indices

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###
#DEFINE FUNCTIONS FOR THE CALCULATION OF ALL GRIDCELLS WITHIN A SPECIFIC DISTANCE PERIMITER OF A SPECIFIC LONGITUDE/LATITUDE COORDINATE
#THIS CALUCLATION CAN BE DONE USING GREAT CIRCLES, OR, MORE ACCURACTELY, GEODESICALLY. 
###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def geodesic_destination_given_distance_and_bearing(start_point, bearing, distance_to_travel):

    '''calculate the destination latitude/longitude from an initial latitude/longitude and a given bearing and distance geodesically.
       latitude/longitude are given in decimal degrees, and distance is given in kilometres.
       the geodesic distance is the shortest distance on the surface of an ellipsoidal model of the earth ('WGS-84' is the model used in this instance).
       the algorithm uses the method given by Karney (2013): https://link.springer.com/article/10.1007%2Fs00190-012-0578-z 
       this is accurate to round-off and always converges. This improves on Vincenty (1975), which is only accurate to 0.2 mm, and for which the distance calculation fails to converge for nearly antipodal points.
    '''
    
    d_instance = distance.geodesic(kilometers = distance_to_travel, ellipsoid='WGS-84')
    destination_point = d_instance.destination(point=start_point, bearing=bearing)  

    return [destination_point.longitude, destination_point.latitude]
    
#----------------------------------------------------------------------------------------#


def great_circle_destination_given_distance_and_bearing(start_point, bearing, distance_to_travel):

    '''calculate the destination latitude/longitude from an initial latitude/longitude and a given bearing and distance using the great-circle distance.
       latitude/longitude are given in decimal degrees, and distance is given in kilometres.
       the great-circle distance uses a spherical model of the earth, using the mean earth radius as defined by the International Union of Geodesy and Geophysics, (2a + b)/3 = 6371.0087714150598 kilometers approx 6371.009 km (for WGS-84), resulting in an error of up to about 0.5%.
    '''

    d_instance = distance.great_circle(kilometers = distance_to_travel)
    destination_point = d_instance.destination(point=start_point, bearing=bearing)  

    return [destination_point.longitude, destination_point.latitude]

#----------------------------------------------------------------------------------------#

def calculate_destination_points_perimeter(lon1,lat1,distance_to_travel,distance_calculation='geodesic'):

    '''given an initial longitude/latitude point, and a given distance, iteratively find destination longitude/latitude points 360 degrees around the initial point. 
       latitude/longitude are given in decimal degrees, and distance is given in kilometres.
       the distance calculation used can be 'great-circle' or 'geodesic' (default is 'geodesic').
    '''
    
    #create geopy Point instance for an initial latitude and longitude
    start_point = geopy.Point(lat1, lon1)
    
    #list for storing points around 360 degree perimeter
    perimeter_points = []

    #setup variables for defining most northernly, southerly, easterly and westerly latitudes and longitudes
    max_north = -999.0
    max_east = -999.0
    max_south = 999.0
    max_west = 999.0

    #iterate through 360 degrees
    for bearing in np.arange(0,360,1.0):
        #calculate destination point, using chosen distance calculation algorithm
        if distance_calculation == 'geodesic':
            destination_point = geodesic_destination_given_distance_and_bearing(start_point, bearing, distance_to_travel)
        elif distance_calculation == 'great-circle':
            destination_point = great_circle_destination_given_distance_and_bearing(start_point, bearing, distance_to_travel)

        #save most northernly and southernly destination point latitudes, and most easterly and westerly longitudes
        if destination_point[1] > max_north:
            max_north = copy.deepcopy(destination_point[1])
    
        if destination_point[0] > max_east:
            max_east = copy.deepcopy(destination_point[0])

        if destination_point[1] < max_south:
            max_south = copy.deepcopy(destination_point[1])
    
        if destination_point[0] < max_west:
            max_west = copy.deepcopy(destination_point[0])

        #append destination point to perimeter points list
        perimeter_points.append(destination_point)

    #group together perimeter point extremities
    perimeter_extreme_points = [max_north,max_east,max_south,max_west]

    return perimeter_points, perimeter_extreme_points

#----------------------------------------------------------------------------------------#

def get_relevant_gridbox_indices(perimeter_points,perimeter_extreme_points,grid_centre_lons,grid_edge_lons,grid_centre_lats,grid_edge_lats):

    '''taking a list of lists containing longitude/latitude points of a certain geodesic distance surrounding a fixed longitude/latitude point, get indices of gridcells which intersect the perimeter of the points. 
       a gridbox is taken to be inside the point extents if any part of the gridbox intersects with the perimeter of points, an gridbox intersection fraction is also returned. 
       a simple reduction in operations is done first by reducing the grid to a square/rectangle of grid cells surrounding the perimeter of points (done by using the most extreme located points N/E/W/S).
       makes assumption grid values run south to north, west to east.
    '''

    #make shapely polygon
    perimeter_polygon = Polygon(perimeter_points)

    #test validity of perimeter polygon
    #if the initial coordinates are very close to poles, there can be some self-intersections. 
    #in these very rare cases, all the gridcells in the square/rectangle surrounding perimeter points are taken as intersecting with 100% fraction 
    perimter_validity = perimeter_polygon.is_valid

    #reduce grid to solely square/rectangle surrounding perimeter points (done by finding closest grid centre longitudes/latitudes to perimeter point extremities N/E/W/S).
    #explicitly handle longitude wrap around
    
    #find closest northerly latitude gridbox centre
    closest_lat_north_index = (np.abs(grid_centre_lats - perimeter_extreme_points[0])).argmin()  
    #find closest southerly latitude gridbox centre
    closest_lat_south_index = (np.abs(grid_centre_lats - perimeter_extreme_points[2])).argmin() 
    #find closest easterly longitude gridbox centre
    closest_lon_east_index = (np.abs(grid_centre_lons - perimeter_extreme_points[1])).argmin()
    #find closest westerly longitude gridbox centre
    closest_lon_west_index = (np.abs(grid_centre_lons - perimeter_extreme_points[3])).argmin() 

    #create arrays of longitude/latitude indices for relevant gridboxes 
    relevant_centre_lat_indices = np.arange(closest_lat_south_index,closest_lat_north_index+0.5,1,dtype='int')
    if closest_lon_east_index < closest_lon_west_index:
        relevant_centre_lon_indices = np.arange(closest_lon_west_index,len(grid_centre_lons),1,dtype='int')
        relevant_centre_lon_indices = np.insert(relevant_centre_lon_indices,0,np.arange(0,closest_lon_east_index+0.5,1,dtype='int'))
    else:
        relevant_centre_lon_indices = np.arange(closest_lon_west_index,closest_lon_east_index+0.5,1,dtype='int')

    #determine which gridcells intersect with points perimeter
    #define list for storing intersecting longitude/latitude grid indices, and a gridbox intersection fraction with perimeter (as lists of indices: [lon_i,lat_i,intersect_fraction])
    relevant_grid_indices = [] 
    #iterate through gridcells
    for lat_i in relevant_centre_lat_indices:
        for lon_i in relevant_centre_lon_indices:
            #if the perimeter polygon is not valid then assume all gridcells in the square/rectangle surrounding the perimeter points all intersect with perimeter with 100% fraction 
            if perimter_validity == False:      
                relevant_grid_indices.append([lon_i,lat_i,1.0])
            else:
                #create Shapely Polygon for current gridcell iterating over   
                gridbox = Polygon([[grid_edge_lons[lon_i],grid_edge_lats[lat_i]], [grid_edge_lons[lon_i],grid_edge_lats[lat_i+1]], [grid_edge_lons[lon_i+1],grid_edge_lats[lat_i+1]], [grid_edge_lons[lon_i+1],grid_edge_lats[lat_i]]])          
                #get area of gridcell
                gridbox_area = gridbox.area
                #get area of intersection inside gridcell with perimeter
                gridbox_intersection_area = perimeter_polygon.intersection(gridbox).area

                #write out if gridbox_intersection_area not equal to 0.0 (i.e. have some intersection)
                if gridbox_intersection_area != 0.0:
                    #get fraction of intersection inside gridcell
                    gridbox_intersection_fraction = 1./(gridbox_area/gridbox_intersection_area)
                    #write out
                    relevant_grid_indices.append([lon_i,lat_i,gridbox_intersection_fraction])

    return relevant_grid_indices

#----------------------------------------------------------------------------------------#