#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#quality_control_functions.py

#define functions pertaining to quality control checks.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import bisect
from calendar import monthrange
from collections import Counter
from crossnetwork_processing_functions import datetime_to_fractional_minutes_from_reference, fractional_minutes_from_reference_to_datetime, local_offset_from_utc, get_unique_longitude_latitude_pairs
import datetime
from GHOST_standards import standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code, standard_classification_flag_name_to_classification_code
import itertools
import numpy as np
import pandas as pd
import scipy.stats as st
import statsmodels.stats.stattools as sm

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def do_qa(run_details, parameter_details, metadata, data):

    '''define pipeline of functions that apply data quality checks to data from each site'''
    
    #get unique longitude and latitude pairs across metadata, with associated metadata/data indices of each pair
    unique_pair_dictionary = get_unique_longitude_latitude_pairs(metadata, data)

    #missing measurements check
    data = check_missing_measurement(parameter_details, data)
    
    #negative measurement check (only do check for gas and aerosol matrices)
    if (parameter_details['matrix'] == 'gas') or (parameter_details['matrix'] == 'aerosol'):
        data = check_negative_measurement(parameter_details, data)
    
    #zero values check (only do check for gas and aerosol matrices)
    if (parameter_details['matrix'] == 'gas') or (parameter_details['matrix'] == 'aerosol'):
        data = check_zero_values(parameter_details, data)
    
    #infinite values check
    data = check_infinite_values(parameter_details, data)

    #preliminary data check
    data = check_preliminary_data(data)
    
    #LOD checks (only do check for gas and aerosol matrices)
    if (parameter_details['matrix'] == 'gas') or (parameter_details['matrix'] == 'aerosol'):
        data = check_lod(parameter_details, metadata, data)
    
    #extreme data - scientifically non-feasible check
    data = check_extreme_scientifically_non_feasible(parameter_details, data)
    
    #extreme data - manually decreed check
    data = check_extreme_manual(run_details, metadata, data)
    
    #recurring value check
    data = check_recurring_values(parameter_details, data)
    
    #measurement resolution check - by reported metadata
    data = check_measurement_resolution_by_metadata(parameter_details, metadata, data)
    
    #run qa checks which process data by year/month (UTC)
    #first temporal data is split out in a standalone module and this module then calls the qa checks - this is done for code optimisation
    #the qa checks called are: 
    #temporal factor checks (non-integer local timezone check, weekday/weekend check)
    #measurement resolution check - empirical
    #extreme data - distributional outlier check
    data = iterate_by_yearmonth(parameter_details, metadata, data, unique_pair_dictionary)
    
    return data 
    
###--------------------------------------------------------------------------------------------------###
#DEFINE QUALITY CHECK FUNCTIONS
###--------------------------------------------------------------------------------------------------###

def check_missing_measurement(parameter_details, data):
    
    '''Check if have missing measurements (i.e NaN), or have network flag stating missing measurement (possible in cases where have data actually reported, but network states it as being missing using a data flag)'''

    #get data flag code for indicating missing values
    missing_flag_code = standard_data_flag_name_to_data_flag_code['Missing Data']

    #check for missing values by evaluating if each measurement value is np.NaN
    invalid_inds_empirical = list(np.arange(len(data[parameter_details['bsc_parameter_name']]))[pd.isnull(data[parameter_details['bsc_parameter_name']])])
    #check missing values by evaluating missing data flags
    invalid_inds_missing_flags = [ii for ii,flag_codes in enumerate(data['flag']) if missing_flag_code in flag_codes] 

    #get unique indices from the multiple lists
    invalid_inds = list(set(invalid_inds_empirical + invalid_inds_missing_flags))

    #update relevant inds with qa codes 
    if len(invalid_inds) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['Missing Measurement']

        for inv in invalid_inds:
            data['qa'][inv].append(qa_flag_code)
            
    return data 

###------------------------------------------------------------------------------------###

def check_negative_measurement(parameter_details, data):
    
    '''Check if have negative measurements (i.e < 0.0), or have network flags stating negative measurements'''
    
    #get data flag codes for indicating negative values (untouched or modified (i.e. estimated) )
    negative_value_flag_code = standard_data_flag_name_to_data_flag_code['Negative Value']
    negative_value_estimated_flag_code = standard_data_flag_name_to_data_flag_code['Estimated Data - Negative Value Detected']

    #check negative values by evaluating if each measurement value is < 0.0
    invalid_inds_empirical = list(np.where(data[parameter_details['bsc_parameter_name']] < 0.0)[0])
    #check negative values by evaluating negative value data flags
    invalid_inds_negative_flags = [ii for ii,flag_codes in enumerate(data['flag']) if negative_value_flag_code in flag_codes]
    invalid_inds_negative_estimated_flags = [ii for ii,flag_codes in enumerate(data['flag']) if negative_value_estimated_flag_code in flag_codes]
    
    #get unique indices from the multiple lists
    invalid_inds = list(set(invalid_inds_empirical + invalid_inds_negative_flags + invalid_inds_negative_estimated_flags))
    
    #update relevantinds with qa codes 
    if len(invalid_inds) > 0:
        #get relevant qa flag code to write
        qa_flag_code = standard_QA_name_to_QA_code['Negative Measurement']

        #set list of qa flags to check have not written first before writing qa flag
        check_do_not_have_flag_names = ['Missing Measurement']
        check_do_not_have_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in check_do_not_have_flag_names]

        for inv in invalid_inds:
            
            #check have not written certain other qa flags before writing
            if not set(data['qa'][inv]).isdisjoint(check_do_not_have_flag_codes) == False:
                data['qa'][inv].append(qa_flag_code)
            
    return data 

###------------------------------------------------------------------------------------###

def check_zero_values(parameter_details, data):

    '''Check if have values == 0, or have network flags stating zero measurements (i.e. no values detected)'''

    #get data flag codes for indicating zero values (untouched or modified (i.e. estimated) )
    zero_value_flag_code = standard_data_flag_name_to_data_flag_code['No Value Detected']
    zero_value_estimated_flag_code = standard_data_flag_name_to_data_flag_code['Estimated Data - No Value Detected']

    #check zero values by evaluating if each measurement value is == 0.0
    invalid_inds_empirical = list(np.where(data[parameter_details['bsc_parameter_name']] == 0.0)[0])
    #check zero values by evaluating zero value data flags
    invalid_inds_zero_flags = [ii for ii,flag_codes in enumerate(data['flag']) if zero_value_flag_code in flag_codes]
    invalid_inds_zero_estimated_flags = [ii for ii,flag_codes in enumerate(data['flag']) if zero_value_estimated_flag_code in flag_codes]
    
    #get unique indices from the multiple lists
    invalid_inds = list(set(invalid_inds_empirical + invalid_inds_zero_flags + invalid_inds_zero_estimated_flags))
    
    #update relevant inds with qa codes 
    if len(invalid_inds) > 0:
        #get relevant qa flag code to write 
        qa_flag_code = standard_QA_name_to_QA_code['Zero Measurement']

        #set list of qa flags to check have not written first before writing qa flag
        check_do_not_have_flag_names = ['Missing Measurement','Negative Measurement']
        check_do_not_have_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in check_do_not_have_flag_names]

        for inv in invalid_inds:
            
            #check have not written certain other qa flags before writing
            if not set(data['qa'][inv]).isdisjoint(check_do_not_have_flag_codes) == False:
                data['qa'][inv].append(qa_flag_code)
            
    return data 

###------------------------------------------------------------------------------------###

def check_infinite_values(parameter_details, data):

    '''Check if have infinite values (i.e. np.inf).
       This happens when values are outside of the range that the float32 data type can handle (-3.4E+38 to +3.4E+38)
    '''

    invalid_inds = np.arange(len(data[parameter_details['bsc_parameter_name']]))[np.isinf(data[parameter_details['bsc_parameter_name']])]
    
    #update relevant invalid inds with qa codes 
    if len(invalid_inds) > 0:
        #get relevant qa flag code to write
        qa_flag_code = standard_QA_name_to_QA_code['Infinite Value']

        #set list of qa flags to check have not written first before writing qa flag
        check_do_not_have_flag_names = ['Missing Measurement','Negative Measurement','Zero Measurement']
        check_do_not_have_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in check_do_not_have_flag_names]

        for inv in invalid_inds:
            
            #check have not written certain other qa flags before writing
            if not set(data['qa'][inv]).isdisjoint(check_do_not_have_flag_codes) == False:
                data['qa'][inv].append(qa_flag_code)
             
    return data 

###------------------------------------------------------------------------------------###

def check_preliminary_data(data):

    '''Check all data which is flagged in the data provider flags as preliminary.
       Flag all of this data with its own QA flag.
    '''

    #get data flag code for "Preliminary Data"
    preliminary_data_flag_code = standard_data_flag_name_to_data_flag_code['Preliminary Data']

    #get all instances where data flags contain code for 'Preliminary Data'
    invalid_inds = [ii for ii,flag_codes in enumerate(data['flag']) if preliminary_data_flag_code in flag_codes]

    #update relevant inds with qa code
    if len(invalid_inds) > 0:
        #get relevant qa flag code to write 
        qa_flag_code = standard_QA_name_to_QA_code['Preliminary Data']

        for inv in invalid_inds:
            data['qa'][inv].append(qa_flag_code)

    return data

###------------------------------------------------------------------------------------###

def check_lod(parameter_details, metadata, data):

    '''iterate through measured data, and check if the values are <=/>= the reported/documented limits of detection (LODs)
       documented LODs are static and apply for the whole use time of measurement instrument
       the network reported limits of detection are either variable in time, or as a fixed value for an entire data range.
       additionally network providers often give flags stating if data is <=/>= detection limits
       these reported flags are first used in preference over testing if measurements are above/below reported static/variable detection limits
       variable reported LODs are then taken preferentially over reported static LODs where both are available.
       preferential checks are evaluated by first trying to evaluate measurements with reported LODs, or if not available, then with documented LODs 
       if have no lower/upper reported or documented LODs or flags for a given period, these periods are flagged.
       any measurements which have NaN values are not flagged with any flags by this function
    '''
    
    #get relevant qa flags
    below_lod_documented_qa_flag_code = standard_QA_name_to_QA_code['Below Documented Lower Limit of Detection']
    below_lod_reported_qa_flag_code = standard_QA_name_to_QA_code['Below Reported Lower Limit of Detection']
    below_lod_preferential_qa_flag_code = standard_QA_name_to_QA_code['Below Preferential Lower Limit of Detection']
    no_below_lod_metadata_qa_flag_code = standard_QA_name_to_QA_code['No Documented/Reported Lower Limit of Detection Metadata']
    above_lod_documented_qa_flag_code = standard_QA_name_to_QA_code['Above Documented Upper Limit of Detection']
    above_lod_reported_qa_flag_code = standard_QA_name_to_QA_code['Above Reported Upper Limit of Detection']
    above_lod_preferential_qa_flag_code = standard_QA_name_to_QA_code['Above Preferential Upper Limit of Detection']
    no_upper_lod_metadata_qa_flag_code = standard_QA_name_to_QA_code['No Documented/Reported Upper Limit of Detection Metadata']

    #get needed data/metadata
    static_lower_lod_documented = metadata['measuring_instrument_documented_lower_limit_of_detection']
    static_lower_lod_reported = metadata['measuring_instrument_reported_lower_limit_of_detection']
    static_upper_lod_documented = metadata['measuring_instrument_documented_upper_limit_of_detection']
    static_upper_lod_reported = metadata['measuring_instrument_reported_upper_limit_of_detection']
    update_metadata_stamp = metadata['meta_update_stamp']

    #get relevant standard data flag codes
    below_lod_data_flag_code = standard_data_flag_name_to_data_flag_code['Value Below Detection Limit']
    estimated_below_lod_data_flag_code = standard_data_flag_name_to_data_flag_code['Estimated Data - Value Below Detection Limit']
    above_lod_data_flag_code = standard_data_flag_name_to_data_flag_code['Value Above Detection Limit']
    estimated_above_lod_data_flag_code = standard_data_flag_name_to_data_flag_code['Estimated Data - Value Above Detection Limit']

    #iterate through #iterate through all time points
    for ii in range(len(data['time'])):
        
        #get measurement value for timepoint
        current_data = data[parameter_details['bsc_parameter_name']][ii]
        
        #if current data is NaN, skip to next value
        if pd.isnull(current_data) == True:
            continue

        #set flags to state if have evaluated validity of measurements through comparison with reported/documented LODs 
        evaluated_lower_lod_reported = False
        evaluated_upper_lod_reported = False 
        evaluated_lower_lod_documented = False
        evaluated_upper_lod_documented = False 

        #set flags when have written reported/documented flags as measurements are below/above reported/documented LODs
        written_lower_lod_reported = False
        written_upper_lod_reported = False 
        written_lower_lod_documented = False
        written_upper_lod_documented = False 

        #preferentially check if have reported network data flags informing of hitting of LOD
        current_flag_codes = data['flag'][ii]
        
        #if have network flag stating measurement is <=/>= detection limits, write reported qa flag
        #have flag stating measurement is <= detection limit?
        if (below_lod_data_flag_code in current_flag_codes) or (estimated_below_lod_data_flag_code in current_flag_codes):
            data['qa'][ii].append(below_lod_reported_qa_flag_code)
            evaluated_lower_lod_reported = True
            written_lower_lod_reported = True
        #have flag stating measurement is >= detection limit?
        if (above_lod_data_flag_code in current_flag_codes) or (estimated_above_lod_data_flag_code in current_flag_codes):
            data['qa'][ii].append(above_lod_reported_qa_flag_code)
            evaluated_upper_lod_reported = True
            written_upper_lod_reported = True
                    
        #get utc time in minutes
        current_minutes = data['time'][ii]        
        #get index of current valid metadata for timepoint
        current_meta_index = bisect.bisect(update_metadata_stamp,current_minutes)-1
        #use metadata index to get static documented limit metadata
        current_lower_lod_documented = static_lower_lod_documented[current_meta_index]
        current_upper_lod_documented = static_upper_lod_documented[current_meta_index]
        
        #-----------------------------------------------------
        #test measurements by documented LODs

        #is documented lower lod available?
        if pd.isnull(current_lower_lod_documented) == False:
    
            #is measurement below or equal to documented lower limit of detection?
            if current_data <= current_lower_lod_documented:
                data['qa'][ii].append(below_lod_documented_qa_flag_code)
                written_lower_lod_documented = True

            #set flag that have evaluated documented lower LOD
            evaluated_lower_lod_documented = True

        #is documented upper lod available?
        if pd.isnull(current_upper_lod_documented) == False:            

            #is measurement above or equal to documented upper limit of detection?
            if current_data >= current_upper_lod_documented:
                data['qa'][ii].append(above_lod_documented_qa_flag_code)
                written_upper_lod_documented = True  

            #set flag that have evaluated documented upper LOD
            evaluated_upper_lod_documented = True

        #-----------------------------------------------------
        
        #test measurements by reported LODs (if not evaluated yet)
        #preferentially evaluate by variable reported LOD, and then static LOD

        #have not evaluated reported lower LOD yet?
        if evaluated_lower_lod_reported == False:

            #get variable reported lower LOD
            current_variable_lower_lod_reported = data['reported_lower_limit_of_detection_per_measurement'][ii]
        
            #is variable reported lower lod available?
            if pd.isnull(current_variable_lower_lod_reported) == False:
 
                #is measurement below or equal to reported lower limit of detection?
                if current_data <= current_variable_lower_lod_reported:
                    data['qa'][ii].append(below_lod_reported_qa_flag_code)
                    written_lower_lod_reported = True

                #set flag that have evaluated reported lower LOD
                evaluated_lower_lod_reported = True

            #if variable reported LOD not available, evaluate using static reported LOD (if available)
            else:
            
                #get static reported lower LOD
                current_static_lower_lod_reported = static_lower_lod_reported[current_meta_index]

                #is static reported lower lod available?
                if pd.isnull(current_static_lower_lod_reported) == False:

                    #is measurement below or equal to reported lower limit of detection?
                    if current_data <= current_static_lower_lod_reported:
                        data['qa'][ii].append(below_lod_reported_qa_flag_code)
                        written_lower_lod_reported = True

                    #set flag that have evaluated reported lower LOD
                    evaluated_lower_lod_reported = True

        #have not evaluated reported upper LOD yet?
        if evaluated_upper_lod_reported == False:

            #get variable reported upper LOD
            current_variable_upper_lod_reported = data['reported_upper_limit_of_detection_per_measurement'][ii]
        
            #is variable reported upper lod available?
            if pd.isnull(current_variable_upper_lod_reported) == False:
 
                #is measurement above or equal to reported lower limit of detection?
                if current_data >= current_variable_upper_lod_reported:
                    data['qa'][ii].append(above_lod_reported_qa_flag_code)
                    written_upper_lod_reported = True

                #set flag that have evaluated reported upper LOD
                evaluated_upper_lod_reported = True

            #if variable reported LOD not available, evaulate using static reported LOD (if available)
            else:
            
                #get static reported upper LOD
                current_static_upper_lod_reported = static_upper_lod_reported[current_meta_index]

                #is static reported upper lod available?
                if pd.isnull(current_static_upper_lod_reported) == False:

                    #is measurement above or equal to reported upper limit of detection?
                    if current_data >= current_static_upper_lod_reported:
                        data['qa'][ii].append(above_lod_reported_qa_flag_code)
                        written_upper_lod_reported = True

                    #set flag that have evaluated reported upper LOD
                    evaluated_upper_lod_reported = True

        #-----------------------------------------------------  
        #write preferential flags that are set by first trying to evaluate measurements compared with reported LOD, or if not available then with documented LOD
        #also write flags for informing when have no lower/upper LOD information        


        #have evaluated reported lower LOD?
        if evaluated_lower_lod_reported == True:
            #have written reported lower LOD flag?
            if written_lower_lod_reported == True:  
                data['qa'][ii].append(below_lod_preferential_qa_flag_code)
            
        #else, have evaluated documented lower LOD?   
        elif evaluated_lower_lod_documented == True:
            #have written documented lower LOD flag?
            if written_lower_lod_documented == True: 
                data['qa'][ii].append(below_lod_preferential_qa_flag_code)

        #else if have not evaluated measurements with neither of reported/documented lower LODs, write flag
        else:
            data['qa'][ii].append(no_below_lod_metadata_qa_flag_code)
            
        #have evaluated reported upper LOD?
        if evaluated_upper_lod_reported == True:
            #have written reported upper LOD flag?
            if written_upper_lod_reported == True:  
                data['qa'][ii].append(above_lod_preferential_qa_flag_code)
            
        #else, have evaluated documented upper LOD?   
        elif evaluated_upper_lod_documented == True:
            #have written documented upper LOD flag?
            if written_upper_lod_documented == True: 
                data['qa'][ii].append(above_lod_preferential_qa_flag_code)

        #if have not evaluated measurements with neither of reported/documented upper LODs, write flag
        else:
            data['qa'][ii].append(no_upper_lod_metadata_qa_flag_code)

        #-----------------------------------------------------
        
    return data

###------------------------------------------------------------------------------------###

def check_extreme_scientifically_non_feasible(parameter_details, data):

    '''do check for extreme data that is scientifically non-feasible (i.e lower/greater than scientifically feasible).
       for each different measured parameter, minimum and maximums value are set, if data is lower/greater than these values, it is flagged 
       NaNs not caught by this check
    '''

    #get relevant qa flag code
    qa_flag_code = standard_QA_name_to_QA_code['Extreme Data - Scientifically Non-Feasible']

    #iterate through measured values
    for ii in range(len(data[parameter_details['bsc_parameter_name']])):
        
        #if measured value is lower/greater than the set lower/upper limits, flag data
        if (data[parameter_details['bsc_parameter_name']][ii] < parameter_details['extreme_lower_limit']) or (data[parameter_details['bsc_parameter_name']][ii] > parameter_details['extreme_upper_limit']):
            #write out qa flag
            data['qa'][ii].append(qa_flag_code)
            
    return data
    
###-----------------------------------------------------------------------------------###

def check_extreme_manual(run_details, metadata, data):

    '''function for manually flagging extreme data
       if data has been manually found to be extreme, and has not been flagged by other extreme flags, then it should be manually flagged. 

       input is in the following format:
       extremes = {'parameter-station_reference':'YYYYMMDDHHMM-YYYYMMDDHHMM', 'station_reference2':YYYYMMDDHHMM-YYYYMMDDHHMM;YYYYMMDDHHMM-YYYYMMDDHHMM}
      
       if more than 1 bad period for a reference, it is separated by a semicolon.
       all data inside the given datetime ranges for a station reference is flagged as extreme (if have data for this window)
    '''

    #get relevant qa flag code
    qa_flag_code = standard_QA_name_to_QA_code['Extreme Data - Manually Decreed']

    #dictionary for definition of manually decreed extreme data windows
    extremes = {}
    
    #get all unique parameter-station references keys
    unique_keys = list(extremes.keys())
    
    #iterate through keys
    if len(unique_keys) > 0:
        for key in unique_keys:
            #separate into parameter, station reference 
            key_split = key.split('-')
            current_parameter, current_station_reference = key[0],key[1]
            
            #if are currently processing parameter, and station reference proceed
            if (current_parameter == run_details['parameter']) & (current_station_reference == metadata['station_reference'][0]):
                
                #get all reference times for current station_reference/parameter processed
                reference_time = data['time']
                
                #split extreme window ranges by ; (if more than 1 extreme window)
                extreme_ranges = extremes[key].split(';')
                
                #iterate through extreme window ranges
                for extreme_range in extreme_ranges:
                    #break apart range string
                    extreme_range_split = extreme_range.split('-')
                    start_dt_string, end_dt_string = extreme_range_split[0],extreme_range_split[1]
                    #convert dt strings to python datetimes
                    start_dt = datetime.datetime(start_dt_string[:4],start_dt_string[4:6],start_dt_string[6:8],start_dt_string[8:10],start_dt_string[10:])
                    end_dt = datetime.datetime(end_dt_string[:4],end_dt_string[4:6],end_dt_string[6:8],end_dt_string[8:10],end_dt_string[10:])
                    #get start time/end time in reference time
                    start_reference_time = datetime_to_fractional_minutes_from_reference(start_dt) 
                    end_reference_time = datetime_to_fractional_minutes_from_reference(end_dt) 
                
                    #flag all indices with start times inside extreme window range
                    extreme_inds = np.where((reference_time >= start_reference_time) & (reference_time < end_reference_time))[0]
                    
                    #if have times inside extreme window, then write out flags for all measurements inside window
                    if len(extreme_inds) > 0:
                        for inv in extreme_inds:
                            data['qa'][inv].append(qa_flag_code)
                        
    return data
    
###------------------------------------------------------------------------------------###

def check_recurring_values(parameter_details, data):

    '''do check for persistently recurring values 
       persistently recurring values are symptomatic of when an instrument hits the detection limit, or is malfunctioning.

       check is done by using a moving window of 9 measurements. If 7/9 of values in the window are a physical value (i.e. not NaN), and the same value, flag values as recurring.

       check is only done for measurements with a temporal resolution of 1 hour or finer, and when there is a gap between measurements of no greater than 3 hours across all whole moving window of measurements.
    '''

    #remove data with a temporal resolution of coarser than 1 hour
    valid_inds = np.array([False]*len(data['temporal_resolution']))
    for ii, temporal_resolution in enumerate(data['temporal_resolution']):         
        temporal_resolution = int(temporal_resolution)
        if temporal_resolution <= (60):
            valid_inds[ii] = True 
    valid_data = data[parameter_details['bsc_parameter_name']][valid_inds]
    
    #get the gaps between each measurement start/endpoints
    valid_data_startpoints = data['time'][valid_inds]
    valid_temporal_resolution = data['temporal_resolution'][valid_inds]
    valid_data_endpoints = valid_data_startpoints + valid_temporal_resolution
    valid_data_measurement_window_gaps = valid_data_startpoints[1:] - valid_data_endpoints[:-1]
    valid_data_measurement_window_gaps = np.insert(valid_data_measurement_window_gaps, 0, 0)

    #generate list of lists of moving window indices
    window_width = 9
    total_n = len(valid_data)    
    moving_window_indices = [np.arange(n,n+window_width) for n in range((total_n-window_width)+1)]

    #iterate though windows
    invalid_inds = []
    for window in moving_window_indices:
        #Ensure total gaps between between measurements in moving window does sum 3 hours or more
        if np.sum(valid_data_measurement_window_gaps[window][1:]) <= 180:

            #get maximum count of values in valid value list - np.NaN values are not counted additively (i.e. will be multiple nan counts of 1)
            max_occuring_value,n_occurances = Counter(valid_data[window]).most_common(1)[0]
            
            #if 7 or more of values are duplicates, flag values
            if n_occurances >= 7:
                invalid_inds.append(window[valid_data[window]==max_occuring_value])
             
    #remove duplicate values from invalid inds array
    invalid_inds = list(itertools.chain.from_iterable(invalid_inds))
    invalid_inds = list(set(invalid_inds))
    
    #update relevant invalid inds with qa codes 
    if len(invalid_inds) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['Recurring Value']
        for inv in invalid_inds:
            data['qa'][inv].append(qa_flag_code)
            
    return data
    
###-----------------------------------------------------------------------------------###

def check_measurement_resolution_by_metadata(parameter_details, metadata, data):

    '''check if measurement resolution is coarser than a set limit (variable by parameter measured). This is checked in 3 ways: by taking documented and reported metadata, and also empirically. This is handled separately by 2 different functions.
       this function handles the documented and reported metadata checks.
       function simply takes documented and reported measurement resolutions and checks if the resolutions are coarser than a set limit (variable by parameter measured). Any occasions where these are coarser are flagged.  
       also any occasions where there is no documented or reported metadata are flagged.
       a preferential check is also done where first data is flagged by evaluating by reported measurement resolutions, and if not available, then by documented measurement resolutions
    '''
    
    #get key metadata
    reported_measurement_resolutions = metadata['measuring_instrument_reported_measurement_resolution']
    documented_measurement_resolutions = metadata['measuring_instrument_documented_measurement_resolution']
    update_metadata_stamps = metadata['meta_update_stamp']

    invalid_inds_reported = []
    invalid_inds_documented = []
    invalid_inds_preferential = []
    invalid_inds_nometa = []

    for ii in range(len(reported_measurement_resolutions)):

        #set flags to state when have evaluated reported/documented measurement resolutions
        evaluated_reported_measurement_resolution = False
        evaluated_documented_measurement_resolution = False
        
        #set flags to state when have written reported/documented measurement resolution qa flags
        written_reported_measurement_resolution = False 
        written_documented_measurement_resolution = False 

        #get reported/documented measurement resolutions
        reported_measurement_resolution = reported_measurement_resolutions[ii]
        documented_measurement_resolution = documented_measurement_resolutions[ii]

        #get data indices for period (in case period needs flagging)
        #is it last index? 
        if ii == (len(reported_measurement_resolutions)-1):
            invalid_inds = range(bisect.bisect_left(data['time'], update_metadata_stamps[ii]), len(data['time']))
        else:
            invalid_inds = range(bisect.bisect_left(data['time'], update_metadata_stamps[ii]), bisect.bisect_left(data['time'], update_metadata_stamps[ii+1]))
                
        #do you have valid reported measurement resolution?
        if pd.isnull(reported_measurement_resolution) == False:

            #is reported measurement_resolution coarser than the set minimum resolution for the parameter being processed?
            if reported_measurement_resolution > parameter_details['minimum_resolution']:
                invalid_inds_reported.extend(invalid_inds)
                written_reported_measurement_resolution = True
            
            #set flag to state you have evaluated reported measurement resolution 
            evaluated_reported_measurement_resolution = True

        #do you have valid documented measurement resolution?
        if pd.isnull(documented_measurement_resolution) == False:

            #is documented measurement_resolution coarser than the set minimum resolution for the parameter being processed?
            if documented_measurement_resolution > parameter_details['minimum_resolution']:
                invalid_inds_documented.extend(invalid_inds)
                written_documented_measurement_resolution = True

            #set flag to state you have evaluated documented measurement resolution 
            evaluated_documented_measurement_resolution = True

        #--------------------------------------

        #do preferential evaluation, first evaluating the reported measurement resolution if available, and if not then by documented measurement resolution 
        #have evaluated by reported measurement resolution?
        if evaluated_reported_measurement_resolution == True:
            #have written reported measurement resolution flag?
            if written_reported_measurement_resolution == True:
                invalid_inds_preferential.extend(invalid_inds)

        #else if, have evaluated by documented measurement resolution?
        elif evaluated_documented_measurement_resolution == True:
            #have written documented measurement resolution flag?
            if written_documented_measurement_resolution == True:
                invalid_inds_preferential.extend(invalid_inds)
                
        #else, do you have no valid documented or reported measurement resolution metadata?
        else:
            invalid_inds_nometa.extend(invalid_inds)
                
    #----------------------------------------
      
    #update relevant invalid inds with qa codes 
      
    if len(invalid_inds_reported) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['Insufficient Measurement Resolution - Reported']
        for inv in invalid_inds_reported:
            data['qa'][inv].append(qa_flag_code)
      
    if len(invalid_inds_documented) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['Insufficient Measurement Resolution - Documented']
        for inv in invalid_inds_documented:
            data['qa'][inv].append(qa_flag_code)

    if len(invalid_inds_preferential) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['Insufficient Measurement Resolution - Preferential']
        for inv in invalid_inds_preferential:
            data['qa'][inv].append(qa_flag_code)

    if len(invalid_inds_nometa) > 0:
        #get relevant qa flag code
        qa_flag_code = standard_QA_name_to_QA_code['No Documented/Reported Measurement Resolution Metadata']
        for inv in invalid_inds_nometa:
            data['qa'][inv].append(qa_flag_code)


    return data

###------------------------------------------------------------------------------------###

def check_temporal_factors(metadata, data, day_monthly_inds, daytohour_inds, unique_pair_dictionary):

    '''functions that checks many temporally associated factors of measurements:

       non-integer local timezone
       -----------------------------
       check that determines if the local timezone of measurement station is non-integer - important for the temporal alignment of periodic processes (e.g. when using spectral analysis)
       check is done in 2 ways:
 
       A 
       check alignment of starts of measurement of measurement periods with UTC hours (all those with a temporal resolution of equal to or coarser than 1 hour, and finer than 1 day) 
       this is done by getting remainder of start times divided by 60
       all times with remainder not equal to 0 are not aligned with UTC start hours, this implies the local timezone is non-integer.
       HOWEVER, sometimes times are manually shifted by small amounts by the data providers (i.e. 30 minute) to align perfectly with UTC hours. These cases should be attempted to be flagged also, which is what the next test does

       B
       a timezone package is used to determine the local timezone UTC offset via latitude/longitude pairings. The check is performed in the following steps:
       1. iterates through all lat/lon pairings, getting all relevant measurement start times within each metadata window 
       2. groups all relevant start times by year+month, and for each group get local timezone offset (in hours) relative to a UTC datetime (set as 00:00 UTC on 1st of each month)
       3. if offset is non-integer, writes out a flag for all relevant indices in the monthly window.
       The check is only done for data with a measurement resolution finer than 1 day
       ** it is important to note that it is possible that flagged measurements taken in places with non-integer local timezones are recorded truly aligned with UTC time, however this information is impossible to know.
       -----------------------------

       weekday/weekend
       -----------------------------
       checks if the time of measurement is a weekday or weekend (by local time). The check is performed in the following steps:
       1. iterates through all lat/lon pairings, getting all relevant measurement start times within each metadata window 
       2. groups all relevant start times by year+month, and for each group get local timezone offset (in hours) relative to a UTC datetime (set as 00:00 UTC on 1st of each month)
       3. iterate through all start measurement times in year/month and get centre of measurement window time
       4. Take UTC offset from centre measurement time to get a local time
       5. Find what day of week the centre local time is in, and if the day is weekday or weekend (as it is centre of measurement window local time, if 51% of measurement window is weekday - the window will be classed as weekday)
       The check is only done for data with a measurement resolution finer than 1 day
    '''

    #get relevant qa flags
    qa_flag_ni_code = standard_QA_name_to_QA_code['Non-Integer Local Timezone (relative to UTC)']
    #get relevant classification flags
    classification_flag_weekday_code = standard_classification_flag_name_to_classification_code['Weekday']
    classification_flag_weekend_code = standard_classification_flag_name_to_classification_code['Weekend']

    #--------------------------------------------------------------#
    #check alignment of starts of measurement of measurement periods with UTC hours (all those with a temporal resolution of equal to or coarser than 1 hour, and finer than 1 day) 
    
    #get remainder of entire time array divided by 60
    time_modulus = np.mod(data['time'][daytohour_inds],60)
    #all times with remainder not equal to 0 are not aligned with UTC hour starts and ends, therefore flag
    non_matching_time_inds = np.where(time_modulus != 0)[0]
    if len(non_matching_time_inds) != 0:
        for inv in non_matching_time_inds:
            data['qa'][inv].append(qa_flag_ni_code)
            
    #--------------------------------------------------------------#
    #iterate through unique longitude/latitude pairs
    for unique_pair_ii, unique_pair in unique_pair_dictionary.items(): 

        #get current unique longitude/latitude
        current_longitude = unique_pair['longitude']
        current_latitude = unique_pair['latitude']

        #get data indices associated with current unique longitude/latitude pair
        relevant_data_indices = unique_pair['data_indices']

        #iterate through year/month datetime groups 
        #cut day_monthly_inds groups to be those that are also valid inds in the metadata window 
        for monthly_inds in day_monthly_inds:
        
            #if first relevant data index is after the last ind of monthly window, skip to next month
            if relevant_data_indices[0] > monthly_inds[-1]:
                continue
            #if last relevant data index is before the start ind of the monthly window, break out of processing
            if relevant_data_indices[-1] < monthly_inds[0]:
                break
            
            #get relevant data inds in monthly window
            start_relevant_ind = bisect.bisect_left(relevant_data_indices,monthly_inds[0])
            end_relevant_ind = bisect.bisect(relevant_data_indices,monthly_inds[-1])
            #cut out relevant data indices in month
            relevant_monthly_data_indices = relevant_data_indices[start_relevant_ind:end_relevant_ind]
            if len(relevant_monthly_data_indices) != 0:

                #convert relevant times in month to datetimes     
                relevant_datetimes = []
                for current_minutes in data['time'][relevant_monthly_data_indices]:
                    relevant_datetimes.append(fractional_minutes_from_reference_to_datetime(float(current_minutes)))
                
                #now check if the local UTC offset is integer or not

                #create datetime at 00:00 UTC on 1st of month
                first_of_month_dt = relevant_datetimes[0].replace(hour=0, day=1)
        
                #get local UTC offset in hours relative to the 00:00 UTC on 1st of month for each latitude/longitude pairing
                pytz_str, local_utc_offset = local_offset_from_utc(first_of_month_dt,longitude=current_longitude,latitude=current_latitude)

                #is the hour offset integer or not?
                if local_utc_offset.is_integer() == False:
                
                    #limit relevant inds to those that do not have qa_flag from the first hour alignment check
                    rel_i = np.array([ii for ii,val in enumerate(data['qa'][relevant_monthly_data_indices]) if qa_flag_ni_code not in val])

                    #write qa flag
                    for inv in rel_i:
                        data['qa'][inv].append(qa_flag_ni_code)

                #check if each time is weekday/weekend and check season (Northern Hemisphere seasons)
                #iterate through each relevant datetime in month
                for ii,relevant_ind in enumerate(relevant_monthly_data_indices):
                    relevant_datetime = relevant_datetimes[ii]
                    relevant_temporal_resolution = data['temporal_resolution'][relevant_ind]
                    centre_timedelta = datetime.timedelta(minutes=float(relevant_temporal_resolution)/2.)      
                    #get datetime in centre of measurement window
                    centre_datetime = relevant_datetime + centre_timedelta

                    #add the previously calculated UTC offset to the centre window datetime
                    local_datetime = centre_datetime + datetime.timedelta(hours=local_utc_offset)   

                    #determine if centre window datetime is weekday or weekend
                    #get n_day of week, where Monday = 0 and Sunday = 6
                    n_day_of_week = local_datetime.weekday()
                    #weekday
                    if n_day_of_week < 5:
                        #write out qa flag code
                        data['classification'][relevant_ind].append(classification_flag_weekday_code)
                    #weekend
                    else:
                        #write out qa flag code
                        data['classification'][relevant_ind].append(classification_flag_weekend_code)
        
    return data

###------------------------------------------------------------------------------------###

def check_measurement_resolution_empirical(parameter_details, data, hour_monthly_inds):

    '''check if measurement resolution is coarser than a set limit (variable by parameter measured). This is checked in 2 ways: by reported metadata, and empirically. Handled separately by 2 different functions.
       this function is the empirical check
       empirically determine measurement resolution of measurements in grouped monthly windows (UTC time).     
       if the minimum difference between observations in the monthly window is coarser than a set limit (variable by measured parameter), flag the measurements
       check accounts for the improvement of measurement resolution over time (as happens with improved methods and instrumentation over time)
       **check is only done for measurements resolutions of 1 hour or finer, and when have >= 108 valid values (i.e. not NaN) and not decreed invalid by qa flags, in the monthly window (~15% of hourly values in a month), to ensure have sufficient values to fairly return a true minimum difference value
       measurements that start in one group and end in the next are treated as belonging to the first group consistently.
    '''

    #get relevant qa flag code
    qa_flag_code = standard_QA_name_to_QA_code['Insufficient Measurement Resolution - Empirical']
    
    #iterate through grouped data and determine minimum difference between observations
    for xx, grouped_inds in enumerate(hour_monthly_inds):
        grouped_inds = np.array(grouped_inds)
        
        #if have valid data proceed (i.e. have data with sufficient temporal resolution) 
        if len(grouped_inds) > 0:
                
            #get relevant data
            valid_monthly_data = data[parameter_details['bsc_parameter_name']][grouped_inds]
            
            #remove invalid data (decreed from qa flags)
            invalid_qa_flag_names = ['Missing Measurement', 'Infinite Value', 'Negative Measurement', 'Zero Measurement', 'Preliminary Data', 'Invalid Data Provider Flags - GHOST Decreed', 'Recurring Value', 'Extreme Data - Scientifically Non-Feasible', 'Extreme Data - Manually Decreed', 'Insufficient Measurement Resolution - Preferential', 'Below Preferential Lower Limit of Detection', 'Above Preferential Upper Limit of Detection', 'Erroneous Primary Sampling', 'Erroneous Sample Preparation', 'Erroneous Measurement Methodology']
            invalid_qa_flag_codes = np.array([standard_QA_name_to_QA_code[invalid_flag_name] for invalid_flag_name in invalid_qa_flag_names], dtype=np.uint8)
            #find all indices where none of invalid qa codes are associated with measurements
            valid_mask = np.array([all(x not in qa_flag_codes for x in invalid_qa_flag_codes) for qa_flag_codes in data['qa'][grouped_inds]])
            valid_monthly_data = valid_monthly_data[valid_mask]
            
            #check number of total measurements reported in the annual/month window
            #if < 108 values (~15% of hourly values in a month - with an average 30 days)
            if len(valid_monthly_data) < 108:
                continue
        
            #otherwise, check minimum difference between observations
            else:
            
                #initiate minimum difference
                minimum_difference = np.inf
        
                #remove duplicate values from array
                valid_monthly_data = list(set(valid_monthly_data))
    
                #sort array in ascending order
                valid_monthly_data = sorted(valid_monthly_data)                    
 
                #Find the min diff by comparing adjacent pairs in sorted array
                for ii in range(len(valid_monthly_data)-1):
                    #is difference smaller than current minimum difference
                    if (valid_monthly_data[ii+1] - valid_monthly_data[ii]) < minimum_difference:
                        #if so set new minimum difference
                        minimum_difference = valid_monthly_data[ii+1] - valid_monthly_data[ii]

                #if minimum difference between all points for the monthly window is coarser than the set minimum resolution for the parameter being processed, flag the measurements.
                if minimum_difference > parameter_details['minimum_resolution']:
                    for inv in grouped_inds:
                        data['qa'][inv].append(qa_flag_code)
    
    return data

###------------------------------------------------------------------------------------###

def check_extreme_distributional_outlier(parameter_details, data, hour_monthly_inds):
    
    '''do check for extreme data based on distribution of data.
       check is done by each UTC month
       check is only done for measurements resolutions of 1 hour or finer, and when have >= 108 valid values (i.e. not NaN) and not decreed invalid by qa flags, in the monthly window (~15% of hourly values in a month), to ensure have sufficient values to get extreme values
       measurements that start in one group and end in the next are treated as belonging to the first group consistently.

       1. apply adjusted boxplot to determine outliers, robustly for normal and non-normal distributions.

       Technique designed to identify outliers based on an adapted version of the classical Tukey boxplot.
       Tukey's boxplot is very popular tool for detection of outliers. It reveals the location, spread and skewness of the data.
       It works nicely for detection of outliers when the data are symmetric.
       When the data are skewed it covers boundary away from the whisker on the compressed side while declares erroneous outliers on the extended side of the distribution.
       Hubert and Vandervieren (2008) made adjustment to Tukey's technique to overcome this problem.
       They apply a robust measure of skewness to Tukey's technique: the 'Medcouple', incorporated as a power of exponential times some constant on left and right as -3.5 and 4 changing position depending upon sign of medcouple. Incorporating this function condenses the interval from narrow side and extends the interval towards the puffy tail.
       However another problem arises that is the adjusted boxplot constructs the interval of critical values which even exceeds from the extremes of the data.
       In this situation the adjusted boxplot is unable to detect outliers.
       Adil and Irshad (2015) provide a solution for this:

       Where the lower and upper critical limits are defined as:
       [Lf Uf] = [Q1 - 1.5*IQR*e-SK*|MC|, Q3 + 1.5*IQR*eSK*|MC|]
       where, Q1 = the 25th percentile, Q3 = the 75th percentile, IQR = the interquartile range, SK = the classical skewness, MC = the medcouple

       **A restriction is also imposed here that if the classical skewness, SK, is greater than 3.5 then it should be treated as 3.5.
       The reason to fix maximum level of skewness 3.5 is to avoid the problem of constructing the large interval of critical values due classical skewness that might be higher than 3.5.
       Not allowing the skewness statistic to exceed 3.5 synchronize the interval of critical value with the data sets as against the adjusted box plot and prevents the interval to be very large in case of highly skewed distributions.
       It also constructs smaller interval in case of moderately skewed distributions
       The 1.5 factor is found to be too conservative for air quality data, and is multiplied by 7 to give a factor of 10.5, which is found empirically to be much more appropriate. 

       See References here:
       http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.882.3197&rep=rep1&type=pdf
       https://en.wikipedia.org/wiki/Medcouple
       https://en.wikipedia.org/wiki/Box_plot#Variations

       2. Take data flagged as outlier by the adjusted boxplot
    '''

    #get relevant qa flag code
    qa_flag_code = standard_QA_name_to_QA_code['Extreme Data - Distributional Outlier']
    
    #iterate through grouped data 
    for xx, grouped_inds in enumerate(hour_monthly_inds):
        grouped_inds = np.array(grouped_inds)
        
        #get relevant data
        valid_monthly_data = data[parameter_details['bsc_parameter_name']][grouped_inds]
        
        #remove invalid data (decreed from qa flags)
        invalid_qa_flag_names = ['Missing Measurement', 'Infinite Value', 'Negative Measurement', 'Zero Measurement', 'Preliminary Data', 'Invalid Data Provider Flags - GHOST Decreed', 'Recurring Value', 'Extreme Data - Scientifically Non-Feasible', 'Extreme Data - Manually Decreed', 'Insufficient Measurement Resolution - Preferential', 'Insufficient Measurement Resolution - Empirical', 'Below Preferential Lower Limit of Detection', 'Above Preferential Upper Limit of Detection', 'Erroneous Primary Sampling', 'Erroneous Sample Preparation', 'Erroneous Measurement Methodology']
        invalid_qa_flag_codes = [standard_QA_name_to_QA_code[invalid_flag_name] for invalid_flag_name in invalid_qa_flag_names]
        valid_mask = np.array([all(x not in qa_flag_codes for x in invalid_qa_flag_codes) for qa_flag_codes in data['qa'][grouped_inds]])
        valid_monthly_data = valid_monthly_data[valid_mask]
        valid_monthly_inds = grouped_inds[valid_mask]
        
        #check number of total measurements reported in the annual/month window
        #if < 108 values (~15% of hourly values in a month - with an average 30 days), continue to next month
        if len(valid_monthly_data) < 108:
            continue
    
        #otherwise, check for distributional extreme values
        else:    

            #caluclate medcouple
            mc = sm.medcouple(valid_monthly_data)
            #calculate absolute value of medcouple
            mc_abs = np.abs(mc)
            

            #caluclate the 25th percentile (p1)
            p25 = np.percentile(valid_monthly_data, 25)
            #caluclate the 75th percentile (p1)
            p75 = np.percentile(valid_monthly_data, 75)

            #calculate the interquartile range
            iqr = p75-p25

            #caluclate classical skewness
            skew = st.skew(valid_monthly_data)
            #limit skewness to to bet between -3.5 and 3.5
            if (skew > 3.5):
                skew = 3.5
            elif (skew < -3.5):
                skew = -3.5

            #calculate lower/upper critical limits
            lower_critical = p25 - 10.5 * iqr * np.exp(-skew*mc_abs)
            upper_critical = p75 + 10.5 * iqr * np.exp(skew*mc_abs)

            #check which data is below adjusted boxplot upper limit and above lower limit
            invs = np.where(np.logical_or(valid_monthly_data<lower_critical, valid_monthly_data>upper_critical))[0]
            invalid_inds = valid_monthly_inds[invs]
            
            #write out invalid indices
            if len(invalid_inds) > 0:
                for inv in invalid_inds:
                    data['qa'][inv].append(qa_flag_code)
            
    return data
                    
###------------------------------------------------------------------------------------###

def iterate_by_yearmonth(parameter_details, metadata, data, unique_pair_dictionary):
 
    '''function that groups time by year/month (UTC) for passing to multiple qa functions.
       also handles filtering of times which do not match up the required temporal resolutions required for the specific functions
    '''

    hour_monthly_inds = []
    day_monthly_inds = []
    daytohour_inds = []
    current_year = -9999
    current_month = -99

    for ii,start_time in enumerate(data['time']):
        #append values for 3 separate temporal resolution instances, for temporal resolutions of 1 hour or finer (hour), for temporal resolutions of finer than 1 day (day), and for resolutions as fine as 1 hour and less coarse than 1 day (daytohour)
        #the first two are grouped by year/month (UTC), and the last is appended all together

        #get datetime
        current_dt = fractional_minutes_from_reference_to_datetime(float(start_time))
        #is datetime in a new year/monthly window?
        if (current_dt.year != current_year) or (current_dt.month != current_month):
            hour_monthly_inds.append([])
            day_monthly_inds.append([])
            current_year = current_dt.year
            current_month = current_dt.month
        
        #evaluate temporal resolution groupings
        temporal_resolution = data['temporal_resolution'][ii]
        if temporal_resolution < (60*24):
            day_monthly_inds[-1].append(ii)
        if (temporal_resolution >= 60) or (temporal_resolution < (60*24)):
            daytohour_inds.append(ii)
        if temporal_resolution <= (60):
            hour_monthly_inds[-1].append(ii)

    #remove empty lists from hour_monthly_inds/day_monthly_inds
    hour_monthly_inds = [x for x in hour_monthly_inds if x != []]
    day_monthly_inds = [x for x in day_monthly_inds if x != []]
    
    #-----------------------------------------

    #pass grouped year/month indices to qa functions for processing
    
    #first convert daytohour_inds to numpy array
    daytohour_inds = np.array(daytohour_inds, dtype=np.uint32)
    
    #multiple temporal factor checks (non-integer local timezone check, weekday/weekend check)
    data = check_temporal_factors(metadata, data, day_monthly_inds, daytohour_inds, unique_pair_dictionary)
    
    #measurement resolution check
    data = check_measurement_resolution_empirical(parameter_details, data, hour_monthly_inds)
    
    #extreme data - distributional outlier check
    data = check_extreme_distributional_outlier(parameter_details, data, hour_monthly_inds)

    return data
        
    
###------------------------------------------------------------------------------------###

def check_representation(n_minutes_represented, all_gaps, period, qa_data, relevant_inds, current_dt):

    '''function for temporal representation checks of data (after screening by key QA flags). 
       representation checks are handled for 5 different temporal periods: hourly, diurnal, weekly, monthly, annual
       hourly period:  start/end NN:00 UTC
       daily period:   start/end NN 00:00 UTC
       weekly period:  start/end defined by UTC isocalendar
       monthly period: start/end NN:01 00:00 UTC
       annual period:  start/end NNNN:01:01 00:00 UTC

       two types are checks are done to flag the representativity of each of the different periods.
       first is Data Completeness, i.e. what percentage across the period is represented by valid data (after screening by key QA flags)?
       second is Maximum Data Gap, i.e. what is the maximum data gap percentage in the provided valid data across the period (after screening by key QA flags)? 
    
       multiple separate limits are evaluated by, for each of the checks, to give flexibility in the definitions of temporal representativity.
       For % data completeness these are if the % completeness is < : 90%, 75%, 66%, 50%, 25%
       For % maximum data gap, these are if the % maximum data gap is >= : 10%, 25%, 33%, 50%, 75%
    '''

    #set data completeness percentage limits to check by
    #i.e. if in data period, < than limit % is represented in window, data is flagged 
    data_completeness_percentage_limits = [25,50,66,75,90]
    
    #set maximum data gap percentage limits to check by
    #i.e. if have max data gap percentage of >= than limit % in window, data is flagged
    maximum_data_gap_percentage_limits = [10,25,33,50,75] 
    
    #get maximum gap in all_gaps array (if a gap exists)
    if len(all_gaps) == 0:
        max_temporal_gap = 0
    else:   
        max_temporal_gap = np.max(all_gaps)
    
    #get maximum possible minutes by period 
    if period == 'hourly':
        max_minutes_possible = 60
    elif period == 'daily':
        max_minutes_possible = 60*24
    elif period == 'weekly':
        max_minutes_possible = (60*24)*7
    elif period == 'monthly':
        current_year = current_dt.year
        current_month = current_dt.month
        days_in_month = monthrange(current_year, current_month)[1]
        max_minutes_possible = (60*24)*days_in_month
    elif period == 'annual':
        current_year = current_dt.year
        d1 = datetime.date(current_year, 1, 1) 
        d2 = datetime.date(current_year + 1, 1, 1) 
        days_in_year = (d2 - d1).days 
        max_minutes_possible = (60*24)*days_in_year

    #100./max_minutes_possible (for multiplying by)
    percent_max_minutes = 100./max_minutes_possible

    #calculate % of period represented by valid data
    percent_period_represented = percent_max_minutes*n_minutes_represented

    #calculate % max data gap across period
    percent_period_max_data_gap = percent_max_minutes*max_temporal_gap

    #-------------------------------------------------------------------------------------

    #iterate through data completeness % limits
    for data_completeness_percentage_limit in data_completeness_percentage_limits:
        
        #does percent of period represented < data completeness % limit?
        if percent_period_represented < data_completeness_percentage_limit:
            
            #get relevant qa flag code
            qa_flag_code = standard_QA_name_to_QA_code['{} Window Data Completeness < {}%'.format(period.capitalize(), data_completeness_percentage_limit)]
            for inv in relevant_inds:
                qa_data[inv].append(qa_flag_code)

    #-------------------------------------------------------------------------------------
    
    #iterate through maximum data gap % limits
    for maximum_data_gap_percentage_limit in maximum_data_gap_percentage_limits:
        
        #is maximum data gap % in period >= maximum data gap % limit?
        if percent_period_max_data_gap >= maximum_data_gap_percentage_limit:

            #get relevant qa flag code
            qa_flag_code = standard_QA_name_to_QA_code['{} Window Maximum Data Gap >= {}%'.format(period.capitalize(), maximum_data_gap_percentage_limit)]
            for inv in relevant_inds:
                qa_data[inv].append(qa_flag_code)

    #-------------------------------------------------------------------------------------
    
            
    return qa_data
    
###------------------------------------------------------------------------------------###