#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#crossnetwork_processing_functions.py

#module which stores functions which handle general GHOST processing routines across networks. 

###--------------------------------------------------------------------------------------------------###
###IMPORT CORE MODULES
###--------------------------------------------------------------------------------------------------###

import bisect
import copy
import csv
import datetime
import glob
import numpy as np
import os
import pandas as pd
import re
import subprocess 
import sys
import time
import logging

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def create_greasy_arguments_file(GHOST_level, run_details, arguments, GHOST_stage=''):
        
    '''function which creates greasy arguments text file storing all different tasks to run by greasy'''

    #define list to store chunked argument files
    argument_files = []

    #define greasy arguments filenames
    if GHOST_level == 'pipeline':
        argument_files.append('{}/run/pipeline_arguments/{}.txt'.format(run_details['process_root'],run_details['slurm_job_ID'])) 
    elif GHOST_level == 'stage':
        #get the CPU chunk size -- set initially as miniumum number of arguments per file 
        N_arguments_per_file_minimum = copy.deepcopy(run_details['chunk_size'])
        N_arguments_per_file = copy.deepcopy(run_details['chunk_size'])
        #divide the number of arguments by the CPU chunk size, to determine how many argument files will be needed to submit all jobs
        N_submit_files = int(np.ceil(len(arguments)/run_details['chunk_size']))
        #set argument remainder as 0 initially
        argument_remainder = 0
        #if the number of argument files is greater than the job array limit (i.e the limit on the number of argument files that can be processed simultaneously)
        #then add adjust minimum N arguments per file appropriately (i.e. split extra arguments across the maximum number of argument files evenly)        
        if N_submit_files > run_details['job_array_limit']:
            #update the minimum number of arguments per file
            N_arguments_per_file_minimum = int(np.floor(len(arguments)/run_details['job_array_limit']))
            N_arguments_per_file = copy.deepcopy(N_arguments_per_file_minimum)
            #if the number of extra arguments does not divide equally between all files get the remainder
            argument_remainder = int(len(arguments)%run_details['job_array_limit'])
            #if have argument remainder then update N_arguments_per_file variable to be 1 greater than minimum for first file written (and for all files thereafter until  remainder is accounted for)
            if argument_remainder > 0:
                N_arguments_per_file = N_arguments_per_file_minimum + 1
                #subtract 1 from the argument remainder
                argument_remainder -= 1
            #set N submit files as N of job array limit
            N_submit_files = copy.deepcopy(run_details['job_array_limit'])

        #create file which will store a list of all stage argument filenames
        greasy_file = open('{}/run/stage_arguments/{}_{}/{}.grz'.format(run_details['process_root'],run_details['network'],run_details['parameter'],GHOST_stage), 'w')
        #create all stage argument filenames
        for ii in range(N_submit_files):
            stage_argument_fname = '{}/run/stage_arguments/{}_{}/{}_{}.txt'.format(run_details['process_root'],run_details['network'],run_details['parameter'],GHOST_stage,ii)
            argument_files.append(stage_argument_fname)
            greasy_file.write('{}\n'.format(stage_argument_fname))
        greasy_file.close()

    #initialise variables to count N argument files written and N lines written
    file_ii = 0
    n_lines_written = 0
    #define special characters that will need to be escaped in string written        
    special_characters = ['(',')']

    #open first arguments file for writing
    arguments_file = open(argument_files[file_ii], 'w')

    #iterate through different arguments, writing line by line to arguments file
    for arguments_ii, line in enumerate(arguments):
        if type(line) == list:
            str_to_write = ' '.join([str(elem) for elem in line])
        else:
            str_to_write = '{}'.format(line)

        #escape certain special characters in str_to_write
        for ch in special_characters:
            str_to_write = str_to_write.replace(ch,'\{}'.format(ch))
        
        #write arguments str to current file
        if GHOST_level == 'pipeline':
            arguments_file.write('python -u {}/processing_modules/GHOST_pipeline.py {}\n'.format(run_details['process_root'],str_to_write))
        elif GHOST_level == 'stage':
            #affix slurm_job_ID, network and parameter to str_to_write 
            str_to_write = '{} {} {} {}'.format(run_details['slurm_job_ID'],run_details['network'],run_details['parameter'],str_to_write) 
            arguments_file.write('python -u {}/processing_modules/execute_{}.py {}\n'.format(run_details['process_root'],GHOST_stage,str_to_write))

            #iterate n lines written    
            n_lines_written += 1
        
            #if have written sufficient arguments to file (and not currently processing last argument)
            #then close current file and open another
            if (n_lines_written == N_arguments_per_file) & (arguments_ii < (len(arguments)-1)):
                #reset n lines written
                n_lines_written = 0
                #close current arguments file
                arguments_file.close()
                #iterate n arguments files written
                file_ii += 1
                #if have argument remainder, write an extra argument to the next file
                if argument_remainder > 0:
                    N_arguments_per_file = N_arguments_per_file_minimum + 1
                    #subtract 1 from the argument remainder
                    argument_remainder -= 1
                #otherwise N arguments to write to next file should be minimum
                else:
                    N_arguments_per_file = copy.deepcopy(N_arguments_per_file_minimum)
                #open new arguments file
                arguments_file = open(argument_files[file_ii], 'w')
                
    #close current arguments file
    arguments_file.close()

###------------------------------------------------------------------------------------###

def submit_greasy_job(GHOST_level, job_name, run_details, GHOST_stage='', pipeline_start_time='', pipeline_logfile=''):

    '''function which submits greasy job to process GHOST pipeline/stage'''

    #define directory where jobs are run from 
    run_dir = run_details['process_root']+'/run'

    #define submit/log/argument directories
    if GHOST_level == 'pipeline':
        submit_dir = '{}/run/submit_pipeline'.format(run_details['process_root'])
        log_dir =    '{}/run/pipeline_logs'.format(run_details['process_root'])
        arguments_dir = '{}/run/pipeline_arguments'.format(run_details['process_root'])
    elif GHOST_level == 'stage':
        submit_dir = '{}/run/submit_stage/{}_{}'.format(run_details['process_root'],run_details['network'],run_details['parameter'])
        log_dir =    '{}/run/stage_logs/{}_{}/{}'.format(run_details['process_root'],run_details['network'],run_details['parameter'],GHOST_stage)
        arguments_dir = '{}/run/stage_arguments/{}_{}'.format(run_details['process_root'],run_details['network'],run_details['parameter'])

    #write slurm script for submission of greasy job
    slurm_fname = create_slurm_submission_script(GHOST_level, job_name, run_dir, submit_dir, arguments_dir, run_details, GHOST_stage)

    #start timer on job
    start_job_time = time.time()

    #submit slurm script
    submit_complete = False
    while submit_complete == False:

        submit_process =  subprocess.Popen(['sbatch', '-A', 'bsc32', '{}'.format(slurm_fname)], cwd=submit_dir, stdout=subprocess.PIPE)
        submit_status = submit_process.communicate()[0]
        submit_return_code = submit_process.returncode

        #if sbatch fails, time out for 60 seconds and then try again
        if submit_return_code != 0:
            time.sleep(60)
            continue
        else:
            submit_complete = True

        #take a 1 second pause after submittal (to help slurm scheduler)
        time.sleep(1)

    #now job has been submitted
    #monitor job in queue (every 10 seconds) until there are 0 left in the squeue
    all_tasks_finished = False

    while all_tasks_finished == False:
        cmd = ['squeue', '-h', '-n', '{}'.format(job_name)]
        squeue_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, encoding='utf8')
        squeue_status = squeue_process.communicate()[0]
        squeue_return_code = squeue_process.returncode

        #if squeue fails, time out for 60 seconds and then try again
        if squeue_return_code != 0:
            time.sleep(60)
            continue
        else:
            n_jobs_in_queue = len(squeue_status.split('\n')[:-1])
            #if number of jobs in queue > 0, then sleep for 10 seconds and then check again how many jobs there are in queue
            if n_jobs_in_queue > 0:
                time.sleep(10)
                continue
            else:
                all_tasks_finished = True

    #get total job time
    end_job_time = time.time()
    total_job_time = (end_job_time - start_job_time)/60.

    #if job no longer in the squeue, then now check the outcome of the job (different dependent if it is a pipeline or stage job)
   
    if GHOST_level == 'pipeline':
        check_GHOST_pipeline_outcome(total_job_time, run_details, log_dir)

    elif GHOST_level == 'stage':
        #if GHOST stage == T1, the valid temporal resolution codes of the processed jobs are returned
        if GHOST_stage == 'T1': 
            stage_process_time, valid_temporal_resolution_codes = check_GHOST_stage_outcome(GHOST_stage, pipeline_logfile, arguments_dir, log_dir, pipeline_start_time, end_job_time, total_job_time)
            stage_queue_time = total_job_time - stage_process_time
            return stage_process_time, stage_queue_time, valid_temporal_resolution_codes
        else:
            stage_process_time = check_GHOST_stage_outcome(GHOST_stage, pipeline_logfile, arguments_dir, log_dir, pipeline_start_time, end_job_time, total_job_time)
            stage_queue_time = total_job_time - stage_process_time
            return stage_process_time, stage_queue_time

###------------------------------------------------------------------------------------###

def create_slurm_submission_script(GHOST_level, job_name, run_dir, submit_dir, arguments_dir, run_details, GHOST_stage):

    '''function that writes a slurm submission shell script that submits a greasy job'''

    #get current BSC machine
    machine = os.environ['BSC_MACHINE']

    #create slurm_fname (job_name + 'sh.')
    slurm_fname = job_name+'.sh'
    
    #get all relevant argument files for current GHOST level 
    if GHOST_level == 'pipeline':
        argument_files = ['{}/{}.txt'.format(arguments_dir,run_details['slurm_job_ID'])]
    elif GHOST_level == 'stage':
        argument_files = sorted(glob.glob('{}/{}_*.txt'.format(arguments_dir,GHOST_stage)), key=natural_key)

    #read how many lines are in first arguments file
    with open(argument_files[0]) as f: 
        for ii, line in enumerate(f):
            pass
        N_arguments = ii + 1

    #set the number of simultaneous tasks to run based on the GHOST level 
    #if the GHOST level == 'pipeline', then cap the number of simultaneously running tasks to be 5 
    if GHOST_level == 'pipeline':
        max_tasks = 5
    #else, if the GHOST level == 'stage', then cap the number of simultaneously running tasks to be the defined CPU chunk size  
    elif GHOST_level == 'stage':
        max_tasks = run_details['chunk_size']

    #if the number of arguments is > capped max tasks, then set N simultaneous tasks to be the max tasks permitted
    if N_arguments > max_tasks:
        n_simultaneous_tasks = copy.deepcopy(max_tasks) 
    #else, if the number of arguments is <= capped max tasks, then set N simultaneous tasks to be N arguments
    else:
        n_simultaneous_tasks = copy.deepcopy(N_arguments)

    #if processing get maximum walltime for specific GHOST level
    if GHOST_level == 'pipeline':
        #if QOS == 'debug', then it is always 2 hours
        if run_details['qos_pipeline'] == 'debug':
            max_walltime = '02:00:00'
        #otherwise, it should be 48 hours
        elif run_details['qos_pipeline'] in ['bsc_es','prace']:
            max_walltime = '48:00:00'
    elif GHOST_level == 'stage':
        #if QOS == 'debug', then it is always 2 hours
        if run_details['qos_stage'] == 'debug':
            max_walltime = '02:00:00'
        elif run_details['qos_stage'] in ['bsc_es','prace']:
            max_walltime = max_walltime_for_GHOST_stage(GHOST_stage)

    #create slurm submission script for specific GHOST stage
    file = open(submit_dir+'/'+slurm_fname,"w")

    file.write("#!/bin/bash\n")
    file.write("\n")
    file.write("#SBATCH --job-name={}\n".format(job_name))
    file.write("#SBATCH --ntasks={}\n".format(n_simultaneous_tasks))
    #fix number of nodes to be 1 (for faster execution)
    file.write("#SBATCH --nodes 1\n")
    file.write("#SBATCH --time={}\n".format(max_walltime))
    if GHOST_level == 'pipeline':
        file.write("#SBATCH --qos={}\n".format(run_details['qos_pipeline']))
    elif GHOST_level == 'stage':
        file.write("#SBATCH --qos={}\n".format(run_details['qos_stage']))
        file.write("#SBATCH --array=1-{}\n".format(len(argument_files)))
    #if machine is power9, then there are 4 threads per physical CPU, which show as 4 separate CPUs. 
    #if multithreading is enabled, ensure 4 virtual cores are actually seen as 1 CPU.
    if (machine == 'power') & (run_details['multithreading'] == True):
        file.write("#SBATCH --cpus-per-task=4\n")
    file.write("##SBATCH --output=/dev/null\n")
    file.write("##SBATCH --error=/dev/null\n")      
    file.write("\n")
    file.write("source {}/load_modules.sh\n".format(run_dir))
    file.write("export GREASY_ENGINE=basic\n")
    file.write("export GREASY_NWORKERS=$SLURM_NPROCS\n") 
    if GHOST_level == 'pipeline':
        file.write("export GREASY_LOGFILE={}/{}.log\n".format(submit_dir,job_name))
        file.write("\n")
        file.write("greasy {}".format(argument_files[0]))
    elif GHOST_level == 'stage':
        file.write("export GREASY_LOGFILE={}/{}_$SLURM_ARRAY_TASK_ID.log\n".format(submit_dir,job_name))
        file.write("arguments_store={}/{}.grz\n".format(arguments_dir,GHOST_stage))
        file.write("argument_file=$(cat $arguments_store | awk -v var=$SLURM_ARRAY_TASK_ID 'NR==var {print $1}')\n")
        file.write("\n")
        file.write("greasy $argument_file")
    
    #close file
    file.close()

    #return slurm fname
    return slurm_fname

###------------------------------------------------------------------------------------###

def max_walltime_for_GHOST_stage(GHOST_stage):
    
    '''function that returns the max walltime for a specific GHOST stage'''

    max_walltime_per_stage = {'S1':'02:00:00',
                              'S2':'02:00:00',
                              'S3':'02:00:00',
                              'S4':'02:00:00',
                              'S5':'02:00:00',
                              'S6':'02:00:00',
                              'T1':'02:00:00',
                              'T2':'04:00:00'
                             }
                        
    return max_walltime_per_stage[GHOST_stage]

###------------------------------------------------------------------------------------###

def check_GHOST_pipeline_outcome(total_pipeline_time, run_details, pipeline_log_dir):
        
    '''function that checks the outcome of all submitted GHOST pipeline jobs'''

    #create lists for storing pipeline references relating to pipeline outcomes
    success = []
    no_data = []
    not_finished_stage = [] 
    not_finished_pipeline = []   
    failed_caught = []
    failed_uncaught = []

    #get all output pipeline log filenames
    output_pipeline_log_files = [] 
    for network in run_details['networks']:
        for parameter in run_details['parameters']:
            output_pipeline_log_files.append('{}/{}_{}.out'.format(pipeline_log_dir,network, parameter))
            
    #iterate though files
    for output_pipeline_log_file in output_pipeline_log_files:
        #get specific pipeline reference
        pipeline_reference = output_pipeline_log_file.split('/')[-1].split('.out')[0]

        #test if output log file exists for pipeline (if not write to not finished pipeline list)
        if os.path.isfile(output_pipeline_log_file) == False:
            not_finished_pipeline.append(pipeline_reference)
         
        #otherwise, output log file exists
        else:
            #get second last line of output
            line = subprocess.check_output(['tail', '-2', output_pipeline_log_file], encoding='utf8').strip()
        
            #analyse pipeline outcome

            #second last line should have the following string inside: 'PROCESSING OF ALL DESIRED STAGES FOR' if the pipeline has completed successfully 
            if 'PROCESSING OF ALL DESIRED STAGES FOR' in line:
                success.append(pipeline_reference)

            #if the second last line contains the string: 'NOT PROCESSED AS HAVE NO', then the pipeline has terminated early due to a lack of data
            elif 'NOT PROCESSED AS HAVE NO' in line:
                no_data.append(pipeline_reference)
    
            #if the second last line contains the string: 'BUT 0 STATIONS REMAIN', then the pipeline has terminated early due to a lack of data 
            elif 'BUT 0 STATIONS REMAIN' in line:
                no_data.append(pipeline_reference)

            #if the second last line contains the string: 'WITH CAUGHT ERROR, then the pipeline has terminated early due to caught error/s
            elif 'WITH CAUGHT ERROR' in line:
                failed_caught.append(pipeline_reference)

            #if the second last line contains the string: 'UNCAUGHT ERROR', then the pipeline has terminated early due to uncaught error/s
            elif 'UNCAUGHT ERROR' in line:
                failed_uncaught.append(pipeline_reference)

            #if the second last line contains the string: 'DID NOT FINISH DUE TO UNFINISHED JOB', then the pipeline has terminated early due stage job/s not terminating before wallclock
            elif 'DID NOT FINISH DUE TO UNFINISHED JOB' in line:
                not_finished_stage.append(pipeline_reference)

    #open a summary log file to report on GHOST processing output
    summary_file= open("{}/run/summary_logs/{}.out".format(run_details['process_root'],run_details['slurm_job_ID']),"w")
    summary_file.write("SUBMITTED JOB TO PROCESS DATA ({}-{}) FOR {} DIFFERENT PARAMETER/S ({}), IN {} DIFFERENT NETWORK/S ({}).\nTHESE WERE PROCESSED IN {} UNIQUE PIPELINE/S.\n".format(run_details['start_date'], run_details['end_date'], len(run_details['parameters']), run_details['parameters'], len(run_details['networks']), run_details['networks'], len(run_details['parameters'])*len(run_details['networks'])))
    summary_file.write("\nTOTAL RUN TIME: {:.2f} MINUTES.\n".format(total_pipeline_time)) 
    summary_file.write("\n{} PIPELINE/S COMPLETED SUCCESSFULLY\n".format(len(success))) 
    if len(no_data) > 0:
        summary_file.write("{} PIPELINE/S TERMINATED EARLY DUE TO A LACK OF DATA.\nTHESE ARE: {}\n".format(len(no_data),no_data))
    if len(failed_caught) > 0:
        summary_file.write("{} PIPELINE/S TERMINATED EARLY DUE TO CAUGHT ERROR/S.\nTHESE ARE: {}\n".format(len(failed_caught),failed_caught))
    if len(failed_uncaught) > 0:
        summary_file.write("{} PIPELINE/S TERMINATED EARLY DUE TO UNCAUGHT ERROR/S.\nTHESE ARE: {}\n".format(len(failed_uncaught),failed_uncaught))
    if len(not_finished_stage) > 0:
        summary_file.write("{} PIPELINE/S TERMINATED EARLY DUE TO STAGE JOBS NOT FINISHING BEFORE WALLCLOCK.\nTHESE ARE: {}\n".format(len(not_finished_stage),not_finished_stage))           
    if len(not_finished_pipeline) > 0:
        summary_file.write("{} PIPELINE/S TERMINATED EARLY DUE TO PIPELINE NOT FINISHING BEFORE WALLCLOCK.\nTHESE ARE: {}\n".format(len(not_finished_pipeline),not_finished_pipeline))           

    #close summary log file
    summary_file.close()

###------------------------------------------------------------------------------------###

def check_GHOST_stage_outcome(GHOST_stage, pipeline_logfile, arguments_dir, stage_log_dir, pipeline_start_time, stage_end_time, total_stage_time):

    '''function that checks the outcome of a GHOST stage.

       all tasks processed in a GHOST stage return a log file prefixed with an appropriate process code

       these process codes are:
       0: Process completed without issue
       1: Process completed, but station no longer continuing in pipeline
       2: Caught error in process
       3: Uncaught error in process

       The current pipeline is terminated if there are any 2 or 3 process codes.
       The current pipeline is terminated if all parallel tasks return a process code of 1.
       Additionally the current pipeline is terminated if a submitted task has not returned a log file  
       Otherwise, the pipeline is continued.

       Also the process time and queue time of stage are calculated.
    '''

    #gather all returned process codes in a list from the output log files 
    process_codes = []

    #create list to store process times per successful task (only of successful jobs)
    task_process_times = []

    #create lists to store references of jobs that fail for different reasons
    success_job_references = []
    non_continuing_job_references = []
    caught_error_job_references = []
    non_caught_error_job_references = []
    non_finished_job_references = []

    #create list of all output stage log root filenames should be output for stage
    #gather all of the argument files used to submit jobs and extract job references used to prefix log files for specific stage
    argument_files = sorted(glob.glob('{}/{}**.txt'.format(arguments_dir, GHOST_stage)))     
    job_references = []
    for argument_file in argument_files:
        with open(argument_file) as f:
            csv_reader = csv.reader(f, delimiter=' ') 
            for line in csv_reader:
                if (GHOST_stage == 'T1') or (GHOST_stage == 'T2'):
                    job_references.append('{}_{}'.format(line[6], line[7]))
                else:
                    job_references.append('{}'.format(line[6]))

    #iterate though derived job references
    for job_reference in job_references:

        #get actual output stage log filename 
        output_stage_log_file = glob.glob('{}/*{}.out'.format(stage_log_dir,job_reference))

        #if output log file does not exists, write to non finished job list
        if len(output_stage_log_file) == 0:
            non_finished_job_references.append(job_reference)
        #otherwise, a log file has been output (i.e. job has finished)
        else:
            output_stage_log_file = output_stage_log_file[0]
            output_stage_log_file_name = output_stage_log_file.split('/')[-1] 
            #get process_code 
            process_code = int(output_stage_log_file_name[:1])
            process_codes.append(process_code)

            #append task references to appropriate lists by process codes
            if process_code == 0:
                success_job_references.append(job_reference)
                process_time = subprocess.check_output(['tail', '-1', output_stage_log_file], encoding='utf8').strip()
                task_process_times.append(float(process_time))
            elif process_code == 1:
                non_continuing_job_references.append(job_reference)
            elif process_code == 2:
                caught_error_job_references.append(job_reference)
            elif process_code == 3:
                non_caught_error_job_references.append(job_reference) 
    
    #convert list to numpy array
    process_codes = np.array(process_codes, dtype=np.int)       

    #if all the process codes returned are 1 (no stations continuing in pipeline)
    if np.all(process_codes == 1):
        pipeline_logfile.write('{} COMPLETED ON {:.2f} MINUTES (BUT 0 STATIONS REMAIN).\nTERMINATING PIPELINE HERE'.format(GHOST_stage,(stage_end_time-pipeline_start_time)/60.))
        pipeline_logfile.close()
        sys.exit()

    #if a process code of 2 is returned in any of the jobs (caught error)
    elif 2 in process_codes:
        pipeline_logfile.write('{} COMPLETED ON {:.2f} MINUTES (WITH CAUGHT ERROR/S). TERMINATING PIPELINE HERE\nJOB REFERENCES THAT FAILED ARE: {}'.format(GHOST_stage,(stage_end_time-pipeline_start_time)/60.,caught_error_job_references))
        pipeline_logfile.close()
        sys.exit()

    #if one of the process codes are 3 (uncaught error)
    elif 3 in process_codes:
        pipeline_logfile.write('{} DID NOT FINISH DUE TO UNCAUGHT ERROR/S. TERMINATING PIPELINE HERE\nJOB REFERENCES THAT FAILED ARE: {}'.format(GHOST_stage,non_caught_error_job_references))
        pipeline_logfile.close()
        sys.exit()

    #if one of the jobs did not finish
    elif len(non_finished_job_references) > 0:
        pipeline_logfile.write('{} DID NOT FINISH DUE TO UNFINISHED JOB/S. TERMINATING PIPELINE HERE\nJOB REFERENCES THAT DID NOT FINISH ARE: {}'.format(GHOST_stage,non_finished_job_references))
        pipeline_logfile.close()
        sys.exit()

    #if one of the process codes are 1 (i.e. one of stations no longer continues in the pipeline)
    elif 1 in process_codes:
        pipeline_logfile.write('{} COMPLETED ON {:.2f} MINUTES ({:.2f} MINUTES TO PROCESS STAGE: {:.2f} MINUTES PROCESSING, {:.2f} MINUTES QUEUING).\nSOME STATIONS ARE NOT CONTINUING IN PIPELINE: {}\n'.format(GHOST_stage,(stage_end_time-pipeline_start_time)/60.,total_stage_time,np.max(task_process_times),total_stage_time-np.max(task_process_times),non_continuing_job_references))

    #otherwise, all process codes are 0 (all processed successfully)
    else:
        pipeline_logfile.write('{} COMPLETED SUCCESSFULLY ON {:.2f} MINUTES ({:.2f} MINUTES TO PROCESS STAGE: {:.2f} MINUTES PROCESSING, {:.2f} MINUTES QUEUING)\n'.format(GHOST_stage,(stage_end_time-pipeline_start_time)/60.,total_stage_time,np.max(task_process_times),total_stage_time-np.max(task_process_times)))

    #calculate the process time for the stage (taken to be the maximum process time across all successful tasks)
    stage_process_time = np.max(task_process_times)

    #if stage == T1, get all the temporal output resolutions that some station data has been successfully processed for 
    if GHOST_stage == 'T1':
        valid_temporal_resolution_codes = np.unique([success_job_reference.split('_')[-1] for success_job_reference in success_job_references])
        return stage_process_time, valid_temporal_resolution_codes

    else:
        return stage_process_time


###------------------------------------------------------------------------------------###

def create_GHOST_stage_logfile(GHOST_stage, run_details, process_code, message, param_to_parallelise_by):
    
    '''function which creates a logfile for stating outcome of a GHOST stage.
       the filename is prefixed with a code referencing the stage outcome + param_to_parallelise_by: process_code_param_to_parallelise_by.out
       the process codes are:
       0: Process completed without issue
       1: Process completed, but station no longer continuing in pipeline
       2: Caught error in process
       3: Uncaught error in process
    '''
    f=open('{}/run/stage_logs/{}_{}/{}/{}_{}.out'.format(run_details['process_root'],run_details['network'],run_details['parameter'],GHOST_stage,process_code,param_to_parallelise_by),"w")
    f.write(str(message))
    f.close() 

    #exit from current process after writing logfile
    sys.exit()
    
###------------------------------------------------------------------------------------###

def get_run_dictionaries(slurm_job_ID, network, parameter):

    '''return relevant specific dictionaries for specific network/parameter configuration'''

    from GHOST_standards import standard_parameters, standard_networks
    import importlib

    #Read configuration file (as was originally set when pipeline was launched)
    cwd = os.getcwd()
    if '/run' in cwd:
        sys.path.append(cwd.split('/run')[0]+'/run/GHOST_configurations')
    elif '/processing_modules' in cwd:
        sys.path.append(cwd.split('/processing_modules')[0]+'/run/GHOST_configurations')
    GHOST_configuration = importlib.import_module('configuration_{}'.format(slurm_job_ID))

    #get run details
    start_year =  GHOST_configuration.start_date[:4]
    end_year =    GHOST_configuration.end_date[:4]
    start_month = GHOST_configuration.start_date[4:]
    end_month =   GHOST_configuration.end_date[4:]
    run_details = {'network':network,'parameter':parameter,'start_date':GHOST_configuration.start_date,'end_date':GHOST_configuration.end_date,'start_year':start_year,'end_year':end_year,'start_month':start_month,'end_month':end_month,'ghost_version':GHOST_configuration.ghost_version,'data_root':GHOST_configuration.data_root,'process_root':GHOST_configuration.process_root,'qos_pipeline':GHOST_configuration.qos_pipeline,'qos_stage':GHOST_configuration.qos_stage,'chunk_size':GHOST_configuration.chunk_size,'job_array_limit':GHOST_configuration.job_array_limit,'multithreading':GHOST_configuration.multithreading,'temporal_resolutions':GHOST_configuration.temporal_resolutions,'stages_to_process':GHOST_configuration.stages_to_process,'slurm_job_ID':slurm_job_ID}

    #get standard parameter details
    parameter_details = standard_parameters[parameter]
        
    #get standard network details
    network_details = standard_networks[network]

    return run_details, parameter_details, network_details

###------------------------------------------------------------------------------------###

def check_GHOST_output_directories(run_details, parameter_details):

    '''test for presence of native and output resolution directories for all defined output resolutions, for a specific network/parameter.
       if directories do not exist, create them.
    '''

    from GHOST_standards import standard_temporal_resolutions

    #set native directory for specific network/parameter
    native_dir = run_details['data_root'] + '{}/{}/native/{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
    #check if directory already exists
    if os.path.isdir(native_dir) == False:
        #if not, make directory
        os.makedirs(native_dir)
        #give 770 permissions to directory
        os.chmod(native_dir, 0o770)

    #iterate through all defined temporal resolutions to process
    for temporal_resolution_code in run_details['temporal_resolutions']:

        #get temporal resolution specific variables
        temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

        #set output resolution directory for specific network/output resolution/parameter
        output_resolution_dir = run_details['data_root'] + '{}/{}/{}/{}/'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'])
        
        #check if directory already exists
        if os.path.isdir(output_resolution_dir) == False:
            #if not, make directory
            os.makedirs(output_resolution_dir)
            #give 770 permissions to directory
            os.chmod(output_resolution_dir, 0o770)
    
###------------------------------------------------------------------------------------###

def check_GHOST_stage_temporary_directories(run_details, parameter_details):

    '''test for presence of defined stages' temporary data file directories for a specific network/parameter,
       used to store .npz station files between each of the S stages,
       and year/month chunked station files between the T stages. 
       if directories do not exist, create them.
       if directories do exist, delete all current temporary data files in directories
    '''

    from GHOST_standards import standard_temporal_resolutions

    #get all defined S stages to process
    S_GHOST_stages = sorted([stage for stage in run_details['stages_to_process'] if 'S' in stage])
    #should not have a temporary directory for S6
    if 'S6' in S_GHOST_stages:
        S_GHOST_stages.remove('S6')
    #iterate through all temporary S GHOST stage directories 
    for stage in S_GHOST_stages:
        #define temporary data directory for GHOST stage    
        temp_dir = run_details['data_root']+'{}/{}/native/{}/temporary_{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],stage)
        #check if directory already exists
        does_temporary_exist = os.path.isdir(temp_dir)
        #if does not exist, then create it
        if does_temporary_exist == False:
            os.makedirs(temp_dir)
            #give 770 permissions to directory
            os.chmod(temp_dir, 0o770)
        #if it does exist, then remove all files currently in the directory
        else:
            for zippath in glob.iglob(os.path.join(temp_dir, '*.npz')):
                os.remove(zippath)    

    #check if 'T1' stage is defined to be processed
    if 'T1' in run_details['stages_to_process']:
        #if so, check year/month temporary directories for all defined temporal resolutions to output
        
        #iterate through defined temporal_resolutions to output
        for temporal_resolution_code in run_details['temporal_resolutions']:

            #get all years to process
            process_years = range(int(run_details['start_year']),int(run_details['end_year'])+1)
            for year in process_years:

                #is it first process year?
                if year == process_years[0]:
                    months = np.arange(int(run_details['start_month']), 13)
    
                #is it last process year?
                elif year == process_years[-1]:
                    months = np.arange(1,int(run_details['end_month']))
    
                #full year of months otherwise
                else:
                    months = np.arange(1,13)
    
                for month in months:
                    if month < 10:
                        month_string = '0'+str(month)
                    else:
                        month_string = str(month)
                    #define temporary data directory for T1 stage  
                    temp_dir = run_details['data_root']+'{}/{}/{}/{}/temporary/{}/{}/'.format(run_details['network'],run_details['ghost_version'],standard_temporal_resolutions[temporal_resolution_code]['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string)
                    #check if directory already exists
                    does_temporary_exist = os.path.isdir(temp_dir)
                    #if does not exist, then create it
                    if does_temporary_exist == False:
                        os.makedirs(temp_dir)
                        #give 770 permissions to directory
                        os.chmod(temp_dir, 0o770)
                    #if it does exist, then remove all files currently in the directory
                    else:
                        for zippath in glob.iglob(os.path.join(temp_dir, '*.npz')):
                            os.remove(zippath)
                        for zippath in glob.iglob(os.path.join(temp_dir, '*.nc')):
                            os.remove(zippath)

###------------------------------------------------------------------------------------###
def check_GHOST_stage_arguments_logs_submit(run_details):

    '''clean all existing argument/log/submit directories for defined GHOST stages to process.
       if directories do not already exist, create them.'''

    #iterate through defined GHOST stages to process
    for GHOST_stage in run_details['stages_to_process']:

        #define directory where submittal logs will be written
        log_dir = run_details['process_root']+'/run/stage_logs/{}_{}/{}'.format(run_details['network'],run_details['parameter'], GHOST_stage)

        #clean log directory (if exists already)
        if os.path.isdir(log_dir) == True:
            for zippath in glob.iglob(os.path.join(log_dir, '*.out')):
                os.remove(zippath)
            for zippath in glob.iglob(os.path.join(log_dir, '*.err')):
                os.remove(zippath)
        #if log directory does not exist already, create it
        else:
            os.makedirs(log_dir)
            #give 770 permissions to directory
            os.chmod(log_dir, 0o770)

        #-----------------------------#

        #define directory where job arguments will be written
        arguments_dir = run_details['process_root']+'/run/stage_arguments/{}_{}'.format(run_details['network'],run_details['parameter'])

        #clean arguments directory (if exists already)
        if os.path.isdir(arguments_dir) == True:
            for zippath in glob.iglob(os.path.join(arguments_dir, '*.txt')):
                os.remove(zippath)
            for zippath in glob.iglob(os.path.join(arguments_dir, '*.grz')):
                os.remove(zippath)
        #if arguments directory does not exist already, create it
        else:
            os.makedirs(arguments_dir)
            #give 770 permissions to directory
            os.chmod(arguments_dir, 0o770)

        #-----------------------------#

        #define directory where jobs a from 
        submit_dir = run_details['process_root']+'/run/submit_stage/{}_{}'.format(run_details['network'],run_details['parameter'])

        #clean submit directory (if exists already)
        if os.path.isdir(submit_dir) == True:
            for zippath in glob.iglob(os.path.join(submit_dir, '*.sh')):
                os.remove(zippath)
            for zippath in glob.iglob(os.path.join(submit_dir, '*.rst')):
                os.remove(zippath)
            for zippath in glob.iglob(os.path.join(submit_dir, '*.log')):
                os.remove(zippath)
        #if submit directory does not exist already, create it
        else:
            os.makedirs(submit_dir)
            #give 770 permissions to directory
            os.chmod(submit_dir, 0o770)

###------------------------------------------------------------------------------------###

def clean_native_directory(run_details, parameter_details):
        
    '''remove all existing station specific native resolution .npz files in native file directory for a specific network/parameter'''

    #set native directory for specific network/parameter
    native_dir = run_details['data_root'] + '{}/{}/native/{}/'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'])
    #iterate through all files in directory, removing each of them
    for zippath in glob.iglob(os.path.join(native_dir, '*.npz')):
        os.remove(zippath)

###------------------------------------------------------------------------------------###

def clean_output_resolution_directories(run_details, parameter_details):
    
    '''remove all existing final output resolution netCDF files for a specific network/parameter, for all defined output resolutions'''

    from GHOST_standards import standard_temporal_resolutions

    #iterate through all defined temporal resolutions to process
    for temporal_resolution_code in run_details['temporal_resolutions']:
    
        #get temporal resolution specific variables
        temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

        #set output resolution directory for specific network/output resolution/parameter
        output_resolution_dir = run_details['data_root'] + '{}/{}/{}/{}/'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'])
        #iterate through all files in directory, removing each of them
        for zippath in glob.iglob(os.path.join(output_resolution_dir, '*.nc')):
            os.remove(zippath)

###------------------------------------------------------------------------------------###

def get_data_files(run_details, parameter_details, specific_station='', specific_year=''):
    '''gets list of all relevant data files for network/parameter (and an optional specific station/year) combination in set date range'''

    from dateutil.relativedelta import relativedelta

    # create datetime objects of defined start date/end date.
    start_date_datetime = datetime.datetime(int(run_details['start_year']), int(run_details['start_month']), 1)
    end_date_datetime = datetime.datetime(int(run_details['end_year']), int(run_details['end_month']), 1)

    # create empty list to store valid data files
    valid_data_files = []

    if run_details['network'] == 'CANADA_NAPS':
        #handle complicated aerosol parameters
        #PM10, PM2.5, PM1
        if (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (run_details['parameter'] == 'PM1'):
            data_files = glob.glob(run_details['data_root']+'{}/original_files/{}/{}/mass/*.hly'.format(run_details['network'],parameter_details['matrix'],run_details['parameter']))
        #all other parameters
        else:
            data_files = glob.glob(run_details['data_root']+'{}/original_files/{}/{}/*.hly'.format(run_details['network'],parameter_details['matrix'],run_details['parameter']))

        #list to append file data years
        file_data_years = []

        #limit data files to just those with data inside defined data range
        for data_file in data_files:
            data_year = int(data_file.split('/')[-1][:4])
            start_date = datetime.datetime(data_year, 1, 1, 0, 0)
            end_date = datetime.datetime(data_year+1, 1, 1, 0, 0)
            if (start_date_datetime <= start_date < end_date_datetime) or (start_date_datetime < end_date <= end_date_datetime) or (start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_file)
                file_data_years.append(str(data_year))

    elif run_details['network'] == 'EBAS':
        #special case for NO2 (due to NOx being given in joint files - all of which are stored under NO)
        if run_details['parameter'] == 'NO2':
            nox_files = glob.glob(
                run_details['data_root'] + '{}/original_files/gas/NO/*.nas'.format(run_details['network']))
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/gas/{}/*.nas'.format(run_details['network'],
                                                                                   run_details['parameter']))
            # append files in NO directory which contain NO2
            for f in nox_files:
                if 'nitrogen_monoxide' not in f:
                    data_files.append(f)
        # handle complicated aerosol parameters
        # PM10, PM2.5, PM1
        elif (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (
                run_details['parameter'] == 'PM1'):
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/aerosol/{}/mass/*.nas'.format(run_details['network'],
                                                                                            run_details['parameter']))
        # bulk speciated
        elif run_details['parameter'] in ['Al', 'As', 'BC', 'C', 'C_corrected', 'Ca++', 'Cd', 'Cl-', 'Co', 'Cr', 'Cu',
                                          'EC', 'Fe', 'Hg', 'K+', 'Mg++', 'Mn', 'Na+', 'NH4+', 'Ni', 'NO3-', 'OC',
                                          'OC_corrected', 'Pb', 'Se', 'SO4--', 'SO4--_NSS', 'V', 'Zn']:
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/aerosol/bulk/speciated/{}/*.nas'.format(
                    run_details['network'], run_details['parameter']))
        # PM10/PM2.5/PM1 speciated
        elif ('PM10_' in run_details['parameter']) or ('PM2.5_' in run_details['parameter']) or (
                'PM1_' in run_details['parameter']):
            aerosol_type = run_details['parameter'].split('_')[0]
            param = run_details['parameter'].split('_')[-1]
            data_files = glob.glob(run_details['data_root'] + '{}/original_files/aerosol/{}/speciated/{}/*.nas'.format(
                run_details['network'], aerosol_type, param))
        # all other parameters
        else:
            data_files = glob.glob(
                run_details['data_root'] + '{}/original_files/{}/{}/*.nas'.format(run_details['network'],
                                                                                  parameter_details['matrix'],
                                                                                  run_details['parameter']))

        # get start and end timepoints of data files
        data_files_without_root = [data_file.split('/')[-1] for data_file in data_files]
        data_files_start_dates = [data_file.split('.')[1] for data_file in data_files_without_root]
        data_files_start_dates = [
            datetime.datetime(int(start_date[:4]), int(start_date[4:6]), int(start_date[6:8]), int(start_date[8:10]),
                              int(start_date[10:12])) for start_date in data_files_start_dates]
        data_files_file_lengths = [data_file.split('.')[6] for data_file in data_files_without_root]
        for file_ii, start_date in enumerate(data_files_start_dates):
            file_length_int = int(re.findall(r'\d+', data_files_file_lengths[file_ii])[0])
            file_length_str = re.findall(r'\D+', data_files_file_lengths[file_ii])[0]
            if file_length_str == 'h':
                end_date = start_date + relativedelta(hours=file_length_int)
            elif file_length_str == 'd':
                end_date = start_date + relativedelta(days=file_length_int)
            elif file_length_str == 'w':
                end_date = start_date + relativedelta(weeks=file_length_int)
            elif file_length_str == 'mo':
                end_date = start_date + relativedelta(months=file_length_int)
            elif file_length_str == 'y':
                end_date = start_date + relativedelta(years=file_length_int)

            # limit data files to just those with data inside defined data range
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])

        #get start and end timepoints of data files
        data_files_without_root = [data_file.split('/')[-1] for data_file in data_files]
        data_files_start_dates = [data_file.split('.')[1] for data_file in data_files_without_root]
        data_files_start_dates = [datetime.datetime(int(start_date[:4]), int(start_date[4:6]), int(start_date[6:8]), int(start_date[8:10]), int(start_date[10:12])) for start_date in data_files_start_dates]      
        data_files_file_lengths = [data_file.split('.')[6] for data_file in data_files_without_root]
        for file_ii, start_date in enumerate(data_files_start_dates):
            file_length_int = int(re.findall(r'\d+', data_files_file_lengths[file_ii])[0])
            file_length_str =  re.findall(r'\D+', data_files_file_lengths[file_ii])[0]
            if file_length_str == 'h':
                end_date = start_date + relativedelta(hours=file_length_int)
            elif file_length_str == 'd':
                end_date = start_date + relativedelta(days=file_length_int)
            elif file_length_str == 'w':
                end_date = start_date + relativedelta(weeks=file_length_int)
            elif file_length_str == 'mo':
                end_date = start_date + relativedelta(months=file_length_int)
            elif file_length_str == 'y':
                end_date = start_date + relativedelta(years=file_length_int)

            #limit data files to just those with data inside defined data range
            if (start_date_datetime <= start_date < end_date_datetime) or (start_date_datetime < end_date <= end_date_datetime) or (start_date <= start_date_datetime < end_date) or (start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])

        #get station references from all valid data files
        file_station_references = [f.split('/')[-1].split('.')[0] for f in valid_data_files] 

    if run_details['network'] == 'EEA_AQ_eReporting':
        # handle complicated aerosol parameters
        # PM10, PM2.5, PM1
        if (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (
                run_details['parameter'] == 'PM1'):
            stem = run_details['data_root'] + '{}/original_files/aerosol/{}/mass'.format(run_details['network'],
                                                                                         run_details['parameter'])
        # bulk speciated
        elif run_details['parameter'] in ['Al', 'As', 'BC', 'C', 'C_corrected', 'Ca++', 'Cd', 'Cl-', 'Co', 'Cr', 'Cu',
                                          'EC', 'Fe', 'Hg', 'K+', 'Mg++', 'Mn', 'Na+', 'NH4+', 'Ni', 'NO3-', 'OC',
                                          'OC_corrected', 'Pb', 'Se', 'SO4--', 'SO4--_NSS', 'V', 'Zn']:
            stem = run_details['data_root'] + '{}/original_files/aerosol/bulk/speciated/{}'.format(
                run_details['network'], run_details['parameter'])
        # PM10/PM2.5/PM1 speciated
        elif ('PM10_' in run_details['parameter']) or ('PM2.5_' in run_details['parameter']) or (
                'PM1_' in run_details['parameter']):
            aerosol_type = run_details['parameter'].split('_')[0]
            param = run_details['parameter'].split('_')[-1]
            stem = run_details['data_root'] + '{}/original_files/aerosol/{}/speciated/{}'.format(run_details['network'],
                                                                                                 aerosol_type, param)
        # all other parameters
        else:
            stem = run_details['data_root'] + '{}/original_files/{}/{}'.format(run_details['network'],
                                                                               parameter_details['matrix'],
                                                                               run_details['parameter'])
        # gather data files
        data_files = []
        for root, dirs, files in os.walk(stem):
            for file in files:
                if file.endswith(".csv"):
                    data_files.append(os.path.join(root, file))

        # get start and end timepoints of data files
        data_files_start_dates = [datetime.datetime(int(data_file.split('/')[-1].split('_')[3]), 1, 1) for
                                  data_file in data_files]
        for file_ii, start_date in enumerate(data_files_start_dates):
            end_date = start_date + relativedelta(years=1)

            # limit data files to just those with data inside defined data range
            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (
                    start_date < end_date_datetime <= end_date):
                valid_data_files.append(data_files[file_ii])

        #get station references from all valid data files
        f_split = [f.split('/')[-1].split('_') for f in valid_data_files]
        file_station_references = ['{}_{}_{}'.format(spl[0],spl[1],spl[2]) for spl in f_split]

    if run_details['network'] == 'NCDC_ISD':
        data_files = glob.glob(
            run_details['data_root'] + '{}/original_files/meteo/2016/*'.format(run_details['network']))

        file_station_references = []
        for file in data_files:
            stationref = str(file).split('/')[-1][:6] + str(file).split('/')[-1][7:12]
            stationyear = str(file).split('/')[-1][-4:]

            start_date = datetime.datetime(int(stationyear),1,1,0,0)
            end_date = datetime.datetime(int(stationyear)+1,1,1,0,0)

            if (start_date_datetime <= start_date < end_date_datetime) or (
                    start_date_datetime < end_date <= end_date_datetime) or (
                    start_date <= start_date_datetime < end_date) or (
                    start_date < end_date_datetime <= end_date):
                valid_data_files.append(file)
                file_station_references.append(stationref)

    # if a specific station variable has been defined, limit data files to be just those associated with the defined station reference
    if specific_station != '':
        valid_data_files = [data_file for data_file, file_station_reference in
                            zip(valid_data_files, file_station_references) if
                            str(specific_station) == str(file_station_reference)]
    # if a specific data year variable ha been defined, limit data files to be just those associated with the defined data year
    if specific_year != '':
        valid_data_files = [data_file for data_file, file_data_year in zip(valid_data_files, file_data_years) if
                            str(specific_year) == str(file_data_year)]

    return sorted(valid_data_files)


###------------------------------------------------------------------------------------###

def get_station_references_from_filenames(data_files, run_details):

    '''get unique list of raw station references provided in data filenames.
       process for this differs by network.
    '''

    station_references = []

    if run_details['network'] == 'EBAS':
        for i, f in enumerate(data_files):
            f = f.split('/')[-1]
            station_references.append(f.split('.')[0])

    if run_details['network'] == 'EEA_AQ_eReporting':
        for i, f in enumerate(data_files):
            f_split = f.split('/')[-1]
            f_split = f_split.split('_')

            station_references.append('{}_{}_{}'.format(f_split[0],f_split[1],f_split[2]))

    if run_details['network'] == 'NCDC_ISD':
        for f in enumerate(data_files):
            f_split = f[-1][-17:].split('-')

            station_references.append('{}{}'.format(f_split[0], f_split[1]))

    #sort station references in a human-intuitive order
    station_references = sorted(remove_duplicates_preserve_order(station_references), key=natural_key)

    return station_references

###------------------------------------------------------------------------------------###

class process_nasa_ames(object):

    '''class that handles processing of a nasa ames file'''

    def __init__(self,filename,log_file_str,station_reference,run_details):
        
        self.filename = filename

        self.log_file_str = log_file_str   
        self.station_reference = station_reference
        self.run_details = run_details

        #core header meta
        self.n_header = ''          #number of header lines
        self.ffi = ''               #file format index
                                    #1st digit indicates the number of independent variables (dimensions)
                                    #2nd digit indicates number of necessary auxiliary variable
                                    #Remaining differences cover the way the independent variables increment and the nature (numeric or character) of the variables
        self.originator_name = ''   #data originator name
        self.originator_affil = ''  #data originator affiliation
        self.provider_name = ''     #data provider name
        self.campaign_name = ''     #campaign name
        self.volume_number = ''     #current volume number (dated field - should be 1)
        self.n_volumes = ''         #number of volumes in dataset (dated field - should be 1)
        self.date_start = ''        #date for which record starts
        self.file_creation = ''     #date of file creation


        #independent variable meta (e.g. time)
        self.independent_interval = ''  #independent variable interval (if changes should be zero)
        self.independent_name = ''      #independent variable name (e.g. days from file reference point)


        #primary variable meta (quantities that have been observed or derived)
        self.n_primary = ''               #number of primary variables
        self.primary_scale_factors = ''   #scale factors for each of primary variables
        self.primary_missing_values = ''  #value imposed for missing data for each primary variable
        self.primary_names = []           #primary variable names (e.g. ozone)
        self.primary_units = []           #primary variable units
        self.primary_details = []         #extra details about variables (e.g. limit of detection)

        #auxiliary variable meta
        self.n_auxiliary = ''    #number of auxiliary variables

        #comments meta
        self.n_comments = ''     #number of lines of comments
        self.comments = []       #extra meta comments about data. Can also include column headings

        #important line counts
        self.endline_primary_count = -999
        self.start_auxiliary = -999
        self.start_comments = -999
        self.start_data = -999
        self.row_count = 0

    def read_header(self):
        from GHOST_standards import standard_parameters

        with open(self.filename) as f:
            self.content = f.readlines()
            self.content = [x.strip('\n') for x in self.content]

            for row in self.content:

                #--read core header meta--
                if self.row_count == 0:
                    line_split = row.split(' ')
                    self.n_header,self.ffi = line_split[0],line_split[1]

                elif self.row_count == 1:
                    self.originator_name = row

                elif self.row_count == 2:
                    self.originator_affil = row

                elif self.row_count == 3:
                    self.provider_name = row

                elif self.row_count == 4:
                    self.campaign_name = row

                elif self.row_count == 5:
                    line_split = row.split(' ')
                    self.volume_name, self.n_volumes = line_split[0],line_split[1]

                elif self.row_count == 6:
                    line_split = row.split(' ')
                    self.date_start, self.file_creation = line_split[0]+line_split[1]+line_split[2], line_split[3]+line_split[4]+line_split[5]

                #--read independent variable meta--
                elif self.row_count == 7:
                    self.independent_interval = row

                elif self.row_count == 8:
                    self.independent_name = row

                #--read primary variable meta--
                elif self.row_count == 9:
                    self.n_primary = int(row)
                    self.endline_primary_count = 11+self.n_primary
                    self.start_auxiliary = self.endline_primary_count+1

                elif self.row_count == 10:
                    self.primary_scale_factors = row.split(' ')

                elif self.row_count == 11:
                    self.primary_missing_values = row.split(' ')

                elif (self.row_count > 11) & (self.row_count <= self.endline_primary_count):
                    #handle line splitting with quotes (with commas inside)
                    line_split = []
                    split_by_quotes = row.split('"')
                    for split_ii, split_element in enumerate(split_by_quotes):
                        if (split_ii % 2) == 0:
                            line_split = np.append(line_split, split_element.split(','))
                        else:
                            if split_element[:7] == 'Comment':
                                new_str = 'Comment="{}"'.format(split_element[8:])
                            else: 
                                new_str = '{}"{}"'.format(line_split[-1], split_element)
                                line_split = line_split[:-1]
                            line_split = np.append(line_split, new_str)
                    line_split = [element.strip() for element in line_split]
                    line_split = [element for element in line_split if element != '']
                    if len(line_split) == 0:
                        line_split = [row]

                    #only read line if first value is one of accepted primary variable names
                    #otherwise line is appended to primary details - this deals with badly formatted files
                    #also add 1 to auxiliary start count - so don't start reading too soon

                    #create a list of all potentially valid primary variables
                    valid_primary_variables = ['end_time of measurement','numflag']
                    for param in list(standard_parameters.keys()):
                        ebas_parameter_name = standard_parameters[param]['ebas_parameter_name']
                        #only process variables with a valid EBAS parameter name (i.e. not empty string)
                        if ebas_parameter_name != '':
                            valid_primary_variables=np.append(valid_primary_variables,ebas_parameter_name)
                            #create numflag variable for all parameters
                            valid_primary_variables=np.append(valid_primary_variables,'numflag {}'.format(ebas_parameter_name))                    
    
                    if line_split[0] in valid_primary_variables:
                        self.primary_names.append(line_split[0])
                        self.primary_units.append(line_split[1])
                        if len(line_split) > 2:
                            self.primary_details.append(line_split[2:])
                        else:
                            self.primary_details.append([])
                    else:
                        self.primary_details[-1] = self.primary_details[-1]+line_split
                        self.endline_primary_count+=1
                        self.start_auxiliary+=1

                #--read auxiliary variable meta--
                #first checks if row is a valid number or relates to the previous primary details not correctly formatted
                #if so then append to primary details
                elif self.row_count == self.start_auxiliary:
                    try:
                        self.n_auxiliary = int(row)
                        #if is not 0 then break processing and panic - as have never seen this be used and don't know format
                        if self.n_auxiliary != 0:
                            self.log_file_str += 'Using auxilary variables. {} of them! How Strange. Now need to code for this. In file: {} .Stopping processing.'.format(self.n_auxiliary, self.filename)
                            create_GHOST_stage_logfile('S1',self.run_details,2,self.log_file_str,self.station_reference)
                        self.start_comments = self.start_auxiliary+1
                    except:
                        line_split = row.split(',')
                        self.primary_details[-1] = np.append(self.primary_details[-1],line_split)
                        self.endline_primary_count+=1
                        self.start_auxiliary+=1

                #--read comments meta--
                elif self.row_count == self.start_comments:
                    self.n_comments = int(row)
                    self.end_comments = self.start_comments+self.n_comments
                    self.start_data = self.end_comments+1

                elif (self.row_count > self.start_comments) & (self.row_count <= self.end_comments):
                    self.comments.append(row)

                #manually iterate row count
                self.row_count+=1

    def read_data(self):
        #cut file read to just be data columns
        data = self.content[self.start_data:]

        #make list of lists to store data columns (1 independent variable + primary variables)
        self.structured_data = [[] for _ in range(self.n_primary+1)]

        #iterate through lines, appending column data into a list of lists
        for line in data:
            line_split = line.split()
            for n,var in enumerate(line_split):
                self.structured_data[n].append(var)

###------------------------------------------------------------------------------------###

"""A dictionary, where key is the ISD identifier in ASCII and value
is a list that contains [variable_name, number_of_chars]. The variable
name is as in ISD format. The number of chars indicate the number of
characters data (without considering the 3 chars of the identifier"""


isd_map = {'AA1': ['LIQUID-PRECIPITATION', 8],
           'AA2': ['LIQUID-PRECIPITATION', 8],
           'AA3': ['LIQUID-PRECIPITATION', 8],
           'AA4': ['LIQUID-PRECIPITATION', 8],
           'AB1': ['LIQUID-PRECIPITATION MONTHLY TOTAL', 7],
           'AC1': ['PRECIPITATION-OBSERVATION-HISTORY', 3],
           'AD1': ['LIQUID-PRECIPITATION GREATEST AMOUNT IN 24 HOURS, FOR THE MONTH', 19],
           'AE1': ['LIQUID-PRECIPITATION, NUMBER OF DAYS WITH SPECIFIC AMOUNTS, FOR THE MONTH', 12],
           'AG1': ['PRECIPITATION-ESTIMATED-OBSERVATION', 4],
           'AH1': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH2': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH3': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH4': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH5': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH6': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI1': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI2': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI3': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI4': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI5': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI6': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AJ1': ['SNOW-DEPTH', 14],
           'AK1': ['SNOW-DEPTH-GREATEST-ON-GROUND', 12],
           'AL1': ['SNOW-ACCUMULATION', 7],
           'AL2': ['SNOW-ACCUMULATION', 7],
           'AL3': ['SNOW-ACCUMULATION', 7],
           'AL4': ['SNOW-ACCUMULATION', 7],
           'AM1': ['SNOW-ACCUMULATION-GREATEST-IN-24H', 18],
           'AN1': ['SNOW-ACCUMULATION-FOR-MONTH', 9],
           'AO1': ['LIQUID-PRECIPITATION', 8],
           'AO2': ['LIQUID-PRECIPITATION', 8],
           'AO3': ['LIQUID-PRECIPITATION', 8],
           'AO4': ['LIQUID-PRECIPITATION', 8],
           'AP1': ['15MIN-LIQUID-PRECIP', 6],
           'AP2': ['15MIN-LIQUID-PRECIP', 6],
           'AP3': ['15MIN-LIQUID-PRECIP', 6],
           'AP4': ['15MIN-LIQUID-PRECIP', 6],
           'AT1': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],             # Weather Occurrence Data
           'AT2': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT3': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT4': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT5': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT6': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT7': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT8': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AU1': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU2': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU3': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU4': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU5': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU6': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU7': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU8': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU9': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AW1': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AW2': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AW3': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AX1': ['PAST-WEATHER-OBSERVATION', 6],
           'AX2': ['PAST-WEATHER-OBSERVATION', 6],
           'AX3': ['PAST-WEATHER-OBSERVATION', 6],
           'AX4': ['PAST-WEATHER-OBSERVATION', 6],
           'AX5': ['PAST-WEATHER-OBSERVATION', 6],
           'AX6': ['PAST-WEATHER-OBSERVATION', 6],
           'AY1': ['PAST-WEATHER-OBSERVATION manual', 5],
           'AY2': ['PAST-WEATHER-OBSERVATION manual', 5],
           'AZ1': ['PAST-WEATHER-OBSERVATION automated', 5],
           'AZ2': ['PAST-WEATHER-OBSERVATION automated', 5],
           'CB1': ['Subhourly Liquid Precipitation Section', 10],       # Climate Reference Network Unique Data
           'CB2': ['Subhourly Liquid Precipitation Section', 10],
           'CF1': ['Hourly Fan Speed Section', 6],
           'CF2': ['Hourly Fan Speed Section', 6],
           'CF3': ['Hourly Fan Speed Section', 6],
           'CG1': ['PRIMARY-SENSOR', 8],
           'CG2': ['PRIMARY-SENSOR', 8],
           'CG3': ['PRIMARY-SENSOR', 8],
           'CH1': ['HOURLY-RELATIVE-HUMIDITY', 15],
           'CH2': ['HOURLY-RELATIVE-HUMIDITY', 15],
           'CI1': ['RELATIVE-HUMIDITY-TEMPERATURE', 28],
           'CN1': ['HOURLY-BATTERY-VOLTAGE', 18],
           'CN2': ['HOURLY-DIAGNOSTIC', 18],
           'CN3': ['SECONDARY-HOURLY-DIAGNOSTIC', 16],
           'CN4': ['SECONDARY-HOURLY-DIAGNOSTIC', 19],
           'CO1': ['US-NETWORK-METADATA', 5],                           # Network Metadata
           'CO2': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO3': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO4': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO5': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO6': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO7': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO8': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO9': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CR1': ['CONTROL-SECTION', 7],
           'CT1': ['SUBHOURLY-TEMPERATURE', 7],
           'CT2': ['SUBHOURLY-TEMPERATURE', 7],
           'CT3': ['SUBHOURLY-TEMPERATURE', 7],
           'CU1': ['HOURLY-TEMPERATURE', 13],
           'CU2': ['HOURLY-TEMPERATURE', 13],
           'CU3': ['HOURLY-TEMPERATURE', 13],
           'CV1': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CV2': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CV3': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CW1': ['SUBHOURLY-WETNESS', 14],
           'CX1': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'CX2': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'CX3': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'ED1': ['RUNWAY-VISUAL-RANGE', 8],                           # Runway Visual Range Data
           'GA1': ['SKY-COVER-LAYER', 13],                              # Cloud and Solar Data
           'GA2': ['SKY-COVER-LAYER', 13],
           'GA3': ['SKY-COVER-LAYER', 13],
           'GA4': ['SKY-COVER-LAYER', 13],
           'GA5': ['SKY-COVER-LAYER', 13],
           'GA6': ['SKY-COVER-LAYER', 13],
           'GD1': ['SKY-COVER-SUMMATION', 12],
           'GD2': ['SKY-COVER-SUMMATION', 12],
           'GD3': ['SKY-COVER-SUMMATION', 12],
           'GD4': ['SKY-COVER-SUMMATION', 12],
           'GD5': ['SKY-COVER-SUMMATION', 12],
           'GD6': ['SKY-COVER-SUMMATION', 12],
           'GE1': ['SKY-CONDITION', 19],
           'GF1': ['SKY-CONDITION', 23],
           'GG1': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG2': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG3': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG4': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG5': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG6': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GH1': ['HOURLY-SOLAR-RADIATION', 28],
           'GJ1': ['SUNSHINE', 5],
           'GK1': ['SUNSHINE-OBSERVATION', 4],
           'GL1': ['SUSHINE-OBSERVATION-FOR-MONTH', 6],
           'GM1': ['SOLAR-IRRADIANCE', 30],
           'GN1': ['SOLAR-RADIATION', 28],
           'GO1': ['NET-SOLAR-RADIATION', 19],
           'GP1': ['MODELED-SOLAR-IRRADIANCE', 31],
           'GQ1': ['HOURLY-SOLAR-ANGLE', 14],
           'GR1': ['HOURLY-EXTRATERRESTRIAL-RADIATION', 14],
           '???': ['HAIL', 4],                                          # Hail Data - Missing ID in the Doc
           'IA1': ['GROUND-SURFACE', 3],                                # Ground Surface Data
           'IA2': ['GROUND-SURFACE', 9],
           'IB1': ['HOURLY-SURFACE', 27],
           'IB2': ['HOURLY-SURFACE', 13],
           'IC1': ['GROUND-SURFACE', 25],
           'KA1': ['EXTREME-AIR-TEMPERATURE', 10],                      # Temperature Data
           'KA2': ['EXTREME-AIR-TEMPERATURE', 10],
           'KA3': ['EXTREME-AIR-TEMPERATURE', 10],
           'KA4': ['EXTREME-AIR-TEMPERATURE', 10],
           'KB1': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KB2': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KB3': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KC1': ['EXTREME-AIR-TEMP-FOR-MONTH', 14],
           'KC2': ['EXTREME-AIR-TEMP-FOR-MONTH', 14],
           'KD1': ['HEATING-COOLING-DEGREE-DAYS', 9],
           'KD2': ['HEATING-COOLING-DEGREE-DAYS', 9],
           'KE1': ['EXTREME-TEMPS-EXCEEDING-CRITERIA', 12],
           'KF1': ['HOURLY-CALCULATED-TEMP', 6],
           'KG1': ['AVERAGE-DEW_POINT', 11],
           'KG2': ['AVERAGE-DEW_POINT', 11],
           'MA1': ['ATMOSPHERIC-PRESSURE-OBSERVATION', 12],             # Pressure Data
           'MD1': ['ATMOSPHERIC-PRESSURE-CHANGE', 11],
           'ME1': ['GEOPOTENTIAL-HEIGHT', 6],
           'MF1': ['ATMOSPHERIC-PRESSURE-STP-SLP', 12],
           'MG1': ['ATMOSPHERIC-PRESSURE', 12],
           'MH1': ['ATMOSPHERIC-PRESSURE-FOR-MONTH', 12],
           'MK1': ['ATMOSPHERIC-PRESSURE-FOR-MONTH', 24],
           'MV1': ['PRESENT-WEATHER-IN-VICINITY', 3],                   # Weather Occurrence Data (Again?)
           'MV2': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV3': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV4': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV5': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV6': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV7': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MW1': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW2': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW3': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW4': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW5': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW6': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW7': ['PRESENT-WEATHER-OBSERVATION', 3],
           'OA1': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],                # Wind Data
           'OA2': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],
           'OA3': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],
           'OB1': ['Hourly/Sub-Hourly Wind Section', 28],
           'OB2': ['Hourly/Sub-Hourly Wind Section', 28],
           'OC1': ['WIND-GUST-OBSERVATION', 5],
           'OD1': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'OD2': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'OD3': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'RH1': ['RELATIVE HUMIDITY', 9],
           'RH2': ['RELATIVE HUMIDITY', 9],
           'RH3': ['RELATIVE HUMIDITY', 9],
           'SA1': ['SEA-SURFACE-TEMPERATURE', 5],                       # Sea Surface Temperature Data
           'ST1': ['SOIL-TEMPERATURE', 17],                             # Soil Temperature Data
           'UA1': ['WAVE-MEASUREMENT', 10],                             # Marine Data
           'UG1': ['WAVE-MEASUREMENT-PRIMARY', 9],
           'UG2': ['WAVE-MEASUREMENT-SECONDARY', 9],
           'WA1': ['PLATFORM-ICE-ACCRETION', 6],
           'WD1': ['WATER-SURFACE-ICE', 20],
           'WG1': ['WATER-SURFACE-ICE-HISTORICAL', 11],
           'WJ1': ['WATER-LEVEL-OBSERVATION', 19]}

class ISDLineParser(object):

    def __init__(self, filename, row):

        self.filename = filename
        self.prec_period = None
        self.add_offset = 105  # always fixed
        self.section_chars = 3
        self.missing = 9999

        self.varmap = isd_map

        # Include all variables that should appear to every file in __init__ for readability
        # Initialize all to None, then call parsing methods

        self.num_char = None

        # Control section attributes
        self.station_idu = None
        self.station_idn = None
        self.date_ascii = None
        self.time_ascii = None
        self.data_source_flag = None
        self.latitude = None
        self.longitude = None
        self.code = None
        self.elevation = None
        self.call_letter = None
        self.qc_name = None

        # Mandatory section attributes
        # wind-obdervation
        self.direction_angle = None
        self.direction_angle_qc = None
        self.type_code = None
        self.speed = None
        self.speed_qc = None

        # sky-condition
        self.ceiling = None
        self.ceiling_qc = None
        self.ceiling_dc = None

        # visibility
        self.distance = None
        self.distance_qc = None
        self.variability = None
        self.variability_qc = None

        # air-temperature"
        self.air_temp = None
        self.air_temp_qc = None

        # air-temperature dew point"
        self.dpoint_temp = None
        self.dpoint_temp_qc = None

        # atmospheric pressure
        self.sea_level_pressure = None
        self.sea_level_pressure_qc = None

        # Call the first method to start the initilization of the attributes
        # self.open_file()
        self.parse_file_line(row)

        # Remarks Section
        self.rem_content = None

        # Original Observation Section
        self.has_qnn = False

    def parse_file_line(self, line):
        # First read how many chars the line has
        self.num_char = int(line[0:4]) + self.add_offset
        # Read control and mandatory section
        self.parse_control_section(line)
        self.parse_mandatory_section(line)

        # Additional Data Section
        additional = line[105:108]
        pos = 105
        if additional == "ADD":
            # try to read
            pos += self.section_chars
            while pos < self.num_char:
                try:
                    read = self.parse_add_section(line, pos)
                    pos += read
                except Exception as e:
                    # reached REM
                    # logging.exception(e)
                    break
        while pos < self.num_char:
            if line[pos:pos + self.section_chars] == 'REM':
                try:
                    rem_text = self.parse_rem_section(line[pos:self.num_char])
                    pos += len(rem_text)
                except Exception as e:
                    break
            else:
                break
        if line[pos:pos + self.section_chars] == 'EQD':
            # while pos < num_char:
            try:
                read = self.parse_eqd_section(line[pos:self.num_char])
                pos += read
            except Exception as e:
                print(line[pos:pos + self.section_chars])  # change to logging
        if line[pos:pos + self.section_chars] == 'QNN':
            self.parse_qnn_section()
            print(pos, self.num_char)
        # if pos < self.num_char:


    def parse_control_section(self, line):
        """Function to read the Control Data Section"""

        # scaling factors
        scale_lats = 1000

        self.station_idu = line[4:10]  # station, USAF catalog identifier
        self.station_idn = line[10:15]  # station, NCEI catalog identifier
        self.date_ascii = line[15:23]  # date of observation
        self.time_ascii = line[23:27]  # time on UTC
        self.datetime_ascii = line[15:27]
        self.data_source_flag = line[17]  #
        self.latitude = float(line[28:34]) / scale_lats
        if self.latitude == 99.999:
            self.latitude = np.NaN
        self.longitude = float(line[34:41]) / scale_lats  # 000000 to 179999 are signed negative
        if self.longitude == 999.999:
            self.longitude = np.NaN
        self.code = line[41:46]  # type of geophysical surface observation
        self.elevation = float(line[46:51])  # relative to MSL
        if self.elevation == 9999:
            self.elevation = np.NaN
        # self.call_letter = line[51 - 56]  # no idea
        # self.qc_name = line[56:60]  # quality control process name

    def parse_mandatory_section(self, line):
        """Function to read the Mandatory Data Section"""

        # scaling factors
        scale_speed = 10
        scale_temp = 10
        scale_slp = 10
        
        # constant to convert to kelvin
        kelvin = 273.15

        "Wind-Observation"
        self.direction_angle = float(line[60:63])
        self.direction_angle_qc = line[63]
        if self.direction_angle == 999:
            self.direction_angle_qc = '3'
        self.speed = float(line[65:69]) / scale_speed
        self.speed_qc = line[69]
        if self.speed == 999.9:
            self.speed_qc = '3'

        "Sky-Condition-Observation"
        self.ceiling = float(line[70:75])
        self.ceiling_qc = line[75]
        if self.ceiling == 99999:
            self.ceiling_qc = '3'

        "Visibility-Observation"
        self.distance = float(line[78:84])
        self.distance_qc = line[84]
        if self.distance == 999999 or self.distance_qc == 9999:
            self.distance_qc = '3'

        "Air-Temperature-Observation"
        self.air_temp = (float(line[87:92]) / scale_temp) + kelvin
        self.air_temp_qc = line[92]
        if self.air_temp == 999.9:
            self.air_temp_qc = '3'

        "Air-Temperature-Observation dew point"
        self.dpoint_temp = (float(line[93:98]) / scale_temp) + kelvin
        self.dpoint_temp_qc = line[98]
        if self.dpoint_temp == 999.9:
            self.dpoint_temp_qc = '3'

        "Atmospheric-Pressure-Observation"
        self.sea_level_pressure = float(line[99:104]) / scale_slp
        self.sea_level_pressure_qc = line[104]
        if self.sea_level_pressure == 9999.9:
            self.sea_level_pressure_qc = '3'

    def parse_add_section(self, line, position):
        """ Function to parse the Additional Data Section"""

        identifier = line[position:position + self.section_chars]

        # where the section of data start and end according to the id
        start_block = position + self.section_chars
        end_block = start_block + isd_map[identifier][1]
        block = line[start_block:end_block]

        # the data from the additional section
        # call function find_add_id
        self.find_add_id(identifier, block)

        # send back how many characters were read
        read = isd_map[identifier][1] + self.section_chars
        return read

    def find_add_id(self, identifier, block, print_info=False, print_block=False):
        """ According to the input identifier, the function reads respectively
        the variables under the identifier, following the isd format document.
        Unfortunately, this is by far a not elegant way. Suggestions are welcome."""

        map_value = 0  # position of the name in isd_map
        # scale factor - similar for all values thankfully
        scale_factor = 10
        kelvin = 273.15

        if identifier in ('AA1', 'AA2', 'AA3', 'AA4', 'AO1', 'AO2', 'AO3', 'AO4'):
            if identifier[-1] in ['2', '3', '4']:
                return
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes precipitation depth
            self.prec_period = float(block[0:2]) * 60  # because we want minutes
            if self.prec_period == 0:
                return
            self.prec_depth = float(block[2:6]) / scale_factor
            self.prec_depth_qc = block[7]
            if self.prec_depth == 999.9:
                self.prec_depth_qc = '3'

        elif identifier == 'AL1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes snow accumulation
            self.snow_acc_period = float(block[0:2])  * 60  # because we want minutes
            if self.snow_acc_period == 0:
                return
            self.snow_acc = float(block[2:5])
            self.snow_acc_qc = block[6]
            if self.snow_acc == 999:
                self.snow_acc_qc = '3'

        elif identifier == 'AJ1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes snow depth
            self.si = float(block[0:4])
            self.si_qc = block[5]
            if self.si == 9999:
                self.si_qc = '3'

        elif identifier in ('CH1', 'CH2'):
            if print_info:
                print(isd_map[identifier][map_value])
            self.avg_rh_period = float(block[0:2])  # already in minutes
            self.avg_rh = float(block[9:13]) / scale_factor
            self.avg_rh_qc = block[13]
            if self.avg_rh == 999.9:
                self.avg_rh_qc = '3'

        elif identifier == 'MA1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes station pressure rate
            self.station_p = float(block[6:11]) / scale_factor
            self.station_p_qc = block[11]
            if self.station_p == 9999.9:
                self.station_p_qc = '3'

        elif identifier in ('GA1', 'GA2', 'GA3', 'GA4', 'GA5', 'GA6', 'GD1'):
            # if we have already read a value for cloud coverage in the same line,
            # then keep the field with maximum value
            if identifier == 'GA1':
                self.cloud = int(block[0:2])
                self.cloud_qc = block[2]
                # keep flag to know that line has GA
                self.has_ga = True
                if self.cloud == 99:
                    self.cloud_qc = '3'
            elif identifier in ('GA2', 'GA3', 'GA4', 'GA5', 'GA6'):
                new_value = int(block[0:2])
                if new_value != 99 and new_value > self.cloud:
                    self.cloud = int(block[0:2])
                    self.cloud_qc = block[2]
                    self.has_ga = True
            elif identifier == 'GD1':
                # First of all, read the GD parameter
                # read part that includes sky coverage summation code
                self.cloud_gd = int(block[1:3])
                self.cloud_gd_qc = block[3]
                # if code 2 is missing, take code 1 and convert to oktas
                if self.cloud_gd == 99 or int(self.cloud_cov) > 8:
                    code1 = int(block[0])
                    # convert code to oktas
                    if code1 == 0:
                        self.cloud_gd = 0
                    elif code1 == 1:
                        self.cloud_gd = 2
                    elif code1 == 2:
                        self.cloud_gd = 4
                    elif code1 == 3:
                        self.cloud_gd = 6
                    elif code1 == 4:
                        self.cloud_gd = 8
                    else:
                        self.cloud_gd = 99
                        self.cloud_gd_qc = '3'
                # Now that we have the code, check what if GA exists, using the flag
                if self.cloud:
                    # if it has ga, check if ga == 99
                    if self.cloud == 99:
                        # then replace with the value we keep from gd
                        self.cloud = self.cloud_gd
                        self.cloud_qc = self.cloud_gd_qc
                elif not self.has_ga:
                    self.cloud = self.cloud_gd
                    self.cloud_qc = self.cloud_gd_qc
            # at this point we should have the final value that we keep for cloud
            # so following, we can compute the cloud fraction
            if self.cloud == 99:
                self.cloud_frac = 999
                self.cloud_frac_qc = '3'
            elif self.cloud == 1:
                self.cloud_frac = 12.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 2:
                self.cloud_frac = 25
                self.cloud_frac_qc = '0'
            elif self.cloud == 3:
                self.cloud_frac = 37.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 4:
                self.cloud_frac = 50
                self.cloud_frac_qc = '0'
            elif self.cloud == 5:
                self.cloud_frac = 62.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 6:
                self.cloud_frac = 75
                self.cloud_frac_qc = '0'
            elif self.cloud == 7:
                self.cloud_frac = 87.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 8:
                self.cloud_frac = 100
                self.cloud_frac_qc = '0'

        elif identifier == 'SA1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that contains sea surface temperature
            self.sst = (float(block[0:4]) / scale_factor) + kelvin
            self.sst_qc = block[4]
            if self.sst == 99.9:
                self.sst_qc = '3'

        elif identifier == 'ST1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that contains soil temperature data
            self.soil_temp = (float(block[1:6]) / scale_factor) + kelvin
            self.soil_temp_qc = block[6]
            if self.soil_temp == 999.9:
                self.soil_temp_depth_qc = '3'
            self.soil_temp_depth = float(block[7:11]) / scale_factor
            self.soil_temp_depth_qc = block[11]
            if self.soil_temp_depth == 999.9:
                self.soil_temp_depth_qc = '3'

        if print_block:
            print('read following block: {}'.format(block))

    def parse_rem_section(self, block):
        """Function to read the Remarks Data Section"""

        # initialize position, after REM and fist id
        info = self.section_chars * 2
        # read how many chars the REM has
        to_read = int(block[info:info + self.section_chars]) + self.section_chars * 3  # the ids we read already
        # start position
        pos = 0
        self.rem_content = block[pos:to_read]

        # # sanity check
        # print('block: {}'.format(block))
        # print('rem content: {}'.format(rem_content))
        return self.rem_content

    def parse_eqd_section(self, block):
        """Function to read the Element Quality Data Section"""

        param_code = 6  # length of the original value text & parameters' code
        rcode_length = 1  # reason code length
        pos = self.section_chars  # starting from 3, after reading section id
        read = self.section_chars

        while pos < len(block):
            # quality identifier
            qid = block[pos:pos + self.section_chars]
            # if qid == 'QNN':
            #     raise ValueError
            # update position
            pos += self.section_chars
            # original value text
            orig_text = block[pos:pos + param_code]
            # update position
            pos += param_code
            # reason code
            reason_code = block[pos:pos + rcode_length]
            # update position
            pos += rcode_length
            # parameter identifier
            parameter = block[pos:pos + param_code]
            # update position
            pos += param_code

            # characters read so far
            read += len(qid + orig_text + reason_code + parameter)

            if 'WOSPD' in parameter:
                self.speed_orig_value = orig_text
                self.speed_reason_code = reason_code

            if 'WODIR' in parameter:
                self.direction_orig_value = orig_text
                self.direction_reason_code = reason_code

            if 'ATOD' in parameter:
                self.dpoint_orig_value = orig_text
                self.dpoint_reason_code = reason_code

            if 'ATOT' in parameter:
                self.temp_orig_value = orig_text
                self.temp_reason_code = reason_code

            if 'APOSLP' in parameter:
                self.pressure_orig_value = orig_text
                self.pressure_reason_code = reason_code

            if 'SCOCIG' in parameter:
                self.ceiling_orig_value = orig_text
                self.ceiling_reason_code = reason_code

            if 'VODIS' in parameter:
                self.distance_orig_value = orig_text
                self.distance_reason_code = reason_code

        return read

    def parse_qnn_section(self):
        self.has_qnn = True
        # logging.info('You will not believe this! This line has a QNN section! Filename: '.format(self.filename[-17:0]))

###------------------------------------------------------------------------------------###

def get_molecular_mass(species):

    '''get molecular mass of a specific molecule in kg mol-1'''

    #define atomic masses per element in g mol-1
    atomic_masses = {
    "H":1.00794,"He":4.002602,"Li":6.941,"Be":9.012182,"B":10.811,"C":12.011,"N":14.00674,"O":15.9994,"F":18.9984032,"Ne":20.1797,"Na":22.989768,"Mg":24.3050,"Al":26.981539,"Si":28.0855,"P":30.973762,"S":32.066,"Cl":35.4527,"Ar":39.948,
    "K":39.0983,"Ca":40.078,"Sc":44.955910,"Ti":47.88,"V":50.9415,"Cr":51.9961,"Mn":54.93805,"Fe":55.847,"Co":58.93320,"Ni":58.6934,"Cu":63.546,"Zn":65.39,"Ga":69.723,"Ge":72.61,"As":74.92159,"Se":78.96,"Br":79.904,"Kr":83.80,
    "Rb":85.4678,"Sr":87.62,"Y":88.90585,"Zr":91.224,"Nb":92.90638,"Mo":95.94,"Tc":98,"Ru":101.07,"Rh":102.90550,"Pd":106.42,"Ag":107.8682,"Cd":112.411,"In":114.82,"Sn":118.710,"Sb":121.757,"Te":127.60,"I":126.90447,"Xe":131.29,
    "Cs":132.90543,"Ba":137.327,"La":138.9055,"Ce":140.115,"Pr":140.90765,"Nd":144.24,"Pm":145,"Sm":150.36,"Eu":151.965,"Gd":157.25,"Tb":158.92534,"Dy":162.50,"Ho":164.93032,"Er":167.26,"Tm":168.93421,"Yb":173.04,"Lu":174.967,
    "Hf":178.49,"Ta":180.9479,"W":183.85,"Re":186.207,"Os":190.2,"Ir":192.22,"Pt":195.08,"Au":196.96654,"Hg":200.59,"Tl":204.3833,"Pb":207.2,"Bi":208.98037,"Po":209,"At":210,"Rn":222,"Fr":223,"Ra":226.0254,"Ac":227,"Th":232.0381,
    "Pa":213.0359,"U":238.0289,"Np":237.0482,"Pu":244,"Am":243,"Cm":247,"Bk":247,"Cf":251,"Es":252,"Fm":257,"Md":258,"No":259,"Lr":260,"Rf":261,"Db":262,"Sg":263,"Bh":262,"Hs":265,"Mt":266}

    s = re.findall('([A-Z][a-z]?)([0-9]*)', species)
    compoundweight = 0

    #take sum of all element atomic masses (in g mol-1)
    for element, count in s:
        count = int(count or '1')
        compoundweight += atomic_masses[element] * count

    #convert to kg mol-1
    compoundweight = compoundweight/1e3

    return compoundweight

###------------------------------------------------------------------------------------###

def get_N_atoms(molecule, element):

    '''get integer number of atoms of a desired element in a molecule.'''

    parse_list = re.findall(r'[A-Z][a-z]*|\d+', re.sub('[A-Z][a-z]*(?![\da-z])', r'\g<0>1', molecule))
    try:
        n_atoms_element = parse_list[parse_list.index(element)+1]
    except:
        n_atoms_element = 1

    return int(n_atoms_element)

###------------------------------------------------------------------------------------###

def list_duplicates(seq):

    '''Function for finding and returning values/indices of all duplicates in a string/list
       Returns list of tuples.
       The first value in each tuple is the duplicated value
       The second value in each tuple is a list of duplicated indices
       Returns empty list if no duplicates are found
       Function is optimised for speed
    '''

    from collections import defaultdict

    tally = defaultdict(list)
    for i,item in enumerate(seq):
        tally[item].append(i)
    return sorted(((key,locs) for key,locs in tally.items()
                            if len(locs)>1))

###------------------------------------------------------------------------------------###

#time rounding functions

def round_to_minute(dt):

    '''rounds datetime object to the nearest minute'''

    dt_start_of_minute = dt.replace(second=0, microsecond=0)
    dt_half_minute = dt.replace(second=30, microsecond=0)

    if dt >= dt_half_minute:
        # round up
        dt = dt_start_of_minute + datetime.timedelta(minutes=1)
    else:
        # round down
        dt = dt_start_of_minute

    return dt

def round_to_hour(dt):

    '''rounds datetime object to the nearest hour'''

    dt_start_of_hour = dt.replace(minute=0, second=0, microsecond=0)
    dt_half_hour = dt.replace(minute=30, second=0, microsecond=0)

    if dt >= dt_half_hour:
        # round up
        dt = dt_start_of_hour + datetime.timedelta(hours=1)
    else:
        # round down
        dt = dt_start_of_hour

    return dt

###------------------------------------------------------------------------------------###

#functions to interconvert between datetimes and fractional times

def datetime_to_fractional_days_from_reference(dateFuture):

    '''get reference start datetime'''

    datePast = datetime.datetime(1,1,1,0,0)
    difference = dateFuture - datePast
    return difference.total_seconds() / datetime.timedelta(days=1).total_seconds()

def datetime_to_fractional_minutes_from_reference(dateFuture):
    '''get reference start datetime'''

    datePast = datetime.datetime(1,1,1,0,0)
    difference = dateFuture - datePast
    return difference.total_seconds() / datetime.timedelta(minutes=1).total_seconds()

def fractional_days_from_reference_to_datetime(dateFuture):
    '''get reference start datetime'''

    datePast = datetime.datetime(1,1,1,0,0)
    future_timedelta = datetime.timedelta(days=dateFuture)
    return datePast + future_timedelta

def fractional_minutes_from_reference_to_datetime(dateFuture):
    '''get reference start datetime'''

    datePast = datetime.datetime(1,1,1,0,0)
    future_timedelta = datetime.timedelta(minutes=dateFuture)
    return datePast + future_timedelta

###------------------------------------------------------------------------------------###

def remove_duplicates_preserve_order(seq):

    '''remove duplicates from list while preserving order'''

    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

###------------------------------------------------------------------------------------###

def local_offset_from_utc(reference_utc_dt,latitude=-999.0,longitude=-999.0,pytz_local_tz=None):

    '''get pytz string timezone, and local timezone offset (in hours) of lat/lon pairing, relative to a reference UTC datetime
       if you give the function a pytz string timezone, it skips the calculation of the pytz string timezone by lat/lon (very costly) and proceeds to calculate local timezone offset
    '''

    import pytz
    from timezonefinder import TimezoneFinder
    utc = pytz.utc
    timezone_finder_obj = TimezoneFinder()

    #if are not provided
    if pytz_local_tz == None:

        latitude, longitude = float(latitude), float(longitude)

        #first try get local timezone within defined timezone polygons in timezonefinder module
        #all polygons (with shortcuts in that area) are checked until one polygon is matched.
        #None is returned without any match.
        #NOTE: The timezone polygons do NOT follow the shorelines any more. Just because you do not get None, the point could still lie off land!

        local_tz = timezone_finder_obj.certain_timezone_at(lat=latitude,lng=longitude)

        #try get pytz timezone object from local timezone string

        if local_tz != None:
            try:
                pytz_local_tz = pytz.timezone(local_tz)
            except:
                local_tz = None

        #if don't have local_tz or pytz has had a fit and will not recognise a timezone
        #force it to find different timezones until it recognises one
        #this is done using function that finds closest local timezone.
        #simply computes and compares the distances to the polygon boundaries!.
        #This by default returns the closest timezone of all polygons within +-1 degree lng and +-1 degree lat, or returns None if no matches.
        #Optionally can increase radius search using the delta_degree option.
        #Iteratively increase by 1 deg. until have a match
        #NOTE: force_evaluation=True, is set as some polygons might not be tested (for example when a zone is found to be the closest already), and this prevents this

        if local_tz == None:
            current_radius = 1

            timer = time.time() + 60
            while pytz_local_tz == None:
                # try to find the local timezone for ~1', otherwise break
                if time.time() > timer:
                    pytz_local_tz = None
                    break

                local_tz = timezone_finder_obj.closest_timezone_at(lat=latitude,lng=longitude,delta_degree=current_radius,force_evaluation=True)

                #get pytz timezone object from local timezone string
                #occasionally pytz will have a fit and not recognise a timezone
                #force it to run until it recognises one
                if local_tz != None:
                    try:
                        pytz_local_tz = pytz.timezone(local_tz)
                    except:
                        pass
                current_radius+=1

    #reference utc pytz datetime
    reference_pytz_dt = utc.localize(reference_utc_dt)

    #local time pytz datetime (relative to reference UTC datetime)
    if pytz_local_tz is not None:
        local_pytz_dt = pytz_local_tz.localize(reference_utc_dt)
    else:
        return np.NaN, np.NaN

    return pytz_local_tz, (reference_pytz_dt - local_pytz_dt).total_seconds() / 60. / 60.

###------------------------------------------------------------------------------------###

def natural_key(string_):

    '''key for sorting in human natural order'''

    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]

###------------------------------------------------------------------------------------###

def get_unique_processed_references(current_GHOST_stage, run_details, parameter_details):

    '''get unique processed station references from files saved out in the native/temporary file directories based on GHOST stage you are currently in'''

    processed_references = []

    #if current GHOST stage is in S stages, gather station files from (GHOST_stage-1)_temporary directory
    if 'S' in current_GHOST_stage:
        previous_GHOST_stage = 'S{}'.format(int(current_GHOST_stage[1:])-1)
        station_files = glob.glob(run_details['data_root']+'{}/{}/native/{}/temporary_{}/*.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],previous_GHOST_stage))
    #otherwise gather finalised station files from native directory
    else:
         station_files = glob.glob(run_details['data_root']+'{}/{}/native/{}/*.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name']))

    for file_i, file in enumerate(station_files):
        split = file.split('/')[-1].split('_')
        station_reference = '{}_{}'.format(split[0],split[1])
        processed_references.append(station_reference)

    #sort references in alphabetical order
    processed_references = sorted(np.unique(processed_references), key=natural_key)

    return processed_references

###------------------------------------------------------------------------------------###

def process_non_network_specific_metadata(site_metadata, data_file):

    '''process non-network specific metadata'''

    #get station_timezone (at reference 2000-01-01 00:00 UTC)
    if (pd.isnull(site_metadata['latitude']['value'][-1]) == False) & (pd.isnull(site_metadata['longitude']['value'][-1]) == False):
        dt_ref = datetime.datetime(2000,1,1,0,0)
        pytz_str, site_metadata['station_timezone']['value'][-1] = local_offset_from_utc(dt_ref,longitude=site_metadata['longitude']['value'][-1],latitude=site_metadata['latitude']['value'][-1])

    #save data_download_stamp --> rounded to nearest minute
    data_download_datetime = datetime.datetime.strptime(time.ctime(os.path.getmtime(data_file)), "%a %b %d %H:%M:%S %Y")
    data_download_datetime = round_to_minute(data_download_datetime)
    site_metadata['data_download_stamp']['value'][-1] = datetime_to_fractional_minutes_from_reference(data_download_datetime)

    return site_metadata

###------------------------------------------------------------------------------------###

def check_geographical_position(longitude,latitude):


    '''Check if longitude/latitude outside valid bounds: -180 <-> +180 for longitude, -90 <-> +90 for latitude;
       or both longitude/latitude == 0.0 --> in the middle of the ocean.
       if so, return longitude/latitude as NaN
    '''

    #convert longitude/latitude to float, if cannot, return pair of NaNs
    try:
        longitude = np.float64(longitude)
        latitude = np.float64(latitude)
    except:
        return np.NaN, np.NaN

    #set geographical position check flag
    geographical_position_check_flag = False

    #check if longitude is inside valid bounds (-180 <-> +180)
    if (longitude < -180) or (longitude > 180):
        geographical_position_check_flag = True

    #check if latitude is inside valid bounds (-90 <-> +90)
    if (latitude < -90) or (latitude > 90):
        geographical_position_check_flag = True

    #check if longitude and latitude are not both == 0.0, --> in the middle of the ocean
    if (longitude == 0.0) & (latitude == 0.0):
        geographical_position_check_flag = True

    #if geographical position check flag has been set as True, return pair of NaNs
    if geographical_position_check_flag == True:
        return np.NaN, np.NaN
    #otherwise return original longitude/latitude
    else:
        return longitude, latitude
###------------------------------------------------------------------------------------###

def process_overlapping_and_duplicate_times(metadata,data):

    '''Function to check if any times are duplicate or overlapping across the time record from a specific data file.
       Start times are first sorted in ascending order to resolve potential overlapping times (also metadata is sorted by timestamp).
       Firstly, if a temporal window is identically duplicated (i.e. same start and end time), data which is flagged as GHOST decreed invalid is first screened out. If this does not resolve the duplication, then the first remaining temporal window is kept preferentially.
       Secondly, after removing the duplicate windows, it is checked if any window end times overlap with the next start time. If an overlap is found, again data which is flagged as GHOST decreed invalid is first screened out. If this does not resolve the overlap, the remaining windows with a temporal resolution equal to the finest temporal resolution across the remaining windows are kept. If this still does not resolve the duplication, then the first remaining temporal window is kept preferentially.
       These checks are done recursively until each window has no other windows which are duplicates or overlap with it.
       Function also deals with re-ordering/modification of all variables which are time variant
    '''

    from GHOST_standards import standard_QA_name_to_QA_code

    #get QA flags that may be used
    duplicate_time_GHOST_decreed_valid_qa_flag = standard_QA_name_to_QA_code['Duplicate Time - GHOST Decreed Valid Flagged Values Kept']
    duplicate_time_first_value_qa_flag = standard_QA_name_to_QA_code['Duplicate Time - First Value Kept']
    overlap_time_GHOST_decreed_valid_qa_flag = standard_QA_name_to_QA_code['Overlapping Time - GHOST Decreed Valid Flagged Values Kept']
    overlap_time_finest_resolution_qa_flag = standard_QA_name_to_QA_code['Overlapping Time - Finest Temporal Resolutions Kept']
    overlap_time_first_value_qa_flag = standard_QA_name_to_QA_code['Overlapping Time - First Value Kept']

    invalid_GHOST_decreed_data_qa_flag = standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed']

    #sort measurement start time (and associated time variant variables) in ascending order to resolve potential overlapping times

    #get data parameter names
    data_keys = data.dtype.names

    #get indices of sorted time array
    sorted_time_indices = np.array(sorted(range(len(data['time'])), key=lambda k: data['time'][k]))

    #sort time array and other data arrays
    data = data[sorted_time_indices]

    #get indices of sorted metadata update stamp
    sorted_meta_update_stamp_indices = np.array(sorted(range(len(metadata['meta_update_stamp'])), key=lambda k: metadata['meta_update_stamp'][k]))

    #sort metadata update stamp array and other metadata arrays
    metadata = metadata[sorted_meta_update_stamp_indices]    

    #-----------------------
    #Add temporal resolutions to start times to get measurement end times
    #check for measurement windows which are duplicated or overlap iteratively with each individual window

    #iterate through time
    inds_to_remove = np.empty(0, dtype=np.uint32)
    for ii, start_time in enumerate(data['time'][:-1]):

        #only test for current time if not in list of bad indices to be removed
        if ii not in inds_to_remove:

            duplicate_inds = np.array([ii], dtype=np.uint32)
            overlap_inds = np.array([ii], dtype=np.uint32)

            #get window end time by adding temporal resolution as timedelta
            timedelta = data['temporal_resolution'][ii]

            #get measurement window end time
            end_time = start_time + timedelta

            #1. check if any measurement window is duplicate,
            #2. check if window end time overlaps with next start time,
            #keep checking until there are no overlaps/ duplicates with current window
            for jj, next_start_time in enumerate(data['time'][ii+1:]):

                #only test for next time if not in list of bad indices to be removed
                if ii+(jj+1) not in inds_to_remove:

                    #get window end time by adding temporal resolution as timedelta
                    next_timedelta = data['temporal_resolution'][ii+(jj+1)]

                    #get measurement window end time
                    next_end_time = next_start_time + next_timedelta

                    #duplicate check
                    if (start_time == next_start_time) & (end_time == next_end_time):
                        duplicate_inds=np.append(duplicate_inds,ii+(jj+1))
                    #overlap check
                    elif end_time > next_start_time:
                        overlap_inds=np.append(overlap_inds,ii+(jj+1))
                    #if no duplicates or overlaps, proceed to next start time
                    else:
                        break

            #resolve duplicate windows
            if len(duplicate_inds) > 1:

                #set qa flag write variables
                duplicate_first_value_write = False

                #first screen out GHOST decreed invalid flagged data
                duplicate_inds_valid = np.array([duplicate_ind for duplicate_ind in duplicate_inds if invalid_GHOST_decreed_data_qa_flag not in data['qa'][duplicate_ind]])

                #if removing GHOST decreed flagged data fails to resolve duplication, proceed to keep first duplicate value preferentially
                if len(duplicate_inds_valid) == 0:
                    duplicate_first_value_write = True
                    #get first index of original duplicate inds
                    chosen_ind = duplicate_inds[0]
                elif len(duplicate_inds_valid) > 1:
                    duplicate_first_value_write = True
                    #get first index of valid duplicate inds
                    chosen_ind = duplicate_inds_valid[0]
                else:
                    #get only remaining index of valid duplicate inds
                    chosen_ind = duplicate_inds_valid[0]

                #get bad inds to drop
                bad_inds = np.array([duplicate_ind for duplicate_ind in duplicate_inds if duplicate_ind != chosen_ind])

                #save out indices to remove
                inds_to_remove = np.append(inds_to_remove, bad_inds)

                #remove bad inds from overlap inds
                overlap_inds = np.array([overlap_ind for overlap_ind in overlap_inds if overlap_ind not in bad_inds])

                #write necessary qa flags

                #write out qa flag to state have attempting to break duplication by removing of GHOST decreed invalid flagged data
                data['qa'][chosen_ind].append(duplicate_time_GHOST_decreed_valid_qa_flag)

                if duplicate_first_value_write == True:
                    #write out qa flag to state have preferentially kept first duplicate value
                    data['qa'][chosen_ind].append(duplicate_time_first_value_qa_flag)


            #resolve overlapping windows
            if len(overlap_inds) > 1:

                #set qa flag write variables
                overlap_finest_resolutions_write = False
                overlap_first_value_write = False

                #first screen out GHOST decreed invalid flagged data
                overlap_inds_valid = np.array([overlap_ind for overlap_ind in overlap_inds if invalid_GHOST_decreed_data_qa_flag not in data['qa'][overlap_ind]])

                #if removing GHOST decreed flagged data fails to resolve overlap, proceed to attempt to resolve it by keeping finest temporal resolution data
                if len(overlap_inds_valid) == 0:
                    overlap_inds_valid = copy.deepcopy(overlap_inds)

                if len(overlap_inds_valid) > 1:
                    overlap_finest_resolutions_write = True

                    #get temporal resolutions of overlapping time periods
                    overlapping_window_resolutions = data['temporal_resolution'][overlap_inds_valid]

                    #filter overlap inds, only keeping indices == finest temporal resolution
                    overlap_inds_valid = overlap_inds_valid[np.where(overlapping_window_resolutions == np.min(overlapping_window_resolutions))[0]]

                    #if still have multiple overlap inds, the first window is preferentially kept
                    if len(overlap_inds_valid) > 1:
                        overlap_first_value_write = True

                #get first index of overlap inds
                chosen_ind = overlap_inds_valid[0]

                #get bad inds to drop
                bad_inds = np.array([overlap_ind for overlap_ind in overlap_inds if overlap_ind != chosen_ind])

                #save out indices to remove
                inds_to_remove = np.append(inds_to_remove, bad_inds)

                #write necessary qa flags

                #write out qa flag to state have attempting to break overlap by removing of GHOST decreed invalid flagged data
                data['qa'][chosen_ind].append(overlap_time_GHOST_decreed_valid_qa_flag)

                if overlap_finest_resolutions_write == True:
                    #write out qa flag to state have attempting to break overlap by keeping finest temporal resolution data
                    data['qa'][chosen_ind].append(overlap_time_finest_resolution_qa_flag)

                if overlap_first_value_write == True:
                    #write out qa flag for overlapping time window that is preferentially kept
                    data['qa'][chosen_ind].append(overlap_time_first_value_qa_flag)


    #if have data indices to remove (either duplicates/or overlaps), do so now
    #if metadata is reported line by line, then also delete metadata columns
    if len(inds_to_remove) > 0:
        data = np.delete(data, inds_to_remove)
        if len(metadata) > 1:
            metadata = np.delete(metadata, inds_to_remove)

    return metadata,data

###------------------------------------------------------------------------------------###

def view_fields(a, names):
    """
    `a` must be a numpy structured array.
    `names` is the collection of field names to keep.

    Returns a view of the array `a` (not a copy).
    """
    dt = a.dtype
    formats = [dt.fields[name][0] for name in names]
    offsets = [dt.fields[name][1] for name in names]
    itemsize = a.dtype.itemsize
    newdt = np.dtype(dict(names=names,
                          formats=formats,
                          offsets=offsets,
                          itemsize=itemsize))
    b = a.view(newdt)
    return b

def remove_fields(a, names):
    """
    `a` must be a numpy structured array.
    `names` is the collection of field names to remove.

    Returns a view of the array `a` (not a copy).
    """
    dt = a.dtype
    keep_names = [name for name in dt.names if name not in names]
    return view_fields(a, keep_names)

###------------------------------------------------------------------------------------###

def determine_key_changing_metadata(old_metadata, new_metadata):

    '''function for determining if certain key metadata fields have changed between files
       returns True if metadata has changed, False if not
    '''

    #define key metadata fields to check
    key_metadata_fields = ['latitude', 'longitude', 'altitude', 'sampling_height', 'measurement_altitude', 'distance_to_building', 'distance_to_kerb', 'distance_to_junction', 'distance_to_source', 'street_width', 'street_type', 'daytime_traffic_speed', 'daily_passing_vehicles', 'city', 'representative_radius',
                           'standardised_network_provided_area_classification', 'standardised_network_provided_station_classification', 'standardised_network_provided_main_emission_source', 'standardised_network_provided_land_use', 'standardised_network_provided_terrain', 'standardised_network_provided_measurement_scale',
                           'primary_sampling_instrument_name', 'primary_sampling_type', 'primary_sampling_instrument_reported_flow_rate',
                           'sample_preparation_types', 'sample_preparation_techniques', 
                           'measuring_instrument_name', 'measurement_methodology', 'measuring_instrument_sampling_type', 'measuring_instrument_reported_flow_rate', 'measuring_instrument_reported_lower_limit_of_detection', 'measuring_instrument_reported_upper_limit_of_detection', 'measuring_instrument_reported_uncertainty', 'measuring_instrument_reported_accuracy', 'measuring_instrument_reported_precision', 'measuring_instrument_reported_measurement_resolution', 'measuring_instrument_calibration_scale']

    #iterate through key metadata fields and check to see if there are differences between old/new metadata
    for key_metadata_field in key_metadata_fields:
        old_value = old_metadata[key_metadata_field]
        new_value = new_metadata[key_metadata_field]

        #if both values are NaN, continue to next field
        if (pd.isnull(old_value) == True) & (pd.isnull(new_value) == True):
            continue

        #test equivalence, if not equal, break out of function, as already know key metadata not equal between files
        if old_value != new_value:
            return True

    #if have made it through iteration of key metadata fields, key metadata is therefore equivalent
    return False

###------------------------------------------------------------------------------------###

def resolve_duplicate_metadata_from_file(metadata):

    '''function that removes duplicate metadata read from a data file.
       this happens when a network reports metadata line by line, and therefore metadata is collected and processed line by line.
       a consequence of this is therefore multiple columns are duplicated needlessly, where no key fields are changing between columns. 
      
       this function checks column by column and if all key metadata metadata values are identical to the last column, the column is removed.
    '''

    #if number of metadata column is already 1, do not need to run function
    if len(metadata) == 1:
        return metadata
    else:
        #initialise list to store metadata columns to remove
        metadata_column_indices_to_remove = []

        #iterate through metadata columns, check if one of key fields are different from previous column
        previous_metadata_column = copy.deepcopy(metadata[0])
        for metadata_column_ii, metadata_column in enumerate(metadata[1:]):

            #has one of key metadata values changed between columns?
            has_column_changed = determine_key_changing_metadata(previous_metadata_column, metadata_column)
            
            #if not append column index to remove
            if has_column_changed == False:
                metadata_column_indices_to_remove=np.append(metadata_column_indices_to_remove, metadata_column_ii+1)
            #otherwise take current column as previous column to test against
            else:
                previous_metadata_column = copy.deepcopy(metadata_column)

        #if have metadata columns to remove, do so now
        if len(metadata_column_indices_to_remove) > 0:
            metadata = np.delete(metadata, metadata_column_indices_to_remove)

    return metadata

###------------------------------------------------------------------------------------###

def convert_to_numpy_dataframe(data_dict):

    '''function that processes data dictionary (i.e. metadata/data) processed from raw data files to numpy dataframe
       this process also standardises the data formats of the raw data
       also does string formatting (i.e. upper, lower case strings).
       all missing elements are conserved as np.NaN values throughout processing, regardless of type
    '''

    #create lists for standardising data formats of the input data dictionary
    standard_dtype = [(key,data_dict[key]['data_type']) for key in data_dict]

    #create numpy dataframe to store standardised forms of variables inside data dictionary
    standardised_dataframe = np.empty(len(data_dict[list(data_dict.keys())[0]]['value']), dtype=standard_dtype)

    #fill numpy dataframe with data
    for key in standardised_dataframe.dtype.names:

        #cut data for data_key
        param_array = data_dict[key]['value']

        #do string formatting for relevant fields before filling dataframe
        if (standardised_dataframe[key].dtype == np.dtype('O')) & (key not in ['flag', 'qa', 'classification']):

            #upper case string formatting
            if 'upper' in data_dict[key]['string_format']:
                param_array = [np.str(x).replace(",", "-").upper() if pd.isnull(x) == False else x for x in param_array]

            #lower case string formatting
            elif 'lower' in data_dict[key]['string_format']:
                param_array = [np.str(x).replace(",", "-").lower() if pd.isnull(x) == False else x for x in param_array]

            #titalling string formatting
            elif 'title' in data_dict[key]['string_format']:
                param_array = [np.str(x).replace(",", "-").title() if pd.isnull(x) == False else x for x in param_array]

            #standard string formatting
            else:
                param_array = [np.str(x).replace(",", "-") if pd.isnull(x) == False else x for x in param_array]

        #fill dataframe
        standardised_dataframe[key] = param_array

    return standardised_dataframe

###------------------------------------------------------------------------------------###

def redo_standardisations(dataframe, fields_to_standardise, parameter_details):

    '''take numpy dataframe for metadata/data and a list of fields that have been updated/modified ensure are standardised correctly.
       this is imperative for string formatting, e.g. to ensure capitalisations are correct, but the function will check standards of all passed fields.
    '''
    
    from GHOST_standards import get_standard_data, get_standard_metadata

    #get standard formats for dataframe
    if 'latitude' in dataframe.dtype.names:
        standard_dict = get_standard_metadata(parameter_details)
    else:
        standard_dict = get_standard_data(parameter_details)

    #iterate through provided fields
    for key in fields_to_standardise:

        #get list of values for field
        param_array = dataframe[key]

        #get standard data type for field
        standard_dtype = dataframe[key].dtype

        #get string format for field
        string_format = standard_dict[key]['string_format']

        #do string formatting for relevant fields before filling dataframe
        if (standard_dtype == np.dtype('O')) & (key not in ['flag', 'qa', 'classification']):

            #upper case string formatting
            if 'upper' in string_format:
                param_array = [np.str(x).replace(",", "-").upper() if pd.isnull(x) == False else x for x in param_array]

            #lower case string formatting
            elif 'lower' in string_format:
                param_array = [np.str(x).replace(",", "-").lower() if pd.isnull(x) == False else x for x in param_array]

            #titalling string formatting
            elif 'title' in string_format:
                param_array = [np.str(x).replace(",", "-").title() if pd.isnull(x) == False else x for x in param_array]

            #standard string formatting
            else:
                param_array = [np.str(x).replace(",", "-") if pd.isnull(x) == False else x for x in param_array]

        #rewrite dataframe
        dataframe[key] = param_array

    return dataframe

###------------------------------------------------------------------------------------###

def write_station_specific_numpy(metadata,data,run_details,parameter_details,GHOST_stage,update_file = True):

    '''write out metadata/data to station specific numpy .npz file
       update_file variable is used to overwrite file when want to update it
    '''

    #get station reference
    station_reference = metadata['station_reference'][0]

    #write numpy file

    #get start and end of data ranges from file
    start_time = data['time'][0]
    end_time = data['time'][-1]+int(data['temporal_resolution'][-1])

    #deal with instances where temporary station file already exists and are not updating file - write dupN suffix variable for differentiation
    if update_file == False:
        temporary_file_exists = os.path.isfile(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],GHOST_stage,station_reference,start_time,end_time))

        if temporary_file_exists == True:
            dup_n = 1
            while temporary_file_exists == True:
                if os.path.isfile(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}:dup{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],GHOST_stage,station_reference,start_time,end_time,dup_n)) == False:
                    break
                dup_n +=1
            fname = run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}:dup{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],GHOST_stage,station_reference,start_time,end_time,dup_n)
            np.savez(fname,data=data,metadata=metadata)
            #give 770 permissions to file
            os.chmod(fname, 0o770)
            return

    #write file for the first time or update it
    fname = run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_{}_{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],GHOST_stage,station_reference,start_time,end_time)
    np.savez(fname,data=data,metadata=metadata)
    #give 770 permissions to file
    os.chmod(fname, 0o770)
    return

###------------------------------------------------------------------------------------###

def post_process_file_parse(metadata,data,run_details,parameter_details):

    '''standardise metadata/data formats, check for overlapping/duplicate times, and write out site data to temporary numpy file'''

    #standardise data formats - convert to numpy dataframes
    metadata = convert_to_numpy_dataframe(metadata)
    data = convert_to_numpy_dataframe(data)

    #check for duplicated/overlapping times across whole time record
    metadata,data = process_overlapping_and_duplicate_times(metadata,data)

    #remove duplicated metadata columns read from file
    metadata = resolve_duplicate_metadata_from_file(metadata)

    #write out to temporary station specific numpy file
    write_station_specific_numpy(metadata,data,run_details,parameter_details,'S1',update_file=False)

    return

###------------------------------------------------------------------------------------###

def read_from_numpy(numpy_file, desired_variables):

    '''read numpy .npz files, returning desired variables (input as list) stored in a dictionary: {desired_var_name: var}'''

    numpy_dataframe = np.load(numpy_file, allow_pickle=True)

    loaded_variables = {}

    for desired_variable in desired_variables:
        loaded_variables[desired_variable] = numpy_dataframe[desired_variable]

    return loaded_variables

###------------------------------------------------------------------------------------###

def cut_data_array(data_array,data_type,period_start,period_end):

    '''function which cuts data/metadata arrays by a given data range, provided as start minute from reference + end minute from reference'''

    if data_type == 'data':
        #cut relevant indices
        start_i = bisect.bisect_left(data_array['time'], period_start)
        end_i =   bisect.bisect_left(data_array['time'], period_end)

        #cut data array
        data_array = data_array[start_i:end_i]

    if data_type == 'metadata':

        #cut relevant indices
        start_i = bisect.bisect_left(data_array['meta_update_stamp'], period_start)
        end_i =   bisect.bisect_left(data_array['meta_update_stamp'], period_end)

        #flag for determining if need to update first metadata update stamp
        update_start_stamp = False

        #if only 1 metadata value, and data period starts after when metadata start stamp, then modify start_meta_i to be 0
        #also set flag for updating of meta update stamp
        if len(data_array['meta_update_stamp']) == 1:
            if period_start > data_array['meta_update_stamp']:
                start_i = 0
                update_start_stamp = True

        #if have more than 1 metadata value and data range starts between metadata update periods, then adjust first metadata field to be from previous index
        if len(data_array['meta_update_stamp']) > 1: 
            if period_start < data_array[start_i]['meta_update_stamp']:
                start_i = start_i - 1
                update_start_stamp = True

        #cut metadata array
        data_array = data_array[start_i:end_i]

        #if need to update first metadata time stamp, do so now
        if update_start_stamp == True:
            data_array[0]['meta_update_stamp'] = period_start

    return data_array

###------------------------------------------------------------------------------------###

def get_join_variable_specific_external_metadata(external_metadata, join_field_column_name, join_variable, station_reference, run_details, log_file_str, need_variable=False):

    '''get external metadata specific for a certain join variable.
       if do not have data for a join variable and "need_variable" is set to "True" then stop process and return handled error logfile (2)
       if do not have data for a join variable and "need_variable" is set to "False" then return empty list.
    '''

    #first, check if you have join variable in join field column
    if join_variable not in external_metadata[join_field_column_name]:
        #if do not have join variable in join field column , and "need_variable" is set to "True", write handled error logfile: 2
        if need_variable == True:
            log_file_str += 'JOIN VARIABLE: {} DOES NOT EXIST IN EXTERNAL STATION METADATA JOIN FIELD COLUMN.'.format(join_variable)
            create_GHOST_stage_logfile('S2',run_details,2,log_file_str,station_reference)
        #if do not have join variable in join field column , and "need_variable" is set to "False", return empty list
        else:
            return []            
    #if have join variable in join field column, get join variable specific metadata
    else:
        join_variable_index = np.where(external_metadata[join_field_column_name] == join_variable)[0][0]
        return external_metadata[join_variable_index]

###------------------------------------------------------------------------------------###

def process_manually_compiled_metadata(station_reference,metadata,data,run_details,parameter_details,network_details,log_file_str):

    '''function for handling manually compiled metadata.
       Manually compiled metadata is used to set metadata fields for the entire time record when there is no previously processed valid metadata throughout the time record for a field.
       For some fields (i.e. classifications), multiple manually complied fields are processed together to return standard output.
    '''

    from GHOST_standards import standard_QA_name_to_QA_code, metadata_modification_field_to_QA_name
    import standardisation_mapping_functions

    #determine if have manually compiled metadata for network (if there is the 'MANUAL' string in the raw external filename)
    have_external_metadata_file = any(['MANUAL' in f for f in network_details['external_metadata_files']])

    #list for keeping modified standard field names
    modified_standard_fields = []

    #parse external metadata file if exists
    if have_external_metadata_file == True:

        #read external metadata file into memory (specific for network)
        external_metadata = np.load(run_details['data_root']+'{}/metadata/processed/{}_META.npy'.format(run_details['network'], run_details['network']), allow_pickle=True)

        #get station reference without standard method abbreviation
        station_reference_split = station_reference.split('_')[:-1]
        station_reference_without_method = '_'.join(station_reference_split)

        #get external metadata for the specific station reference being processed (if none exists, then return handled error logile (2))
        external_metadata_station = get_join_variable_specific_external_metadata(external_metadata, network_details['metadata_data_join_variable_name'], station_reference_without_method, station_reference, run_details, log_file_str, need_variable=True)

        #load dictionary that stores mapping information for variable names between manually compiled metadata file and standard fields
        variable_mapping_dictionary = getattr(standardisation_mapping_functions, '{}_manual_external_meta_map'.format(run_details['network']))

        #check if any station classification information is stored within the external file, if so, package it together in a separate dictionary
        classification_standard_variables = ['area_classification', 'station_classification', 'main_emission_source', 'land_use', 'terrain', 'measurement_scale']         
        manual_native_classifications = {} 
        for k,v in list(variable_mapping_dictionary.items()):
            if v in classification_standard_variables:
                manual_native_classifications[v] = external_metadata_station[k]
                #remove all existing classification variables from variable_mapping_dictionary
                del variable_mapping_dictionary[k]

        #if have some station classification information to process, then do do
        if len(manual_native_classifications) > 0:
            standard_classifications = standardisation_mapping_functions.map_network_classifications_to_standard_classifications(manual_native_classifications, metadata, run_details, write_to_metadata_array=False)
            #iterate through standard classification metadata variables 
            for standard_classification_variable in list(standard_classifications.keys()):
                #is standard classification value not NaN?
                if pd.isnull(standard_classifications[standard_classification_variable]) == False:
                    #if you do not have any values saved for current standard classification field, then set it through time record 
                    if all(pd.isnull(metadata[standard_classification_variable])) == True:
                        metadata[standard_classification_variable] = [standard_classifications[standard_classification_variable]] * len(metadata[standard_classification_variable])
                        #add standard variable name to modified_standard_fields list
                        modified_standard_fields.append(standard_classification_variable)
                        #if have modified any of standard station classification fields a QA code should be written to inform of this
                        qa_flag_code = standard_QA_name_to_QA_code['No {} Metadata - Used Manually Compiled Metadata'.format(metadata_modification_field_to_QA_name[standard_classification_variable])]
                        for inv in range(len(data['qa'])):
                            data['qa'][inv].append(qa_flag_code)

        #iterate through rest of variables in manually compiled external metadata file
        for raw_var in list(variable_mapping_dictionary.keys()):

            #get standard variable name
            standard_var = variable_mapping_dictionary[raw_var]

            #is external metadata value not NaN?
            if pd.isnull(external_metadata_station[raw_var]) == False:
                #if do have any values saved for current standard classification field, then set through time record 
                if all(pd.isnull(metadata[standard_var])) == True:
                    metadata[standard_var] = [external_metadata_station[raw_var]] * len(metadata[standard_var])
                    #add standard variable name to modified_standard_fields list
                    modified_standard_fields.append(standard_var)
                    #if have modified a metadata modification field, a QA code should be written to inform of this
                    if standard_var in list(metadata_modification_field_to_QA_name.keys()):
                        qa_flag_code = standard_QA_name_to_QA_code['No {} Metadata - Used Manually Compiled Metadata'.format(metadata_modification_field_to_QA_name[standard_var])]
                        for inv in range(len(data['qa'])):
                            data['qa'][inv].append(qa_flag_code)
                
    return metadata, data, modified_standard_fields

###------------------------------------------------------------------------------------###

def handle_gapped_key_metadata(station_reference, metadata, data, run_details, log_file_str):

    '''function for resolving gaps in chosen metadata fields across time record.
       if one of the metadata fields is NaN at any point in time modify them to be the most recent valid value, or if not possible the next valid value.
       for locational metadata, this stops stations further down the GHOST pipeline being split apart due to missing fields which may have erroneously being left out (or due to metadata reporting standards improving with time)
       it is therefore assumed that the most recent valid metadata value, or next valid value is applicable for the missing period.
       if one of the fields is deemed a 'key metadata' field, appropriate QA flags are written to inform of instances where these assumptions are made.
       the key metadata fields defined for this check are: latitude, longitude, altitude, sampling_height, standardised_network_provided_area_classification, standardised_network_provided_station_classification, standardised_network_provided_main_emission_source, standardised_network_provided_land_use, standardised_network_provided_terrain, standardised_network_provided_measurement_scale, representative_radius
 
       if all of one of the latitude or longitude fields are NaNs, do not run function, and do not proceed with GHOST pipeline for the specific station further.
       if all of altitudes are NaN then these in a later function are overwritten by ETOPO1 altitude estimations, an appropriate QA flag is written to inform of this measure taken for the altitudes.

       no assumptions are made for any measurement process information as it is deemed too big an uncertainty
       no assumptions are also made for any fields that are highly sensitive with time (daytime_traffic_speed, daytime_passing_vehicles, population)
       fields which are always written out are not processed (GHOST_version, station_reference, data_timezone, data_level)
    '''

    from GHOST_standards import standard_QA_name_to_QA_code, metadata_modification_field_to_QA_name

    #get metadata update timestamps
    update_metadata_stamp = metadata['meta_update_stamp']

    #set meta fields to do check upon
    meta_fields = ['latitude','longitude','altitude','sampling_height','measurement_altitude','standardised_network_provided_area_classification',
    'standardised_network_provided_station_classification', 'standardised_network_provided_main_emission_source',
    'standardised_network_provided_land_use', 'standardised_network_provided_terrain', 'standardised_network_provided_measurement_scale',
    'representative_radius', 'station_timezone', 'distance_to_building', 'distance_to_kerb', 'distance_to_junction', 'distance_to_source',
    'street_width', 'street_type', 'horizontal_geodetic_datum', 'climatology', 'station_name', 'city', 'province', 'county', 'country',
    'associated_networks', 'principal_investigator_name', 'principal_investigator_institution', 'principal_investigator_email_address',
    'contact_name', 'contact_institution', 'contact_email_address']

    #iterate through meta fields
    for meta_field in meta_fields:

        #get metadata values for field
        meta_values = metadata[meta_field]

        #if all of one of the latitude or longitude fields are all NaN, then do not proceed further with function for this station.
        #return logfile to inform that station is to be removed from further processing in pipeline (1)
        if (meta_field == 'latitude') or (meta_field == 'longitude'):
            if all(pd.isnull(meta_values) == True):
                log_file_str += '{} -- ALL OF {} VALUES ARE NaN. STATION REMOVED FROM FURTHER PROCESSING IN PIPELINE'.format(station_reference, meta_field.upper())
                create_GHOST_stage_logfile('S2',run_details,1,log_file_str,station_reference)

        #if all of altitudes are NaN, then temporarily leave them be. These will be overwritten by ETOPO1 altitude estimation in one of the further stages.
        if (meta_field == 'altitude'):
            if all(pd.isnull(meta_values) == True):
                #continue to next meta field
                continue

        #if any one of the key metadata field values are NaN values, modify them to be the most recent valid value, or if not possible the next valid value
        #this stops stations further down the GHOST pipeline from being split apart due to missing fields which may have erroneously being left out (or due to standards of metadata reporting improving with time)
        #it is therefore assumed that the most recent valid metadata value, or next valid value is applicable for the missing period.
        #for each instance that a key metadata field value is modified, write a quality control flag to inform of this assumption

        #continue to next key metadata field if all values are NaN, or none are
        if (all(pd.isnull(meta_values) == True)) or (all(pd.isnull(meta_values) == False)):
            continue

        #set arrays for writing new metadata, and keeping indices of values modified
        new_meta = copy.deepcopy(meta_values)
        invalid_inds_backward = []
        invalid_inds_forward = []

        #iterate through all values for field
        for jj,val in enumerate(meta_values):
            #if current value is NaN, first check for most recent valid value (going backwards in time), don't check form index 0 as no values before
            if pd.isnull(val) == True:
                #if not == index 0
                if jj != 0:
                    for check_valid_var_back in meta_values[jj-1::-1]:
                        #if have found valid value overwrite it temporally to the current time window for purposes of this function
                        if pd.isnull(check_valid_var_back) == False:
                            new_meta[jj] = copy.deepcopy(check_valid_var_back)
                            #Save relevant data indices to write QC flag informing that are assuming some key metadata
                            #is it last index?
                            if jj == (len(meta_values)-1):
                                invalid_inds_backward.extend(range(bisect.bisect_left(data['time'], update_metadata_stamp[jj]), len(data['time'])))
                            else:
                                invalid_inds_backward.extend(range(bisect.bisect_left(data['time'], update_metadata_stamp[jj]), bisect.bisect_left(data['time'], update_metadata_stamp[jj+1])))
                            break
            #if have not found valid value going backwards in time, check forwards in time
            if pd.isnull(new_meta[jj]) == True:
                for check_valid_var_forward in meta_values[jj+1:]:
                    #if have found valid value overwrite it temporally to the current time window for purposes of this function
                    if pd.isnull(check_valid_var_forward) == False:
                        new_meta[jj] = copy.deepcopy(check_valid_var_forward)
                        #Save relevant data indices to write QC flag informing that are assuming some key metadata
                        #is it last index?
                        if jj == (len(meta_values)-1):
                            invalid_inds_forward.extend(range(bisect.bisect_left(data['time'], update_metadata_stamp[jj]), len(data['time'])))
                        else:
                            invalid_inds_forward.extend(range(bisect.bisect_left(data['time'], update_metadata_stamp[jj]), bisect.bisect_left(data['time'], update_metadata_stamp[jj+1])))
                        break

        #update metadata field, with modified data
        metadata[meta_field] = copy.deepcopy(new_meta)

        #write quality control flags out (if have made modifications to key meta fields)
        if meta_field in list(metadata_modification_field_to_QA_name.keys()):
            if len(invalid_inds_backward) > 0:
                #get relevant qa flag code
                qa_flag_code = standard_QA_name_to_QA_code['No {} Metadata - Took Most Recent Valid Value'.format(metadata_modification_field_to_QA_name[meta_field])]
                for inv in invalid_inds_backward:
                    data['qa'][inv].append(qa_flag_code)

            if len(invalid_inds_forward) > 0:
                #get relevant qa flag code
                qa_flag_code = standard_QA_name_to_QA_code['No {} Metadata - Took Next Valid Value'.format(metadata_modification_field_to_QA_name[meta_field])]
                for inv in invalid_inds_forward:
                    data['qa'][inv].append(qa_flag_code)

    return metadata, data

###------------------------------------------------------------------------------------###

def join_station_measurements(slurm_job_ID, network, parameter, station_reference):

    '''function for handling of all processed reported station specific data files to create continuous time series (joining measurements using the same methodology).
       this function specifically handles instances of temporally overlapping reported files, in a logical manner, creating consistently time series.

       first, all files are grouped by raw station_reference:measurement_method

       if only 1 file is available for a station_reference:measurement_method, then data file is simply taken as is

       if more than 1 file is available for a station_reference:measurement_method but there is no overlap between any of them, then the files are simply appended together

       if more than 1 file is available for a station_reference:measurement_method and there is overlap between them, then a stepped process is undertaken to determine which files should be kept on the occasions of data overlap.
       The entire potential time series is split into segments, which correspond to unique periods of different reporting files.
       i.e. for 3 months, it is possible 3 files are reporting data for the period, and after that for a month just 1 file, and after that 2 files. In this case 3 segments would be created to represent the variability of files reporting data.
       If segments where just 1 file is reporting data, data is simply taken from the 1 unique file for that period.
       In cases where there are multiple files reporting data in a segment then then a 2 step process is undertaken to determine which data should be kept.
       Firstly the overlap is attempted to be resolved by priority data levels rankings. Data files with higher data level rankings are preferentially kept over data files with lower data level rankings.
       If overlaps still exist after this check, then only data files with data level rankings equal to the max data level rank continue to be considered, and a ranking algorithm is used to determine which data file should be chosen for the period from the remaining files.
       The logic applied in this algorithm takes into account the parameters: latitude, longitude, altitude, sampling height, measurement altitude, and standard instrument name.
       Data from all the different station_reference:measurement_method specific files is first read and counts are made of the modally occurring values of the paramaters stated above (excluding NaNs).
       The algorithm favours consistency, that is to say taking data for which the metadata states is was measured and reported in the near as possible conditions as was accepted in the previous time step, and are most utilised through the time record.
       In each segment where there are multiple reporting files, multiple rankings are made based on certain information (described below) provided in each data file:
       The ranking counts used for in ranking algorithm are:
       if latitude == last accepted latitude --> +1 Count (check not counted if value == NaN)
       if longitude == last accepted longitude --> +1 Count(check not counted if value == NaN)
       if altitude == last accepted altitude  --> +1 Count (check not counted if value == NaN)
       if sampling_height == last accepted sampling_height --> +1 Count (check not counted if value == NaN)
       if measurement_altitude == last accepted measurement_altitude --> +1 Count (check not counted if value == NaN)
       if measurement_methodology == last accepted measurement_methodology --> +1 Count (check not counted if value == NaN)
       if measuring_instrument_name == last accepted measuring_instrument_name  --> +1 Count (check not counted if value == NaN)
       sorted count latitude --> +N count (where N is the number of different latitudes through time record, with the latitude matching that with the highest count given the largest N, working down)
       sorted count longitude --> +N count (where N is the number of different longitudes through time record, with the longitude matching that with the highest count given the largest N, working down)
       sorted count altitude --> +N count (where N is the number of different altitudes through time record, with the altitude matching that with the highest count given the largest N, working down)
       sorted count sampling_height --> +N count (where N is the number of different sampling_heights through time record, with the sampling_height matching that with the highest count given the largest N, working down)
       sorted count measurement_altitude --> +N count (where N is the number of different measurement_altitudes through time record, with the measurement_altitude matching that with the highest count given the largest N, working down)
       sorted count measurement_methodology --> +N count (where N is the number of different measurement_methodologies through time record, with the measurement_methodology matching that with the highest count given the largest N, working down)
       sorted count measuring_instrument_name --> +N count (where N is the number of different measuring_instrument_names through time record, with the instrument_name matching that with the highest count given the largest N, working down)
       highest average measurement resolution in relevant segment of file --> +N count (where N is the number of different reporting files)
       number of valid measurement points in relevant segment of file --> +N count (where N is the number of different reporting files)
       The total summative ranking is taken for all files after all the ranks are done, and the file which garners the highest total ranking count is kept.
       If there is a tiebreak, this is first attempted to be broken by ranking the number of valid measurement points across the whole files.
       If there is still a tiebreak, this is then attempted top be broken by counting the amount of non-NaN metadata variables provided.
       If there is still a tiebreak, the first index is preferentially taken.

       after the unification of all data/metadata by station, empty metadata fields are attempted to be filled (through time record) by the incorporation of manually compiled metadata

       next, gaps in selected metadata fields (i.e. those not too sensitive temporally) are attempted to be filled through interpolation either backwards (priority) and forwards through record, assuming values hold through time (writing QA codes to inform of assumptions for key meta fields)
       during this procedure, if all of one of the latitude or longitude fields are NaNs, do not proceed with GHOST pipeline for the specific station further.

       ultimately unique files are generated per station_reference:measurement_method, containing data/metadata for the entire time record (with the initial files removed)
    '''

    from GHOST_standards import standard_measurement_methodologies, standard_QA_name_to_QA_code
    import scipy.stats as ss

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- JOIN STATION SPECIFIC MEASUREMENTS\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get list of data files specific for station_reference:method
    data_files = sorted(glob.glob(run_details['data_root']+'{}/{}/native/{}/temporary_S1/{}_*'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],station_reference)))

    #------------------------------------------------------------------------------------------------------------------------#
    #------------------------------------------------------------------------------------------------------------------------#

    #if have just one data file, nothing to process, just read file into memory
    if len(data_files) == 1:
        log_file_str += '{} -- ONLY 1 FILE. NOTHING TO PROCESS.\n'.format(station_reference)
        loaded_variables = read_from_numpy(data_files[0], ['data', 'metadata'])
        existing_data = loaded_variables['data']
        existing_metadata = loaded_variables['metadata']

    #------------------------------------------------------------------------------------------------------------------------#
    #------------------------------------------------------------------------------------------------------------------------#

    #if have more than 1 file, determine if have data overlaps between files
    #this is done by simply looking at the ordered files (by start minute and then end minute),
    #if the start of the next file begins before the end of the previous file, then overlaps exist.
    elif len(data_files) > 1:

        relevant_file_dictionary = {}

        have_overlaps = False

        for data_i, data_file in enumerate(data_files):
            split = data_file.split('/')[-1].split(':')[0].split('.')[0].split('_')
            file_start = int(split[2])
            file_end = int(split[3])
            if data_i > 0:
                if file_start < previous_end:
                    have_overlaps = True
            previous_end = copy.deepcopy(file_end)

            #load data/metadata for file, and store in dictionary for later access (preventing re-loading)
            loaded_variables = read_from_numpy(data_file, ['data', 'metadata'])
            data = loaded_variables['data']
            metadata = loaded_variables['metadata']
            relevant_file_dictionary[data_file] = {'data':data,'metadata':metadata,'file_start':file_start,'file_end':file_end,'non_checked_indices':range(len(data))}

        #--------------------------------------------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------------------------------------------#

        #if there are no overlapping files, simply append them together
        if have_overlaps == False:
            log_file_str += '{} -- MULTIPLE FILES, BUT NO OVERLAPS. APPENDING FILES.\n'.format(station_reference)
            for data_ii, data_file in enumerate(data_files):
                if data_ii == 0:
                    #if first file to put data in time series, create variables for continuous time series
                    existing_data = copy.deepcopy(relevant_file_dictionary[data_file]['data'])
                    existing_metadata = copy.deepcopy(relevant_file_dictionary[data_file]['metadata'])
                else:
                    #append current metadata/data to existing metadata/data
                    data = copy.deepcopy(relevant_file_dictionary[data_file]['data'])
                    metadata = copy.deepcopy(relevant_file_dictionary[data_file]['metadata'])
                    existing_data_n_rows = existing_data.shape[0]
                    data_n_rows = data.shape[0]
                    existing_data.resize(existing_data_n_rows+data_n_rows)
                    existing_data[-data_n_rows:] = data
                    #check if each current metadata column is same as last key appended metadata column
                    #if different, a new metadata block is appended from column onwards
                    for metadata_ii, metadata_column in enumerate(metadata):
                        metadata_has_changed = determine_key_changing_metadata(existing_metadata[-1],metadata_column)
                        #append metadata
                        if metadata_has_changed == True:
                            metadata_block = metadata[metadata_ii:]
                            existing_metadata_n_rows = existing_metadata.shape[0]
                            metadata_n_rows = metadata_block.shape[0]
                            existing_metadata.resize(existing_metadata_n_rows+metadata_n_rows)
                            existing_metadata[-metadata_n_rows:] = metadata_block
                            break
        #--------------------------------------------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------------------------------------------#

        #if have overlapping files, split the entire potential time series into series of segments where different combinations of files provide data
        #determine which parts of time series can be taken from which file. If only have 1 file for a certain portion, take data from that file.
        #in periods where have duplicate data from different files, first preferentially keep data with highest ranked data level, and if still have overlap then use ranking algorithm to determine which file to take data from
        else:
            log_file_str += '{} -- MULTIPLE FILES, WITH OVERLAPS. CHOOSING BEST FILE FOR EACH OVERLAPPING PERIOD USING ALGORITHM.\n'.format(station_reference)
            #----------------------------------------------------------------------------------------------------------------#
            #----------------------------------------------------------------------------------------------------------------#

            #before iteration of overlap instances, get counts (in terms of minutes represented) of certain key metadata values across entire time series (from all possible files)
            #do not allow NaNs to be counted
            metadata_vars = ['latitude','longitude','altitude','sampling_height','measurement_altitude','measurement_methodology','measuring_instrument_name']
            metadata_count_dicts = {'latitude':{}, 'longitude':{}, 'altitude':{}, 'sampling_height':{}, 'measurement_altitude':{}, 'measurement_methodology':{}, 'measuring_instrument_name':{}}
            metadata_rank_dicts = {'latitude':{}, 'longitude':{}, 'altitude':{}, 'sampling_height':{}, 'measurement_altitude':{}, 'measurement_methodology':{}, 'measuring_instrument_name':{}}
            #iterate through data files
            for data_i, data_file in enumerate(data_files):
                metadata = relevant_file_dictionary[data_file]['metadata']
                #iterate through all metadata columns
                for metadata_ii, metadata_column in enumerate(metadata):
                    #ger minutes that metadata column represents
                    if (metadata_ii+1) == len(metadata):
                        meta_representativity = relevant_file_dictionary[data_file]['file_end'] - metadata_column['meta_update_stamp'] 
                    else:
                        meta_representativity = metadata[metadata_ii+1]['meta_update_stamp'] - metadata_column['meta_update_stamp']
                    #iterate through all metadata fields
                    for metadata_var in metadata_vars:
                        metadata_value = metadata_column[metadata_var]
                        if (pd.isnull(metadata_value) == False):
                            #if metadata value is already registered in count dictionary, add the minutes represented of metadata fields
                            if metadata_value in metadata_count_dicts[metadata_var]:
                                metadata_count_dicts[metadata_var][metadata_value] = metadata_count_dicts[metadata_var][metadata_value] + meta_representativity
                            #otherwise add metadata value to count dictionary
                            else:
                                metadata_count_dicts[metadata_var][metadata_value] = meta_representativity

            #give each of the counts a ranking based on numbers of minutes represented (more minutes represented = higher rank)
            for metadata_var in metadata_vars:
                metadata_rank_dicts[metadata_var] = {key:rank for key,rank in zip(list(metadata_count_dicts[metadata_var].keys()), ss.rankdata(list(metadata_count_dicts[metadata_var].values())))}

            #----------------------------------------------------------------------------------------------------------------#
            #----------------------------------------------------------------------------------------------------------------#

            #split time series into segments of periods where different combinations of files provide data
            #this is done by iterating through all measurement windows, of all data files and checking for overlaps with measurement windows of all other data files
            #all periods where there are overlaps across multiple data files are counted as the same segments until there are no more overlaps and the contributing files change.
            #where there are overlaps between files, the edges of the relevant measurement windows are taken as the borders of the overlapping setion.

            #create timeseries dict for storing unique segment start/ends associated with files providing data in the segments
            timeseries_compilation_dict = {}

            #first do quick and dirty check to group together files which overlap based on file start/end times

            #list to store grouped lists of associated overlapping files
            overlap_groups = []

            #initialise list wrote_lists to store all the files you have associated with a list
            wrote_files = []

            #iterate through all files
            for base_data_ii, base_data_file in enumerate(data_files):

                #if have already associated base file, continue to next file
                if base_data_file in wrote_files:
                    continue

                #set current start of overlap group as beginning/end of iterated base file
                current_group_start = relevant_file_dictionary[base_data_file]['file_start']
                current_group_end = relevant_file_dictionary[base_data_file]['file_end']

                #initialise current group files as current base file
                current_group_files = [base_data_file]
                #add base file to wrote files
                wrote_files.append(base_data_file)

                #set flag to keep iterating over files until current group start/end boundaries are unmodified
                keep_iterating = True
                while keep_iterating == True:
                    #set keep iterating flag to False, i.e. will stop iterating over list of next files, if current group start/end is not modified on this pass
                    keep_iterating = False

                    #iterate through all files after current base file
                    for next_data_ii, next_data_file in enumerate(data_files[base_data_ii+1:]):

                        #if have already iterated associated file, continue to next file
                        if next_data_file in wrote_files:
                            continue

                        #get start/end times of iterated file
                        next_file_start = relevant_file_dictionary[next_data_file]['file_start']
                        next_file_end = relevant_file_dictionary[next_data_file]['file_end']

                        #determine if iterated file has overlap with current_group
                        have_overlap = (current_group_start <= next_file_start < current_group_end)

                        #if have overlap, then append iterated file to current group files and modify current end of group overlap
                        #also write filename to list of files you have already associated
                        if have_overlap == True:
                            current_group_files.append(next_data_file)
                            wrote_files.append(next_data_file)

                            #determine if group start/end need modifying
                            #if either do, reset keep_iterating to true
                            if next_file_start < current_group_start:
                                current_group_start = copy.deepcopy(next_file_start)
                                keep_iterating = True
                            if next_file_end > current_group_end:
                                current_group_end = copy.deepcopy(next_file_end)
                                keep_iterating = True

                #append associated overlapping data files together as list to overlap groups list
                overlap_groups.append(current_group_files)

            #iterate through overlapping group data files and compare measurement windows with all other files (the files in this iteration which be called base files)
            for overlap_group_data_files in overlap_groups:
                #if only have 1 file in group, no overlaps with that file therefore simply write file start/end as segment start/end
                if len(overlap_group_data_files) == 1:
                    timeseries_compilation_dict['{}-{}'.format(relevant_file_dictionary[overlap_group_data_files[0]]['file_start'],relevant_file_dictionary[overlap_group_data_files[0]]['file_end'])] = overlap_group_data_files

                #else if have overlapping files, resolve overlaps window by window
                else:

                    for base_data_ii, base_data_file in enumerate(overlap_group_data_files):
                        #get data for base file
                        data = relevant_file_dictionary[base_data_file]['data']

                        #get base file start/end times
                        base_file_start = relevant_file_dictionary[base_data_file]['file_start']
                        base_file_end = relevant_file_dictionary[base_data_file]['file_end']

                        #get indices of measurement windows in base file that have not been cross-checked with other file windows
                        base_non_checked_indices = relevant_file_dictionary[base_data_file]['non_checked_indices']

                        #get base data file non-checked window start times and measurement resolutions
                        base_file_times = data['time'][base_non_checked_indices]
                        base_file_temporal_resolutions = data['temporal_resolution'][base_non_checked_indices]

                        if len(base_file_times) > 0:

                            #initialise previous iteration variables
                            previous_iteration_end = 0
                            previous_iteration_files = []

                            #initialise previous/current segment variables
                            current_segment_start = base_file_times[0]
                            current_segment_end = current_segment_start + base_file_temporal_resolutions[0]
                            current_segment_files = []

                            #initialise segment iteration variables
                            segment_iteration_starts = []
                            segment_iteration_ends = []

                            #initialise list for handling duplicate files
                            duplicates_to_add_to_iteration_files = []

                            #iterate over base file measurement windows
                            for base_window_ii, (base_window_start, base_temporal_resolution) in enumerate(zip(base_file_times,base_file_temporal_resolutions)):

                                #get end of base measurement window
                                base_window_end = base_window_start + base_temporal_resolution

                                #initialise current iteration variables
                                current_iteration_start = copy.deepcopy(base_window_start)
                                current_iteration_end = copy.deepcopy(base_window_end)
                                if len(duplicates_to_add_to_iteration_files) > 0:
                                    current_iteration_files = [base_data_file] + duplicates_to_add_to_iteration_files
                                else:
                                    current_iteration_files = [base_data_file]

                                #update indices of non-checked windows for base file (drop current index)
                                relevant_file_dictionary[base_data_file]['non_checked_indices'] = relevant_file_dictionary[base_data_file]['non_checked_indices'][1:]

                                #iterate through all other files and through all file measurement windows. Check overlap initially with current base file measurement window, and then iteratively with all other file windows
                                #set flag to indicate if you want to keep iterating  (this will be set to False when have no more overlaps)
                                keep_iterating = True

                                while keep_iterating == True:

                                    #set flag for keep_iterating to be False -- this will be modified if find any overlaps are found with current segment through pass of all other files
                                    keep_iterating = False

                                    #set flag for forcing write on this iteration if window in void between current segment end and current iteration start
                                    #this reverts back to False each time current iteration segment is modified
                                    force_write = False

                                    #check overlap of current segment with all measurement windows in other files
                                    for next_data_ii, next_data_file in enumerate(overlap_group_data_files[base_data_ii+1:]):
                                        #if iterated file has been decreed as a duplicate, continue to next file immediately
                                        if next_data_file in duplicates_to_add_to_iteration_files:
                                            continue

                                        #get iterated file start/end times
                                        next_file_start = relevant_file_dictionary[next_data_file]['file_start']
                                        next_file_end = relevant_file_dictionary[next_data_file]['file_end']

                                        #get iterated file data
                                        next_data = relevant_file_dictionary[next_data_file]['data']

                                        #get non-checked indices of measurement windows in iterated file that have not been cross-checked with other file windows
                                        next_non_checked_indices = relevant_file_dictionary[next_data_file]['non_checked_indices']

                                        #get iterated data file non-checked window start times and measurement resolutions
                                        next_file_times = next_data['time'][next_non_checked_indices]
                                        next_file_temporal_resolutions = next_data['temporal_resolution'][next_non_checked_indices]

                                        #if both iterated file times and iterated file temporal resolutions are exactly same as the base file, this is because they are exact duplicates
                                        #set flag to add duplicate file to base file on each iteration for rest of iteration through base file windows
                                        #also set all indices to be checked, and add duplicated file to current_iteration_files
                                        if (base_file_start == next_file_start) & (base_file_end == next_file_end):
                                            if np.all((base_file_times == next_file_times) & (base_file_temporal_resolutions == next_file_temporal_resolutions)):
                                                duplicates_to_add_to_iteration_files.append(next_data_file)
                                                relevant_file_dictionary[next_data_file]['non_checked_indices'] = []
                                                current_iteration_files.append(next_data_file)
                                                continue

                                        #set list for appending checked indices to drop in iterated file
                                        next_file_indices_to_drop = []

                                        #iterate over file windows
                                        for next_window_ii, (next_window_start, next_temporal_resolution) in enumerate(zip(next_file_times,next_file_temporal_resolutions)):

                                            #get end of iterated file measurement window
                                            next_window_end = next_window_start + next_temporal_resolution

                                            #determine if there are any windows in void between current_segment_end and current_iteration_start
                                            #if there are, force write on this iteration (if current iteration window is not subsequently modified - in which case this will be reevaluated)
                                            if (current_segment_end <= next_window_start < current_iteration_start) & ((current_segment_end < next_window_end <= current_iteration_start)):
                                                force_write = True
                                                continue

                                            #if next_window_end <= current_iteration_start, then continue to next file window
                                            if next_window_end <= current_iteration_start:
                                                continue

                                            #if next_window_start >= current_iteration_end, break to next file
                                            if next_window_start >= current_iteration_end:
                                                break

                                            #else, have overlap with current iteration segment!
                                            #add iterated file to list of current iteration files (if not already there)
                                            if next_data_file not in current_iteration_files:
                                                current_iteration_files.append(next_data_file)
                                            #if next_window_start < current_iteration_start, update current_iteration_start
                                            #also set keep_iterating flag to True, as have modified current iteration segment boundaries
                                            if next_window_start < current_iteration_start:
                                                current_iteration_start = copy.deepcopy(next_window_start)
                                                keep_iterating = True
                                            #if next_window_end > current_iteration_end, update current_iteration_end
                                            #also set keep_iterating flag to True, as have modified current iteration segment boundaries
                                            if next_window_end > current_iteration_end:
                                                current_iteration_end = copy.deepcopy(next_window_end)
                                                keep_iterating = True

                                            #append checked iterated file index to list of indices to drop (after have gone through all relevant file windows on this pass)
                                            next_file_indices_to_drop.append(next_non_checked_indices[next_window_ii])

                                        #update indices of non-checked windows for iterated file (drop current index)
                                        relevant_file_dictionary[next_data_file]['non_checked_indices'] = [x for x in relevant_file_dictionary[next_data_file]['non_checked_indices'] if x not in next_file_indices_to_drop]

                                #now have found all overlapping windows with one base file window in all other files
                                #now determine if will write out segment on this iteration

                                #1. determine if you have overlap between current iteration start and current segment end
                                    #if have overlap do not write out on this iteration
                                #AND
                                #2. #determine if current segment files are different from current iteration files
                                    #if are not different do not write

                                #OR
                                #1. force_write flag == True

                                if ((current_segment_end <= current_iteration_start) & (sorted(current_segment_files) != sorted(current_iteration_files))) or (force_write == True):

                                    #WRITE OUT SEGMENT!

                                    #test intersections between current segment start/end points with already written segments through time record. If have any intersection, split current segment files into chunks split by intersecting written segment start points.
                                    #this is for dealing with cases where you have segments in files (which are not the initial base file) entirely disconnected from all other segments

                                    intersect_break_points = []
                                    for written_segment in list(timeseries_compilation_dict.keys()):
                                        written_segment_split = written_segment.split('-')
                                        written_segment_start, written_segment_end = int(written_segment_split[0]), int(written_segment_split[1])
                                        have_intersect = (current_segment_start <= written_segment_start < current_segment_end) or (written_segment_start <= current_segment_start < written_segment_end)
                                        if have_intersect == True:
                                            intersect_break_points.append(written_segment_start)
                                    if len(intersect_break_points) > 0:
                                        #iterate through break points
                                        for break_point in sorted(intersect_break_points):
                                            #chunk segment_iteration_starts and segment_iteration_ends into separate groups split by breakpoints
                                            chunked_group_inds = [i for i in range(len(segment_iteration_starts)) if segment_iteration_starts[i] < break_point]
                                            #check have some indices to chunk, if not continue to next break point
                                            if len(chunked_group_inds) > 0:
                                                #filter segment_iteration_starts and segment_iteration_ends by valid inds
                                                chunked_group_starts = [segment_iteration_starts[i] for i in chunked_group_inds]
                                                chunked_group_ends = [segment_iteration_ends[i] for i in chunked_group_inds]
                                                #cut segment_iteration_starts and segment_iteration_starts lists
                                                segment_iteration_starts = [segment_iteration_starts[i] for i in range(len(segment_iteration_starts)) if i not in chunked_group_inds]
                                                segment_iteration_ends = [segment_iteration_ends[i] for i in range(len(segment_iteration_ends)) if i not in chunked_group_inds]

                                                #write out chunked groups (take first chunked group start and last chunked group end as boundaries)
                                                timeseries_compilation_dict['{}-{}'.format(chunked_group_starts[0],chunked_group_ends[-1])] = copy.deepcopy(current_segment_files)
                                                #log_file_str += '------b {}-{} {}\n'.format(chunked_group_starts[0],chunked_group_ends[-1], current_segment_files)

                                        #write out final chunked group after last breakpoint - if have data remaining (take first remaining group start and last group end as boundaries)
                                        if len(segment_iteration_starts) > 0:
                                            timeseries_compilation_dict['{}-{}'.format(segment_iteration_starts[0],segment_iteration_ends[-1])] = copy.deepcopy(current_segment_files)
                                            #log_file_str += '------c {}-{} {}\n'.format(segment_iteration_starts[0],segment_iteration_ends[-1],current_segment_files)

                                    else:
                                        timeseries_compilation_dict['{}-{}'.format(current_segment_start,current_segment_end)] = copy.deepcopy(current_segment_files)
                                        #log_file_str += '------a {}-{} {}\n'.format(current_segment_start,current_segment_end,current_segment_files)

                                    #set current segment files to be current iteration files
                                    current_segment_files = copy.deepcopy(current_iteration_files)

                                    #reset current segment start/end to current iteration start/end
                                    current_segment_start = copy.deepcopy(current_iteration_start)
                                    current_segment_end = copy.deepcopy(current_iteration_end)

                                    #reset segment_iteration_starts with current_iteration_start
                                    segment_iteration_starts = [current_iteration_start]

                                    #reset segment_iteration_ends with current_iteration_end
                                    segment_iteration_ends = [current_iteration_end]


                                #if not writing out, update current segment variables
                                else:
                                    #if current iteration start < current segment start, update it
                                    if current_iteration_start < current_segment_start:
                                        current_segment_start = copy.deepcopy(current_iteration_start)

                                    #if current iteration end > current segment end, update it
                                    if current_iteration_end > current_segment_end:
                                        current_segment_end = copy.deepcopy(current_iteration_end)

                                    #update current segment files (with all new files that are not in current group files)
                                    for iteration_file in current_iteration_files:
                                        if iteration_file not in current_segment_files:
                                            current_segment_files.append(iteration_file)

                                    #append segment_iteration_starts with current_iteration_start
                                    segment_iteration_starts.append(current_iteration_start)

                                    #append to segment_iteration_ends with current_iteration_end
                                    segment_iteration_ends.append(current_iteration_end)


                                #set previous iteration end as current_iteration_end
                                previous_iteration_end = copy.deepcopy(current_iteration_end)


                            #now have cross-checked all windows of one file with all other file windows, write out what remains in current_segment files which has start data before last written segment start

                            #WRITE OUT SEGMENT!

                            #test intersections between current segment start/end points with already written segments through time record. If have any intersection, split current segment files into chunks split by intersecting written segment start points.
                            #this is for dealing with cases where you have segments in files (which are not the initial base file) entirely disconnected from all other segments

                            intersect_break_points = []
                            for written_segment in list(timeseries_compilation_dict.keys()):
                                written_segment_split = written_segment.split('-')
                                written_segment_start, written_segment_end = int(written_segment_split[0]), int(written_segment_split[1])
                                have_intersect = (current_segment_start <= written_segment_start < current_segment_end) or (written_segment_start <= current_segment_start < written_segment_end)
                                if have_intersect == True:
                                    intersect_break_points.append(written_segment_start)
                            if len(intersect_break_points) > 0:
                                #iterate through break points
                                for break_point in sorted(intersect_break_points):
                                    #chunk segment_iteration_starts and segment_iteration_ends into separate groups split by breakpoints
                                    chunked_group_inds = [i for i in range(len(segment_iteration_starts)) if segment_iteration_starts[i] < break_point]
                                    #check have some indices to chunk, if not continue to next break point
                                    if len(chunked_group_inds) > 0:
                                        #filter segment_iteration_starts and segment_iteration_ends by valid inds
                                        chunked_group_starts = [segment_iteration_starts[i] for i in chunked_group_inds]
                                        chunked_group_ends = [segment_iteration_ends[i] for i in chunked_group_inds]
                                        #cut segment_iteration_starts and segment_iteration_starts lists
                                        segment_iteration_starts = [segment_iteration_starts[i] for i in range(len(segment_iteration_starts)) if i not in chunked_group_inds]
                                        segment_iteration_ends = [segment_iteration_ends[i] for i in range(len(segment_iteration_ends)) if i not in chunked_group_inds]

                                        #write out chunked groups (take first chunked group start and last chunked group end as boundaries)
                                        timeseries_compilation_dict['{}-{}'.format(chunked_group_starts[0],chunked_group_ends[-1])] = copy.deepcopy(current_segment_files)
                                        #log_file_str += '------e {}-{} {}\n'.format(chunked_group_starts[0],chunked_group_ends[-1],current_segment_files)

                                #write out final chunked group after last breakpoint - if have data remaining (take first remaining group start and last group end as boundaries)
                                if len(segment_iteration_starts) > 0:
                                    timeseries_compilation_dict['{}-{}'.format(segment_iteration_starts[0],segment_iteration_ends[-1])] = copy.deepcopy(current_segment_files)
                                    #log_file_str += '------f {}-{} {}\n'.format(segment_iteration_starts[0],segment_iteration_ends[-1],current_segment_files)

                            else:
                                timeseries_compilation_dict['{}-{}'.format(current_segment_start,current_segment_end)] = copy.deepcopy(current_segment_files)
                                #log_file_str += '------d {}-{} {}\n'.format(current_segment_start,current_segment_end,current_segment_files)

            #-----------------------------------------------------#

            #now have finished all cross-checking of files
            #for key in sorted(list(timeseries_compilation_dict.keys())):
            #    split_key = key.split('-')
            #    log_file_str += '{} {} {} {} {}\n'.format(fractional_minutes_from_reference_to_datetime(float(split_key[0])), fractional_minutes_from_reference_to_datetime(float(split_key[1])), split_key[0], split_key[1], timeseries_compilation_dict[key])
            #    log_file_str += '{}\n'.format([np.unique(relevant_file_dictionary[period_file]['data']['temporal_resolution']) for period_file in timeseries_compilation_dict[key]])

            #iterate through segmented periods
            for period in sorted(timeseries_compilation_dict):
                period_split = period.split('-')
                current_period_start = int(period_split[0])
                current_period_end = int(period_split[1])
                current_period_files = timeseries_compilation_dict[period]

                #------------------------------------------------------------------------------------------------------------#
                #------------------------------------------------------------------------------------------------------------#

                #if only 1 file provides data in period, simply append relevant data from it to current timeseries (if already exists)
                if len(current_period_files) == 1:
                    chosen_file = current_period_files[0]

                #------------------------------------------------------------------------------------------------------------#
                #------------------------------------------------------------------------------------------------------------#

                #if have file overlap for period, first attempt to resolve overlap by priority data levels, then by using ranking algorithm
                else:

                    #set some variables for storing file data level ranks, and for calculation of ranking algorithm
                    file_data_level_ranks = np.array([],dtype=np.object)
                    algorithm_ranks = np.array([0.0]*len(current_period_files))
                    average_temporal_resolutions = np.array([],dtype=np.float32)
                    valid_data_N_relevant = np.array([],dtype=np.float32)
                    valid_data_N_full = np.array([],dtype=np.float32)
                    valid_metadata_vars_N = np.array([],dtype=np.float32)
                    chosen_file_index = -99

                    #iterate through all overlapping files in current period
                    for period_i, period_file in enumerate(current_period_files):

                        full_data = copy.deepcopy(relevant_file_dictionary[period_file]['data'])
                        full_metadata = copy.deepcopy(relevant_file_dictionary[period_file]['metadata'])
                        #check if want all of data from file or it needs cutting
                        if (str(current_period_start) not in period_file) or (str(current_period_end) not in period_file):
                            relevant_data = cut_data_array(full_data, 'data', current_period_start, current_period_end)
                            relevant_metadata = cut_data_array(full_metadata, 'metadata', current_period_start, current_period_end)
                        else:
                            relevant_data = copy.deepcopy(full_data)
                            relevant_metadata = copy.deepcopy(full_metadata)

                        #---------------------------------------------------------#
                        #get data level hierarchy ranking of overlapping file for network (higher being better)

                        all_data_level_ranks = []
                        #iterate through metadata columns
                        for metadata_column in relevant_metadata:
                            all_data_level_ranks.append(network_details['data_level_hierarchy'][metadata_column['data_level']])
                        #append average data level rank across metadata columns
                        file_data_level_ranks = np.append(file_data_level_ranks,np.average(all_data_level_ranks))

                        #---------------------------------------------------------#
                        #calculate algorithm rankings

                        #process modal occurring metadata ranks and previously accepted metadata ranks

                        #iterate through all metadata fields
                        for metadata_var in metadata_vars:

                            all_algorithm_ranks = []                                        

                            #iterate through metadata columns
                            for metadata_column_ii, metadata_column in enumerate(relevant_metadata):
                                #get current metadata value
                                current_meta_value = metadata_column[metadata_var]
                                #process modally occurring metadata ranks (only process if have current metadata is not np.NaN)
                                if pd.isnull(current_meta_value) == False:
                                    all_algorithm_ranks.append(metadata_rank_dicts[metadata_var][current_meta_value])
                                    #process last accepted metadata ranks (only process if have current metadata is not np.NaN, and is first column index)
                                    if metadata_column_ii == 0:
                                        try:
                                            previously_accepted_meta_value = existing_metadata[metadata_var][-1]
                                            #if current metadata value == previously accepted metadata value, add 1 to file rank
                                            if previously_accepted_meta_value == current_meta_value:
                                                algorithm_ranks[period_i] += 1.0
                                        except:
                                            pass
                            #append average algorithm ranks across metadata columns (if have values for metadata variable)
                            if len(all_algorithm_ranks) > 0:
                                algorithm_ranks[period_i] += np.average(all_algorithm_ranks)

                        #calculate average temporal resolution in relevant segment of file, and append to list
                        average_temporal_resolutions = np.append(average_temporal_resolutions, np.average(relevant_data['temporal_resolution']))

                        #calculate number of valid data in relevant segment of file (after removing data by flags - GHOST decreed)
                        invalid_qa_flag_code = standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed']
                        valid_data_N_relevant = np.append(valid_data_N_relevant, len([0 for qa_codes in relevant_data['qa'] if invalid_qa_flag_code not in qa_codes]))

                        #calculate number of valid data across all of file (after removing data by flags - GHOST decreed) --> used as primary tiebreaker if file ranks are equal
                        valid_data_N_full = np.append(valid_data_N_full, len([0 for qa_codes in full_data['qa'] if invalid_qa_flag_code not in qa_codes]))

                        #calculate number of valid metadata fields provided (i.e. not np.NaN) --> used as secondary tiebreaker if file ranks are equal
                        valid_metadata_vars = 0
                        for metadata_column_ii, metadata_column in enumerate(relevant_metadata):
                            for meta_name in metadata_column.dtype.names:
                                if pd.isnull(metadata_column[meta_name]) == False:
                                    valid_metadata_vars += 1.0
                        #append number of average valid metadata fields across data columns 
                        valid_metadata_vars_N = np.append(valid_metadata_vars_N, valid_metadata_vars/len(relevant_metadata))

                    #get ranks of average temporal resolutions of files (higher resolution (i.e. smaller number) == higher rank)
                    #first convert temporal resolutions to be negative to ensure that the smallest numbers are given the biggest ranks
                    average_temporal_resolutions = [-1.0*res for res in average_temporal_resolutions]
                    average_temporal_resolutions_ranks = ss.rankdata(average_temporal_resolutions)
                    #update algorithm ranks
                    algorithm_ranks += average_temporal_resolutions_ranks

                    #get ranks of number of valid data in relevant sections of files (more data == higher rank)
                    valid_data_N_relevant_ranks = ss.rankdata(valid_data_N_relevant)

                    #update algorithm ranks (by valid data ranks in relevant section of file)
                    algorithm_ranks += valid_data_N_relevant_ranks

                    #----------------------------------------#
                    #now have iterated through all overlapping files and have ranked data levels of each file, and have processed ranking algorithm
                    #first attempt to choose between duplicate files by keeping file with the highest priority data level.

                    #get max ranking of data levels across the overlapping files
                    max_rank = np.max(file_data_level_ranks)
                    #get indices of overlapping files which have a ranking == max rank
                    max_rank_indices = np.where(file_data_level_ranks == max_rank)[0]

                    #log_file_str += '{}\n'.format(file_data_level_ranks)
                    #log_file_str += '{}\n'.format(algorithm_ranks)
                    #log_file_str += '{}\n'.format(valid_data_N_relevant)
                    #log_file_str += '{}\n'.format(valid_data_N_full)
                    #log_file_str += '{}\n'.format(valid_metadata_vars_N)

                    #if have multiple files with the same max rank, drop any files which have do not have the max rank from consideration, and proceed to resolve overlap using ranking algorithm
                    if (len(max_rank_indices) > 1) & (len(max_rank_indices) != len(file_data_level_ranks)):
                        #drop files which do not have max data level rank from consideration when resolving overlap using ranking algorithm
                        current_period_files = np.array(current_period_files)[max_rank_indices]
                        algorithm_ranks = algorithm_ranks[max_rank_indices]
                        valid_data_N_full = valid_data_N_full[max_rank_indices]
                        valid_metadata_vars_N = valid_metadata_vars_N[max_rank_indices]
                    #else if have only 1 file with max rank, use this file preferentially for current period and do not use ranking algorithm
                    elif len(max_rank_indices) == 1:
                        #log_file_str += 'resolved using data levels\n'
                        chosen_file_index = max_rank_indices[0]
                    #----------------------------------------#
                    #if still have overlapping files, now resolve using ranking algorithm

                    if chosen_file_index < 0:

                        #get max algorithm ranking across the overlapping files
                        max_rank = np.max(algorithm_ranks)
                        #get indices of overlapping files which have a ranking == max rank
                        max_rank_indices = np.where(algorithm_ranks == max_rank)[0]

                        #if have multiple files with the same rank, tiebreak by N valid data in whole data file, and then N metadata values provided (that are not np.NaN)
                        if len(max_rank_indices) > 1:
                            #reduce variables for valid files
                            current_period_files = np.array(current_period_files)[max_rank_indices]
                            valid_data_N_full = valid_data_N_full[max_rank_indices]
                            valid_metadata_vars_N = valid_metadata_vars_N[max_rank_indices]

                            #tiebreak preferentially by N valid data in whole data file
                            #get ranks of number of valid data in across all of files (more data == higher rank)
                            valid_data_N_full_ranks = ss.rankdata(valid_data_N_full)
                            max_rank_indices = np.where(valid_data_N_full_ranks == np.max(valid_data_N_full_ranks))[0]

                            #if have broken tiebreak, set preferential file index here
                            if len(max_rank_indices) == 1:
                                chosen_file_index = max_rank_indices[0]
                                #log_file_str += 'broke tiebreak using full file valid N\n'
                            else:
                                #tiebreak secondly by N metadata values provided (that are not np.NaN)
                                #reduce variables for valid files
                                current_period_files = np.array(current_period_files)[max_rank_indices]
                                valid_metadata_vars_N = valid_metadata_vars_N[max_rank_indices]

                                #get ranks of number of metadata fields values are provided for (i.e. not NaN) in relevant section of file (more fields == higher rank)
                                valid_metadata_vars_N_ranks = ss.rankdata(valid_metadata_vars_N)
                                #finally, if N metadata values are equal also, then simply first index is taken
                                chosen_file_index = np.argmax(valid_metadata_vars_N_ranks)
                                #log_file_str += 'broke tiebreak using valid meta fields\n'
                        else:
                            chosen_file_index = max_rank_indices[0]
                            #log_file_str += 'resolved using algorithm\n'

                    #----------------------------------------#
                    #define chosen file to keep for data period
                    chosen_file = current_period_files[chosen_file_index]

                #------------------------------------------------------------------------------------------------------------#
                #------------------------------------------------------------------------------------------------------------#

                #log_file_str += '{}\n'.format(chosen_file)
                #add chosen file data/metadata for period to existing time series (if already exists)
                data = copy.deepcopy(relevant_file_dictionary[chosen_file]['data'])
                metadata = copy.deepcopy(relevant_file_dictionary[chosen_file]['metadata'])
                #check if want all of data from file or it needs cutting
                if (str(current_period_start) not in chosen_file) or (str(current_period_end) not in chosen_file):
                    data = cut_data_array(data, 'data', current_period_start, current_period_end)
                    metadata = cut_data_array(metadata, 'metadata', current_period_start, current_period_end)

                try:
                    #check if have some continuous time series data stored already - if so, append to these data/metadata variables
                    existing_data

                    #append current metadata/data to existing metadata/data
                    existing_data_n_rows = existing_data.shape[0]
                    data_n_rows = data.shape[0]
                    existing_data.resize(existing_data_n_rows+data_n_rows)
                    existing_data[-data_n_rows:] = data
                    #if the file is the same as previous, metadata continues to be the same, therefore do not need to append metadata
                    if previous_chosen_file != chosen_file:
                        #otherwise, check if each current metadata column is same as last key appended metadata column
                        #if different, a new metadata block is appended from column onwards
                        for metadata_ii, metadata_column in enumerate(metadata):
                            metadata_has_changed = determine_key_changing_metadata(existing_metadata[-1],metadata_column)
                            #append metadata
                            if metadata_has_changed == True:
                                metadata_block = metadata[metadata_ii:]
                                existing_metadata_n_rows = existing_metadata.shape[0]
                                metadata_n_rows = metadata_block.shape[0]
                                existing_metadata.resize(existing_metadata_n_rows+metadata_n_rows)
                                existing_metadata[-metadata_n_rows:] = metadata_block
                                break

                except:
                    #if first file to put data in time series, create variables for continuous time series
                    existing_data = copy.deepcopy(data)
                    existing_metadata = copy.deepcopy(metadata)

                #update previous chosen file to be current chosen file
                previous_chosen_file = copy.deepcopy(chosen_file)

    #------------------------------------------------------------------------------------------------------------------------#
    #------------------------------------------------------------------------------------------------------------------------#

    #do quick sanity check to see if time has been processed correctly.
    #that check is simply to see if the time array is ordered correctly, i.e. all in ascending order?
    #if not ordered correctly, return caught error logfile (2)
    if all(existing_data['time'] == np.sort(existing_data['time'])) == False:
        log_file_str += 'TIME COLUMN IS NOT IN ASCENDING ORDER'
        create_GHOST_stage_logfile('S2',run_details,2,log_file_str,station_reference)

    #process manually compiled metadata for station (if exists)
    existing_metadata, existing_data, modified_standard_fields = process_manually_compiled_metadata(station_reference,existing_metadata,existing_data,run_details,parameter_details,network_details,log_file_str)
    #update standardisations for modified metadata fields (if any)
    if len(modified_standard_fields) > 0:
        existing_metadata = redo_standardisations(existing_metadata, modified_standard_fields, parameter_details)

    #handle gapped key metadata
    existing_metadata,existing_data = handle_gapped_key_metadata(station_reference,existing_metadata,existing_data,run_details,log_file_str)

    #write out processed continuous data/metadata
    write_station_specific_numpy(existing_metadata,existing_data,run_details,parameter_details,'S2')

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('S2',run_details,0,log_file_str,station_reference)

###------------------------------------------------------------------------------------###

def get_data_from_reference(current_GHOST_stage,station_reference,run_details,parameter_details,read='all'):

    '''take station reference and return loaded metadata/data
       the type of data returned ('read') can be configured. By default both final stage metadata and data read and returned.
       read = 'all' --> read and return data + metadata (and filename); read = 'data' --> just read and return data (and filename); read = 'metadata' --> just read and return metadata (and filename)
    '''

    #if GHOST stage is in S stages, gather station filename from (GHOST_stage-1)_temporary directory
    if 'S' in current_GHOST_stage:
        previous_GHOST_stage = 'S{}'.format(int(current_GHOST_stage[1:])-1)
        station_numpy_filename = glob.glob(run_details['data_root']+'{}/{}/native/{}/temporary_{}/{}_*'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],previous_GHOST_stage,station_reference))[0]
     #otherwise gather filename from native directory
    else:
        station_numpy_filename = glob.glob(run_details['data_root']+'{}/{}/native/{}/{}_*'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],station_reference))[0]

    #set desired variables to read
    if read == 'metadata':
        desired_variables = ['metadata']
    elif read == 'data':
        desired_variables = ['data']
    elif read == 'all':
        desired_variables = ['metadata', 'data']

    #read variables
    loaded_variables = read_from_numpy(station_numpy_filename, desired_variables)

    #return variables
    if read == 'metadata':
        return loaded_variables['metadata'], station_numpy_filename
    elif read == 'data':
        return loaded_variables['data'], station_numpy_filename
    elif read == 'all':
        return loaded_variables['metadata'], loaded_variables['data'], station_numpy_filename


###------------------------------------------------------------------------------------###

def changing_measurement_position_new_stations(slurm_job_ID, network, parameter, station_reference):

    '''check if key metadata (Latitude, Longitude, Measurement Altitude (Altitude + Sampling Height) changes significantly through time record.
       Split station data if necessary into separate stations on the occasions of significant changes in key metadata.

       significant change in Latitude/Longitude coordinates is defined to be >= 0.0001 degrees (~11m)
       significant change in Measurement Altitude is defined to be >= 11m.

       Measurement Altitude (if not already defined) is defined as station altitude + measurement height.
       if sampling heights are not available, measurement altitude is simply taken to be altitude. 

       check that Latitude, Longitude, Measurement Altitude does not change then revert back at later date.
       therefore group different collections of Latitudes, Longitudes & Measurement Altitudes together through entire time record - outputting these groupings as 1 unique station
       this prevents writing of multiple new stations needlessly.

       finally split out data/metadata into new stations if required - appending each new split station reference with a --S%N string
    '''

    from GHOST_standards import standard_QA_name_to_QA_code
    import itertools

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- EVALUATING VARIABILITY OF REPORTED MEASUREMENT POSITION.\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get metadata/data for reference
    metadata, data, station_numpy_filename = get_data_from_reference('S3',station_reference,run_details,parameter_details,read='all')

    #tolerance limit for changing significant lat/lon coordinate changes
    latlon_tolerance = 0.0001 #~11m
    #tolerance limit for changing significant measurement altitude changes
    measurement_alt_tolerance = 11.0 #11m

    #load required metadata
    lats = metadata['latitude']
    lons = metadata['longitude']
    alts = metadata['altitude']
    sampling_heights = metadata['sampling_height']
    measurement_alts = metadata['measurement_altitude']
    update_metadata_stamp = metadata['meta_update_stamp']

    #update measurement altitude field 
    for ii in range(len(measurement_alts)):
        #do you already have a measurement altitude?
        if pd.isnull(measurement_alts[ii]) == True:
            #if not, if have valid sampling height and altitude, create a measurement altitude by adding them
            if (pd.isnull(sampling_heights[ii]) == False) & (pd.isnull(alts[ii]) == False):
                measurement_alts[ii] = sampling_heights[ii] + alts[ii]
            #else, if have a valid altitude, then set measurement altitude as just altitude
            elif (pd.isnull(alts[ii]) == False):
                measurement_alts[ii] = alts[ii]

    #overwrite potentially modified measurement altitudes to metadata array
    metadata['measurement_altitude'] = copy.deepcopy(measurement_alts)

    #if all measurement altitudes are NaN still, set them all to be -99999 for this function 
    if all(pd.isnull(measurement_alts)) == True:
        measurement_alts = np.full(len(lats), -99999)

    #get all unique measurement positions across time record (i.e. longitude, latitude & measurement altitude)
    all_measurement_positions = np.array([[lon, lats[ii], measurement_alts[ii]] for ii,lon in enumerate(lons)])
    unique_measurement_positions, unique_measurement_positions_indices = np.unique(all_measurement_positions, axis=0, return_index=True)
    unique_measurement_positions = [x for _,x in sorted(zip(unique_measurement_positions_indices,unique_measurement_positions))]

    #if have just 1 unique measurement position, terminate function
    if len(unique_measurement_positions) == 1:

         #write data/metadata to new file
         write_station_specific_numpy(metadata,data,run_details,parameter_details,'S3')

    #else, if have more than 1 unique measurement position, evaluate if they are significantly different
    else:

        #group all unique measurement positions that are not significantly different 
        grouped_unique_measurement_positions = []
        grouped_measurement_positions = []

        for ii, unique_measurement_position in enumerate(unique_measurement_positions):
            if ii not in grouped_measurement_positions:
                current_group = [unique_measurement_position]
                grouped_measurement_positions.append(ii)
                for jj, next_unique_measurement_position in enumerate(unique_measurement_positions[1:]):
                    if jj+1 not in grouped_measurement_positions:
                        lon_diff = np.abs(next_unique_measurement_position[0] - unique_measurement_position[0])
                        lat_diff = np.abs(next_unique_measurement_position[1] - unique_measurement_position[1])
                        measurement_alt_diff = np.abs(next_unique_measurement_position[2] - unique_measurement_position[2])
                        if (lon_diff < latlon_tolerance) & (lat_diff < latlon_tolerance) & (measurement_alt_diff < measurement_alt_tolerance):
                            current_group.append(next_unique_measurement_position)
                            grouped_measurement_positions.append(jj+1)
                grouped_unique_measurement_positions.append(current_group)

        #iterate through grouped unique measurement positions and get indices of all relevant metadata 
        unique_measurement_position_metadata_indices = []
        for unique_measurement_positions_group in grouped_unique_measurement_positions:
            current_group_inds = []
            for ii, unique_measurement_position in enumerate(unique_measurement_positions_group):
                for jj, measurement_position in enumerate(all_measurement_positions):
                    if np.array_equal(unique_measurement_position, measurement_position) == True:
                        current_group_inds.append(jj)
            unique_measurement_position_metadata_indices.append(current_group_inds)
                 
        #if have significantly different measurement positions, then split out metadata/data into different stations
        #first list of indices always contains the zero index (i.e. the first reported metadata), so this remains as the original station
        if len(unique_measurement_position_metadata_indices) > 1:
 
            log_file_str += '{} -- SPLITTING INTO {} DIFFERENT STATIONS.\n'.format(station_reference,len(unique_measurement_position_metadata_indices))
 
            for unique_measurement_position_ii, metadata_indices in enumerate(unique_measurement_position_metadata_indices):
                
                data_indices = []
                for metadata_index in metadata_indices: 
                    start_data_ind = bisect.bisect_left(data['time'], update_metadata_stamp[metadata_index])
                    if (metadata_index + 1) == len(update_metadata_stamp):
                        end_data_ind = len(data['time'])
                    else: 
                        end_data_ind = bisect.bisect_left(data['time'], update_metadata_stamp[metadata_index+1])
                    data_indices.extend(range(start_data_ind, end_data_ind))
                    
                #cut new metadata/data from original numpy dataframe
                cut_metadata = metadata[metadata_indices]
                cut_data = data[data_indices]

                #change station reference metadata (if not first station in time)
                if unique_measurement_position_ii > 0:
                    for qq in range(len(cut_metadata['station_reference'])):
                        cut_metadata['station_reference'][qq] = '{}--S{}'.format(station_reference,unique_measurement_position_ii)

                #remove duplicated key metadata columns
                cut_metadata = resolve_duplicate_metadata_from_file(cut_metadata)

                #save cut data/metadata to new file
                write_station_specific_numpy(cut_metadata,cut_data,run_details,parameter_details,'S3')

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('S3',run_details,0,log_file_str,station_reference)

###------------------------------------------------------------------------------------###

def weighted_mode(list_of_values, weights):

    '''taking a list of values, and a list of weights associated with the values, calculate the weighted modal value
       this simply works by calculating the unique values in the list, then separately summing the weights associated with each unique value, and then determining which unique value is associated with the highest cumulative weights.
       if there is a tie, the lowest sorted value will be kept preferentially (i.e. a instead of b, 1 instead of 2)
       if there is only 1 unique value, then modal value is simply the only unique value
    '''

    #get unique values in list
    unique_values = np.unique(list_of_values)

    #if only 1 unique value, then modal value is simply the only unique value, return function
    if len(unique_values) == 1:
        return unique_values[0]

    #add all unique values to weight counter dictionary as keys, and give an initial value of 0.0
    weight_counter = {}
    for unique_value in unique_values:
        weight_counter[unique_value] = 0.0

    #iterate through all values in the list of values and associated weights
    for value, weight in zip(list_of_values,weights):
        #add weight associated with value to weight counter dictionary
        weight_counter[value] += weight

    #the unique value key associated the the greatest sum of weights in the weight counter dictionary is the modal value
    #if there is a tie, the lowest sorted key will be kept preferentially (i.e. a instead of b, 1 instead of 2)
    modal_value = max(weight_counter.keys(), key=(lambda key: weight_counter[key]))

    return modal_value

###--------------------------------------------------------------------------------------------------###

def get_unique_longitude_latitude_pairs(metadata, data):

    '''get unique longitude and latitude pairs across metadata, and return associated metadata/data indices of each pair'''

    #get unique longitude/latitude pairs across metadata
    all_measurement_pairs = np.array([[metadata[ii]['longitude'], metadata[ii]['latitude']] for ii in range(len(metadata))])
    unique_measurement_pairs = np.unique(all_measurement_pairs, axis=0)
    
    #get associated metadata/data indices with each unique longitude/latitude pair
    unique_pair_dictionary = {}
    for ii, unique_measurement_pair in enumerate(unique_measurement_pairs):
        unique_pair_dictionary[ii] = {'metadata_indices':[], 'data_indices':[], 'longitude':unique_measurement_pair[0], 'latitude':unique_measurement_pair[1]} 
        for jj, measurement_pair in enumerate(all_measurement_pairs):   
            if np.array_equal(unique_measurement_pair, measurement_pair) == True:
                unique_pair_dictionary[ii]['metadata_indices'].append(jj)

        for metadata_index in unique_pair_dictionary[ii]['metadata_indices']: 
            start_data_ind = bisect.bisect_left(data['time'], metadata['meta_update_stamp'][metadata_index])
            if (metadata_index + 1) == len(metadata['meta_update_stamp']):
                end_data_ind = len(data['time'])
            else: 
                end_data_ind = bisect.bisect_left(data['time'], metadata['meta_update_stamp'][metadata_index+1])
            unique_pair_dictionary[ii]['data_indices'].extend(range(start_data_ind, end_data_ind))

    #return dictionary of all unique longitude/latitude pairs with associated metadata/data indices
    return unique_pair_dictionary

###------------------------------------------------------------------------------------###

def globally_gridded_metadata_processing(slurm_job_ID, network, parameter, station_reference):

    '''get globally gridded metadata for a specific station'''

    import globally_gridded_metadata_functions 

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- GLOBALLY GRIDDED METADATA\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get metadata/data for reference
    metadata, data, station_numpy_filename = get_data_from_reference('S4',station_reference,run_details,parameter_details,read='all')

    #get globally gridded metadata for station
    metadata, data, log_file_str, metadata_fields_to_update = globally_gridded_metadata_functions.get_globally_gridded_metadata(run_details, metadata, data, log_file_str)

    #update standardisations for modified metadata fields
    metadata = redo_standardisations(metadata, metadata_fields_to_update, parameter_details)

    #re-save metadata/data for station
    write_station_specific_numpy(metadata,data, run_details, parameter_details, 'S4')

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('S4',run_details,0,log_file_str,station_reference)

###------------------------------------------------------------------------------------###

def process_reported_and_derived_uncertainty_per_measurement(parameter_details, metadata, data):

    '''function processes reported uncertainty per measurement from metadata reported (or if not available, documented) uncertainties, if not already given per measurement by network.

       function also derives +/- uncertainty on each measurement from the reported (or if not available, documented) accuracy and precision metadata.
       precision and accuracy are added in quadrature to generate uncertainties per measurement.
       calculation is only done if have both valid precision and accuracy information
    '''

    #get key data
    reported_uncertainty_per_measurement = data['reported_uncertainty_per_measurement']

    #get key metadata
    reported_uncertainty = metadata['measuring_instrument_reported_uncertainty']
    documented_uncertainty = metadata['measuring_instrument_documented_uncertainty']
    reported_precision = metadata['measuring_instrument_reported_precision']
    documented_precision = metadata['measuring_instrument_documented_precision']
    reported_accuracy = metadata['measuring_instrument_reported_accuracy']
    documented_accuracy = metadata['measuring_instrument_documented_accuracy']
    update_metadata_stamp = metadata['meta_update_stamp']

    #------------------------------------------------------------#
    #process reported uncertainties

    #iterate through each metadata parameter
    for ii in range(len(reported_uncertainty)):

        #do you have valid reported metadata uncertainty?
        if pd.isnull(reported_uncertainty[ii]) == False:
            uncertainty = reported_uncertainty[ii]
        #do you have valid documented metadata uncertainty?
        elif pd.isnull(documented_uncertainty[ii]) == False:
            uncertainty = documented_uncertainty[ii]
        #no valid uncertainty value, continue to next metadata index
        else:
            continue

        #get relevant range of inds to update reported_uncertainty_per_measurement data field
        #is it last index?
        if ii == (len(reported_precision)-1):
            relevant_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time']))
        else:
            relevant_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1]))

        #if uncertainty is in percentage terms, convert to absolute terms (for all values across relevant range)
        #determine type of way uncertainty is reported, and deal with processing accordingly

        #get all string characters (no numbers)
        uncertainty_char = ''.join([i for i in uncertainty if not i.isdigit()]).replace('.','')

        #put all separate numerical values in string as elements in new list
        uncertainty_val = np.array(re.findall(r"[+-]?\d+(?:\.\d+)?", uncertainty)).astype(np.float32)

        #1. absolute value
        if uncertainty_char == '':
            uncertainty_absolute = uncertainty_val[0] * np.ones(len(relevant_inds))

        #2. percentage value
        elif uncertainty_char == '%':
            fraction = uncertainty_val[0]/100.0
            uncertainty_absolute = data[parameter_details['bsc_parameter_name']][relevant_inds]*fraction

        #3. the greater of either an absolute value or percentage
        elif (uncertainty_char == '/%') or (uncertainty_char == '%/'):

            if quantity_char == '/%':
                absolute = uncertainty_val[0]
                fraction = uncertainty_val[1]/100.0

            elif quantity_char == '%/':
                absolute = uncertainty_val[1]
                fraction = uncertainty_val[0]/100.0

            uncertainty_absolute = []
            for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                #percentage (as absolute) greater
                if (fraction*val) > absolute:
                    uncertainty_absolute=np.append(uncertainty_absolute,fraction*val)
                #absolute greater
                else:
                    uncertainty_absolute=np.append(uncertainty_absolute,absolute)

        #4. percentage value after a fixed limit (minimum uncertainty taken from this limit)
        elif uncertainty_char == '%>=':
            limit = uncertainty_val[1]
            fraction = uncertainty_val[0]/100.0

            uncertainty_absolute = []
            for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                #percentage (as absolute) higher than limit
                if (fraction*val) > limit:
                    uncertainty_absolute=np.append(uncertainty_absolute,fraction*val)
                #not greater than limit
                else:
                    uncertainty_absolute=np.append(uncertainty_absolute,fraction*limit)

        #5. choice of 2 absolute values under/over fixed limit
        elif uncertainty_char == '< >=':
            absolute_under =  uncertainty_val[0]
            limit =     uncertainty_val[1]
            absolute_over =  uncertainty_val[2]

            uncertainty_absolute = []
            for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                #absolute value higher than limit
                if val >= limit:
                    uncertainty_absolute=np.append(uncertainty_absolute,absolute_over)
                #absolute value under limit
                elif val < limit:
                    uncertainty_absolute=np.append(uncertainty_absolute,absolute_under)

        #6. choice of 2 percentage values under/over fixed limit
        elif uncertainty_char == '%< %>=':
            fraction_under =  uncertainty_val[0]/100.0
            limit =           uncertainty_val[1]
            fraction_over =   uncertainty_val[2]/100.0

            uncertainty_absolute = []
            for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                #percentage (as absolute) >= limit
                if (fraction_over*val) >= limit:
                    uncertainty_absolute=np.append(uncertainty_absolute,fraction_over*val)
                #< limit
                else:
                    uncertainty_absolute=np.append(uncertainty_absolute,fraction_under*val)


        #overwrite 'reported_uncertainty_per_measurement' field if == NaN
        for rr, relevant_ind in enumerate(relevant_inds):
            if pd.isnull(data['reported_uncertainty_per_measurement'][relevant_ind]) == True:
                data['reported_uncertainty_per_measurement'][relevant_ind] = uncertainty_absolute[rr]

    #------------------------------------------------------------#
    #process derived uncertainties

    #iterate through each metadata parameter
    for ii in range(len(reported_precision)):

        #do you have valid reported precision?
        if pd.isnull(reported_precision[ii]) == False:
            precision = reported_precision[ii]
        #do you have valid documented precision?
        elif pd.isnull(documented_precision[ii]) == False:
            precision = documented_precision[ii]
        #no valid precision value, continue to next metadata index
        else:
            continue

        #do you have valid reported accuracy?
        if pd.isnull(reported_accuracy[ii]) == False:
            accuracy = reported_accuracy[ii]
        #do you have valid documented accuracy?
        elif pd.isnull(documented_accuracy[ii]) == False:
            accuracy = documented_accuracy[ii]
        #no valid accuracy value, continue to next metadata index
        else:
            continue

        #get relevant range of inds to update derived_uncertainty_per_measurement data field
        #is it last index?
        if ii == (len(reported_precision)-1):
            relevant_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time']))
        else:
            relevant_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1]))

        #if precision or accuracy are in percentage terms, convert to absolute terms (for all values across relevant range)
        #determine type of way precision/accuracy are reported, and deal with processing accordingly

        quantities = [precision,accuracy]
        quantities_absolute = []

        for quantity in quantities:

            #get all string characters (no numbers)
            quantity_char = ''.join([i for i in quantity if not i.isdigit()]).replace('.','')

            #put all separate numerical values in string as elements in new list
            quantity_val = np.array(re.findall(r"[+-]?\d+(?:\.\d+)?", quantity)).astype(np.float32)

            #1. absolute value
            if quantity_char == '':
                quantity_absolute = quantity_val[0] * np.ones(len(relevant_inds))

            #2. percentage value
            elif quantity_char == '%':
                fraction = quantity_val[0]/100.0
                quantity_absolute = data[parameter_details['bsc_parameter_name']][relevant_inds]*fraction

            #3. the greater of either an absolute value or percentage
            elif (quantity_char == '/%') or (quantity_char == '%/'):

                if quantity_char == '/%':
                    absolute = quantity_val[0]
                    fraction = quantity_val[1]/100.0

                elif quantity_char == '%/':
                    absolute = quantity_val[1]
                    fraction = quantity_val[0]/100.0

                quantity_absolute = []
                for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                    #percentage (as absolute) greater
                    if (fraction*val) > absolute:
                        quantity_absolute=np.append(quantity_absolute,fraction*val)
                    #absolute greater
                    else:
                        quantity_absolute=np.append(quantity_absolute,absolute)

            #4. percentage value after greater or equal to a fixed limit (minimum uncertainty taken from this limit)
            elif quantity_char == '%>=':
                limit = quantity_val[1]
                fraction = quantity_val[0]/100.0

                quantity_absolute = []
                for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                    #percentage (as absolute) higher than limit
                    if (fraction*val) >= limit:
                        quantity_absolute=np.append(quantity_absolute,fraction*val)
                    #not greater than limit
                    else:
                        quantity_absolute=np.append(quantity_absolute,fraction*limit)

            #5. choice of 2 absolute values under/over fixed limit
            elif quantity_char == '< >=':
                absolute_under =  quantity_val[0]
                limit =     quantity_val[1]
                absolute_over =  quantity_val[2]

                quantity_absolute = []
                for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                    #absolute value higher than limit
                    if val >= limit:
                        quantity_absolute=np.append(quantity_absolute,absolute_over)
                    #absolute value under limit
                    elif val < limit:
                        quantity_absolute=np.append(quantity_absolute,absolute_under)

            #6. choice of 2 percentage values under/over fixed limit
            elif quantity_char == '%< %>=':
                fraction_under =  quantity_val[0]/100.0
                limit =           quantity_val[1]
                fraction_over =   quantity_val[2]/100.0

                quantity_absolute = []
                for val in data[parameter_details['bsc_parameter_name']][relevant_inds]:
                    #percentage (as absolute) >= limit
                    if (fraction_over*val) >= limit:
                        quantity_absolute=np.append(quantity_absolute,fraction_over*val)
                    #< limit
                    else:
                        quantity_absolute=np.append(quantity_absolute,fraction_under*val)

            quantities_absolute.append(quantity_absolute)

        #calculate derived uncertainty -- sqrt((precision**2) + (accuracy**2))
        uncertainty = np.sqrt((quantities_absolute[0]**2)+(quantities_absolute[1]**2))

        #update field
        data['derived_uncertainty_per_measurement'][relevant_inds] = uncertainty

    return data


###------------------------------------------------------------------------------------###

def quality_control_processing(slurm_job_ID, network, parameter, station_reference):

    '''read temporary specific station numpy .npz file and pass through multiple quality control checks,
       generating many quality assurance codes characterising observational quality.
    '''

    import quality_control_functions

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- QC\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get metadata/data for reference
    metadata, data, station_numpy_filename = get_data_from_reference('S5',station_reference,run_details,parameter_details,read='all')

    #process reported and derived uncertainty per measurement
    data = process_reported_and_derived_uncertainty_per_measurement(parameter_details,metadata,data)

    #process station data through quality control checks
    data = quality_control_functions.do_qa(run_details,parameter_details,metadata,data)

    #re-save metadata/data for station
    write_station_specific_numpy(metadata,data,run_details,parameter_details,'S5')

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('S5',run_details,0,log_file_str,station_reference)

###------------------------------------------------------------------------------------###

def classification_processing(slurm_job_ID, network, parameter, station_reference):

    '''read temporary station specific .npz file, and process through classification checks, re-saving file with new classification information.'''

    import classification_functions
    import numpy.lib.recfunctions as rf

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- CLASSIFICATIONS\n'.format(station_reference)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get metadata/data for reference
    metadata, data, station_numpy_filename = get_data_from_reference('S6',station_reference,run_details,parameter_details,read='all')

    #process station data through classification processing
    data = classification_functions.do_classifications(metadata,data)

    #remove some data variables from dataframe that no longer need (as have been used when creating QA flags)
    data = rf.drop_fields(data, ['reported_lower_limit_of_detection_per_measurement','reported_upper_limit_of_detection_per_measurement'])

    #get start and end of data ranges
    start_time = data['time'][0]
    end_time = data['time'][-1]+int(data['temporal_resolution'][-1])

    #write out to finalised native file directory
    fname = run_details['data_root']+'{}/{}/native/{}/{}_{}_{}.npz'.format(run_details['network'],run_details['ghost_version'],parameter_details['bsc_parameter_name'],station_reference,start_time,end_time)
    np.savez_compressed(fname,data=data,metadata=metadata)
    #give 770 permissions to file
    os.chmod(fname, 0o770)

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('S6',run_details,0,log_file_str,station_reference)

###------------------------------------------------------------------------------------###

def temporally_grid_data(slurm_job_ID, network, parameter, station_reference, temporal_resolution_code):

    '''function for converting native mixed temporal resolution data onto a grid of times of fixed intervals, between a start and end time in UTC time
       the temporal resolution of the grid can be either hourly (H), daily (D) or monthly (M), or hourly instantaneous (HI)
       data with a resolution coarser than the temporal resolution desired is screened out.
       when converting from higher temporal resolution to lower temporal resolution, some processing is done:
       1. invalid data is first screened out using set of key qa flags (defined in "invalid_qa_flag_names").
       2. numerical data values are averaged
       3. certain qa flags which are no longer relevant are dropped (this includes qa flags used to screen out measurements in step 1, and other flags which make no sense for a temporal resolution different from which they were originally associated with)
       4. other defined qa flags are kept if they appear more often than not over period (i.e. modally)
       5. measurement flags are dropped as are no longer relevant
    '''
    
    from GHOST_standards import standard_temporal_resolutions, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code, standard_classification_flag_name_to_classification_code
    import numpy.lib.recfunctions as rf
    import quality_control_functions

    #initialise log file string
    log_file_str = ''

    log_file_str += '{} -- TEMPORALLY GRID DATA\n'.format(station_reference)

    #create log reference code (station_reference + temporal_resolution_code) 
    log_reference = '{}_{}'.format(station_reference, temporal_resolution_code)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get output temporal resolution specific variables
    temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

    #create list of all minutes in temporal window (representing start of temporal window), with spacing in defined temporal resolution between start_date and end_date 
    start_dt = datetime.datetime(int(run_details['start_year']),int(run_details['start_month']),1,0,0)
    end_dt = datetime.datetime(int(run_details['end_year']),int(run_details['end_month']),1,0,0)
    all_dt = pd.date_range(start_dt, end_dt, freq=temporal_resolution_vars['temporal_resolution_datetime_frequency']).to_pydatetime()
    grid_start_minutes = [datetime_to_fractional_minutes_from_reference(dt) for dt in all_dt] 
    #get start/end stamp of defined processing window in minutes from reference time
    defined_start_stamp = datetime_to_fractional_minutes_from_reference(start_dt)
    defined_end_stamp = datetime_to_fractional_minutes_from_reference(end_dt)
    
    #flags for removing invalid data when averaging
    invalid_qa_flag_names = ['Missing Measurement', 'Infinite Value', 'Negative Measurement', 'Zero Measurement', 'Preliminary Data', 'Invalid Data Provider Flags - GHOST Decreed', 'Recurring Value', 'Extreme Data - Scientifically Non-Feasible', 'Extreme Data - Distributional Outlier', 'Extreme Data - Manually Decreed', 'Insufficient Measurement Resolution - Preferential', 'Insufficient Measurement Resolution - Empirical', 'Below Preferential Lower Limit of Detection', 'Above Preferential Upper Limit of Detection', 'Erroneous Primary Sampling', 'Erroneous Sample Preparation', 'Erroneous Measurement Methodology']
    invalid_qa_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in invalid_qa_flag_names]

    #nonsensical qa flags when concatenated/modal value is taken, therefore are dropped
    nonsensical_qa_flag_names = ['Missing Measurement', 'Infinite Value', 'Negative Measurement', 'Zero Measurement', 'Preliminary Data', 'Invalid Data Provider Flags - GHOST Decreed', 'Invalid Data Provider Flags - Network Decreed', 'Duplicate Time - GHOST Decreed Valid Flagged Values Kept', 'Duplicate Time - First Value Kept', 'Overlapping Time - GHOST Decreed Valid Flagged Values Kept', 'Overlapping Time - Finest Temporal Resolutions Kept', 'Overlapping Time - First Value Kept', 'Recurring Value', 'Extreme Data - Scientifically Non-Feasible', 'Extreme Data - Distributional Outlier', 'Extreme Data - Manually Decreed', 'Insufficient Measurement Resolution - Preferential', 'Insufficient Measurement Resolution - Empirical', 'Below Reported Lower Limit of Detection', 'Below Documented Lower Limit of Detection', 'Below Preferential Lower Limit of Detection', 'Above Reported Upper Limit of Detection', 'Above Documented Upper Limit of Detection', 'Above Preferential Upper Limit of Detection']
    nonsensical_qa_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in nonsensical_qa_flag_names]

    #qa flags to take modal value of across averaging period to write out (i.e. if flag appears more than not over period then write it out)
    modal_qa_flag_names = ['Duplicate Station', 'Assumed Gas Volume', 'No Latitude Metadata - Took Most Recent Valid Value', 'No Latitude Metadata - Took Next Valid Value', 'No Latitude Metadata - Used Manually Compiled Metadata', 'No Longitude Metadata - Took Most Recent Valid Value', 'No Longitude Metadata - Took Next Valid Value', 'No Longitude Metadata - Used Manually Compiled Metadata', 'No Altitude Metadata - Took Most Recent Valid Value', 'No Altitude Metadata - Took Next Valid Value', 'No Altitude Metadata - Used Manually Compiled Metadata', 'No Altitude Metadata - Used ETOPO1 to Estimate Altitude', 'No Sampling Height Metadata - Took Most Recent Valid Value', 'No Sampling Height Metadata - Took Next Valid Value', 'No Sampling Height Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Area Classification Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Area Classification Metadata - Took Next Valid Value', 'No Standardised Network Provided Area Classification Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Station Classification Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Station Classification Metadata - Took Next Valid Value', 'No Standardised Network Provided Station Classification Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Main Emission Source Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Main Emission Source Metadata - Took Next Valid Value', 'No Standardised Network Provided Main Emission Source Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Land Use Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Land Use Metadata - Took Next Valid Value', 'No Standardised Network Provided Land Use Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Terrain Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Terrain Metadata - Took Next Valid Value', 'No Standardised Network Provided Terrain Metadata - Used Manually Compiled Metadata', 'No Standardised Network Provided Measurement Scale Metadata - Took Most Recent Valid Value', 'No Standardised Network Provided Measurement Scale Metadata - Took Next Valid Value', 'No Standardised Network Provided Measurement Scale Metadata - Used Manually Compiled Metadata', 'No Representative Radius Metadata - Took Most Recent Valid Value', 'No Representative Radius Metadata - Took Next Valid Value', 'No Representative Radius Metadata - Used Manually Compiled Metadata', 'Non-Integer Local Timezone (relative to UTC)', 'Insufficient Measurement Resolution - Reported', 'Insufficient Measurement Resolution - Documented', 'No Documented/Reported Measurement Resolution Metadata', 'No Documented/Reported Lower Limit of Detection Metadata', 'No Documented/Reported Upper Limit of Detection Metadata', 'Methodology Not Mapped', 'Assumed Primary Sampling', 'Assumed Sample Preparation', 'Assumed Measurement Methodology', 'Unknown Primary Sampling Type', 'Unknown Primary Sampling Instrument', 'Unknown Sample Preparation Type', 'Unknown Sample Preparation Technique', 'Unknown Measurement Method', 'Unknown Measuring Instrument', 'Erroneous Primary Sampling', 'Erroneous Sample Preparation', 'Erroneous Measurement Methodology', 'Invalid QA Measurement Methodology']
    modal_qa_flag_codes = [standard_QA_name_to_QA_code[flag_name] for flag_name in modal_qa_flag_names]

    #classification flags to take modal value of across averaging period to write out (i.e. if flag appears more than not over period then write it out)
    modal_classification_flag_names =  ['High Altitude - Metadata Altitude', 'High Altitude - ETOPO1', 'High Altitude - Iwahashi Global Landform Classification', 'High Altitude - Meybeck Global Landform Classification', 'Near Coast - GSFC', 'Rural Station - Lenient Metadata Derived', 'Urban Station - Lenient Metadata Derived', 'Unclassified Station - Lenient Metadata Derived', 'Rural Station - Strict Metadata Derived', 'Urban Station - Strict Metadata Derived', 'Unclassified Station - Strict Metadata Derived', 'Rural Station - Anthrome (Native Resolution)', 'Urban Station - Anthrome (Native Resolution)', 'Rural Station - Anthrome (Mode in 5km Perimeter)', 'Urban Station - Anthrome (Mode in 5km Perimeter)', 'Rural Station - Anthrome (Mode in 25km Perimeter)', 'Urban Station - Anthrome (Mode in 25km Perimeter)', 'Rural Station - TOAR', 'Urban Station - TOAR', 'Unclassified Station - TOAR', 'Rural Station - Joly-Peuch', 'Unclassified Station - Joly-Peuch', "Area Classification Metadata = 'urban'", "Area Classification Metadata = 'urban-centre'", "Area Classification Metadata = 'urban-suburban'", "Area Classification Metadata = 'rural'", "Area Classification Metadata = 'rural-near_city'", "Area Classification Metadata = 'rural-regional'", "Area Classification Metadata = 'rural-remote'", "Station Classification Metadata = 'background'", "Station Classification Metadata = 'point_source'", "Station Classification Metadata = 'point_source-industrial'", "Station Classification Metadata = 'point_source-traffic'", "Spatial Representativity Metadata = '<0.1km'", "Spatial Representativity Metadata = '0.1-0.5km'", "Spatial Representativity Metadata = '0.5-4km'", "Spatial Representativity Metadata = '4-50km'", "Spatial Representativity Metadata = '>=50km'", "Terrain Metadata = 'coastal'", "Terrain Metadata = 'complex'", "Terrain Metadata = 'flat'", "Terrain Metadata = 'mountain'", "Terrain Metadata = 'rolling'", "Land Use Metadata = 'barren'", "Land Use Metadata = 'forest'", "Land Use Metadata = 'open'", "Land Use Metadata = 'snow'", "Land Use Metadata = 'urban'", "Land Use Metadata = 'water'", "Land Use Metadata = 'wetland'", "MODIS MCD12C1 Land Use = 'barren'", "MODIS MCD12C1 Land Use = 'forest'", "MODIS MCD12C1 Land Use = 'open'", "MODIS MCD12C1 Land Use = 'snow'", "MODIS MCD12C1 Land Use = 'urban'", "MODIS MCD12C1 Land Use = 'water'", "MODIS MCD12C1 Land Use = 'wetland'", 'Daytime', 'Nighttime', 'Weekday', 'Weekend', 'Winter', 'Spring', 'Summer', 'Autumn']
    modal_classification_flag_codes = [standard_classification_flag_name_to_classification_code[flag_name] for flag_name in modal_classification_flag_names]

    #get metadata/data for reference
    metadata, data, station_numpy_filename = get_data_from_reference('T1',station_reference,run_details,parameter_details,read='all')

    #get start and end stamps of read station data
    start_stamp = data['time'][0]
    end_time = data['time'][-1]
    end_res = data['temporal_resolution'][-1]
    end_stamp = end_time + end_res

    #if start stamp >= defined end stamp OR end stamp <= defined start_stamp, no valid data in range, therefore do not process further
    if (start_stamp >= defined_end_stamp) or (end_stamp <= defined_start_stamp):

        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'NO VALID DATA IN DESIRED RANGE TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)

    #cut station metadata/data for relevant period
    metadata = cut_data_array(metadata, 'metadata', defined_start_stamp, defined_end_stamp)
    data =     cut_data_array(data, 'data', defined_start_stamp, defined_end_stamp)
    
    #if have no time data after cutting then no valid data in range, therefore do not process further
    if len(data['time']) == 0:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'NO VALID DATA IN DESIRED RANGE TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)

    #re-get start and end stamps of read data (after cutting)
    start_stamp = data['time'][0]
    end_time = data['time'][-1]
    end_res = data['temporal_resolution'][-1]
    end_stamp = end_time + end_res

    #if start of first gridded time period is after end of station data range, no valid data in range, therefore do not process further
    if grid_start_minutes[0] >= data['time'][-1]:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'NO VALID DATA IN DESIRED RANGE TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)
    #if end of last gridded time period is before start of station data range, no valid data in range, therefore do not process further
    if grid_start_minutes[-1] <= data['time'][0]:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'NO VALID DATA IN DESIRED RANGE TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)

    #get data keys
    data_keys = list(data.dtype.names)
    #remove temporal resolution key from list of keys
    data_keys.remove('temporal_resolution')

    #create empty numpy dataframe for storing gridded data
    standard_dtype = [(key,data[key].dtype) for key in data_keys]
    gridded_data = np.empty(len(grid_start_minutes[:-1]),dtype=standard_dtype)

    #iterate through data_keys to instantiate new numpy dataframe with data
    for kk, key in enumerate(data_keys):

        #add gridded time to dataframe
        if key == 'time':
            gridded_data['time'] = grid_start_minutes[:-1]

        #add array of NaNs for np.float32 types
        elif gridded_data[key].dtype == np.dtype('float32'):
            gridded_data[key] = np.array([np.NaN]*len(grid_start_minutes[:-1]))

        #make list of lists of qa codes - all initialised as empty
        elif key == 'qa':
            gridded_data[key] = [[] for i in range(len(grid_start_minutes[:-1]))]

        #make list of list of data flag codes - with a default of Un-flagged data code
        elif key == 'flag':
            valid_data_flag_code = standard_data_flag_name_to_data_flag_code['Valid Data']
            unflagged_data_flag_code = standard_data_flag_name_to_data_flag_code['Un-Flagged Data']
            gridded_data[key] = [[unflagged_data_flag_code] for i in range(len(grid_start_minutes[:-1]))]

         #make list of lists of classification codes - all initialised as empty
        elif key == 'classification':
            gridded_data[key] = [[] for i in range(len(grid_start_minutes[:-1]))]

    #get valid start and end times
    #screen out data that has coarser resolution than desired output temporal resolution
    valid_start_times = []
    valid_end_times = []
    valid_timedeltas = []
    valid_inds = []
    for ii, start_time in enumerate(data['time']):

        process_flag = False

        #get window end time by adding temporal resolution as timedelta
        timedelta = data['temporal_resolution'][ii]
        if temporal_resolution_code == 'HI':
            if timedelta == 0:
                process_flag = True
        elif temporal_resolution_code == 'H':
            if timedelta <= 60:
                process_flag = True
        elif temporal_resolution_code == 'D':
            if timedelta <= (24*60):
                process_flag = True
        elif temporal_resolution_code == 'M':
            if timedelta <= (31*24*60):
                process_flag = True

        #get measurement window end time
        if process_flag == True:
            valid_start_times.append(start_time)
            valid_end_times.append(start_time + timedelta)
            if temporal_resolution_code == 'HI':
                valid_timedeltas.append(1)
            else:
                valid_timedeltas.append(timedelta)
            valid_inds.append(ii)

    #check have some valid data after removing coarse resolution data
    if len(valid_inds) == 0:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'AFTER REMOVING COARSE RESOLUTION DATA, NO VALID DATA TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)
    #next check have some data reported between start_date and end_date - after removing coarse resolutions
    #if start of first gridded time period is after end of station data range, no valid data in range, therefore do not process further
    elif grid_start_minutes[0] >= valid_end_times[-1]:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'AFTER REMOVING COARSE RESOLUTION DATA, NO VALID DATA TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)
    #if end of last gridded time period is before start of station data range, no valid data in range, therefore do not process further
    elif grid_start_minutes[-1] <= valid_start_times[0]:
        #return logfile to inform station no longer continues in pipeline (1)
        log_file_str += 'AFTER REMOVING COARSE RESOLUTION DATA, NO VALID DATA TO TEMPORALLY GRID'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)

    #convert cut arrays to numpy arrays
    valid_start_times = np.array(valid_start_times, dtype=np.uint32)
    valid_end_times = np.array(valid_end_times, dtype=np.uint32)
    valid_timedeltas = np.array(valid_timedeltas , dtype=np.uint32)
    valid_inds = np.array(valid_inds , dtype=np.uint32)

    #flag that determines if have valid data to write for station netCDF file
    data_to_write = False

    #--------------------------------------------------------------------------------#
    #--------------------------------------------------------------------------------#
    #initialise variables for evaluating temporal representation across periods

    #get QA flag for flagging windows which have no measurements provided within them
    missing_measurement_qa_flag_code = standard_QA_name_to_QA_code['Missing Measurement']

    #get QA flag for flagging windows which have no contain no valid data (after screening by key QA flags --> those in invalid_qa_flags)
    no_valid_data_qa_flag_code = standard_QA_name_to_QA_code['No Valid Data']

    #define periods whose representativity will be evaluated in the data
    periods = ['hourly', 'daily', 'weekly', 'monthly', 'annual']
    #define relevant periods for the temporal resolution currently processing
    if temporal_resolution_code == 'M':
        relevant_periods = ['monthly','annual']
    elif temporal_resolution_code == 'D':
        relevant_periods = ['daily','weekly','monthly','annual']
    else:
        relevant_periods = ['hourly','daily','weekly','monthly','annual']

    #setup default temporal resolution representation flags, for setting QA flags for all periods which have some measurement data, but have a finer resolution than the resolution gridding data onto
    #all unrepresentative flags are set for the periods which are finer in resolution than resolution gridding onto.
    coarse_representation_flag_codes = []
    data_completeness_limits = [25,50,66,75,90]
    max_data_gap_limits = [10,25,33,50,75]

    #initialise dictionary for keeping track of minutes represented across different periods, and data gaps
    period_tracker = {}
    for period in periods: 
        period_tracker[period] = {'n_minutes_represented': 0, 'gaps':[], 'inds':[], 'open_gap':False, 'unrepresentative_flag_codes':[], 'new_period':False}
        #initialise lists containing all possible unrepresentative flags for each temporal period
        for data_completeness_limit in data_completeness_limits:
            period_tracker[period]['unrepresentative_flag_codes'].append(standard_QA_name_to_QA_code['{} Window Data Completeness < {}%'.format(period.capitalize(), data_completeness_limit)])
        for max_data_gap_limit in max_data_gap_limits:
            period_tracker[period]['unrepresentative_flag_codes'].append(standard_QA_name_to_QA_code['{} Window Maximum Data Gap >= {}%'.format(period.capitalize(), max_data_gap_limit)])
       
    #set variable that will hold last written qa codes, for the purpose of filling blank windows (i.e. where have no measurements), to ensure qa flags do not jump wildly.
    #makes assumption qa flags hold for blank windows
    most_recent_qa_codes = []

    #set variable that will hold last written classification codes, for the purpose of filling blank windows (i.e. where have no measurements), to ensure classification flags do not jump wildly.
    #makes assumption classification flags hold for blank windows
    most_recent_classification_codes = []

    #if gridding onto daily/monthly grids, set all flags for periods finer in resolution to be unrepresentative (i.e. Monthly --> hourly/daily/weekly periods; Daily --> hourly period)
    if temporal_resolution_code == 'M':
            coarse_representation_flag_codes.extend(period_tracker['hourly']['unrepresentative_flag_codes'])
            coarse_representation_flag_codes.extend(period_tracker['daily']['unrepresentative_flag_codes'])
            coarse_representation_flag_codes.extend(period_tracker['weekly']['unrepresentative_flag_codes'])
    elif temporal_resolution_code == 'D':
            coarse_representation_flag_codes.extend(period_tracker['hourly']['unrepresentative_flag_codes'])

    #get index of first window which contains measurements, and iterate through windows from there
    if temporal_resolution_code == 'HI':
        window_start_index = bisect.bisect(grid_start_minutes-5.0, valid_start_times[0])-1
    else:
        window_start_index = bisect.bisect(grid_start_minutes, valid_start_times[0])-1
    
    #set all qa for windows before first valid datetime to have a QA flag stating there are no measurements in the windows
    for qa_xx in range(window_start_index):
        gridded_data['qa'][qa_xx].append(missing_measurement_qa_flag_code)

    #--------------------------------------------------------------------------------#
    #--------------------------------------------------------------------------------#

    #iterate through gridded time windows
    for ii, window_start_minute in enumerate(grid_start_minutes[window_start_index:-1]):

        #get actual index of window (as have skipped through windows before first valid datetime)
        actual_ii = window_start_index+ii
        if temporal_resolution_code == 'HI':
            window_start_minute = window_start_minute - 5
            window_end_minute = window_start_minute + 10
        else:
            window_end_minute = grid_start_minutes[actual_ii+1]

        #log_file_str += '{} {}\n'.format(fractional_minutes_from_reference_to_datetime(float(window_start_minute)), fractional_minutes_from_reference_to_datetime(float(window_end_minute))

        #reset native variables
        n_minutes_represented = 0

        #update period group indices
        for period in relevant_periods:
            period_tracker[period]['inds'].append(actual_ii)

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #update current window start year/month/week/day/hour
        current_dt = fractional_minutes_from_reference_to_datetime(float(window_start_minute))
        current_year = current_dt.year
        current_month = current_dt.month
        current_week = current_dt.isocalendar()[1]
        current_day = current_dt.day
        current_hour = current_dt.hour

        #update next window start year/month/week/day/hour
        next_dt = fractional_minutes_from_reference_to_datetime(float(window_end_minute))
        next_year = next_dt.year
        next_month = next_dt.month
        next_week = next_dt.isocalendar()[1]
        next_day = next_dt.day
        next_hour = next_dt.hour

        #determine if next window start datetime is a new hour/day/week/month

        #is it a new year
        if (next_year != current_year):
            period_tracker['annual']['new_period'] = True
        else:
            period_tracker['annual']['new_period'] = False

        #is it a new month?
        if (next_year != current_year) or (next_month != current_month):
            period_tracker['monthly']['new_period'] = True
        else:
            period_tracker['monthly']['new_period'] = False

        #is it a new week?
        if (next_week != current_week):
            period_tracker['weekly']['new_period'] = True
        else:
            if (next_year != current_year) & (current_week not in [1,52,53]):
                period_tracker['weekly']['new_period'] = True
            else:
                period_tracker['weekly']['new_period'] = False

        #is it a new day?
        if (next_year != current_year)  or (next_month != current_month) or (next_day != current_day):
            period_tracker['daily']['new_period'] = True
        else:
            period_tracker['daily']['new_period'] = False

        #is it a new hour?
        if (next_year != current_year)  or (next_month != current_month) or (next_day != current_day) or (next_hour != current_hour):
            period_tracker['hourly']['new_period'] = True
        else:
            period_tracker['hourly']['new_period'] = False

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #if last of valid end times is <= left edge of window, will be no more valid times in window, so break out of iteration
        if valid_end_times[-1] <=  window_start_minute:

            #calculate representativity of periods not yet evaluated

            #get start datetime of previous window
            previous_dt = fractional_minutes_from_reference_to_datetime(float(grid_start_minutes[actual_ii-1]))

            #cut off last period groups indices (as this window is not valid)
            for period in relevant_periods:
                period_tracker[period]['inds'] = period_tracker[period]['inds'][:-1]

            #calculate representativity of periods (if were not evaluated in previous step)
            for period in relevant_periods:
                if len(period_tracker[period]['inds']) > 0: 
                    #if have no data in window, set all unrepresentative flags, and write out QA flag to inform have no valid data in the window (if not already written)
                    if period_tracker[period]['n_minutes_represented'] == 0:
                        for no_valid_data_ii in period_tracker[period]['inds']:
                            gridded_data['qa'][no_valid_data_ii].extend(period_tracker[period]['unrepresentative_flag_codes'])
                    #if have data in window, evaluate data representation across previous period
                    else:
                        gridded_data['qa'] = quality_control_functions.check_representation(period_tracker[period]['n_minutes_represented'], period_tracker[period]['gaps'], period, gridded_data['qa'], period_tracker[period]['inds'], previous_dt)

            #set all windows after last valid datetime to have missing measurement QA flag code
            for qa_xx in range(actual_ii,len(grid_start_minutes[:-1])):
                gridded_data['qa'][qa_xx].append(missing_measurement_qa_flag_code)

            #stop iterating through windows
            break


        #if left edge of window is <= first valid start time, set start_window_i to 0
        if window_start_minute <= valid_start_times[0]:
            start_window_i = 0
            left_aligned = True
        #if left edge of window is > last valid start time, set start_window_i to last valid index
        elif window_start_minute > valid_start_times[-1]:
            start_window_i = len(valid_start_times)-1
        #otherwise, check if first valid start time aligns with left edge of window
        else:
            start_window_i = bisect.bisect_left(valid_start_times,window_start_minute)
            #is start time aligned with start of window?
            if window_start_minute != valid_start_times[start_window_i]:
                #if not aligned, minus 1 to index correctly
                start_window_i = start_window_i - 1
                left_aligned = False
            else:
                left_aligned = True

        #if right edge of window is >= last valid end time, set end_window_i to number of valid times
        if window_end_minute >= valid_end_times[-1]:
            end_window_i = len(valid_end_times)
            right_aligned = True
        #otherwise, check if last valid end time aligns with right edge of window
        else:
            end_window_i = bisect.bisect_left(valid_end_times,window_end_minute)
            #is end time aligned with end of window?
            if window_end_minute != valid_end_times[end_window_i]:
                #if not aligned, plus 1 to index correctly
                end_window_i = end_window_i + 1
                right_aligned = False
            else:
                right_aligned = True

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #get difference between relevant indices
        window_index_diff = end_window_i - start_window_i
        #just 1 relevant index
        if window_index_diff == 0:
            window_indices = np.array([start_window_i])
        #multiple relevant indices
        else:
            window_indices = np.arange(start_window_i, end_window_i)

        #if temporal resolution is "hourly instantaneous", then limit the window_indices to be just the index closest to the hour 
        if temporal_resolution_code == 'HI':
            print(valid_start_times[window_indices] - window_start_minute)
            closest_index = np.abs(valid_start_times[window_indices] - window_start_minute).argmin()
            window_indices = np.array([closest_index])

        #get fractional overlaps on edges of windows (if exist)
        if len(window_indices) > 1:
            left_overlap = np.NaN
            right_overlap = np.NaN

            #Determine if the first measurement period end time is within the measurement window, if not slice off first index
            if valid_end_times[window_indices[0]] <= window_start_minute:
                window_indices = window_indices[1:]

            #Determine if the last measurement period start time is within the measurement window, if not slice off last index
            if valid_start_times[window_indices[-1]] >= window_end_minute:
                window_indices = window_indices[:-1]

            #if have 0 window indices left, it is because the measurement window is in a void between measurements
            if len(window_indices) == 0:
                n_minutes_represented = 0
                native_gap = window_end_minute - window_start_minute
                for period in relevant_periods:
                    period_tracker[period]['gaps'].append(native_gap)
                    period_tracker[period]['open_gap'] = True   

                for period in relevant_periods:
                    #if new period then calculate representativity of previous period
                    if period_tracker[period]['new_period'] == True: 
                        #if have no data in window, set all unrepresentative flags, and write out QA flag to inform have no valid data in the window (if not already written)
                        if period_tracker[period]['n_minutes_represented'] == 0:
                            for no_valid_data_ii in period_tracker[period]['inds']:
                                gridded_data['qa'][no_valid_data_ii].extend(period_tracker[period]['unrepresentative_flag_codes'])
                        #update period variables
                        period_tracker[period]['n_minutes_represented'] = 0
                        period_tracker[period]['gaps'] = []
                        period_tracker[period]['inds'] = []
                        period_tracker[period]['open_gap'] = []

                #append QA flags to state have no measurements in window
                gridded_data['qa'][actual_ii].append(missing_measurement_qa_flag_code)

                #update QA flags with coarse representation codes
                gridded_data['qa'][actual_ii].extend(coarse_representation_flag_codes)

                #copy most recent qa/data flag/classification codes to this empty window
                gridded_data['qa'][actual_ii].extend(most_recent_qa_codes)
                gridded_data['classification'][actual_ii].extend(most_recent_classification_codes)

                continue

            if len(window_indices) > 1:
                #if the left edge of measurement window does align with the end/start of a measurement period, determine why this is.
                if left_aligned == False:

                    #Determine if the first measurement period overlaps the left window edge
                    if (valid_start_times[window_indices[0]] - window_start_minute) < 0:
                        #get number of minutes of measurement period overlapping measurement window on left edge
                        left_overlap = valid_end_times[window_indices[0]] - window_start_minute

                #if the right edge of measurement window does align with the end/start of a measurement period, determine why this is.
                if right_aligned == False:

                    #Determine if the last measurement period overlaps the right window edge
                    if (window_end_minute - valid_end_times[window_indices[-1]]) < 0:
                        #get number of minutes of measurement period overlapping measurement window on right edge
                        right_overlap = window_end_minute - valid_start_times[window_indices[-1]]

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #get indices of measurements within window, respective to full data arrays
        data_respective_window_indices = valid_inds[window_indices]

        #if just have 1 index, get number of minutes represented in window and maximum temporal gap (if data is valid with regards to qa flags)
        if len(window_indices) == 1:

            #if have 1 index will always write out 
            write_out = True

            #evaluate data validity by key qa flags
            valid_data_respective_window_indices = data_respective_window_indices[np.array([all(x not in qa_flag_codes for x in invalid_qa_flag_codes) for qa_flag_codes in data['qa'][data_respective_window_indices]])]
            if len(valid_data_respective_window_indices) > 0:

                #set flag to state have valid data (after key qa flag screening)
                have_valid_data = True

                relevant_start_time = valid_start_times[window_indices][0]
                relevant_end_time = valid_end_times[window_indices][0]
                #if relevant start/end time exceed start/end of window, modify relevant start/end time to be that of window start/end
                if relevant_start_time < window_start_minute:
                    relevant_start_time = window_start_minute
                if relevant_end_time > window_end_minute:
                    relevant_end_time = window_end_minute

                #get n minutes represented
                n_minutes_represented = relevant_end_time - relevant_start_time
                #add n_minutes_represented for all relevant temporal resolutions
                for period in relevant_periods:
                    period_tracker[period]['n_minutes_represented'] = period_tracker[period]['n_minutes_represented'] + n_minutes_represented

                #append/process gaps across other temporal resolutions
                left_edge_gap = relevant_start_time - window_start_minute
                right_edge_gap = window_end_minute - relevant_end_time
                
                if left_edge_gap > 0:
                    for period in relevant_periods:
                        if period_tracker[period]['open_gap'] == True:
                            period_tracker[period]['gaps'][-1] = period_tracker[period]['gaps'][-1] + left_edge_gap
                        else:
                            period_tracker[period]['gaps'].append(left_edge_gap)
                #reset open gap flags to be False
                for period in relevant_periods:
                    period_tracker[period]['open_gap'] = False

                if right_edge_gap > 0:
                    for period in relevant_periods:
                        period_tracker[period]['gaps'].append(right_edge_gap)
                        #set open gap to True
                        period_tracker[period]['open_gap'] = True

            #no valid data
            else:
                n_minutes_represented = 0
                window_gap = window_end_minute - window_start_minute
                for period in relevant_periods:
                    if period_tracker[period]['open_gap'] == True:
                        period_tracker[period]['gaps'][-1] = period_tracker[period]['gaps'][-1] + window_gap
                    else:
                        period_tracker[period]['gaps'].append(window_gap)
                        #set open gap to True
                        period_tracker[period]['open_gap'] = True

                #set flag to state have no valid data (after key qa flag screening)
                have_valid_data = False

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #if are averaging data within window (i.e. more than 1 measurement, screen data to remove invalid data
        #also calculate number of minutes represented in window and maximum temporal gap
        elif len(window_indices) > 1:

            #weight data by timedeltas for averaging different temporal resolution data
            window_weights = valid_timedeltas[window_indices]
            #modify weights on left and right edges to adjust for actual time sampled within window, if measurement overlaps edge
            if pd.isnull(left_overlap) == False:
                window_weights[0] = left_overlap
            if pd.isnull(right_overlap) == False:
                window_weights[-1] = right_overlap

            #cut arrays solely for valid values (by screening by key qa flags)
            valid_mask = np.array([all(x not in qa_flag_codes for x in invalid_qa_flag_codes) for qa_flag_codes in data['qa'][data_respective_window_indices]])
            window_indices = window_indices[valid_mask]
            window_weights = window_weights[valid_mask]
            valid_data_respective_window_indices = data_respective_window_indices[valid_mask]

            if len(valid_data_respective_window_indices) > 0:

                #have valid data after screening by key qa flags, therefore write out
                write_out = True

                #set flag to state have valid data (after key qa flag screening)
                have_valid_data = True

                relevant_start_times = valid_start_times[window_indices]
                relevant_end_times = valid_end_times[window_indices]
                #if relevant first start time/ last end time exceed start/end of window, modify relevant start/end time to be that of window start/end
                if relevant_start_times[0] < window_start_minute:
                    relevant_start_times[0] = window_start_minute
                if relevant_end_times[-1] > window_end_minute:
                    relevant_end_times[-1] = window_end_minute

                #get n minutes represented
                n_minutes_represented = np.sum(window_weights)

                #add n_minutes_represented for all relevant temporal resolutions
                for period in relevant_periods:
                    period_tracker[period]['n_minutes_represented'] = period_tracker[period]['n_minutes_represented'] + n_minutes_represented

                #deal with open gaps on left edge of window
                left_edge_gap = relevant_start_times[0] - window_start_minute
                if left_edge_gap > 0:
                    for period in relevant_periods:
                        if period_tracker[period]['open_gap'] == True:
                            period_tracker[period]['gaps'][-1] = period_tracker[period]['gaps'][-1] + left_edge_gap
                        else:
                            period_tracker[period]['gaps'].append(left_edge_gap)
                #reset open gap flags to be False
                for period in relevant_periods:
                    period_tracker[period]['open_gap'] = False

                #deal with open gaps in centre of window
                for obs_start,obs_end in zip(relevant_start_times[1:],relevant_end_times[:-1]):
                    gap = obs_start - obs_end
                    if gap > 0:
                        for period in relevant_periods:
                            period_tracker[period]['gaps'].append(gap)

               #deal with open gaps on right edge of window
                right_edge_gap = window_end_minute - relevant_end_times[-1]
                if right_edge_gap > 0:
                    for period in relevant_periods:
                        period_tracker[period]['gaps'].append(right_edge_gap)
                        #set open gap to True
                        period_tracker[period]['open_gap'] = True

            #no valid data
            else:
                #have no valid data after screening by key qa flags, therefore do not write out
                write_out = False

                n_minutes_represented = 0
                window_gap = window_end_minute - window_start_minute
                for period in relevant_periods:
                    if period_tracker[period]['open_gap'] == True:
                        period_tracker[period]['gaps'][-1] = period_tracker[period]['gaps'][-1] + window_gap
                    else:
                        period_tracker[period]['gaps'].append(window_gap)
                        #set open gap to True
                        period_tracker[period]['open_gap'] = True

                #set flag to state have no valid data (after key qa flag screening)
                have_valid_data = False

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#
        #now process QA flags for the evaluation of the representativeness of periods

        for period in relevant_periods:
            #if new period then calculate representativity of previous period
            if period_tracker[period]['new_period'] == True: 
                #if have no data in window, set all unrepresentative flags, and write out QA flag to inform have no valid data in the window (if not already written)
                if period_tracker[period]['n_minutes_represented'] == 0:
                    for no_valid_data_ii in period_tracker[period]['inds']:
                        gridded_data['qa'][no_valid_data_ii].extend(period_tracker[period]['unrepresentative_flag_codes'])
                #if have data in window, evaluate data representation across previous period
                else:
                    gridded_data['qa'] = quality_control_functions.check_representation(period_tracker[period]['n_minutes_represented'], period_tracker[period]['gaps'], period, gridded_data['qa'], period_tracker[period]['inds'], current_dt)
                #update period variables
                period_tracker[period]['n_minutes_represented'] = 0
                period_tracker[period]['gaps'] = []
                period_tracker[period]['inds'] = []
                period_tracker[period]['open_gap'] = []
            
        #update QA flags with coarse representation codes
        gridded_data['qa'][actual_ii].extend(coarse_representation_flag_codes)

        #if have no valid data (after screening by key qa flags), update with no valid data QA flag
        if have_valid_data == False:
            gridded_data['qa'][actual_ii].append(no_valid_data_qa_flag_code)

        #--------------------------------------------------------------------------------#
        #--------------------------------------------------------------------------------#

        #proceed if have valid indices for window after screening GHOST Decreed invalid data (when have > 1 window indices), or have chosen to filter by QA flags (when have 1 index)
        if write_out == True:

            #update flag stating have some valid data for station
            data_to_write = True

            #iterate through data variables and process data onto desired temporal resolution grid
            for key in data_keys:

                #variables of type: float32
                if data[key].dtype == np.dtype('float32'):

                    #if are averaging data into desired resolution, take weighted average
                    if len(window_indices) > 1:
                        window_weighted_average = np.average(data[key][valid_data_respective_window_indices], weights=window_weights)
                        gridded_data[key][actual_ii] = window_weighted_average
                    else:
                        gridded_data[key][actual_ii] = data[key][data_respective_window_indices][0]

                #qa codes
                elif key == 'qa':

                    #if have more than 1 measurement in window, need to deal with multiple qa flag codes
                    #this involves 2 steps:
                    #1. drop qa codes which are no longer relevant when taken across an averaged period (most of which have already been applied to screen invalid data)
                    #2. other qa codes are retained if they are more frequently reported across period than not, otherwise they are dropped
                    if len(window_indices) > 1:

                        #get all unique qa flag codes across measurements that are averaging
                        unique_qa = list(set(x for l in data[key][valid_data_respective_window_indices] for x in l))
                        #some qa flags in unique list no longer make sense when taking average over period, therefore drop them
                        unique_qa = [x for x in unique_qa if x not in nonsensical_qa_flag_codes]

                        #for some qa codes need to check if they are more frequently reported or not across the period, if more frequently not reported, drop the code
                        for modal_flag in modal_qa_flag_codes:
                            #first only proceed if flag is actually at all reported in the window
                            if modal_flag in unique_qa:
                                reported_count = 0
                                not_reported_count = 0
                                for ww, qa_codes in enumerate(data[key][valid_data_respective_window_indices]):
                                    if modal_flag in qa_codes:
                                        reported_count += window_weights[ww]
                                    else:
                                        not_reported_count += window_weights[ww]
                                if not_reported_count > reported_count:
                                    unique_qa.remove(modal_flag)

                        #append unique relevant qa codes for averaged period
                        gridded_data[key][actual_ii].extend(unique_qa)

                        #make copy of processed qa codes, for copying when have empty window
                        most_recent_qa_codes = copy.deepcopy(unique_qa)

                    else:
                        #add previously processed qa codes for this window to those processed through temporal gridding
                        gridded_data[key][actual_ii].extend(data[key][data_respective_window_indices][0])

                        #make copy of previously processed qa codes for this window, for copying when have empty window
                        most_recent_qa_codes = copy.deepcopy(data[key][data_respective_window_indices][0])

                #classification flags
                elif key == 'classification':

                    #if have more than 1 measurement in window, need to deal with multiple classification codes
                    #classification codes are retained if they are more frequently reported across period than not, otherwise they are dropped
                    if len(window_indices) > 1:

                        #get all unique classification codes across measurements that are averaging
                        unique_classification = list(set(x for l in data[key][valid_data_respective_window_indices] for x in l))

                        #for some classification codes need to check if they are more frequently reported or not across the period, if more frequently not reported, drop the code
                        for modal_flag in modal_classification_flag_codes:
                            #first only proceed if flag is actually at all reported in the window
                            if modal_flag in unique_classification:
                                reported_count = 0
                                not_reported_count = 0
                                for ww, classification_codes in enumerate(data[key][valid_data_respective_window_indices]):
                                    if modal_flag in classification_codes:
                                        reported_count += window_weights[ww]
                                    else:
                                        not_reported_count += window_weights[ww]
                                if not_reported_count > reported_count:
                                    unique_classification.remove(modal_flag)

                        #append unique relevant classification codes for averaged period
                        gridded_data[key][actual_ii].extend(unique_classification)

                        #make copy of processed classification codes, for copying when have empty window
                        most_recent_classification_codes = copy.deepcopy(unique_classification)

                    else:
                        #add previously processed classification codes for this window to those processed through temporal gridding
                        gridded_data[key][actual_ii].extend(data[key][data_respective_window_indices][0])

                        #make copy of previously processed classification codes for this window, for copying when have empty window
                        most_recent_classification_codes = copy.deepcopy(data[key][data_respective_window_indices][0])

                #data flags
                elif key == 'flag':

                    #if have more than 1 measurement in window, then discard flag as will have already taken account of them (through associated qa codes) and do not make sense as an average over a period
                    #instead replace with valid data flag (as will have passed invalid flag screening)
                    if len(window_indices) > 1:

                        #append valid data flag code as will have passed invalid flag screening
                        gridded_data[key][actual_ii] = [valid_data_flag_code]

                    else:
                        gridded_data[key][actual_ii] = data[key][data_respective_window_indices][0]

    #--------------------------------------------------------------------------------#
    #--------------------------------------------------------------------------------#

    #proceed to write out to station specific netCDF file
    if data_to_write == True:

        #drop duplicates, sort and pad flag/qa/classification code arrays
        padded_qa = drop_duplicates_sort_and_pad_codes('qa', gridded_data['qa'])
        padded_flag = drop_duplicates_sort_and_pad_codes('flag', gridded_data['flag'])
        padded_classification = drop_duplicates_sort_and_pad_codes('classification', gridded_data['classification'])

        #drop qa/flag from gridded_data
        gridded_data = rf.drop_fields(gridded_data, ['qa','flag','classification'])

        log_file_str += '{} -- WRITING TO MONTHLY STATION SPECIFIC TEMPORALLY GRIDDED netCDFs\n'.format(station_reference)

        #create netCDF file for station
        write_station_specific_netCDF_by_month(metadata, gridded_data, padded_qa, padded_flag, padded_classification, start_stamp, end_stamp, temporal_resolution_vars, run_details, parameter_details)

        #return valid process logfile (0)
        log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
        create_GHOST_stage_logfile('T1',run_details,0,log_file_str,log_reference)

    #otherwise no valid data has been found
    #return error logfile to inform station no longer continues in pipeline (1)
    else:
        log_file_str += 'NO VALID DATA FOUND AFTER QA FLAG CHECKS'  
        create_GHOST_stage_logfile('T1',run_details,1,log_file_str,log_reference)
###------------------------------------------------------------------------------------###

def drop_duplicates_sort_and_pad_codes(array_name,array,pad_value=255):

    '''take integer code arrays, drop duplicates and sort alphabetically (natural human order)
       then pad arrays with fill_value (255) - converting to 2D numpy arrays for preparation to writing to netCDF
    '''

    from GHOST_standards import standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code, standard_classification_flag_name_to_classification_code

    #get unique codes in each list of codes per measurement and sort the,
    unique_sorted_array = [list(sorted(set(current_list))) for current_list in array]

    #get maximum possible number of codes
    if array_name == 'qa':
        N_codes_max = len(list(standard_QA_name_to_QA_code.keys()))
    elif array_name == 'flag':
        N_codes_max = len(list(standard_data_flag_name_to_data_flag_code.keys()))
    elif array_name == 'classification':
        N_codes_max = len(list(standard_classification_flag_name_to_classification_code.keys()))

    #pad lists with fill value to maximum N codes - converting all into 2D numpy array strcture
    padded_unique_sorted_array = np.array([i + [pad_value]*(N_codes_max-len(i)) for i in unique_sorted_array], dtype=np.uint8)

    return padded_unique_sorted_array

###------------------------------------------------------------------------------------###

def write_station_specific_netCDF_by_month(metadata,data,qa,flag,classification,start_stamp,end_stamp,temporal_resolution_vars,run_details,parameter_details):

    '''write netCDF file, month by month, in desired temporal range for specific station'''

    from collections import Counter
    from dateutil.relativedelta import relativedelta
    from netCDF4 import Dataset
    import numpy.lib.recfunctions as rf
    from GHOST_standards import get_standard_data, get_standard_metadata, standard_temporal_resolutions, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code, standard_classification_flag_name_to_classification_code

    #get station reference
    station_reference = metadata['station_reference'][0]

    #get metadata updata stamp variable
    update_metadata_stamp = metadata['meta_update_stamp']
    
    #dictionary to map python types to netcdf types
    type_map = {np.uint32:'u4', np.object:str, np.float32:'f4', np.float64:'f8'}

    #some metadata variables make no sense when taken as a modal value across the month (i.e. timestamps)
    #therefore do write/process variables of this type
    #remove from metadata record array
    metadata = rf.drop_fields(metadata, ['meta_update_stamp','data_download_stamp'])

    #get metadata/data keys
    metadata_keys = list(metadata.dtype.names)
    data_keys =     list(data.dtype.names)

    #get start and end datetimes of raw data
    start_dt = fractional_minutes_from_reference_to_datetime(float(start_stamp))
    end_dt = fractional_minutes_from_reference_to_datetime(float(end_stamp))

    #get data start_year/start_month and end_year/end month
    data_start_year = start_dt.year
    data_start_month = start_dt.month
    data_end_year = end_dt.year
    data_end_month = end_dt.month
    data_end_day = end_dt.day
    data_end_hour = end_dt.hour
    #adjust data end year (if end date is 20150101 0000 for example, the end year should be 2014, not 2015)
    if (data_end_month == 1) & (data_end_day == 1) & (data_end_hour == 0):
        data_end_year = data_end_year - 1
        data_end_month = 12
    #adjust data end month (if end date is 20150301 0000 for example, the end month should be 2, not 3)
    elif (data_end_day == 1) & (data_end_hour == 0):
        data_end_month = data_end_month - 1

    #create list of years to process from provided start_date and end_date
    desired_start_year = int(run_details['start_year'])
    desired_end_year = int(run_details['end_year'])
    desired_start_month = int(run_details['start_month'])
    desired_end_month = int(run_details['end_month'])
    process_years = np.arange(desired_start_year, desired_end_year+1)

    #cut list of years to process for years which are >= data_start_year and <= data_end_year
    process_years = process_years[process_years >= data_start_year]
    process_years = process_years[process_years <= data_end_year]

    #loop through years and months to process, and write to monthly station specific .npz files
    #iterate through process_years
    for year in process_years:

        #create list of months to process

        #if year == data start year
        if year == desired_start_year:
            process_months = np.arange(desired_start_month, 13)

        #if year == desired end year
        elif year == desired_end_year:
            process_months = np.arange(1,desired_end_month)

        #full year of months otherwise
        else:
            process_months = np.arange(1,13)

        #if in year that data finishes or ends then cut list of months to process for months which are >= data_start_month and <= data_end_month
        if year == data_start_year:
            process_months = process_months[process_months>=data_start_month]

        elif year == data_end_year:
            process_months = process_months[process_months<=data_end_month]

        #iterate through process_months
        for month in process_months:

            if month < 10:
                month_string = '0'+str(month)
            else:
                month_string = str(month)

            #create yearmonth string
            yearmonth_str = '{}{}'.format(year,month_string)

            #get minutes of start and end of monthly window
            start_dt = datetime.datetime(year,month,1,0,0)
            end_dt = start_dt + relativedelta(months=+1)
            start_window_minutes = datetime_to_fractional_minutes_from_reference(start_dt)
            end_window_minutes = datetime_to_fractional_minutes_from_reference(end_dt)

            #cut data,metadata for monthly period
            #after gridding, data times are guaranteed to align to window edges
            start_data_i = bisect.bisect_left(data['time'], start_window_minutes)
            end_data_i = bisect.bisect_left(data['time'], end_window_minutes)
            cut_data = data[start_data_i:end_data_i]
            #check you have some valid data for the month (i.e. not all NaNs), if not continue to next month
            if all(pd.isnull(cut_data[parameter_details['bsc_parameter_name']])) == True:
                continue

            #cut qa/flags/classification
            cut_qa = qa[start_data_i:end_data_i,:]
            cut_flag = flag[start_data_i:end_data_i,:]
            cut_classification = classification[start_data_i:end_data_i,:]

            #make time array (hours/days/months) from start of month to end of month
            time_array = np.arange(len(cut_data['time']), dtype=cut_data['time'].dtype)
            #overwrite old time array
            cut_data['time'] = time_array

            #meta stamp times are not guaranteed however to align with window edges
            start_meta_i = bisect.bisect(update_metadata_stamp,start_window_minutes)-1
            #if meta starts partially through window, start index will be returned as -1, change this to 0 (as want all from start)
            if start_meta_i < 0:
                start_meta_i = 0
            end_meta_i = bisect.bisect_left(update_metadata_stamp,end_window_minutes)
            cut_metadata = metadata[start_meta_i:end_meta_i]
            cut_update_metadata_stamp = update_metadata_stamp[start_meta_i:end_meta_i]

            #get modal metadata in month
            #if have more than 1 row of metadata in month, at least 1 field has multiple metadata values, take modal value across month for each field (by most minutes occurring)
            if cut_metadata.shape[0] > 1:

                #create numpy dataframe to place modal metadata in month
                #create lists for standardising data formats of the input data dictionary
                standard_dtype = [(key,metadata[key].dtype) for key in metadata_keys]
                #create numpy dataframe to store standardised forms of variables inside data dictionary
                modal_metadata = np.empty(1, dtype=standard_dtype)

                for key in metadata_keys:

                    #metadata values for field
                    metadata_values = cut_metadata[key]

                    #take mode of weighted metadata (by number of minutes that it is relevant for) that does not equal NaN
                    #if are multiple items with same number of instances, the lowest sorted instance will be returned (i.e. 'a' rather 'b'; 1 rather than 2)
                    weighted_meta = {}
                    for xx, metadata_value in enumerate(metadata_values):
                        #last index
                        if (xx+1) == len(metadata_values):
                            minutes_represented = end_window_minutes - cut_update_metadata_stamp[xx]
                        else:
                            minutes_represented = cut_update_metadata_stamp[xx+1] - cut_update_metadata_stamp[xx]

                        if pd.isnull(metadata_value) == False:
                            if metadata_value in list(weighted_meta.keys()):
                                weighted_meta[metadata_value] = weighted_meta[metadata_value]+minutes_represented
                            else:
                                weighted_meta[metadata_value] = minutes_represented

                    #if have no valid metadata, modal value is np.NaN
                    if len(weighted_meta) == 0:
                         modal_value = np.NaN
                    else:
                        counter_instance = Counter(weighted_meta)
                        modal_value = counter_instance.most_common(1)[0][0]

                    #append model value to new metadata frame
                    modal_metadata[key] = modal_value

                modal_metadata = modal_metadata[0]

            #only 1 value in month for each metadata variable, so existing array is model metadata array
            else:
                modal_metadata = cut_metadata[0]
    
            #Create netCDF file for specific yearmonth, and output temporal resolution code
            netCDF_fname = run_details['data_root'] + '{}/{}/{}/{}/temporary/{}/{}/{}.nc'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string,station_reference)
    
            #create netCDF root
            root_grp = Dataset(netCDF_fname, 'w', format="NETCDF4")

            #auto mask arrays
            root_grp.set_auto_mask(True)

            #file contents
            root_grp.title = 'Surface {} data in the {} network in {}-{}.'.format(parameter_details['long_parameter_name'],run_details['network'], year, month_string)
            root_grp.institution = 'Barcelona Supercomputing Center'
            root_grp.source = 'Surface observations'
            root_grp.creator_name = 'Dene R. Bowdalo'
            root_grp.creator_email = 'dene.bowdalo@bsc.es'
            root_grp.conventions = 'CF-1.7'
            root_grp.data_version= '1.0'

            #get time array
            time_array = cut_data['time'][:]

            #get names of all metadata/data variable fields
            metadata_keys = list(modal_metadata.dtype.names)
            data_keys =     list(cut_data.dtype.names)
            #add qa/flag/classification to data_keys
            data_keys.extend(['qa','flag','classification'])    

            #get dicts with standard formats for metadata/data
            metadata_format_dict = get_standard_metadata(parameter_details)
            data_format_dict     = get_standard_data(parameter_details)

            #netcdf dimensions (station as unlimited/record dimension)
            root_grp.createDimension('station', None)
            root_grp.createDimension('time', len(time_array))
            root_grp.createDimension('N_qa_codes', len(list(standard_QA_name_to_QA_code.keys())))
            root_grp.createDimension('N_flag_codes', len(list(standard_data_flag_name_to_data_flag_code.keys())))
            root_grp.createDimension('N_classification_codes', len(list(standard_classification_flag_name_to_classification_code.keys())))

            #set variables for data parameters
            for data_key in data_keys:
                if data_key in ['qa','flag','classification']:
                    current_data_type = 'u1'
                else:
                    #current_data_type = type_map[data[data_key].dtype]
                    current_data_type = type_map[data_format_dict[data_key]['data_type']]

                if data_key == 'time':
                    var = root_grp.createVariable(data_key, current_data_type, ('time',))
                elif data_key == 'qa':
                    var = root_grp.createVariable(data_key, current_data_type, ('station','time','N_qa_codes'), fill_value=255)
                elif data_key == 'flag':
                    var = root_grp.createVariable(data_key, current_data_type, ('station','time','N_flag_codes'), fill_value=255)
                elif data_key == 'classification':
                    var = root_grp.createVariable(data_key, current_data_type, ('station','time','N_classification_codes'), fill_value=255)
                else:
                    var = root_grp.createVariable(data_key, current_data_type, ('station','time'))

                #set variable attributes
                var.standard_name = data_format_dict[data_key]['standard_name']
                var.long_name = data_format_dict[data_key]['long_name']
                var.units = data_format_dict[data_key]['units']
                var.description = data_format_dict[data_key]['description']

                #variable specific attributes
                if data_key == 'time':
                    var.units = '{} since {}-{}-01 00:00:00'.format(temporal_resolution_vars['descriptive_temporal_resolution'],year,month_string)
                    var.description = 'Time in {} since {}-{}-01 00:00 UTC. Time given refers to the start of the time window the measurement is representative of (temporal resolution).'.format(temporal_resolution_vars['descriptive_temporal_resolution'],year,month_string)
                    var.axis = 'T'
                    var.calendar = 'standard'
                    var.tz = 'UTC'

            #set variables for metadata parameters
            for metadata_key in metadata_keys:

                #current_data_type = type_map[metadata[metadata_key].dtype]
                current_data_type = type_map[metadata_format_dict[metadata_key]['data_type']]
                var = root_grp.createVariable(metadata_key, current_data_type, ('station',))

                #set variable attributes
                var.standard_name = metadata_format_dict[metadata_key]['standard_name']
                var.long_name = metadata_format_dict[metadata_key]['long_name']
                var.units = metadata_format_dict[metadata_key]['units']
                var.description = metadata_format_dict[metadata_key]['description']

                #variable specific attributes
                if metadata_key == 'longitude':
                    var.axis = 'X'
                elif metadata_key == 'latitude':
                    var.axis = 'Y'

            #write station data to netCDF
            for data_key in data_keys:
                if data_key == 'time':
                    root_grp[data_key][:] = time_array
                elif data_key == 'qa':
                    root_grp[data_key][0,:,:] = cut_qa
                elif data_key == 'flag':
                    root_grp[data_key][0,:,:] = cut_flag
                elif data_key == 'classification':
                    root_grp[data_key][0,:,:] = cut_classification
                else:
                    root_grp[data_key][0,:] = cut_data[data_key]

            #write station metadata to netCDF
            for metadata_key in metadata_keys:
                #convert float NaNs to str for string metadata fields
                if root_grp[metadata_key].dtype == str:
                    root_grp[metadata_key][0] = str(modal_metadata[metadata_key])
                else:
                    root_grp[metadata_key][0] = modal_metadata[metadata_key]

            #close writing to netCDF 
            root_grp.close()

            #give 770 permissions to file
            os.chmod(netCDF_fname, 0o770)
            
###------------------------------------------------------------------------------------###

def get_relevant_yearmonths(run_details, parameter_details, temporal_resolution_code):

    '''create list of yearmonth strings, for which in those time periods exists valid processed station data, 
    ''' 

    from GHOST_standards import standard_temporal_resolutions

    #get output temporal resolution specific variables
    temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

    #create list for storing relevant yearmonth strings
    yearmonth_strings = []

    #create list of all years data has been processed for
    years = np.arange(int(run_details['start_year']),int(run_details['end_year'])+1)

    #iterate through years
    for year in years:

        #is there less than 1 year of data?
        if len(years) == 1:
            months = np.arange(int(run_details['start_month']), int(run_details['end_month']))

        #else, is it first process year?
        elif year == years[0]:
            months = np.arange(int(run_details['start_month']), 13)

        #else, is it last process year?
        elif year == years[-1]:
            months = np.arange(1,int(run_details['end_month']))

        #full year of months otherwise
        else:
            months = np.arange(1,13)

        #iterate through months
        for month in months:

            if month < 10:
                month_string = '0'+str(month)
                yearmonth_str = str(year)+month_string
            else:
                month_string = str(month)
                yearmonth_str = str(year)+month_string

            #check if have station files in year/month directory
            year_month_files = sorted(glob.glob(run_details['data_root'] + '{}/{}/{}/{}/temporary/{}/{}/*.nc'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'],year,month_string)))

            #if so, then append yearmonth_str to list
            if len(year_month_files) > 0:
                yearmonth_strings.append(yearmonth_str)

    return yearmonth_strings

###------------------------------------------------------------------------------------###

def concatenate_yearmonth_station_netCDFs(slurm_job_ID, network, parameter, yearmonth_str, temporal_resolution_code):

    from GHOST_standards import standard_temporal_resolutions

    #initialise log file string
    log_file_str = ''

    log_file_str += 'CONCATENATING {} STATION NETCDFs\n'.format(yearmonth_str)

    #create log reference code (yearmonth_str + temporal_resolution_code) 
    log_reference = '{}_{}'.format(yearmonth_str, temporal_resolution_code)

    #Get dictionaries holding run specific variables, parameter specific variables and network specific variables
    run_details, parameter_details, network_details = get_run_dictionaries(slurm_job_ID, network, parameter)
    #time start of stage processing
    run_details['stage_process_start_time'] = time.time()

    #get output temporal resolution specific variables
    temporal_resolution_vars = standard_temporal_resolutions[temporal_resolution_code]

    #get specific year and month
    year = yearmonth_str[:4]
    month = yearmonth_str[-2:]

    #get all relevant station netCDF files in yearmonth
    station_files = sorted(glob.glob(run_details['data_root'] + '{}/{}/{}/{}/temporary/{}/{}/*.nc'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'], year, month)))

    #create new netCDF file for yearmonth
    netCDF_fname = run_details['data_root'] + '{}/{}/{}/{}/{}_{}.nc'.format(run_details['network'],run_details['ghost_version'],temporal_resolution_vars['temporal_resolution_path'],parameter_details['bsc_parameter_name'],parameter_details['bsc_parameter_name'],yearmonth_str)
    
    #concatenate station netCDFs
    submit_list = ['ncrcat', '-L', '1']
    submit_list.extend(station_files)
    submit_list.append(netCDF_fname)
    concat_process =  subprocess.Popen(submit_list, stdout=subprocess.PIPE)
    concat_status = concat_process.communicate()[0]
    concat_return_code = concat_process.returncode

    #change record (unlimited) station dimensionto be fixed  
    submit_list = ['ncks', '-O', '--fix_rec_dmn', 'station', netCDF_fname, netCDF_fname]
    record_process =  subprocess.Popen(submit_list, stdout=subprocess.PIPE)
    record_status = record_process.communicate()[0]
    record_return_code = record_process.returncode

    #give 770 permissions to file
    os.chmod(netCDF_fname, 0o770)

    #remove station specific netCDFs
    for station_file in station_files:
        os.remove(station_file)

    #return valid process logfile (0)
    log_file_str += str((time.time() - run_details['stage_process_start_time'])/60.)
    create_GHOST_stage_logfile('T2',run_details,0,log_file_str,log_reference)

###------------------------------------------------------------------------------------###