#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#standardisation_mapping_functions.py

#module which defines functions which map network provided metadata/data to GHOST standardised variables 

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import numpy as np

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

###--------------------------------------------------------------------------------------------------###
#DEFINE FUNCTION TO MAP DATA PROVIDER STATION CLASSIFICATIONS TO STANDARD STATION CLASSIFICATIONS
###--------------------------------------------------------------------------------------------------###

def map_network_classifications_to_standard_classifications(native_classifications, site_metadata, run_details, write_to_metadata_array=True):

    '''takes multiple types of reported site classifications from multiple reporting networks and produces standardised classifications across all networks

       STANDARD CLASSIFICATION DEFINITIONS:
       ----------------------------------------------------

       AREA_CLASSIFICATION: Describes type of area a measurement station is situated in
        
       urban: All areas with a level of urban influence are defined as "urban"
              2 subdivisions to separate between urban centres and suburban areas:
       ------centre: Continuously built-up urban area meaning complete (or at least highly predominant) building-up of the street front side by buildings with at least two floors or large detached buildings with at least two floors.
                     With the exception of city parks, the built-up area is not mixed with non-urbanised areas.     
       ------suburban: Largely built-up urban area. ‘Largely built-up’ means contiguous settlement of detached buildings of any size with a building density less than for ‘continuously built-up’ area.
                       The built-up area is mixed with non-urbanised areas (e.g. agricultural, lakes, woods).
                       It must also be noted that ‘suburban’ as defined here has a different meaning than in every day English i.e. ‘an outlying part of a city or town’ suggesting that a suburban area is always associated to an urban area.
                       A suburban area can be suburban on its own without any urban part.
         
       rural: All areas, that do not fulfil the criteria for an urban area are defined as "rural" 
              3 subdivisions to indicate the distance to the nearest built-up urban area
       ------near_city: area within 10 km from the border of an urban or suburban area.
       ------regional: 10-50 km from major sources/source areas.
       ------remote: > 50 km from major sources/source areas.   

       ----------------------------------------------------
       STATION_CLASSIFICATION: Categorises the dominant type of air measured by a station

       background : Located such that its pollution levels are representative of the average exposure of the general population within the type of area under assessment. The pollution level should not be dominated by a single source type (e.g. traffic), unless that source type is typical within the area
                    under assessment. The station should usually be representative of a wider area of at least several square kilometers.

       point_source: Measurements influenced by a major stationary emissions source (i.e power plant),
                     or influenced by traffic, off-road vehicles and engines, rail, marine, or aviation sources.
                     Subdivisions of 'industrial' & 'traffic' given to allow for greater speciation given by EEA_AIRBASE & EEA_AQ_eReporting)
       --------industrial: Located in close proximity to single industrial sources or industrial areas, in a location that should represent the highest concentrations to which the population within the zone are exposed to. 
                           A wide range of industrial sources can be considered here, including: a) thermal power generation, b) district heating plants, c) refineries, d) waste incineration/treatment plants, dump sites, e) mining, including gravel, oil, natural gas, f) airports or g) ports.
       --------traffic: Located in close proximity to a road, in a location that should represent the highest concentrations to which the population are exposed to within the zone.
                        Located such that the pollution level for the specific pollutant is determined predominantly by the emissions from road traffic (i.e. not ship, railway, airplane, off-road) on distinct major roads.

       ----------------------------------------------------
       MAIN_EMISSION_SOURCE: Describes the main source of the emissions influencing air at a station.   
                              **Note - this describes the main source of emissions, but does not necessarily mean these emissions dominate the type of air measured at a station (given by STATION_CLASSIFICATION). 

       agriculture: emissions associated with agriculture

       commercial_and_residential_combustion: emissions associated with commercial and residential combustion

       extraction_of_fossil_fuels: emissions associated with the extraction of fossil fuels 

       industrial_combustion: emissions associated with industrial combustion

       natural: emissions from natural sources (e.g terpenes from trees)

       other_mobile_sources_and_machinery: emissions from all other mobile sources than traffic, and from off-road vehicles and engines 

       production_processes: emissions from processes associated with production and assembly

       power_production: emissions from processes associated with the generation of power

       road_transport: emissions associated with emissions from traffic

       solvents: emissions associated with use of solvents

       waste_treatment_and_disposal: emissions associated with waste treatment and disposal

       ----------------------------------------------------
       LAND_USE: Dominant land use in the area of the reporting station 
                 categories broadly follow IGBP land cover classification system, followed by MODIS/VIIRS land use maps
                 source: http://www.eomf.ou.edu/static/IGBP.pdf

       barren: Lands with exposed soil, sand or rocks and never have more than 10% vegetated cover during any time of the year.
       --------beach: landform alongside a body of water which consists of loose particles. The particles composing a beach are typically made from rock, such as sand, gravel, shingle, pebbles, or cobblestones.
       --------desert: a barren area of landscape where little precipitation occurs and consequently living conditions are hostile for plant and animal life. 
                       The lack of vegetation exposes the unprotected surface of the ground to the processes of denudation.
       --------rock: characterised by areas of bedrock exposure, scarps, talus, slides, volcanic material, rock glaciers, and other accumulations of rock without vegetative cove 
       --------soil: thin soil, without vegetation

       forest: Lands dominated by woody vegetation or trees, with a percent cover >60% and height exceeding 2 m. 
               Includes all evergreen needleleaf, evergreen broadleaf, deciduous needleleaf, deciduous broadleaf types

       open: Lands with herbaceous, other understory systems or woody vegetation less than 2 m tall
       --------grassland: Lands with herbaceous types of cover. Tree and shrub cover is less than 10%.
       --------savanna: Lands with herbaceous and other understory systems, and with forest canopy cover between 10% and 60%. The forest cover height exceeds 2 m.
       --------shrubland: Lands with woody vegetation less than 2 m tall and with shrub canopy cover > 10%. The shrub foliage can be either evergreen or deciduous.

       snow: Lands under snow/ice cover throughout the year.

       urban: Land covered by buildings and other man-made structures.
       --------agricultural: Lands covered with temporary crops followed by harvest and a bare soil period (e.g., single and multiple cropping systems).
                             Also includes lands used for farming and raising of livestock
                             It is important to note that also agricultural is classed under urban, this class also includes naturally occuring crops.
                             **Sites with agricultural land_use can have an AREA_CLASSIFICATION of 'rural'. 
                             **They are defined as 'urban' under LAND_USE as they would not be there without human activity, but can be located extremely far from urban areas, with no major emission sources nearby (with the exception of pesticides)
       --------blighted: A blighted area is defined as: by reason of deterioration, faulty planning, inadequate or improper facilities, deleterious land use or the existence of unsafe structures, 
                          or any combination of these factors, are detrimental to the safety, health or welfare of the community.
       --------commercial: A commercial area is real estate intended for use by for-profit businesses, such as office complexes, shopping malls, service stations and restaurants. 
                            It may be purchased outright by a developer for future projects or leased through a real estate broker.
       --------industrial: Industrial land use is the use of a piece of land for industrial purposes, eg. manufacturing.
       --------military: Land used for solely military purposes
       --------park: A large public garden or area of land used for recreation.
       --------residential: A residential area is a land used in which housing predominates, as opposed to industrial and commercial areas. Housing may vary significantly between, and through, residential areas.
                             These include single-family housing, multi-family residential, or mobile homes.
       --------transportation: All types of land use used for human transportation. This includes airports, roads, railway lines and shipping ports.
       **possibly some overlap between between 'industrial' and 'transportation' classes, as some networks may class things such as airports as industrial (e.g EEA_AIRBASE), and others not

       water: Oceans, seas, lakes, reservoirs, and rivers. Can be either fresh or salt- water bodies.

       wetland: Lands with a permanent mixture of water and herbaceous or woody vegetation. The vegetation can be present either in salt, brackish, or fresh water.


       ----------------------------------------------------
       TERRAIN: Dominant terrain in the area of the reporting station 

       coastal: A coastal area is where land meets the sea or ocean.

       complex: A region having irregular topography (not including mountains or coastal). 
                 Complex terrain can include variations in land use, such as urban, irrigated, and unirrigated.

       flat: Open terrain, open country or open ground is terrain which is mostly flat and free of obstructions such as trees and buildings. 
             Examples include farmland, grassland and specially cleared areas such as an airport.

       mountain: A mountain is a large landform that stretches above the surrounding land in a limited area, usually in the form of a peak.

       rolling: Terrain where the natural slopes consistently rise and fall across a horizontal plane 


       ----------------------------------------------------
       MEASUREMENT_SCALE: A denotation of the geographic scope of the air quality measurements made by the monitor. 
                          The implication is that the same measurement made elsewhere within the measurement scale 
                          would produce an equivalent result to that produced at the monitoring site
                          **Neighbourhood and City scales have potential to overlap in applications that concern secondarily formed or homogeneously distributed air pollutants
                    
       micro: representative for: several metres - ~100m

       middle: representative for: ~100m - 0.5km, several city blocks

       neighbourhood: representative for: 0.5km - 4km, some extended area of city that has relatively uniform land use

       city: representative for: 4km - 50km, city like dimensions. 

       regional: representative for: 50km - 100s km, rural area of reasonably homogeneous geography, without large sources
    '''

    #iterate through values in native_classifications dictionary and make sure all values are lower-cased and all white spaces are stripped.
    for key in list(native_classifications.keys()):
        native_classifications[key] = native_classifications[key].replace(" ", "").lower()

#------------------------------------------------
#NOW MAP INDIVIDUAL NATIVE NETWORK CLASSIFICATIONS TO STANDARD CLASSIFICATIONS 
#------------------------------------------------
#CANADA_NAPS

#CANADA_NAPS NATIVE CLASSIFICATIONS = STATION_CLASSIFICATION, LAND_USE, MEASUREMENT_SCALE

#STATION_CLASSIFICATION = pe (General Population Exposure), rb (Regional Background), t (Transportation Source Influenced), ps (Point Source Influenced)
#LAND_USE = a (Agricultural), c (Commercial), f (Forested), i (Industrial), o (Open), p (Parks), r (Residential), w (Water)
#MEASUREMENT_SCALE = mi (Micro), md (Middle), ne (Neighbourhood), sr (Sub-Regional), r (Regional)

    if run_details['network'] == 'CANADA_NAPS':
    
        #AREA CLASSIFICATION - not given directly, but derive it from station_classification
                              #all 'General Population Exposure', 'Transportation Source Influenced' and 'Point Source Influenced' classes defined as being within 'urban' area
                              #'Regional background' is defined as being outside 'urban' area --> 'rural' 
        
        mapping_area_classification = {'pe':'urban','t':'urban','ps':'urban','rb':'rural','unknown':'','na':'','':''}
        area_classification = mapping_area_classification[native_classifications['station_classification']]
        
        #STATION CLASSIFICATION
        mapping_station_classification = {'pe':'background','rb':'background','t':'point_source','ps':'point_source','unknown':'','na':'','':''}
        station_classification = mapping_station_classification[native_classifications['station_classification']]

        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''

        #LAND USE
        mapping_land_use = {'a':'urban','c':'urban','f':'forest','i':'urban','o':'open','p':'urban','r':'urban','w':'water','unknown':'','na':'','':''}
        mapping_land_use_urban_subdivision = {'a':'agricultural','c':'commercial','i':'industrial','p':'park','r':'residential'}
        
        land_use = mapping_land_use[native_classifications['land_use']]
        
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
            
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE 
        mapping_measurement_scale = {'mi':'micro','md':'middle','ne':'neighbourhood','sr':'city','r':'regional','unknown':'','na':'','':''}
        measurement_scale = mapping_measurement_scale[native_classifications['measurement_scale']]

#------------------------------------------------
#CAPMoN

#CAPMoN NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = rural
    
    elif run_details['network'] == 'CAPMoN':
    
        #AREA CLASSIFICATION
        if native_classifications['area_classification'] == 'rural':
            area_classification = 'rural'
        else:
            area_classification = ''    
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
    
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
            
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EANET

#EANET NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = urban, rural, remote
    
    elif run_details['network'] == 'EANET':
    
        #AREA CLASSIFICATION - Enough detail reported in data manuals to subdivide rural classes
                               #'rural' class defined as more than 20km from large emission sources --> 'regional' 
                               #'remote' class defined as more than 50km from large emission sources --> 'remote' 
                               
        mapping_area_classification = {'urban':'urban','rural':'rural','remote':'rural','unknown':'','':''}
        mapping_area_classification_rural_subdivision = {'rural':'regional','remote':'remote'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision)  
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume representativeness of 'urban' sites' 
        
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EBAS

#EMEP NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, LAND_USE

#AREA_CLASSIFICATION = mountain, polar, coastal, rural, suburban, urbanandcentercity, other
#LAND_USE = agricultural, airport, blightedarea, commercial, desert, forest, grassland,
#gravelandstone, industrial, militaryreservation, remotepark, residential, singlepointsourcearea,
#snowfield, urbanpark, wetland, other

    elif run_details['network'] == 'EBAS':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'mountain':'rural','polar':'rural','coastal':'rural','rural':'rural','suburban':'urban','urbanandcentercity':'urban','other':'','notavailable':'','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urbanandcentercity':'centre','suburban':'suburban'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision)     


        #STATION CLASSIFICATION - Derive this class from mixture of AREA CLASSIFICATION AND LAND_USE
                                #Where land use is 'singlepointsourcearea', the station is classed as 'point_source'
                                #Where land use is 'airport', the station is classed as 'point_source-industrial'
                                #Otherwise, if area_classification is 'rural', then set class as 'background'
        
        #if land use is 'singlepointsourcearea' then classify station as 'point_source'
        if (native_classifications['land_use'] == 'singlepointsourcearea'):
            station_classification = 'point_source'
         #if land use is 'airport' then classify as 'point_source-industrial'
        elif (native_classifications['land_use'] == 'airport'):
            station_classification = 'point_source-industrial'
        #otherwise if area classification is 'rural', make assumption the station is a 'background' site
        elif area_classification == 'rural':
            station_classification = 'background'
        #can't be sure of station classification - set as blank
        else:
            station_classification = ''
            
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
            
        #LAND USE - Only use 'grassland' to classify land_use as 'open', know from experience it is used to describe both 'shrublands' and 'grasslands'
        mapping_land_use = {'agricultural':'urban','airport':'urban','blightedarea':'urban','commercial':'urban','desert':'barren','forest':'forest','grassland':'open','gravelandstone':'barren','industrial':'urban','militaryreservation':'urban','remotepark':'urban','residential':'urban','singlepointsourcearea':'urban','snowfield':'snow','urbanpark':'urban','wetland':'wetland','other':'','notavailable':'','unknown':'','':''}
        mapping_land_use_barren_subdivision = {'desert':'desert','gravelandstone':'rock'}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','airport':'transportation','blightedarea':'blighted','commercial':'commercial','industrial':'industrial','militaryreservation':'military','remotepark':'park','residential':'residential','singlepointsourcearea':'','urbanpark':'park'}

        land_use = mapping_land_use[native_classifications['land_use']]
        
        if land_use == 'barren':
            land_use_barren_subdivision = mapping_land_use_barren_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            #take care not to needlessly append empty string associated with 'singlepointsourcearea', too vague to give specific land use
            if land_use_urban_subdivision != '':
                land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
    
        #TERRAIN 
        terrain = ''
        if native_classifications['area_classification'] == 'mountain':
            terrain = 'mountain'
        elif native_classifications['area_classification'] == 'coastal':
            terrain = 'coastal'
        
        #if terrain is empty string still, if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if terrain == '':    
            if ('urban' in area_classification):
                terrain = 'complex'

        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EEA_AIRBASE

#EEA_AIRBASE NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, STATION_CLASSIFICATION, MAIN_EMISSION_SOURCE, LAND_USE, AREA_CLASSIFICATION_RURAL_SUBDIVISION, REPRESENTATIVE_RADIUS

#AREA_CLASSIFICATION = suburban, urban, rural 
#STATION_CLASSIFICATION = background, industrial, traffic
#MAIN_EMISSION_SOURCE = agriculture, "commercial, institutional and residential combustion", "extraction and distribution of fossil fuels", "industrial combustion", nature, other, "other mobile sources and machinery (specify)", "production processes", "public power, co-generation and district heating", "road transport", "use of solvents", "waste treatment and disposal"
#LAND_USE = agricultural, commercial, industrial, natural, residential
#AREA_CLASSIFICATION_RURAL_SUBDIVISION = nearcity, remote, regional
#REPRESENTATIVE_RADIUS = representative radius of measurements in kilometres

    elif run_details['network'] == 'EEA_AIRBASE':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        mapping_area_classification_rural_subdivision = {'nearcity':'near_city','regional':'regional','remote':'remote'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification_rural_subdivision']]
            area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
        

        #STATION CLASSIFICATION
        mapping_station_classification = {'background':'background','industrial':'point_source','traffic':'point_source','unknown':'','':''}
    
        station_classification = mapping_station_classification[native_classifications['station_classification']]
    
        if native_classifications['station_classification'] == 'industrial':
            station_classification_point_source_subdivision = 'industrial'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision) 
        if native_classifications['station_classification'] == 'traffic':
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision) 
    
        #MAIN_EMISSION_SOURCE
        mapping_main_emission_source_classification = {'agriculture':'agriculture', 'commercial,institutionalandresidentialcombustion':'commercial_and_residential_combustion', 'extractionanddistributionoffossilfuels':'extraction_of_fossil_fuels', 'industrialcombustion':'industrial_combustion', 'nature':'natural', 'other':'', 'othermobilesourcesandmachinery(specify)':'other_mobile_sources_and_machinery', 'productionprocesses':'production_processes', 'publicpower,co-generationanddistrictheating':'power_production', 'roadtransport':'road_transport', 'useofsolvents':'solvents', 'wastetreatmentanddisposal':'waste_treatment_and_disposal', 'unknown':'', '':''}
        main_emission_source = mapping_main_emission_source_classification[native_classifications['main_emission_source']]
    
    
        #LAND USE - mixed land uses are sometimes given - process all into one standardised string separated by '/'
                   #'natural' land use is too vague to map to a land use (without given details) - do not use
        
        mapping_land_use = {'agricultural':'urban','commercial':'urban','industrial':'urban','residential':'urban','natural':'','mountain':'barren-rock','unknown':'','':''}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','commercial':'commercial','industrial':'industrial','residential':'residential'}
    
        native_land_use = native_classifications['land_use'].split('/')
    
        land_use = ''
        if native_land_use != ['']:
            for nlu in native_land_use:
                if land_use != '':
                    land_use = land_use+'/'
                iter_land_use = mapping_land_use[nlu]
                if iter_land_use == 'urban':
                    iter_land_use_urban_subdivision = mapping_land_use_urban_subdivision[nlu]
                    iter_land_use = '{}-{}'.format(iter_land_use,iter_land_use_urban_subdivision)
                land_use = land_use + iter_land_use
   
        #TERRAIN - if the land_use is 'barren-rock' this is because the raw provided land use is mountain, therefore give terrain this classification
        if land_use == 'barren-rock':
            terrain = 'mountain'
        #else if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        elif ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
          
        #MEASUREMENT SCALE - use REPRESENTATIVE_RADIUS to determine standardised measurement scale
        if (native_classifications['representative_radius'] != '') & (native_classifications['representative_radius'] != 'unknown'):
            if float(native_classifications['representative_radius']) < 0.1:
                measurement_scale = 'micro'
            elif (float(native_classifications['representative_radius']) >= 0.1) & (float(native_classifications['representative_radius']) < 0.5):
                measurement_scale = 'middle'
            elif (float(native_classifications['representative_radius']) >= 0.5) & (float(native_classifications['representative_radius']) < 4.0):
                measurement_scale = 'neighbourhood'
            elif (float(native_classifications['representative_radiuss']) >= 4.0) & (float(native_classifications['representative_radius']) < 50.0):
                measurement_scale = 'city'
            elif float(native_classifications['representative_radius']) >= 50.0:
                measurement_scale = 'regional'
        else:
            measurement_scale = ''

#------------------------------------------------
#EEA_AQ_eReporting

#EEA_AQ_eReporting NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION (EEA_AQ_eReporting OR EEA_AIRBASE), STATION_CLASSIFICATION (EEA_AQ_eReporting OR EEA_AIRBASE), MAIN_EMISSION_SOURCE (EEA_AIRBASE), LAND_USE (EEA_AIRBASE), AREA_CLASSIFICATION_RURAL_SUBDIVISION (EEA_AIRBASE), REPRESENTATIVE_RADIUS (EEA_AIRBASE)

#AREA_CLASSIFICATION = suburban, urban, rural, rural-nearcity, rural-regional, rural-remote 
#STATION_CLASSIFICATION = background, industrial, traffic
#MAIN_EMISSION_SOURCE = agriculture, "commercial, institutional and residential combustion", "extraction and distribution of fossil fuels", "industrial combustion", nature, other, "other mobile sources and machinery (specify)", "production processes", "public power, co-generation and district heating", "road transport", "use of solvents", "waste treatment and disposal"
#LAND_USE = agricultural, commercial, industrial, natural, residential
#AREA_CLASSIFICATION_RURAL_SUBDIVISION = nearcity, remote, regional
#REPRESENTATIVE_RADIUS = representative radius of measurements in kilometres

    elif run_details['network'] == 'EEA_AQ_eReporting':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','rural-nearcity':'rural','rural-near_city':'rural','rural-regional':'rural','rural-remote':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        mapping_area_classification_rural_subdivision = {'rural-nearcity':'near_city','rural-near_city':'near_city','rural-regional':'regional','rural-remote':'remote','rural':''}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification']]
            if area_classification_rural_subdivision != '':
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
            #if do not have rural subdivisions, then try use AIRBASE only available AREA_CLASSIFICATION_RURAL_SUBDIVISION
            #check AREA_CLASSIFICATION_RURAL_SUBDIVISION not empty string or 'unknown'
            elif (native_classifications['area_classification_rural_subdivision'] != '') & (native_classifications['area_classification_rural_subdivision'] != 'unknown'):
                mapping_area_classification_rural_subdivision = {'nearcity':'near_city','regional':'regional','remote':'remote'}
                area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification_rural_subdivision']]
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
        
        #STATION CLASSIFICATION
        mapping_station_classification = {'background':'background','industrial':'point_source','traffic':'point_source','unknown':'','':''}
    
        station_classification = mapping_station_classification[native_classifications['station_classification']]
    
        if native_classifications['station_classification'] == 'traffic':
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)
            
        if native_classifications['station_classification'] == 'industrial':
            station_classification_point_source_subdivision = 'industrial'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)  

        #MAIN_EMISSION_SOURCE
        mapping_main_emission_source_classification = {'agriculture':'agriculture', 'commercial,institutionalandresidentialcombustion':'commercial_and_residential_combustion', 'extractionanddistributionoffossilfuels':'extraction_of_fossil_fuels', 'industrialcombustion':'industrial_combustion', 'nature':'natural', 'other':'', 'othermobilesourcesandmachinery(specify)':'other_mobile_sources_and_machinery', 'productionprocesses':'production_processes', 'publicpower,co-generationanddistrictheating':'power_production', 'roadtransport':'road_transport', 'useofsolvents':'solvents', 'wastetreatmentanddisposal':'waste_treatment_and_disposal', 'unknown':'', '':''}
        main_emission_source = mapping_main_emission_source_classification[native_classifications['main_emission_source']]
    
    
        #LAND USE - mixed land uses are sometimes given - process all into one standardised string separated by '/'
                   #'natural' land use is too vague to map to a land use (without given details) - do not use
        
        mapping_land_use = {'agricultural':'urban','commercial':'urban','industrial':'urban','residential':'urban','natural':'','mountain':'barren-rock','unknown':'','':''}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','commercial':'commercial','industrial':'industrial','residential':'residential'}
    
        native_land_use = native_classifications['land_use'].split('/')
    
        land_use = ''
        if native_land_use != ['']:
            for nlu in native_land_use:
                if land_use != '':
                    land_use = land_use+'/'
                iter_land_use = mapping_land_use[nlu]
                if iter_land_use == 'urban':
                    iter_land_use_urban_subdivision = mapping_land_use_urban_subdivision[nlu]
                    iter_land_use = '{}-{}'.format(iter_land_use,iter_land_use_urban_subdivision)
                land_use = land_use + iter_land_use

        #TERRAIN - if the land_use is 'barren-rock' this is because the raw provided land use is mountain, therefore give terrain this classification
        if land_use == 'barren-rock':
            terrain = 'mountain'
        #else if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        elif ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''

        #MEASUREMENT SCALE - use REPRESENTATIVE_RADIUS to determine standardised measurement scale
        if (native_classifications['representative_radius'] != '') & (native_classifications['representative_radius'] != 'unknown'):
            if float(native_classifications['representative_radius']) < 0.1:
                measurement_scale = 'micro'
            elif (float(native_classifications['representative_radius']) >= 0.1) & (float(native_classifications['representative_radius']) < 0.5):
                measurement_scale = 'middle'
            elif (float(native_classifications['representative_radius']) >= 0.5) & (float(native_classifications['representative_radius']) < 4.0):
                measurement_scale = 'neighbourhood'
            elif (float(native_classifications['representative_radius']) >= 4.0) & (float(native_classifications['representative_radius']) < 50.0):
                measurement_scale = 'city'
            elif float(native_classifications['representative_radius']) >= 50.0:
                measurement_scale = 'regional'
        else:
            measurement_scale = ''

#------------------------------------------------
#NCDC_ISD

#SEARCH NATIVE CLASSIFICATIONS = NONE
    
    elif run_details['network'] == 'NCDC_ISD':
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''
    
#------------------------------------------------
#SEARCH

#SEARCH NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = urban, suburban, rural
    
    elif run_details['network'] == 'SEARCH':
    
        #AREA CLASSIFICATION
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume background nature of 'urban' sites' 
        
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#US_EPA_AQS

#US_EPA_AQS NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, LAND_USE, MEASUREMENT_SCALE

#AREA_CLASSIFICATION = urbanandcentercity, suburban, rural
#LAND_USE = agricultural, blightedareas, commercial, desert, forest, industrial, militaryreservation, mobile, residential
#MEASUREMENT_SCALE = microscale, middlescale, neighborhood, urbanscale, regionalscale 

    elif run_details['network'] == 'US_EPA_AQS':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urbanandcentercity':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urbanandcentercity':'centre','suburban':'suburban'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 


        #STATION CLASSIFICATION - no direct category that gives this, but can derive it
                                 #where measurement_scale is 'micro' or 'middle', assume this point source dominated (<0.5km)
                                 #other larger scales (>0.5km) are assumed to be representative of background  
    
        mapping_station_classification = {'microscale':'point_source','middlescale':'point_source','neighborhood':'background','urbanscale':'background','regionalscale':'background','unknown':'','':''}
        station_classification = mapping_station_classification[native_classifications['measurement_scale']]
    
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
    
        #LAND USE
        mapping_land_use = {'agricultural':'urban','blightedareas':'urban','commercial':'urban','desert':'barren','forest':'forest','industrial':'urban','militaryreservation':'urban','mobile':'urban','residential':'urban','unknown':'','':''}
        mapping_land_use_barren_subdivision = {'desert':'desert'}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','blightedareas':'blighted','commercial':'commercial','industrial':'industrial','militaryreservation':'military','mobile':'transportation','residential':'residential'}

        land_use = mapping_land_use[native_classifications['land_use']]
    
        if land_use == 'barren':
            land_use_barren_subdivision = mapping_land_use_barren_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
    
        #MEASUREMENT SCALE 
        mapping_measurement_scale = {'microscale':'micro','middlescale':'middle','neighborhood':'neighbourhood','urbanscale':'city','regionalscale':'regional','unknown':'','':''}
        measurement_scale = mapping_measurement_scale[native_classifications['measurement_scale']]

#------------------------------------------------
#US_EPA_CASTNET 

#US_EPA_CASTNET NATIVE CLASSIFICATIONS = LAND_USE, TERRAIN

#LAND_USE = agric, arid, coastal, coastal/marsh, desert, forest, forest/arid, forest/marsh, forest/tropical, marsh, prairie, range, tropical, urban, urban/agric
#TERRAIN = coastal, complex, flat, flat/water, mountaintop, rolling, water

    elif run_details['network'] == 'US_EPA_CASTNET':
    
        #AREA CLASSIFICATION - derive from certain LAND_USE cateogories. 
                              #Specifically the 'urban', 'urban/agric' are classed as 'urban'
                              #The rest are assumed to be 'rural'
    
        mapping_area_classification = {'agric':'rural','arid':'rural','coastal':'rural','coastal/marsh':'rural','desert':'rural','forest':'rural','forest/arid':'rural','forest/marsh':'rural','forest/tropical':'rural','marsh':'rural','prairie':'rural','range':'rural','tropical':'rural','urban':'urban','urban/agric':'urban','unknown':'','':''}
        area_classification = mapping_area_classification[native_classifications['land_use']]
    
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume representativeness of 'urban' sites' 
        
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
            
        #LAND USE - mixed land uses are sometimes given, put these mixed values into just category (unlike done with AIRBASE), as way classifications are reported lends itself to this 
                    #i.e. they are are mixture of biome and land covers, and often these merge naturally into 1 category
                    #make assumption for 'range', that lands grazed by animals are naturally occurring -->, 'open'. Can also be lands grazed by domestic animals --> 'agriculture', but this is not accounted for 
                   
        mapping_land_use = {'agric':'urban','arid':'barren','coastal':'wetland','coastal/marsh':'wetland','desert':'barren','forest':'forest','forest/arid':'forest','forest/marsh':'forest','forest/tropical':'forest','marsh':'wetland','prairie':'open','range':'open','tropical':'forest','urban':'urban','urban/agric':'urban','unknown':'','':''}
        land_use = mapping_land_use[native_classifications['land_use']]
        
        if native_classifications['land_use'] == 'desert':
            land_use_barren_subdivision = 'desert'
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
            
        if native_classifications['land_use'] == 'prairie':
            land_use_open_subdivision = 'grassland'
            land_use = '{}-{}'.format(land_use,land_use_open_subdivision) 
        
        if (native_classifications['land_use'] == 'agric') or (native_classifications['land_use'] == 'urban/agric'):
            land_use_urban_subdivision = 'agricultural'
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
            
        #TERRAIN - first use given 'terrain' classification
        mapping_terrain_classification = {'coastal':'coastal','complex':'complex','flat':'flat','flat/water':'flat','mountaintop':'mountain','rolling':'rolling','water':'','unknown':'','':''}
        terrain = mapping_terrain_classification[native_classifications['terrain']]
        #if terrain is empty string still, if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if terrain == '':
            if ('urban' in area_classification):
                terrain = 'complex'

        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#WMO_GAW 

#WMO GAW NATIVE CLASSIFICATIONS = NONE

    elif run_details['network'] == 'WMO_GAW':
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''
        
#------------------------------------------------
    #OUTPUT STANDARDISED CLASSIFICATION INFORMATION

    #map classification variables to GHOST metadata variable names
    standard_classes = {'standardised_network_provided_area_classification':area_classification,
                        'standardised_network_provided_station_classification':station_classification,
                        'standardised_network_provided_main_emission_source':main_emission_source,
                        'standardised_network_provided_land_use':land_use,
                        'standardised_network_provided_terrain':terrain,
                        'standardised_network_provided_measurement_scale':measurement_scale}

    #iterate through each standard class field
    for class_name in list(standard_classes.keys()):    
        #if any of the standard classifications are empty strings, convert to NaN
        if standard_classes[class_name] == '':
            standard_classes[class_name] = np.NaN

        #add standard classifications to station metadata array if desired (write_to_metadata_array == True)
        if write_to_metadata_array == True:
            site_metadata[class_name]['value'].append(standard_classes[class_name])
            
    #if have written to station metadata array, return the modified array
    if write_to_metadata_array == True:
        return site_metadata 
    #otherwise, return standard classification dictionary
    elif write_to_metadata_array == False:
        return standard_classes

###--------------------------------------------------------------------------------------------------###
#DEFINE DATA PROVIDER DATA FLAGS TO STANDARD DATA FLAG MAPPING
###--------------------------------------------------------------------------------------------------###

#define nested dictionary which defines mapping of data provider flag codes to defined standard data flags, per network
#some definitions:
#standard_data_flag_name = standard data flag name
#standard_data_flag_code = standard data flag code
#network_decreed_validity = do the providing data network deem the data to be valid (V = Valid, I = Invalid) 
#GHOST_decreed_validity = does GHOST deem the data to be valid (V = Valid, I = Invalid)
#network_provided_flag_description = network stated description of raw network provided data flag

network_data_flag_to_standard_data_flag_mapping = {

#--------------------------------------------------------------
#CANADA NAPS
#--------------------------------------------------------------
'CANADA_NAPS':{
    #Basic Flags
    'V0':{'standard_data_flag_name':'Valid Data',      'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid value'},
    'M1':{'standard_data_flag_name':'Missing Data',    'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing value - no value available'},
    'H1':{'standard_data_flag_name':'Un-Flagged Data', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Historical data that have not been assessed or validated (data prior to 2015)'},
    #Extreme/Irregular data
    'V1':{'standard_data_flag_name':'Value Below Detection Limit',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Valid value - below detection limit'},
    'V2':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'extreme or unusual value assessed and considered valid'},
    'M2':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing value - invalidated'}
},

#--------------------------------------------------------------
#CAPMoN
#--------------------------------------------------------------
'CAPMoN':{        
    #Basic Flags
    'V0':{'standard_data_flag_name':'Valid Data',   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid measurement'},
    'M1':{'standard_data_flag_name':'Missing Data', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'No value was reported by the field or laboratory instrument (e.g. instrument failure, lost sample).'},
    #Estimated data
    'V7':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'The reported value is below the analytical or method or instrument detection limit and has been substituted by the detection limit or lowest measurable value as described in the EXPLANATION OF REPORTED DETECTION LIMIT VALUES.'},
    #Extreme/Irregular data
    'V1':{'standard_data_flag_name':'Value Below Detection Limit',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'The reported value is below the analytical or method or instrument detection limit. Values may be positive, zero or negative.'},
    'V2':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'The reported value is unusual (e.g., an outlier) but assessed by the Principal investigator who determined that it was valid or there was insufficient evidence for invalidation. Principal investigators may have included a textual comment to indicate possible causes for the unusual value (e.g. forest fire, volcano, dust storm).'},
    'M2':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'invalidated by Principal Investigator'}
},

#--------------------------------------------------------------   
#EANET
#'EANET':{
#    {}
#},

#--------------------------------------------------------------
#EBAS
#--------------------------------------------------------------
'EBAS':{
    #Basic Flags
    '999':{'standard_data_flag_name':'Missing Data',               'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing measurement, unspecified reason'},
    '899':{'standard_data_flag_name':'Missing Data',               'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Measurement undefined, unspecified reason'},
    '750':{'standard_data_flag_name':'Missing Data',               'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'H+ not measured in alkaline sample'},
    '420':{'standard_data_flag_name':'Preliminary Data',           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Preliminary data'},
    '259':{'standard_data_flag_name':'Invalid Data - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Unspecified error expected'},
    '220':{'standard_data_flag_name':'Preliminary Data',           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Preliminary data'},
    '000':{'standard_data_flag_name':'Valid Data',                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid Measurement'},        
    #Estimated data
    '799':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Measurement missing (unspecified reason), data element contains estimated value'},
    '798':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Measurement missing (unspecified reason), data element contains estimated value. Considered valid.'},
    '797':{'standard_data_flag_name':'Estimated Data - Value Substituted from Secondary Monitor', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data element taken from co-located instrument'},
    '784':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low precipitation, concentration estimated'},
    '782':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low precipitation, concentration estimated'},
    '781':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value below detection limit, data element contains detection limit'},
    '780':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value below detection or quantification limit, data element contains estimated or measured value. Use of flag 147 is encouraged.'},
    '771':{'standard_data_flag_name':'Estimated Data - Value Above Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value above range, data element contains upper range limit'},
    '770':{'standard_data_flag_name':'Estimated Data - Value Above Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value above range, data element contains estimated value'},
    '760':{'standard_data_flag_name':'Estimated Data - Multiple Parameters Aggregated',           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Value estimated by summing up the constituents measured'},
    '655':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Estimated value created by averaging or splitting samples'},
    #Measurement issues
    '741':{'standard_data_flag_name':'Measurement Issue - Unspecified',                       'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':"Non refractory AMS concentrations. Don't include compounds that volatalises above 600 deg C"},
    '699':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Mechanical problem, unspecified reason'},
    '685':{'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to secondary standard gas measurement. Used for Level 0.'},
    '681':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low data capture'},
    '680':{'standard_data_flag_name':'Unmeasured Key Meteorological Parameter',               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Undefined wind direction'},
    '677':{'standard_data_flag_name':'Extreme External Meteorological Conditions',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Icing or hoar frost in the intake'},
    '675':{'standard_data_flag_name':'Unmeasured Key Meteorological Parameter',               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'no visibility data available'},
    '674':{'standard_data_flag_name':'Extreme External Meteorological Conditions',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Icing or hoar frost in the intake, considered valid'},
    '670':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Incomplete data acquisition for multi-component data sets'},
    '669':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Moist or wet filter, invalid'},
    '668':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Moist or wet filter, valid'},
    '666':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter damaged, invalid'},
    '665':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Filter damaged, valid'},
    '664':{'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Instrument flow(s) too far off target value, considered invalid'},
    '663':{'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Too high sampling flow, data considered invalid'},
    '662':{'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Too high sampling flow, data considered valid'},
    '660':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unspecified sampling anomaly, considered valid'},
    '659':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Unspecified sampling anomaly'},
    '658':{'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Too small air volume'},
    '656':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wet-only collector failure, operated as bulk collector'},
    '654':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sampling period longer than normal, considered representative. Observed values reported'},
    '653':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sampling period shorter than normal, considered representative. Observed values reported'},
    '650':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Precipitation collector failure'},
    '649':{'standard_data_flag_name':'Power Failure',                                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Temporary power fail has affected sampler operation'},
    '644':{'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low instrument precision and/or calibration issues'},
    '641':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Aerosol filters installed incorrectly'},
    '640':{'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Instrument internal relative humidity above 40%'},
    '635':{'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Internal temperatures too far off target value, considered invalid'},
    '632':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Lid of polyurethane foam (PUF) sampler not closed'},
    '63O':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'POP concentration from the polyurethane foam (PUF) only'},
    '549':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Impure chemicals'},
    '541':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Gold trap passiviated by unknown compound'},
    '540':{'standard_data_flag_name':'Measurement Issue - Unspecified',                       'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Spectral interference in laboratory analysis'},
    '534':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Wrong coated denuder used'},
    '533':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filters mixed up; incorrect analysis'},
    '532':{'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data less accurate than normal due to high field blank value'},
    '531':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low recovery, analysis inaccurate'},
    '530':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to too low or too high recovery'},
    '521':{'standard_data_flag_name':'Extreme Sample Transport Conditions',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Bactericide was added to sample for storage under warm climate. Considered valid'},
    '498':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Gold trap inconsistency in mercury monitor'},
    '478':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to inconsistency discovered through ion balance calculations'},
    '477':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to inconsistency between measured and estimated conductivity'},
    '476':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistency discovered through ion balance calculations, but considered valid'},
    '475':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistency between measured and estimated conductivity, but considered valid'},
    '298':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Gold trap inconsistency in mercury monitor'},
    '278':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to inconsistency discovered through ion balance calculations'},
    '277':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to inconsistency between measured and estimated conductivity'},
    '276':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistency discovered through ion balance calculations, but considered valid'},
    '275':{'standard_data_flag_name':'Chemical Issue',                                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistency between measured and estimated conductivity, but considered valid'},
    '190':{'standard_data_flag_name':'Measurement Issue - Unspecified',                       'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Not corrected for cross-sensitivity to particle scattering'},
    '101':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Denuder capture efficiency < 75%. Valid measurement'},
    #Operational Maintenance checks
    '980':{'standard_data_flag_name':'Operational Maintenance - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing due to calibration or zero/span check'},
    '684':{'standard_data_flag_name':'Zero Check',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to zero/span check. Used for Level 0.'},
    '683':{'standard_data_flag_name':'Calibration',                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to calibration. Used for Level 0.'},
    '682':{'standard_data_flag_name':'Operational Maintenance - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to calibration or zero/span check. Used for Level 0.'},
    #Local Contamination
    '652':{'standard_data_flag_name':'Construction Contamination',        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Construction/acitivity nearby'},
    '651':{'standard_data_flag_name':'Agricultural Contamination',        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Agricultural activity nearby'},
    '647':{'standard_data_flag_name':'Fire/Wood Burning Contamination',   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fire/wood burning nearby'},
    '646':{'standard_data_flag_name':'Traffic Contamination',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Exceptional traffic nearby'},
    '645':{'standard_data_flag_name':'Traffic Contamination',             'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Exceptional traffic nearby'},
    '599':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unspecified contamination or local influence'},
    '593':{'standard_data_flag_name':'Industrial Contamination',          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Industrial contamination, considered invalid'},
    '591':{'standard_data_flag_name':'Agricultural Contamination',        'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Agricultural contamination, considered invalid'},
    '578':{'standard_data_flag_name':'Sea-Salt Contamination',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Large sea salt contribution (ratio between marine and excess sulphate is larger than 2.0). Used for old data only. For newer data use 451/450.'},
    '568':{'standard_data_flag_name':'Dust Contamination',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Dust contamination, considered invalid'},
    '567':{'standard_data_flag_name':'Insect Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Insect contamination, considered invalid'},
    '566':{'standard_data_flag_name':'Bird-Dropping Contamination',       'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Bird droppings, considered invalid'},
    '565':{'standard_data_flag_name':'Pollen/Leaf Contamination',         'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Pollen and/or leaf contamination, considered invalid'},
    '559':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unspecified contamination or local influence, but considered valid'},
    '558':{'standard_data_flag_name':'Dust Contamination',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Dust contamination, but considered valid'},
    '557':{'standard_data_flag_name':'Insect Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Insect contamination, but considered valid'},
    '556':{'standard_data_flag_name':'Bird-Dropping Contamination',       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Bird droppings, but considered valid'},
    '555':{'standard_data_flag_name':'Pollen/Leaf Contamination',         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Pollen and/or leaf contamination, but considered valid'},
    '460':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Contamination suspected'},
    '451':{'standard_data_flag_name':'Sea-Salt Contamination',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Invalid due to large sea salt contribution'},
    '450':{'standard_data_flag_name':'Sea-Salt Contamination',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Considerable sea salt contribution, but considered valid'},
    '260':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Contamination suspected'},
    '251':{'standard_data_flag_name':'Sea-Salt Contamination',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Invalid due to large sea salt contribution'},
    '250':{'standard_data_flag_name':'Sea-Salt Contamination',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Considerable sea salt contribution, but considered valid'},
    '189':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Possible local contamination indicated by wind from contaminated sector (auto)'},
    '188':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Possible local contamination indicated by low wind speed (auto)'},
    '187':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Possible local contamination indicated by occurence of new particles (auto)'},
    '186':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Possible local contamination indicated by single scattering albedo (auto)'},
    '185':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Possible local contamination indicated by wind direction or velocity'},
    #Extreme/Irregular data
    '990':{'standard_data_flag_name':'Extreme External Meteorological Conditions',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Precipitation not measured due to snow-fall. Needed for historic data, should not be needed for new data'},
    '900':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Hidden and invalidated by data originator'},
    '890':{'standard_data_flag_name':'Unmeasured Key Meteorological Parameter',                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Concentration in precipitation undefined, no precipitation'},
    '783':{'standard_data_flag_name':'Unmeasured Key Meteorological Parameter',                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Low precipitation, concentration unknown'},
    '740':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Probably biased gas/particle ratio'},
    '701':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Less accurate than usual, unspecified reason. (Used only with old data, for new data see groups 6 and 5)'},
    '499':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistent with another unspecified measurement'},
    '471':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Particulate mass concentration higher than parallell mass concentration measurement with higher cut off i.e PM1_mass > PM25_mass and PM25_mass > PM10_mass. Considered invalid.'},
    '470':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Particulate mass concentration higher than parallell mass concentration measurement with higher cut off i.e PM1_mass > PM25_mass and PM25_mass > PM10_mass'},
    '459':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Extreme value, unspecified error'},
    '458':{'standard_data_flag_name':'Value Above Acceptable Range',                                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Extremely high value, outside four times standard deviation in a lognormal distribution'},
    '457':{'standard_data_flag_name':'Value Below Acceptable Range',                                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Extremely low value, outside four times standard deviation in a lognormal distribution'},
    '456':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalidated by data originator'},
    '452':{'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid due to large uncertainty'},
    '440':{'standard_data_flag_name':'Reconstructed/Recalculated Data',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Reconstructed or recalculated data'},
    '299':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Inconsistent with another unspecified measurement'},
    '258':{'standard_data_flag_name':'Value Above Acceptable Range',                                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Extremely high value, outside four times standard deviation in a log-normal distribution'},
    '257':{'standard_data_flag_name':'Value Below Acceptable Range',                                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Extremely low value, outside four times standard deviation in a log-normal'},
    '256':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalidated by database co-ordinator'},
    '211':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Irregular data checked and accepted by database co-ordinator. Valid measurement'},
    '210':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Episode data checked and accepted by database co-ordinator. Valid measurement'},
    '147':{'standard_data_flag_name':'Value Below Detection Limit',                                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Below theoretical detection limit or formal Q/A limit, but a value has been measured and reported and is considered valid'},
    '120':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sample reanalysed with similar results. Valid measurement'},
    '111':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Irregular data checked and accepted by data originator. Valid measurement'},
    '110':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Episode data checked and accepted by data originator. Valid measurement'},        
    '103':{'standard_data_flag_name':'High Variability of Data',                                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'CV of replicate ALPHA samplers > 15 %;. Valid measurement'},
    '102':{'standard_data_flag_name':'High Variability of Data',                                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'CV of replicate diffusion tubes > 30 %. Valid measurement'},
    '100':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Checked by data originator. Valid measurement, overrides any invalid flags'},
    #Exceptional Events
    '679':{'standard_data_flag_name':'Exceptional Event - Unspecified',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unspecified meteorological condition'},
    '678':{'standard_data_flag_name':'Tropical Cyclone (Cyclone/Hurricane/Typhoon)', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Hurricane'},
    '676':{'standard_data_flag_name':'Station Inside Cloud',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'station inside cloud (visibility < 1000 m)'},
    '657':{'standard_data_flag_name':'Heavy Rain/Snowfall Shower (Squall)',          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Precipitation collector overflow. Heavy snowfall/rain shower (squall)'},
    '648':{'standard_data_flag_name':'Heavy Rain/Snowfall Shower (Squall)',          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Snow sampler'},
    '410':{'standard_data_flag_name':'Dust Event',                                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sahara dust event'},
    #Aggregation/Representation issues
    '395':{'standard_data_flag_name':'Data Window Completeness < 90%',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data completeness less than 90%'},
    '394':{'standard_data_flag_name':'Data Window Completeness < 90%',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data completeness less than 90%'},
    '393':{'standard_data_flag_name':'Data Window Completeness < 75%',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data completeness less than 75%'},
    '392':{'standard_data_flag_name':'Data Window Completeness < 75%',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data completeness less than 75%'},
    '391':{'standard_data_flag_name':'Data Window Completeness < 50%',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data completeness less than 50%'},
    '390':{'standard_data_flag_name':'Data Window Completeness < 50%',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data completeness less than 50%'},
    '389':{'standard_data_flag_name':'Data Window Completeness < 66%',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data completeness less than 66%'},
    '388':{'standard_data_flag_name':'Data Window Completeness < 66%',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data completeness less than 66%'},
    '382':{'standard_data_flag_name':'>= 75% of Measurements in Window Below Detection Limit', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'More than 75% of the measurements are below detection limit'},
    '380':{'standard_data_flag_name':'>= 50% of Measurements in Window Below Detection Limit', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'More than 50% of the measurements are below detection limit'},
    '370':{'standard_data_flag_name':'Aggregation/Representation Issue - Unspecified',         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'For monthly averages with samples partly in two months, the number of days in each month is used for weighing the sample'},
    #Data Formatting/Processing issues
    '249':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Apparent typing error corrected. Valid measurement'},
    '248':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Illegal flag was removed by the database co-ordinator. Lost flag information (used for historic data only).'},
    '247':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Overlapping sample interval was corrected by the database co-ordinator. Possible wrong sample time (used for historic data only)'},
    '191':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data not truncation corrected - Valid measurement'}
},

#--------------------------------------------------------------
#EEA AIRBASE
#--------------------------------------------------------------
'EEA_AIRBASE':{     
    #Basic Flags
    '0':{'standard_data_flag_name':'Invalid Data - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid'},
    '1':{'standard_data_flag_name':'Valid Data',                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid'},
},

#--------------------------------------------------------------
#EEA AQ eReporting
#--------------------------------------------------------------
'EEA_AQ_eReporting':{
    #Basic Flags
    '1':{'standard_data_flag_name':'Valid Data',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Primary data considered valid by the data provider for its related reporting purposes. Aggregated data is valid following rules set out in Directive and Guidance.'},
    '-1':{'standard_data_flag_name':'Invalid Data - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'For primary data, data is considered to be invalid due to other circumstances or data is simply missing. For aggregated data, data is not valid because of insufficient data capture following rules set out in Directive and Guidance.'},
    #Estimated data
    '2':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data considered valid by the data provider for its related reporting purposes. However, the value has been measured below the detection limit and the value reported is the detection limit (only used for primary data).'},
    '3':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data considered valid by the data provider for its related reporting purposes. However, the value has been measured below the detection limit and the value reported is half of this detection limit (only used for primary data).'},
    #Operational Maintenance checks
    '-99':{'standard_data_flag_name':'Operational Maintenance - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Data is considered to be invalid due to the regular calibration or the normal maintenance of the instrumentation (only used for primary data).'}       
},

#--------------------------------------------------------------   
# NCDC_ISD
'NCDC_ISD':{
    '0':{'standard_data_flag_name':'Valid Data',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed gross limits check'},
    '1':{'standard_data_flag_name':'Valid Data',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed all quality control checks'},
    '2':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Suspect'},
    '3':{'standard_data_flag_name':'Invalid Data - Unspecified',           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Erroneous'},
    '4':{'standard_data_flag_name':'Valid Data',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed gross limits check, data originate from an NCEI data source'},
    '5':{'standard_data_flag_name':'Valid Data',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed all quality control checks, data originate from an NCEI data source'},
    '6':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Suspect, data originate from an NCEI data source'},
    '7':{'standard_data_flag_name':'Invalid Data - Unspecified',           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Erroneous, data originate from an NCEI data source'},
    '9':{'standard_data_flag_name':'Valid Data',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed gross limits check if element is present'},
    'A':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data value flagged as suspect, but accepted as a good value'},
    'C':{'standard_data_flag_name':'Irregular Data Manually Screened and Accepted', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Temperature and dew point received from Automated Weather Observing System (AWOS) are reported in whole degrees Celsius. Automated QC flags these values, but they are accepted as valid.'},
    'I':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue',    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data value not originally in data, but replaced by validator'},
    'M':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue',    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Manual changes made to value based on information provided by NWS'},
    'P':{'standard_data_flag_name':'Corrected Data Formatting/Processing Issue',    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data value not originally flagged as suspect, but replaced by validator'},
    'R':{'standard_data_flag_name':'Reconstructed/Recalculated Data',               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data value replaced with value computed by NCEI software'},
    'U':{'standard_data_flag_name':'Reconstructed/Recalculated Data',               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data value replaced with edited value'}
},

# NCDC_ISD for relative humidity (as they decided to encode same numbers to different codes)
'NCDC_ISD_CH':{
    '1':{'standard_data_flag_name':'Valid Data',                 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Passed all quality control checks'},
    '3':{'standard_data_flag_name':'Invalid Data - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Failed all quality control checks'},
    '9':{'standard_data_flag_name':'Missing Data',               'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing'}
},

#--------------------------------------------------------------
#SEARCH
#--------------------------------------------------------------
#'SEARCH':{
#    {}
#},

#--------------------------------------------------------------
#US EPA AQS
#--------------------------------------------------------------
'US_EPA_AQS':{
    #Basic Flags
    '':{    'standard_data_flag_name':'Valid Data',       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid value'},
    'T':{   'standard_data_flag_name':'Valid Data',       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Multiple PM2.5 Validity Flags (Not Active)'},
    'XX':{  'standard_data_flag_name':'Preliminary Data', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Experimental Data (Legacy=9950)'},
    '9950':{'standard_data_flag_name':'Preliminary Data', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Experimental Data'},
    '8':{   'standard_data_flag_name':'Un-Flagged Data',  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'QA/QC Unknown (Not Active)'},  
    'AF':{  'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Scheduled but not Collected (Legacy=9972)'},
    '9972':{'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Scheduled but not Collected'},
    'BG':{  'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing ozone data not likely to exceed level of standard (Legacy=9966)'},
    '9966':{'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Missing ozone data not likely to exceed level of standard'},
    'AM ':{ 'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Miscellaneous Void (Legacy=9979)'},
    '9979':{'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Miscellaneous Void'},
    'AU':{  'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Monitoring Waived (Legacy=9987)'},
    '9987':{'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Monitoring Waived'},
    'AQ':{  'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Collection Error (Legacy=9983)'},
    '9983':{'standard_data_flag_name':'Missing Data',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Collection Error'},
    #Estimated data
    'ND':{'standard_data_flag_name':'Estimated Data - No Value Detected',                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'No Value Detected, Zero Reported'},
    '9': {'standard_data_flag_name':'Estimated Data - Negative Value Detected',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Negative value detected - zero reported'},
    'MS':{'standard_data_flag_name':'Estimated Data - Value Below Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value reported is 1/2 MDL substituted'},
    'EH':{'standard_data_flag_name':'Estimated Data - Value Above Detection Limit',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Estimated; Exceeds Upper Range'},
    'SS':{'standard_data_flag_name':'Estimated Data - Value Substituted from Secondary Monitor', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Value substituted from secondary monitor'},
    'DN':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'DNPH peak less than NATTS TAD requirement, reported value should be considered an estimate'},
    'LJ':{'standard_data_flag_name':'Estimated Data - Unspecified',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Identification Of Analyte Is Acceptable; Reported Value Is An Estimate'},
    #Measurement issues
    '3':{   'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Field Issue'},
    '4':{   'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Lab Issue'},
    '7':{   'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Below Lowest Calibration Level'},
    'FX':{  'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Filter Integrity Issue'},
    'AV':{  'standard_data_flag_name':'Power Failure',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Power Failure (Legacy=9988)'},
    '9988':{'standard_data_flag_name':'Power Failure',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Power Failure'},
    'AN':{  'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Machine Malfunction (Legacy=9980)'},
    '9980':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Machine Malfunction'},
    'AJ':{  'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Damage (Legacy=9976)'},
    '9976':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Damage'},
    'AK':{  'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Leak (Legacy=9977)'},
    '9977':{'standard_data_flag_name':'Physical Issue With Filter',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Leak'},
    'AR':{  'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Lab Error (Legacy=9984)'},
    '9984':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Lab Error'},
    'AA':{  'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Pressure out of Limits (Legacy=9967)'},
    '9967':{'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Pressure out of Limits'},
    'HT':{  'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sample pick-up hold time exceeded'},
    'AH':{  'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Flow Rate or CV out of Limits (Legacy=9974)'},
    '9974':{'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Flow Rate or CV out of Limits'},
    'SC':{  'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sampler Contamination (Legacy=9948)'},
    '9948':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sampler Contamination'},
    'QP':{  'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Pressure Sensor Questionable'},
    'QT':{  'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Temperature Sensor Questionable'},
    'W':{   'standard_data_flag_name':'Invalid Flow Rate',                                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Flow Rate Average out of Spec.'},
    'X':{   'standard_data_flag_name':'Extreme Internal Instrument Meteorological Conditions', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Filter Temperature Difference or Average out of Spec.'},
    'BK':{  'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Site computer/data logger down (Legacy=9962)'},
    '9962':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Site computer/data logger down'},
    'AB':{  'standard_data_flag_name':'No Technician',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Technician Unavailable (Legacy=9968)'},
    '9968':{'standard_data_flag_name':'No Technician',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Technician Unavailable'},
    'BB':{  'standard_data_flag_name':'No Technician',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Unable to Reach Site (Legacy=9994)'},
    '9994':{'standard_data_flag_name':'No Technician',                                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Unable to Reach Site'},
    'AE':{  'standard_data_flag_name':'Extreme External Meteorological Conditions',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Shelter Temperature Outside Limits (Legacy=9971)'},
    '9971':{'standard_data_flag_name':'Extreme External Meteorological Conditions',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Shelter Temperature Outside Limits'},
    'TS':{  'standard_data_flag_name':'Extreme Sample Transport Conditions',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Holding Time Or Transport Temperature Is Out Of Specs. (Legacy=9956)'},
    '9956':{'standard_data_flag_name':'Extreme Sample Transport Conditions',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Holding Time Or Transport Temperature Is Out Of Specs.'},
    'BJ':{  'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Operator Error (Legacy=9963)'},
    '9963':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Operator Error'},
    'AG':{  'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Time out of Limits (Legacy=9973)'},
    '9973':{'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Time out of Limits'},
    'AI':{  'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Insufficient Data (cannot calculate) (Legacy=9975)'},
    '9975':{'standard_data_flag_name':'Low Data Capture',                                      'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Insufficient Data (cannot calculate)'},
    'AP':{  'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Vandalism (Legacy=9982)'},
    '9982':{'standard_data_flag_name':'Mechanical Issue/Non-Operational Equipment',            'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Vandalism'},
    'BH':{  'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Interference/co-elution/misidentification (Legacy=9965)'},
    '9965':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Interference/co-elution/misidentification'},
    'MC':{  'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Module End Cap Missing (Legacy=9958)'},
    '9958':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Module End Cap Missing'},
    'BI':{  'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Lost or damaged in transit (Legacy=9964)'},
    '9964':{'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Lost or damaged in transit'},
    'SV':{  'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Volume out of limits'},
    'DI':{  'standard_data_flag_name':'Sample Diluted for Analysis',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sample was diluted for analysis'},
    'FB':{  'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Field Blank Value Above Acceptable Limit'},
    'LB':{  'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Lab blank value above acceptable limit'},
    'TB':{  'standard_data_flag_name':'Operational Maintenance Check Issue',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Trip Blank Value Above Acceptable Limit'},
    'TT':{  'standard_data_flag_name':'Extreme Sample Transport Conditions',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Transport Temperaure is Out of Specs.'},
    'Y':{   'standard_data_flag_name':'Erroneous Sampling Operation',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Elapsed Sample Time out of Spec.'},
    'MX':{  'standard_data_flag_name':'Matrix Effect',                                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Matrix Effect'},
    '2':{   'standard_data_flag_name':'Human Error',                                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Operational Deviation'},
    #Operational Maintenance checks
    'AT':{  'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Calibration (Legacy=9986)'},
    '9986':{'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Calibration'},
    'AY':{  'standard_data_flag_name':'Zero Check',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Q C Control Points (zero/span) (Legacy=9991)'},
    '9991':{'standard_data_flag_name':'Zero Check',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Q C Control Points (zero/span)'},
    'AX':{  'standard_data_flag_name':'Precision Check',         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Precision Check (Legacy=9990)'},
    '9990':{'standard_data_flag_name':'Precision Check',         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Precision Check'},
    'AZ':{  'standard_data_flag_name':'Quality Control Audit',   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Q C Audit (Legacy=9992)'},
    '9992':{'standard_data_flag_name':'Quality Control Audit',   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Q C Audit'},
    'BA':{  'standard_data_flag_name':'Instrumental Repair',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Maintenance/Routine Repairs (Legacy=9993)'},
    '9993':{'standard_data_flag_name':'Instrumental Repair',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Maintenance/Routine Repairs'},
    'BC':{  'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Multi-point Calibration (Legacy=9995)'},
    '9995':{'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Multi-point Calibration'},
    'BD':{  'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Auto Calibration (Legacy=9996)'},
    '9996':{'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Auto Calibration'},
    'BF':{  'standard_data_flag_name':'Precision Check',         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Precision/Zero/Span (Legacy=9998)'},
    '9998':{'standard_data_flag_name':'Precision Check',         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Precision/Zero/Span'},
    'BL':{  'standard_data_flag_name':'Quality Control Audit',   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'QA Audit (Legacy=9961)'},
    '9961':{'standard_data_flag_name':'Quality Control Audit',   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'QA Audit'},
    'BM':{  'standard_data_flag_name':'Accuracy Check',          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Accuracy check (Legacy=9960)'},
    '9960':{'standard_data_flag_name':'Accuracy Check',          'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Accuracy check'},
    'CS':{  'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Laboratory Calibration Standard (Legacy=9953)'},
    '9953':{'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Laboratory Calibration Standard'},
    'ST':{  'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Calibration Verification Standard (Legacy=9954)'},
    '9954':{'standard_data_flag_name':'Calibration',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Calibration Verification Standard'},
    'TC':{  'standard_data_flag_name':'Retention Time Check',    'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Component Check & Retention Time Standard (Legacy=9952)'},
    '9952':{'standard_data_flag_name':'Retention Time Check',    'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Component Check & Retention Time Standard'},
    'FI':{  'standard_data_flag_name':'Instrumental Inspection', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Inspection Flag (Legacy=9959)'},
    '9959':{'standard_data_flag_name':'Instrumental Inspection', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Filter Inspection Flag'},
    'DL':{  'standard_data_flag_name':'Detection Limits Check',  'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Detection Limit Analyses (Legacy=9949)'},
    '9949':{'standard_data_flag_name':'Detection Limits Check',  'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Detection Limit Analyses'},
    'MB':{  'standard_data_flag_name':'Blank Check',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Method Blank (Analytical) (Legacy=9955)'},
    '9955':{'standard_data_flag_name':'Blank Check',             'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Method Blank (Analytical)'},
    'QV':{  'standard_data_flag_name':'Quality Control Audit',   'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Quality Control Multi-point Verification'},
    #Local Contamination
    'RF':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fire - Canadian '},
    'RF':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fire - Canadian'},
    'RG':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fire - Mexico/Central America '},
    'IG':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fire - Mexico/Central America'},
    'RM':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Prescribed Fire'},
    'IM':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Prescribed Fire'},
    'Q':{   'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Prescribed Burning (Not Active)'},
    'RP':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Structural Fire'},
    'IP':{  'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Structural Fire'},
    'F':{   'standard_data_flag_name':'Fire/Wood Burning Contamination',              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Structural Fire (Not Active)'},
    'J':{   'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Construction'},
    'L':{   'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Highway Construction (Not Active)'},
    'P':{   'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Roofing Operations (Not Active)'},
    'D':{   'standard_data_flag_name':'Industrial Contamination',                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sandblasting (Not Active)'},
    'N':{   'standard_data_flag_name':'Industrial Contamination',                     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sanding/Salting of Streets (Not Active)'},
    'RI':{  'standard_data_flag_name':'Pollen/Leaf Contamination',                    'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Pollen Count '},
    'II':{  'standard_data_flag_name':'Pollen/Leaf Contamination',                    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Pollen Count '},
    'G':{   'standard_data_flag_name':'Pollen/Leaf Contamination',                    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Pollen Count (Not Active)'},
    'RR':{  'standard_data_flag_name':'Traffic Contamination',                        'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unique Traffic Disruption'},
    'IR':{  'standard_data_flag_name':'Traffic Contamination',                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unique Traffic Disruption'},
    'I':{   'standard_data_flag_name':'Traffic Contamination',                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Unusual Traffic Congestion (Not Active)'},
    'M':{   'standard_data_flag_name':'Traffic Contamination',                        'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Rerouting of Traffic (Not Active)'},
    'K':{   'standard_data_flag_name':'Agricultural Contamination',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Agricultural Tilling (Not Active)'},
    'AC':{  'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Construction/Repairs in Area (Legacy=9969)'},
    '9969':{'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Construction/Repairs in Area'},
    'NS':{  'standard_data_flag_name':'Local Contamination - Unspecified',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Influenced by nearby source'},
    'BE':{  'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Building/Site Repair (Legacy=9997)'},
    '9997':{'standard_data_flag_name':'Construction Contamination',                   'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Building/Site Repair'},
    'CC':{  'standard_data_flag_name':'Internal Laboratory/Instrument Contamination', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Clean Canister Residue'},
     #Extreme/Irregular data
    'MD':{  'standard_data_flag_name':'Value Below Detection Limit',                                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Value less than MDL'},
    'BR':{  'standard_data_flag_name':'Value Below Acceptable Range',                                 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Value Below Acceptable Range'},
    'PQ':{  'standard_data_flag_name':'Value Close to Detection Limit',                               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Values Between PQL And MDL (Not Active)'},
    'SQ':{  'standard_data_flag_name':'Value Close to Detection Limit',                               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Values Between SQL and MDL '},
    '1':{   'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Deviation from a CFR/Critical Criteria Requirement'},
    '1V':{  'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Data reviewed and validated'},
    '5':{   'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Outlier'},
    'LK':{  'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Analyte Identified; Reported Value May Be Biased High'},
    'LL':{  'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Analyte Identified; Reported Value May Be Biased Low'},
    'QX':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Does not meet QC criteria'},
    '6':{   'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'QAPP Issue'},
    'V':{   'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Validated Value'},
    '1C':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'A 1-Point QC check exceeds acceptance criteria but there is compelling evidence that the analyzer data is valid.'},
    'VB':{  'standard_data_flag_name':'Irregular Data Manually Screened and Accepted',                'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Value below normal; no reason to invalidate'},
    'EC':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Exceeds Critical Criteria'},
    'CL':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Surrogate Recoveries Outside Control Limits'},
    'CB':{  'standard_data_flag_name':'Reconstructed/Recalculated Data',                              'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Values have been Blank Corrected'},
    'SX':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Does Not Meet Siting Criteria'},
    'AS':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Poor Quality Assurance Results (Legacy=9985)'},
    '9985':{'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Poor Quality Assurance Results'},
    'AL':{  'standard_data_flag_name':'Irregular Data Manually Screened and Rejected',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Voided by Operator (Legacy=9978)'},
    '9978':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Voided by Operator'},
    'DA':{  'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Aberrant Data (Corrupt Files, Aberrant Chromatography, Spikes, Shifts) (Legacy=9951)'},
    '9951':{'standard_data_flag_name':'Extreme/Irregular Data - Unspecified',                         'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Aberrant Data (Corrupt Files, Aberrant Chromatography, Spikes, Shifts)'},
    'BN':{  'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Value Exceeds Media Limit (Legacy=9947)'},
    '9947':{'standard_data_flag_name':'Data Does Not Meet Internal Network Quality Control Criteria', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sample Value Exceeds Media Limit'},
    #Exceptional Events
    'RA':{  'standard_data_flag_name':'Dust Event',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'African Dust'},
    'IA':{  'standard_data_flag_name':'Dust Event',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'African Dust'},
    'U':{   'standard_data_flag_name':'Dust Event',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Sahara Dust (Not Active)'},
    'RB':{  'standard_data_flag_name':'Dust Event',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Asian Dust'},
    'IB':{  'standard_data_flag_name':'Dust Event',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Asian Dust'},
    'RC':{  'standard_data_flag_name':'Chemical Spill/Industrial Accident',  'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Chem. Spills & Indust. Accidents'},
    'IC':{  'standard_data_flag_name':'Chemical Spill/Industrial Accident',  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Chem. Spills & Indust Accidents'},
    'H':{   'standard_data_flag_name':'Chemical Spill/Industrial Accident',  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Chem. Spills & Indust. Accidents (Not Active)'},
    'RD':{  'standard_data_flag_name':'Cleanup After a Major Disaster',      'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Cleanup After a Major Disaster'},
    'ID':{  'standard_data_flag_name':'Cleanup After a Major Disaster',      'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Cleanup After a Major Disaster'},
    'R':{   'standard_data_flag_name':'Cleanup After a Major Disaster',      'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Cleanup After a Major Disaster (Not Active)'},
    'RE':{  'standard_data_flag_name':'Demolition',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Demolition'},
    'IE':{  'standard_data_flag_name':'Demolition',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Demolition'},
    'RH':{  'standard_data_flag_name':'Fireworks',                           'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fireworks'},
    'IH':{  'standard_data_flag_name':'Fireworks',                           'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Fireworks'},
    'RJ':{  'standard_data_flag_name':'High Winds',                          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Winds'},
    'IJ':{  'standard_data_flag_name':'High Winds',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Winds'},
    'A':{   'standard_data_flag_name':'High Winds',                          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'High Winds (Not Active)'},
    'RK':{  'standard_data_flag_name':'Infrequent Large Gathering',          'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Infrequent Large Gatherings'},
    'IK':{  'standard_data_flag_name':'Infrequent Large Gathering',          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Infrequent Large Gatherings'},
    'O':{   'standard_data_flag_name':'Infrequent Large Gathering',          'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Infrequent Large Gatherings (Not Active)'},
    'RL':{  'standard_data_flag_name':'Exceptional Event - Unspecified',     'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Other'},
    'IL':{  'standard_data_flag_name':'Exceptional Event - Unspecified',     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Other'},
    'Z':{   'standard_data_flag_name':'Exceptional Event - Unspecified',     'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Other event (Not Active)'},
    'RN':{  'standard_data_flag_name':'Seismic Activity',                    'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Seismic Activity'},
    'IN':{  'standard_data_flag_name':'Seismic Activity',                    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Seismic Activity'},
    'S':{   'standard_data_flag_name':'Seismic Activity',                    'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Seismic Activity (Not Active)'},
    'RO':{  'standard_data_flag_name':'Stratospheric Ozone Intrusion',       'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Stratospheric Ozone Intrusion'},
    'IO':{  'standard_data_flag_name':'Stratospheric Ozone Intrusion',       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Stratospheric Ozone Intrusion'},
    'B':{   'standard_data_flag_name':'Stratospheric Ozone Intrusion',       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Stratospheric Ozone Intrusion (Not Active)'},
    'RQ':{  'standard_data_flag_name':'Terrorist Act',                       'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Terrorist Act'},
    'IQ':{  'standard_data_flag_name':'Terrorist Act',                       'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Terrorist Act'},
    'RS':{  'standard_data_flag_name':'Volcanic Eruptions',                  'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Volcanic Eruptions'},
    'IS':{  'standard_data_flag_name':'Volcanic Eruptions',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Volcanic Eruptions'},
    'C':{   'standard_data_flag_name':'Volcanic Eruptions',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Volcanic Eruption (Not Active)'},
    'RT':{  'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildfire-USA'},
    'IT':{  'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildfire-USA'},
    'E':{   'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Forest Fire (Not Active)'},
    'RU':{  'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildland Fire Use Fire-USA (Not Active)'},
    'IU':{  'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildland Fire Use Fire-USA (Not Active)'},
    'AO':{  'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Bad Weather (Legacy=9981)'},
    '9981':{'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Bad Weather'},
    'AD':{  'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Shelter Storm Damage (Legacy=9970)'},
    '9970':{'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Shelter Storm Damage'},
    'AW':{  'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildlife Damage (Legacy=9989)'},
    '9989':{'standard_data_flag_name':'Wildfire',                            'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Wildlife Damage'},
    'SA':{  'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Storm Approaching (Legacy=9957)'},
    '9957':{'standard_data_flag_name':'Storm',                               'network_decreed_validity':'I', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Storm Approaching'}
},

#--------------------------------------------------------------
#US EPA CASTNET
#--------------------------------------------------------------
'US_EPA_CASTNET':{
    #Basic flags
    '':{'standard_data_flag_name':'Valid Data',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Valid data'}, 
    'M':{'standard_data_flag_name':'Missing Data',               'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Field data missing'},
    'I':{'standard_data_flag_name':'Invalid Data - Unspecified', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid reading'},
    #Measurement issues
    'F':{'standard_data_flag_name':'Power Failure',                'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Field instrument power failure'},
    'K':{'standard_data_flag_name':'Invalid Flow Rate',            'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Non-nominal flow value identified during automated screening'},
    'N':{'standard_data_flag_name':'Erroneous Sampling Operation', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Sampling time inadequate to calculate hourly average'},
    #Operational Maintenance checks
    'C':{'standard_data_flag_name':'Calibration', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Invalid data due to field calibration of instrument'},
    #Local Contamination
    'A':{'standard_data_flag_name':'Local Contamination - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Anomalous value resulting from local anthropogenic activity'},
    #Extreme/Irregular data
    'U':{'standard_data_flag_name':'Value Below Detection Limit',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Concentration below reporting limit for analyzer'},
    'Z':{'standard_data_flag_name':'Value Below Detection Limit',                   'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'< and U flags apply'},
    'P':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Potential problem identified during automated screening requiring further investigation'},
    'R':{'standard_data_flag_name':'Irregular Data Manually Screened and Rejected', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Indicates that the QC check is valid but that associated ambient measurements are not valid and the check should not be submitted as a 1-point QC check to AQS.'},
    '^':{'standard_data_flag_name':'Value Above Acceptable Range',                  'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Alarm high -- ozone > 120'},
    '_':{'standard_data_flag_name':'Negative Value',                                'network_decreed_validity':'V', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Field reading less than zero'},
    'S':{'standard_data_flag_name':'Reconstructed/Recalculated Data',               'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Field instrument out of DQO at time of 2nd calibration'},
    #Exceptional Events
    'E':{'standard_data_flag_name':'Exceptional Event - Unspecified', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Exceptional event designation requested by state agency'},
    #Aggregation/Representation issues
    '<':{'standard_data_flag_name':'Data Window Completeness < 90%', 'network_decreed_validity':'V', 'GHOST_decreed_validity':'V', 'network_provided_flag_description':'Missing < 25% of hourly sampling period'},
    'B':{'standard_data_flag_name':'Data Window Completeness < 25%', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Field instrument channel down for > 75% of hourly sampling period'},
    'D':{'standard_data_flag_name':'Data Window Completeness < 75%', 'network_decreed_validity':'I', 'GHOST_decreed_validity':'I', 'network_provided_flag_description':'Field instrument channel down for 25% to 75% of hourly sampling period'}
},

#--------------------------------------------------------------
#WMO GAW
#--------------------------------------------------------------
#'WMO GAW':{
#    {}
#}
}

###--------------------------------------------------------------------------------------------------###
#DEFINE DATA PROVIDER MEASUREMENT INFORMATION TO STANDARD MEASUREMENT INFORMATION MAPPING
###--------------------------------------------------------------------------------------------------###

#define mapping of data provider measurement information to standard measurement information, per network, per species
#two types of dictionaries are defined:
# -- mapping --> dictionary mapping all network provided measurement information to standardised measurement information
# -- standard_measurement_information --> dictionary of pre-standardised measurement information

#any information provided by the networks which improves upon the default standardised measurement information is defined

def network_measurement_information_to_standard_measurement_information_mapping(raw_measurement_info, metadata_stamp_dt, run_details):

#--------------------------------------------------------------
#CANADA NAPS
#--------------------------------------------------------------
    if run_details['network'] == 'CANADA_NAPS':
        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}
        elif run_details['parameter'] == 'ISOPRENE':
            standard_measurement_information = {'measurement':{'methodology':'gas chromatography - mass spectrometry'}}

#--------------------------------------------------------------
#CAPMoN
#--------------------------------------------------------------
    elif run_details['network'] == 'CAPMoN':
        if run_details['parameter'] == 'O3':
            #convert fractional minutes to datetime
            metadata_stamp_dt = crossnetwork_processing_functions.fractional_minutes_from_reference_to_datetime(metadata_stamp)
        
            if (metadata_stamp_dt.year >= 1988) & (metadata_stamp_dt.year < 1992):
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Dasibi 1008-AH', 'reported_lower_limit_of_detection':{'O3':'4.0'}}}
            elif(metadata_stamp_dt.year >= 1992) & (metadata_stamp_dt.year < 2003): 
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49'}}
            elif(metadata_stamp_dt.year >= 2003) & (metadata_stamp_dt.year < 2014):  
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49C',     'reported_lower_limit_of_detection':{'O3':'2.0'}}}
            elif(metadata_stamp_dt.year >= 2014):
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49i',     'reported_lower_limit_of_detection':{'O3':'1.0'}}}

#--------------------------------------------------------------
#EANET
#--------------------------------------------------------------
    elif run_details['network'] == 'EANET':
        mapping = {}

#--------------------------------------------------------------
#EBAS
#--------------------------------------------------------------
    elif run_details['network'] == 'EBAS':      
        if run_details['parameter'] == 'O3':
            mapping = {
            'uv_abs':                              {'measurement':{'methodology':'ultraviolet photometry'}},
            'uv_abs,dylec,1100':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Dylec 1100'}},
            'uv_abs,ebara,eg-2001f':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-2001F'}},
            'uv_abs,ebara,eg-2001ftp':             {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-2001FTP'}},
            'uv_abs,ebara,eg-3000f':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-3000F'}},
            'uv_abs,ecotech,ml_9810':              {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ecotech EC9810'}},
            'uv_abs,ansyco,41m':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Environnement SA O341M'}},
            'uv_abs,horiba,apoa-370':              {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Horiba APOA370'}},
            'uv_abs,teledyne,400a':                {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne 400A'}},
            'uv_abs,monitor_labs,9810':            {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne ML9810'}},
            'uv_abs,teledyne_monitor_labs,ml9810b':{'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne ML9810B'}},
            'uv_abs,teledyne,t400':                {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne T400'}},
            'uv_abs,thermo,49':                    {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49'}},
            'uv_abs,thermo,49c':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49C'}},
            'uv_abs,thermo,49c+49i':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49C'}},
            'uv_abs,thermo,49i':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49i'}},
            'uv_abs,thermo,49w':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49w'}},
            'na':                                  {'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {
            'chemiluminescence_molybdenum':                       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}},
            'chemiluminescence_photometer':                       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_photolytic':                       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'uv_fluoresc':                                        {                                                                                                                                                                                                 'measurement':{'methodology':'laser-induced fluorescence',                                                                                                                                                                                       'process_details':'Make assumption this technique is laser-induced fluorescence (proven for NO/NO2), and not ultraviolet fluorescence (used exclusively for SO2).', 'measuring_assumption':True}},
            'continuous_colorimetric':                            {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'colorimetry',                                                                             'sampling_type':'continuous injection'}},
            'chemiluminescence_photolytic,ecophysics,cld700alppt':{                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecophysics CLD700AL',                                           'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_photolytic,ecophysics,cld770alppt':{                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecophysics CLD770AL',                                           'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,ecophysics,cranox':     {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',   'instrument':'Ecophysics CraNOx'}},
            'chemiluminescence_photolytic,ecotech,9841t':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecotech EC9841T'}},
            'chemiluminescence_molybdenum,horiba,apna-360ce':     {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA360'}},
            'chemiluminescence_molybdenum,horiba,apna-370':       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370'}}, 
            'chemiluminescence_molybdenum,horiba,apna-370ce':     {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370'}},
            'chemiluminescence_photolytic,horiba,apna-370':       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370',                                                'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_photolytic,teledyne_api,200e':     {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Teledyne 200E',                                                 'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,teledyne,t200up':       {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',   'instrument':'Teledyne T200UP'}},
            'chemiluminescence_photolytic,thermo,42c-tl':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42C-TL',                                                 'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_molybdenum,thermo,42i':            {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i'}},
            'chemiluminescence_photometer,thermo,42i':            {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i',                                                    'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_molybdenum,thermo,42i-ls':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-LS'}},
            'chemiluminescence_molybdenum,thermo,42i-tl':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL'}},
            'chemiluminescence_photometer,thermo,42i-tl':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL',                                                 'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,thermo,42i-tl':         {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL',                                                 'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photometer,thermo,49i':            {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet photometry',                              'instrument':'Thermo 49i'}},
            'caps,teledyne,t500u':                                {                                                                                                                                                                                                 'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',          'instrument':'Teledyne T500U'}},
            'na':                                                 {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'CO':
            mapping = {
            'infrared_absorption':               {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                                            'measuring_assumption':True}},      
            'ndir':                              {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                                            'measuring_assumption':True}},                            
            'uv_fluoresc':                       {                                                                                                                                                                                                         'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                                                      'process_details':'Make assumption this technique is vacuum ultraviolet resonance fluorescence (proven for CO), and not ultraviolet fluorescence (used exclusively for SO2).', 'measuring_assumption':True}},
            'online_crds':                       {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'filter_1pack':                      {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'}, 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                                                                                                       'measuring_assumption':True}},
            'vuv_fluorescence,aero-laser,al5001':{                                                                                                                                                                                                         'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                   'instrument':'Aerolaser AL5001'}},
            'ndir,horiba,apma-360':              {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'ndir,horiba,apma-370':              {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA370'}},
            'online_crds,picarro,g2301':         {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'Picarro G2301'}},
            'online_crds,picarro,g2401':         {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'Picarro G2401'}},
            'na':                                {                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'ISOPRENE':
            mapping = {
            'online_gc':          {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'measurement':{'methodology':'gas chromatography - unknown detection',       'sampling_type':'continuous injection'}},
            'online_ptr':         {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'measurement':{'methodology':'proton transfer reaction - mass spectrometry', 'sampling_type':'continuous injection', 'measuring_assumption':True}},
            'na':                 {                                                                                                  'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'SO2':
            mapping = {
            'abs_solution':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                                                                                 'measuring_assumption':True}},
            'abs_tube':                                                  {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Detection Tube',       'preparation_assumption':True},                                                                                                'measurement':{'methodology':'colorimetry',                                                                      'sampling_type':'manual',                                                                                                                                                                        'measuring_assumption':True}},
            'ams':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Some additional details in raw files which suggest this is Ion Chromatography',                                                     'measuring_assumption':True}}, 
            'bulk_sampler':                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'denuder':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'doas':                                                      {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'differential optical absorption spectroscopy'}}, 
            'flame_photometry':                                          {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'flame photometric detection'}},
            'filter_abs_solution':                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                       'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                                                                                 'measuring_assumption':True}},                                                                         
            'filterpack':                                                {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_1pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_3pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack,imwm,two_stage_inline':                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},   
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-2000'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-3000':{'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-3000'}}, 
            'filter_3pack,riemer,spf4':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'lvs_teflon':                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'online_ic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                               'sampling_type':'continuous injection'}},
            'trace_level_analyzer':                                      {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                   'process_details':'Make assumption this technique is ultraviolet fluorescence (exclusive for SO2).',                                                   'measuring_assumption':True}},
            'tub_denuder':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                                'measuring_assumption':True}},
            'uv_abs':                                                    {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                   'process_details':'Make assumption this technique is ultraviolet fluorescence (exclusive for SO2), and not ultraviolet photometry (for measuring O3)', 'measuring_assumption':True}},
            'uv_fluoresc':                                               {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv_fluoresc,ecotech,ec9850':                                {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech EC9850'}},
            'uv_fluoresc,teledyne,t100':                                 {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne T100'}},
            'uv_fluoresc,thermo,43i':                                    {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i'}},
            'uv_fluoresc,thermo,43i-tl':                                 {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i-TLE'}},
            'uv_fluoresc,thermo,43i-tle':                                {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i-TLE'}},
            'na':                                                        {                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'NH3':
            mapping = {
            'abs_solution':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                              'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                  'measuring_assumption':True}},
            'abs_tube':                                                  {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Detection Tube',       'preparation_assumption':True}, 'measurement':{'methodology':'colorimetry',                                           'sampling_type':'manual',                                                                                                                    'measuring_assumption':True}}, 
            'ams':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
            'denuder':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'tub_denuder':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder,ecn,amor':                                          {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'conductimetry',      'instrument':'ECN AMOR'}},
            'filter_1pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,riemer,spf4':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography', 'instrument':'Thermo ICS-2000'}},
            'high_vol_sampler':                                          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},                      
            'lvs_teflon':                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                              'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'online_ic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                    'sampling_type':'continuous injection'}},    
            'na':                                                        {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HNO3':
            mapping = {
            'abs_tube':                                                  {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Detection Tube',       'preparation_assumption':True}, 'measurement':{'methodology':'colorimetry',                                          'sampling_type':'manual',                                                                                                                    'measuring_assumption':True}},                                                                                                   
            'ams':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
            'denuder':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_1pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography', 'instrument':'Thermo ICS-2000'}},             
            'filter_3pack,riemer,spf4':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'lvs_teflon':                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'online_ic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                   'sampling_type':'continuous injection'}}, 
            'na' :                                                       {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (run_details['parameter'] == 'PM1'):
            mapping = {
            'abs_solution':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                                           'measuring_assumption':True}},
            'b-attenuation':                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation'}},
            'b-attenuation,opsis,sm200':                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'beta_gauge_particulate_sampler':                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation'}},
            'beta_gauge_particulate_sampler,environnement,mp101m':     {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'beta_gauge_particulate_sampler,thermo,5030sharp':         {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'beta_ray_absorption':                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation'}},
            'dicho':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Some additional details in raw files which suggest this is a gravimetric method', 'measuring_assumption':True}},
            'filterpack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                            'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'filter_1pack':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'filter_3pack':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler':                                        {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,digitel,da-80':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,digitel,dha-80':                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_mb':                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/mb'},                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_msb':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,thermo,vfc-pm10':                        {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Thermo VFC-PM10'},                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler':                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,derenda,pns_16t':                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda PNS 16T'},                                                                                                                                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,fai_instruments,swam_5a_dual_channel':    {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'low_vol_sampler,leckel,lvs3':                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,leckel,mvs6':                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,leckel,seq47_50':                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,rupprecht&patashnick,partisol2025':       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,thermo,partisol2000i':                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2000i'},                                                                                                                                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,thermo,partisol2025':                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'lvs_qbq':                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'lvs_teflon':                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'impactor':                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'impaction',                                                                                    'instrument':'Dekati PM10 Impactor',      'process_details':'Suggestion on EBAS submission portal that it is this specific instrument.',       'measuring_assumption':True}},
            'ivlpm':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'IVL P-Model',            'process_details':'Rough guess it is this specific instrument.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'optical_particle_counter':                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'optical particle counter'}},
            'optical_particle_size_spectrometer,grimm,180':            {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'optical_particle_size_spectrometer,grimm,190':            {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 190'}},
            'radiometric':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'beta-attenuation'}},
            'teom':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'teom_fdms':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom,rupprecht&patashnick,1400':                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom,rupprecht&patashnick,1400a':                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},            
            'thermo':                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                       'measuring_assumption':True}},
            'teom,thermo,1405':                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},       
            'na' :                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] in ['Al', 'As', 'BC', 'C', 'C_corrected', 'Ca++', 'Cd', 'Cl-', 'Co', 'Cr', 'Cu', 'EC', 'Fe', 'Hg', 'K+', 'Mg++', 'Mn', 'Na+', 'NH4+', 'Ni', 'NO3-', 'OC', 'OC_corrected', 'Pb', 'Se', 'SO4--', 'SO4--_NSS', 'V', 'Zn']:   
            mapping = { 
            'abs_solution':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                        'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                 'measuring_assumption':True}},
            'ams':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
            'bulk_sampler':                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},                                                                                                                                                      
            'filter_abs_solution':                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                 'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                 'measuring_assumption':True}},                                                                         
            'filter_1pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,imwm,two_stage_inline':                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                     {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack':                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},              
            'filter_3pack,nilu,three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-3000':{'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-3000'}},
            'filter_3pack,riemer,spf4':                                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Riemer SPF4'},                                           'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'high_vol_sampler':                                          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                        'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},              
            'high_vol_sampler,digitel,da-80':                            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'high_vol_sampler,digitel,dha-80':                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'low_vol_sampler':                                           {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                         'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                               'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},    
            'low_vol_sampler,leckel,seq47_50':                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                             'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'sulphate_monitor':                                          {                                                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'thermal reduction - ultraviolet fluorescence', 'instrument':'Thermo 5020i',      'process_details':'Some details in raw files suggest instrument is Thermo 5020i',                  'measuring_assumption':True}}, 
            'na' :                                                       {                                                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}}} 
#-------------------------------------------                                                                                                                                                                                                                                                                                                                                                          
        elif ('PM10_' in run_details['parameter']) or ('PM2.5_' in run_details['parameter']) or ('PM1_' in run_details['parameter']):
            mapping = {
            'denuder':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                         'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},                                                                                                                                                      
            'filter_1pack':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},
            'filter_2pack':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},  
            'filter_3pack':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},                   
            'high_vol_sampler':                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},              
            'high_vol_sampler,digitel,da-80':                                                                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},                                                                                              
            'high_vol_sampler,digitel,dha-80':                                                                        {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_mb':                                                                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/mb'},                                                                                                                                                                                                                                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_msb':                                                                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},          
            'low_vol_sampler':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},   
            'low_vol_sampler,derenda,mvs_6.1':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda MVS 6.1'},                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},
            'low_vol_sampler,derenda,pns_16t,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l':{'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda PNS 16T'},                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'thermal-optical analysis', 'instrument':'Sunset OCEC 5l'}},  
            'low_vol_sampler,leckel,mvs6':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},    
            'low_vol_sampler,leckel,seq47_50':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},  
            'low_vol_sampler,rupprecht&patashnick,partisol2025':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},  
            'low_vol_sampler,thermo,partisol2000i':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2000i'},                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}}, 
            'low_vol_sampler,thermo,partisol2025':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},     
            'lvs_denuder_tandem':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}}, 
            'lvs_qbq':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},
            'lvs_single':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},         
            'lvs_single_twin':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},         
            'lvs_teflon':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}}, 
            'impactor':                                                                                               {                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'impaction', 'instrument':'Dekati PM10 Impactor', 'process_details':'Suggestion on EBAS submission portal that it is this specific instrument.', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Assume method is ion chromatography after sampling.',                          'measuring_assumption':True}},   
            'nitrate_monitor':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                 'process_details':'Internal file metadata suggests method is ion chromatography after sampling.', 'measuring_assumption':True}},
            'online_ic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'ion chromatography',                                        'sampling_type':'continuous injection'}},                           
            'online_thermal-optical_analysis':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                'measurement':{'methodology':'thermal-optical analysis',                                  'sampling_type':'continuous injection'}}, 
            'na' :                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}}} 

#--------------------------------------------------------------
#EEA AIRBASE
#--------------------------------------------------------------
    elif run_details['network'] == 'EEA_AIRBASE':
        mapping = {}
    
#--------------------------------------------------------------
#EEA AQ eReporting
#--------------------------------------------------------------
    elif run_details['network'] == 'EEA_AQ_eReporting':
        if run_details['parameter'] == 'O3':
            mapping = {
            '':                                                                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'active':                                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'api100a_uv-fl_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'api100e_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                       'measuring_assumption':True}},
            'api200_chemi_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Teledyne 200'}},
            'api200e_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Teledyne 200E'}},
            'api400_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400'}},
            'api400a_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400A'}},
            'api400e_light-abs_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E',            'process_details':'Assume method is not light absorption as stated but is ultraviolet photometry as reported instrument uses.',                                                'measuring_assumption':True}},
            'api400e_ndir_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E',            'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                                            'measuring_assumption':True}},                   
            'api400e_uv-fl_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E',            'process_details':'Assume method is not ultraviolet fluorescence as stated but is ultraviolet photometry as reported instrument uses.',                                        'measuring_assumption':True}},
            'api400e_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E'}},
            'apit300_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},
            'apit400_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne T400',            'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'apit400_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne T400'}},
            'columbia_scientific_ic_3100_uv-p_automatic':                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'CSI 3100'}},
            'dasibi_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'dasibi1008_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Dasibi 1008-AH',           'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                                                    'measuring_assumption':True}},
            'dasibi1008ah_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Dasibi 1008-AH'}},
            'dasibi1008rs_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Dasibi 1008-RS'}},
            'dasibi1108_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Dasibi 1108 W/GEN',        'process_details':'Make assumption measuring instrument is Dasibi 1108 W/GEN',                                                                                                 'measuring_assumption':True}},
            'dasibi5014_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'SIR S-5014',               'process_details':'Make assumption measuring instrument is SIR S-5014',                                                                                                        'measuring_assumption':True}},
            'environnement_s.a._model_o331m_uv_ozone_analyzer_uv-p_automatic':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O331M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_automatic':       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_uv-p_automatic':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_ultraviolet_(uv)_photometry_automatic':        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O342M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_uv-p_automatic':                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O342M'}},
            'enviroo341m_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O341M'}},
            'enviroo342e_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O342M'}},
            'enviroo342m_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Environnement SA O342M'}},
            'horiba_model_apma_360e_co_analyser_uv-p_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)', 'instrument':'Horiba APMA360'}},
            'horiba_model_apoa_300_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA300'}},
            'horiba_model_apoa_350e_o3_analyser_uv-p_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_ultraviolet_(uv)_photometry_automatic':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_360_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_360_o3_analyser_uv-p_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry':                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA370'}},
            'horiba_model_apoa_370_o3_analyser_uv-p_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA370'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA370'}},
            'horibaapoa350_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA350'}},
            'horibaapma360e_ndir_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)', 'instrument':'Horiba APMA360'}},
            'horibaapna360_chemi_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}},
            'horibaapoa300_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA300'}},
            'horibaapoa350e_uv-p_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA350'}},
            'horibaapoa360_other_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA360'}},
            'horibaapoa360_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA360'}},
            'horibaapoa360e_uv-p_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA360'}},
            'horibaapoa370_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Horiba APOA370'}},
            'ion_chromatography':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                     'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'iseo_argopol-sam-xair_uv-p_automatic':                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Cannot find instrument ISEO argopol sam-xair, therefore assume uses instrument with ultraviolet photometry method as stated.',                              'measuring_assumption':True}},
            'mcv48auv_uv-p_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'MCV 48AUV'}},
            'ml2010_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Cannot find instrument ML2010, therefore assume uses instrument with ultraviolet photometry method as stated.',                                             'measuring_assumption':True}},
            'ml8810_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810'}},
            'ml9810_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810'}},
            'ml9810a_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810A'}},
            'ml9810b_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810B'}},
            'ml9812_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9812'}},
            'ml_uv-p_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'monitor_labs_model_8410_nox_analyser_uv-p_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ethylene chemiluminescence',                                 'instrument':'Teledyne ML8410E',         'process_details':'Assume method is not ultraviolet photometry as stated but is ethylene chemiluminescence as reported instrument uses',                                       'measuring_assumption':True}},              
            'monitor_labs_model_8810_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_9810_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9812_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne ML9812'}},
            'monitor_labs_undetermined_uv-p_automatic':                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},                                                                   
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':  {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er120_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_remote':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'orion_uv-p_automatic':                                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet photometry method as stated.',                                              'measuring_assumption':True}},
            'other_automatic':                                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other,_please_specify':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other_ultraviolet_(uv)_photometry':                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'other_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'other_uv-fl_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv-p_automatic':                                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},  
            'philipsk50094_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Philips K50094'}},
            'serinus10_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Ecotech Serinus10'}},     
            'sri8610_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not gas chromatography as the reported SRI 8610 instrument uses',                                  'measuring_assumption':True}},
            'teledyne_api_400_uv_photometric_o3_analyser_uv-p_automatic':                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_other_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_uv-fl_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400',             'process_details':'Assume method is not ultraviolet fluorescence as stated but is ultraviolet photometry as reported instrument uses.',                                        'measuring_assumption':True}},
            'teledyne_api_400a_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400A'}},
            'teledyne_api_400a_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400A'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_uv-fl_automatic':                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne 400E'}},
            'teledyne_api_t300_gas_filter_correlation_co_analyser_ultraviolet_(uv)_photometry':             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},
            'teledyne_api_t400_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Teledyne T400'}},
            'thermo42_chemi_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Thermo 42'}},
            'thermo43c_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Thermo 43C instrument uses',                          'measuring_assumption':True}},
            'thermo43s_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43S'}},
            'thermo49_uv-p_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49'}},
            'thermo49_ndir_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49',                'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                                            'measuring_assumption':True}},
            'thermo49c_chemi_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49C',               'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},  
            'thermo49c_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49C'}},                       
            'thermo49cps_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49CPS'}},
            'thermo49i_chemi_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i',               'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'thermo49i_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i'}},
            'thermo49w_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49w'}},
            'thermo_49_cps_ozone_primary_standard_ultraviolet_(uv)_photometry_automatic':                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49CPS'}},
            'thermo_model_42_no_nox_analyser_ultraviolet_(uv)_photometry_automatic':                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence (internal molybdenum converter) as the reported Thermo 42 instrument uses.', 'measuring_assumption':True}},
            'thermo_model_49_o3_analyser_ultraviolet_(uv)_photometry_automatic':                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49'}}, 
            'thermo_model_49_o3_analyser_uv-p_automatic':                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry_automatic':                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49C'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry':                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49C'}}, 
            'thermo_model_49c_o3_analyser_uv-p_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49C'}}, 
            'thermo_model_49i_o3_analyser_uv-fl_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i',              'process_details':'Assume method is not ultraviolet fluorescnce as stated but is ultraviolet photometry as reported instrument uses.',                                          'measuring_assumption':True}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry':                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i'}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_automatic':                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i'}},
            'thermo_model_49i_o3_analyser_uv-p_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i'}},
            'too_generic_other_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'too_generic_ultraviolet_(uv)_photometry':                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_ultraviolet_(uv)_photometry_active':                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'ultraviolet_(uv)_photometry':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'undetermined_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}},
            'unknown_doas_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_light-abs_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                              'process_details':'Make assumption light absorption refers to ultraviolet photometry.',                                                                                        'measuring_assumption':True}},
            'unknown_ultraviolet_(uv)_photometry':                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},                                    
            'unknown_other_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'unknown_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'uv-p_automatic':                                                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}}}
#------------------------------------------- 
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {
            'active':                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'airponter_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Airpointer-NOx'}},
            'airpointer_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Airpointer-NOx'}},
            'analysis_automation_mod._447_chemiluminescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Cannot find instrument Analysis Automation Mod 447, therefore assumes uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                                             'measuring_assumption':True}},
            'api100e_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                                                           'measuring_assumption':True}},
            'api100e_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Teledyne 100E'}},
            'api200_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'api200a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'api200a_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A',                                                      'process_details':'Assume method is not ultraviolet fluorescence as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                            'measuring_assumption':True}},
            'api200a_gc-fid_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A',                                                      'process_details':'Assume method is not GCFID as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                                               'measuring_assumption':True}},
            'api200au_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200AU'}},
            'api200e_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'api200e_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E',                                                      'process_details':'Assume method is not ultraviolet fluorescence as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                            'measuring_assumption':True}},
            'api200eu_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'api200up_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                      'instrument':'Teledyne T200UP',                                                    'process_details':'Assume instrument referenced is Teledyne T200UP', 'measuring_assumption':True}},
            'api201a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'apim200e_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'apit200_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'aspirator_spectrometry_neda_griess-ylosway_active':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},  
            'aspirator_spectrophotometry_neda_griess-ylosway_active':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'bendix_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-B',                                                      'process_details':'Make assumption instrument used is Bendix 8101-B, as only instrument manufactuer is given (Bendix). Methodology given is chemiluminescence, which the instrument uses (make assumption of internal molybdenum converter)', 'measuring_assumption':True}},
            'bendix_combustion_engineering_model_8101-c_oxides_of_nitrogen_analyze_chemi_automatic':                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-C'}},
            'chemi_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'chemiluminescence':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'csi1600_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'CSI 1600'}},
            'dasibi2108_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_chemi_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_other_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'eco_physics_cld_700_al_chemi_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'ecophysics-cld700al_chemi_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_al_chemiluminescence':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_al_chemiluminescence_active':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'enrivoac30m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'enrivoac31m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'enrivoac32m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'enviroas32m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}},
            'enviroas32m_caps_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_other_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_af21m_so2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Environnement SA AF21M instrument uses',                                                  'measuring_assumption':True}},
            'environnement_s.a._model_af21m_so2_analyzer_uv-fl_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_as32m_nitrogen_dioxide_analyzer_caps':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}}, 
            'environnement_s.a._model_as32m_nitrogen_dioxide_analyzer_chemi_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}},
            'filter_pack_spectrophotometry_spectrophotometry_passive':                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry'}},
            'filter_pack_spectrophotometry_spectrophotometry_active':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry'}},
            'filter_sampling_spectrophotometry_photometry_active':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                     'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}}, 
            'flame_ionization_detection_(fid)_passive':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'flame ionisation detection',                                                                                      'sampling_type':'passive'}},
            'glass_filter_spectrophotometry_spectrophotometry_active':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}}, 
            'horibaapha350e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna300_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horibaapna350_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna350e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna360_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna370_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_atomic_absorption_spectrometry_(aas)_chemi_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_300e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horiba_model_apna_350_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}}, 
            'horiba_model_apna_350e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_350e_nox_analyser_chemiluminescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_360_nox_analyser_chemi_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemiluminescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_ion_chromatography_chemiluminescence_passive':                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemiluminescence_active':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_370_nox_analyser_chemi_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence_active':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_atomic_absorption_spectrometry_(aas)_chemiluminescence_active':             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_uv-p_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'impregnated_filter_spectrometry_spectrophotometry_active':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'impregnated_filter_spectrometry_spectrophotometry_passive':                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_active':                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'impregnated_filter_spectrophotometry_not-automatic':                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_spectrophotometry_active':                                                {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},  
            'impregnated_filter_spectrophotometry_spectrophotometry_passive':                                               {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}}, 
            'ion_chromatography':                                                                                           {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_passive':                                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'lachat_spectrophotometry_spectrophotometry_active':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'mcv30ql_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'MCV 30QL'}},
            'ml2041_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML2041'}},
            'ml8840_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'ml8841_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},
            'ml9841_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'ml9841a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841A'}},
            'ml9841b_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'ml9841t_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'monitor_labs_model_8840_nox_analyser_chemi_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'monitor_labs_model_8840_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'monitor_labs_model_8841_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},         
            'monitor_labs_model_9841_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'monitor_labs_model_9841b_nox_analyser_chemi_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'monitor_labs_model_9841t_nox_analyser_chemiluminescence':                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'opsisar500_doas_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500_doas_remote':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er120_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_with_er_150_differential_optical_absorption_spectroscopy_(doas)':                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'orion_chemi_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Cannot find instrument Orion, therefore assume uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                                                                    'measuring_assumption':True}},
            'other_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'other_chemi_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'other_chemiluminescence':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'other_photometry_sp_automatic':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'other,_please_specify':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},   
            'other_spectrophotometry_sp_automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'other_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence'}},
            'palmes_colorimetry_passive_sampler_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'colorimetry'}},
            'photometry':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'photometry_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'photometry_passive':                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'sequential_air_sampler,_type_ss2000._nai-impregnated_glass_sinters_spectrophotometry_spectrophotometry_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'serinus40_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecotech Serinus40'}},
            'spectrometry_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'spectrometry_passive':                                                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'spectrophotometry_passive':                                                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemi_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_other_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_400_uv_photometric_o3_analyser_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Teledyne 400 instrument uses',                                                              'measuring_assumption':True}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemiluminescence':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'thermo14b_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B'}},
            'thermo14b-e_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo42_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo42c_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo42c-tl_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo_model_42c-tl_(trace_level_nox)_chemi_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo42i_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo42i-bl_other_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i',               'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'thermo42i-tl_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL'}},
            'thermo42s_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo43i-tle_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Thermo 43i-TLE instrument uses',                                                          'measuring_assumption':True}},
            'thermo49_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Thermo 49 instrument uses',                                                                 'measuring_assumption':True}},
            'thermo450i_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Thermo 450i'}},
            'thermo_model_14b_e_chemiluminescence_no-no2-nox_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo_model_14b_e_chemiluminescence_no-no2-nox_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo_model_42_no_nox_analyser_chemi_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},                            
            'thermo_model_42_no_nox_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo_model_42c_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_other_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},           
            'thermo_model_42i_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42s_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo_model_42s_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'too_generic_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence_active':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_other_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'undetermined_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_caps_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy'}},
            'unknown_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_active':                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_doas_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_ion_chromatography_chemiluminescence_passive':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_ion_chromatography_diffusive_sampler_automatic':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'ion chromatography',                                                                                              'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_diffusive_sampler_passive':                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_unknown_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},  
            'unknown_other_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},                                                   
            'unknown_photometry_chemiluminescence_passive':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}}, 
            'unknown_photometry_diffusive_sampler_passive':                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},                        
            'unknown_unknown_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            '':                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'CO':
            mapping = {
            'api100e_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses', 'measuring_assumption':True}},
            'api200a_chemi_automatic':                                                                                   {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',             'instrument':'Teledyne 200A'}},
            'api200_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200 instrument uses',         'measuring_assumption':True}},
            'api300a_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'api300a_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation) .',        'measuring_assumption':True}},
            'api300e_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300e_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300_gc-fid_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300',             'process_details':'Assume method is not GCFID as stated but is NDIR (Gas-Filter Correlation) as reported instrument uses.',                                     'measuring_assumption':True}},                                                  
            'api300_ir-gfc_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_light-abs_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api_ndir_automatic':                                                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'apit300_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'apit300_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'dasibi3008_ndir_automatic':                                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Dasibi 3008'}},
            'enviroco11m_ir-gfc_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},   
            'enviroco11m_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'enviroco12m_ir-gfc_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'enviroco12m_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'environnement_s.a._model_co10m_co_analyzer_ndir_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO10M'}},
            'environnement_s.a._model_co11m_co_analyzer_ir-gfc_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_ndir_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)':                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co12m_co_analyzer_ndir_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},  
            'environnement_s.a._model_co12m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}}, 
            'horibaapma300_ndir_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA300'}},
            'horibaapma350e_ndir_automatic':                                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horibaapma350_ndir_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horibaapma360ce_ndir_automatic':                                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360e_ndir_automatic':                                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360_ndir_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma370_crds_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not CRDS',                                                   'measuring_assumption':True}},
            'horibaapma370_ir-gfc_automatic':                                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horibaapma370_ndir_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_350_co_analyser_ir-gfc_automatic':                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_ndir_automatic':                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_360_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_370_co_analyser_ndir_automatic':                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                             {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'infrared_gas_filter_correlation':                                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},  
            'ir-gfc_automatic':                                                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'maihak_unor_6n_ir-gfc_automatic':                                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},
            'ml8830_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'ml9830b_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'ml9830_ir-gfc_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'ml9830_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},                                             
            'monitor_labs_model_8830_co_analyser_ir-gfc_automatic':                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                 {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_9830b_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9830b_co_analyser_ndir_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9830_co_analyser_ndir_automatic':                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'ndir_automatic':                                                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'non-dispersive_infrared_spectroscopy_(ndir)':                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'orion_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Orion, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',                 'measuring_assumption':True}},
            'other_beta_automatic':                                                                                      {'measurement':{'methodology':'beta-attenuation'}},
            'other_chemi_automatic':                                                                                     {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                'measuring_assumption':True}},
            'other_crds_automatic':                                                                                      {'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'other_ir-gfc_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'other_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'other_automatic':                                                                                           {'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                                     {'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify':                                                                               {'measurement':{'methodology':'unknown'}},
            'philips_k50109_00_gas_filter_correlation_co_analyser_ndir_automatic':                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Philips K50109'}},
            'picarrog2303_crds_automatic':                                                                               {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2303'}},
            'picarrog2401_crds_automatic':                                                                               {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2401'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_ndir_automatic':                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200A instrument uses.',       'measuring_assumption':True}},
            'teledyne_api_200_chemiluminescent_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 200, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',          'measuring_assumption':True}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_ir-gfc_automatic':                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':{'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',         'measuring_assumption':True}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_ndir_automatic':                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_infrared_gas_filter_correlation_automatic':            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':{'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_ndir_automatic':                                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'thermo42_ndir_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42 instrument uses',            'measuring_assumption':True}},
            'thermo42i_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42i instrument uses',           'measuring_assumption':True}},
            'thermo43i_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Thermo 43i instrument uses',    'measuring_assumption':True}},
            'thermo48c_ir-gfc_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48c_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48i_beta_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i',               'process_details':'Make assumption method used is NDIR (Gas-Filter Correlation), as instrument reported uses, and not beta.',                                   'measuring_assumption':True}},
            'thermo48i_ir-gfc_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48i_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48_ir-gfc_automatic':                                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48_ndir_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48w_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Thermo 48w, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',            'measuring_assumption':True}},
            'thermo49_ndir_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49 instrument uses.',      'measuring_assumption':True}},
            'thermo49c_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49C instrument uses.',     'measuring_assumption':True}},
            'thermo_model_48c_co_analyser_ir-gfc_automatic':                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48c_co_analyser_ndir_automatic':                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48c_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48_co_analyser_ndir_automatic':                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo_model_48_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo_model_48i_co_analyser_infrared_gas_filter_correlation_automatic':                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_ndir_automatic':                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'too_generic_ir-gfc_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'too_generic_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'too_generic_non-dispersive_infrared_spectroscopy_(ndir)':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'too_generic_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'undetermined_ndir_automatic':                                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'unknown_light-abs_automatic':                                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption light absorption refers to NDIR (Gas-Filter Correlation)',                                                                   'measuring_assumption':True}},                                                        
            'unknown_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_other_automatic':                                                                                   {'measurement':{'methodology':'unknown'}},
            'unknown_vakuum-ultraviolet_(vuv)_spectrometry_automatic':                                                   {'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence'}},
            'unor6n_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'vakuum-ultraviolet_(vuv)_spectrometry_automatic':                                                           {'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'ISOPRENE':
            mapping = { 
            'active':                                                                                                                                   {                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'af_20_m_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'airmoozone_gc-fid_automatic':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',   'sampling_type':'continuous injection'}},
            'hp-6890_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                                {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'hp-6890_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',   'sampling_type':'continuous injection'}},
            'markes_thermal_desorber_+_agilent_gas_chromatograph_mass_spectrometer_gas_chromatography_+_mass_spectrometry_gc-ms_after_solvent_o_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',            'sampling_type':'continuous injection'}},
            'other_automatic':                                                                                                                          {                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'passive':                                                                                                                                  {                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'perkinelmerc500_gc-fid_automatic':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',   'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_active':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',            'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',   'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_-_mass_spectrometry_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',            'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',            'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_passive':                                                                                                       {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromotography_(gc)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',            'sampling_type':'continuous injection'}},
            'unknown_other_automatic':                                                                                                                  {                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'varian_chrompack_gc-fid_automatic':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',   'sampling_type':'continuous injection'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '':                                                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'airponter_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Airpointer-SO2'}},
            'api100_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'api100a_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'api100e_ndir_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E',                                                      'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'api100e_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'api101a_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101A'}},
            'apit100_cv-afs_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100',                                                      'process_details':'Assume method is not CV-AFS as stated but is ultraviolet fluorescence as reported instrument uses.',                                     'measuring_assumption':True}},
            'apit100_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100'}},
            'bubbler_24_h_titration_titration_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                      'measurement':{'methodology':'titration',                                                                                           'sampling_type':'continuous injection'}}, 
            'columbia_scientific_ic_sa700_other_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'columbia_scientific_ic_sa700_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_active':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_ion_chromatography_active':                              {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'dasibi4108_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'dasibi_4108_so2_analyser_uv-fl_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'diffusion_tube_ion_chromatography_passive':                                                                       {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'diffusion_tube_ion_chromatography_ion_chromatography_passive':                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'enc_37_other,_please_specify_unknown_active':                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'enviroaf20m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF20M'}}, 
            'enviroaf21m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'enviroaf22m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af21m_so2_analyzer_other_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv_fluorescence_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'filter_pack_ion_chromatography_spectrophotometry_active':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection',   'process_details':'Both ion chromatography and spectrophotometry are given as methods, assume method is ion chromatography.',                               'measuring_assumption':True}},
            'filter_sampling_capillary_electrophoresis_(ce)_capillary_electrophoresis_active':                                 {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                'measurement':{'methodology':'capillary electrophoresis',                                                                           'sampling_type':'continuous injection'}},
            'filter_sampling_spectrophotometry_ion_chromatography_active':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection',   'process_details':'Both ion chromatography and spectrophotometry are given as methods, assume method is ion chromatography.',                               'measuring_assumption':True}},
            'filter_spectrophotometry_thorin_active':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                      'preparation_2':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'}, 'measurement':{'methodology':'spectrophotometry',                                                                                   'sampling_type':'continuous injection'}},
            'gas_meter_other,_please_specify_unknown_active':                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}}, 
            'horiba_model_apsa_360_so2_analyser_uv-fl_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360e_so2_analyser_uv_fluorescence_active':                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360e_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horiba_model_apsa_370_so2_analyser_uv-fl_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horibaapda351e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not beta-attenuation as the reported Horiba APDA351 instrument uses.',        'measuring_assumption':True}},
            'horibaapma370_ndir_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)', 'instrument':'Horiba APMA370'}},
            'horibaapna360_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}},
            'horibaapsa350_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},    
            'horibaapsa350e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horibaapsa360_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360a_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa370_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370',                                                     'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'horibaapsa370_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'ion_chromatography':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}}, 
            'ion_chromatography_passive':                                                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'light-scat_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'light scattering photometry'}},
            'mcv64fuv_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'MCV 64FUV'}},
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_uv-fl_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'meloysa700_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'ml2050_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Teledyne ML2050, therefore assume uses instrument with ultraviolet fluorescence method as stated.',               'measuring_assumption':True}},
            'ml8850_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'ml8850s_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850S'}},
            'ml9810_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Teledyne ML9810 instrument uses.', 'measuring_assumption':True}},                                                                                                                                                                   
            'ml9850_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'ml9850b_ndir_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B',                                                   'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'ml9850b_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B'}},
            'ml_uv-fl_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'monitor_labs_model_8850_so2_analyser_ion_chromatography_uv_fluorescence_passive':                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850',                                                    'process_details':'Assume method is continuous ultraviolet fluorescence, as Teledyne ML8850 instrument uses.',                                              'measuring_assumption':True}},
            'monitor_labs_model_8850_so2_analyser_uv-fl_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850_so2_analyser_uv_fluorescence_active':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_9810_o3_analyser_uv_fluorescence':                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Teledyne ML9810 instrument uses.', 'measuring_assumption':True}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv-fl_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_undetermined_uv-fl_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er120_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_remote':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'orion_chemi_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'sulphur chemiluminescence - gas chromatography',                                                                                                'process_details':'Cannot find instrument Orion, therefore assume uses Sulphur Chemiluminescence - Gas Chromatography.',                                    'measuring_assumption':True}},
            'orion_uv-fl_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet fluorescence method as stated.',                         'measuring_assumption':True}},
            'other_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'other_ion_chromatography_other_automatic':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'other_photometry_other_automatic':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},            'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},            'measurement':{'methodology':'spectrophotometry'}},  
            'other_uv-fl_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv_fluorescence':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'sequential_air_sampler,_type_ek._koh-impregnated_whatman_40_filter_ion_chromatography_ion_chromatography_active': {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU EK'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence_active':                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'thermo42i_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Thermo 42i'}},
            'thermo42i_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Thermo 42i instrument uses.',           'measuring_assumption':True}},
            'thermo43_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo43a_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo43a_uv-p_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A',                                                         'process_details':'Assume method is not ultraviolet photometry as stated but is ultraviolet fluorescence as reported instrument uses.',                     'measuring_assumption':True}},
            'thermo43b_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43B'}},
            'thermo43c_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo43ctl_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE'}},
            'thermo43i-tle_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo43i_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo43i_cv-afs_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not CV-AFS as stated but is ultraviolet fluorescence as reported instrument uses.',                                     'measuring_assumption':True}},
            'thermo43i_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo43s_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43S'}},
            'thermo43w_other,_please_specify_uv-fl_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43W'}},
            'thermo43w_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43W'}},
            'thermo450i_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i',                                                        'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo450i_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i'}},
            'thermo48_ndir_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)','instrument':'Thermo 48'}},
            'thermo49_ndir_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49',                                                          'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                         'measuring_assumption':True}},
            'thermo_450i_chemiluminescence':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i',                                                        'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo_450i_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i'}},
            'thermo_model_43_so2_analyser_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43_so2_analyser_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43_so2_analyser_uv_fluorescence_active':                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43a_so2_analyser_other_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43c_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43h_pulsed_fluorescence_so2_analyser_uv-fl_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43H'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv_fluorescence_not-automatic':                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_uv_fluorescence':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_chemiluminescence_active':                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},                                                   
            'thermo_model_43i_so2_analyser_uv_fluorescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'too_generic_other_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'too_generic_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence_active':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'total_ans_filter_pack_ece-emep_ion_chromatography_ion_chromatography_active':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},      
            'total_ans_filter_pack_ece-emep_ion_chromatography_ion_chromatography_passive':                                    {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'undetermined_uv-fl_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_chemi_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'sulphur chemiluminescence - gas chromatography',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses Sulphur Chemiluminescence - Gas Chromatography',           'measuring_assumption':True}},
            'unknown_doas_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_ion_chromatography_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_uv_fluorescence_passive':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown',                                                                                                                                       'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_other_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'unknown_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv-fl_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv_fluorescence':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'NH3':
            mapping = {
            'api201a_chemi_automatic':                                                    {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'chemi_automatic':                                                            {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                                                           'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'ion_chromatography_active':                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',           'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                     'sampling_type':'continuous injection'}}, 
            'other_chemi_automatic':                                                      {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                                                           'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'other_other_automatic':                                                      {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_automatic':                                      {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'spectrophotometry':                                                          {'primary_sampling':{'sampling_type':'passive',               'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry_passive':                                                  {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':{                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}}, 
            'unknown_chemi_automatic':                                                    {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                                                           'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}}}                                                                                                                                                                                                          
#------------------------------------------- 
        elif run_details['parameter'] == 'HNO3':
            mapping = { 
            'active':{'measurement':{'methodology':'unknown'}}} 
#-------------------------------------------     
        elif (run_details['parameter'] == 'PM10') or (run_details['parameter'] == 'PM2.5') or (run_details['parameter'] == 'PM1'):
            mapping = {
            '':                                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'active':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'andersenfh621n_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh62c14_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'andersenfh62in_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH62 I-N'}},
            'andersenfh62ir_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_gravimetric_analysis_beta_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}}, 
            'andersenfh621n_gravimetric_analysis_beta_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh62ir_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersen_gmw_model_fh621-n_beta_monitor_beta_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_beta_ray_attenuation_active':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},  
            'andersen_gmw_model_fh621-n_beta_monitor_beta_ray_attenuation_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_gravimetric_analysis_beta_ray_attenuation_active':                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_light-scat_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_nephelometry_beta_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_opc-cmc_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersengps-1-modif_beta_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Andersen GPS-1 Modified, therefore assume uses Beta-Attenuation as stated.',                                                 'measuring_assumption':True}},
            'api200_chemi_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Teledyne 200'}},
            'api200e_chemi_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Teledyne 200E'}},
            'apit100_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'apit100_opc-cmc_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'apm-1_elecos_beta_automatic':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'apm-1_elecos_nephelometry_automatic':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not nephelometry as stated but uses beta-attenuation.',                                                                            'measuring_assumption':True}},
            'atmoservice_lvs_3d_gravimetric_analysis_gravimetric_analysis_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15',   'process_details':'Make assumption sampling instrument is Atmoservice PNS3D15.',   'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'atmoservice_pns3d15_gravimetric_analysis_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}},
            'bam1020_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_gravimetric_analysis_beta_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020heated_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020heated_teom-fdms_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'bam1020_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_teom_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020unheated_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_beta_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_ray_attenuation':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_ray_attenuation_active':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'cav_-_a_gravimetric_analysis_gravimetry_active':                                                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'MCV CAV-A/mb',          'process_details':'Make assumption sampling instrument is MCV CAV-A/mb.',          'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'dasibi7001_beta_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Dasibi 7001'}},
            'derenda_gravimetric_analysis_active':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'derenda_gravimetric_analysis_gravimetric_analysis_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'derenda_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'desconocido_gravimetric_analysis_gravimetric_analysis_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'dha-80_gravimetry_active':                                                                                                                    {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_active':                                                                                                   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_gravimetric_analysis_active':                                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_not-automatic':                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_gravimetric_analysis_gravimetry_active':                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_active':                                                   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_dha80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_dha-80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hivol_gravimetry_active':                                                                                                             {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_active':                                                    {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                      {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hivol_other_automatic':                                                                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hvs_dha80_gravimetric_analysis_active':                                                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},                 
            'digitel_hvs_dha80_gravimetric_analysis_gravimetric_analysis_active':                                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_gravimetric_analysis_not-automatic':                                                                                        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hvs_gravimetric_analysis_active':                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'eberlinefh62-1_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'eberlinefh62-1_teom_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'eberline_model_fh_62-1_beta_ray_attenuation_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'enviromp101m_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_gravimetric_analysis_beta_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_nephelometry_beta_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_teom_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'enviromp2.5m_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                                             'measuring_assumption':True}},
            'envirompsi100_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation':                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation_active':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation_automatic':                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_other_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_teom_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_ray_attenuation_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_ray_attenuation_active':                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_gravimetric_analysis_beta_ray_attenuation_active':                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_nephelometry_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not nephelometry as stated but uses beta-attenuation as reported instrument uses.',                                                'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_opc-cmc_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'environnement_s.a._model_mpsi_100_gravimetric_analysis_beta_ray_attenuation_active':                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'f-701-20_beta_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_verewa___durag_beta_ray_attenuation':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_verewa___durag_beta_ray_attenuation_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'fidas200_beta_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_opc-cmc_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_light-scat_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'fidas200_opc-cmc_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200s_light-scat_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200s',                                                                                                                                                                     'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'gravimetric_analysis':                                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'gravimetric_analysis_active':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'gravimetric_analysis_beta_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                                                                                                                                                                                         'measuring_assumption':True}},
            'grimm-edm180_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180_gravimetric_analysis_light-scat_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_edm_180_gravimetric_analysis_opc-cmc_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_gravimetric_analysis_opc-cmc_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_light-scat_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180mc_gravimetric_analysis_opc-cmc_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Cannot find instrument Grimm 180-MC, therefore assume instrument is Grimm 180.',                                                                    'measuring_assumption':True}},                             
            'grimm-edm180mc_opc-cmc_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Cannot find instrument Grimm 180-MC, therefore assume instrument is Grimm 180.',                                                                    'measuring_assumption':True}},                                                                                                            
            'grimm-edm180_nephelometry_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},                                              
            'grimm+edm+180_nephelometry_automatic':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'grimm_edm_180_opc-cmc_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_opc-cmc_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_teom_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm-edm365_light-scat_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm107_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm107_light-scat_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_differential_optical_absorption_spectroscopy_(doas)_active':                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},             
            'grimm_model_edm_180_for_pm10_and_pm2.5_beta_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_gravimetric_analysis_light_scattering_active':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_gravimetric_analysis_scattering_of_laser_light_active':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_laser_spectroscopy_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not laser spectroscopy as stated but uses optical particle counting as reported instrument uses.',                                 'measuring_assumption':True}},          
            'grimm_model_edm_180_for_pm10_and_pm2.5_light_scattering_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_light-scat_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_nephelometry_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_opc-cmc_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optical_particle_counter_+_conversion_to_mass_concentration':                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optical_particle_counter_+_conversion_to_mass_concentration_active':                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_other_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_scattering_of_laser_light_active':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_scattering_of_laser_light_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180-mc_for_pm10,_pm2.5,_pm1,_tc_and_31_size_channels_optical_particle_counter_+_conversion_to_mass_concentration_active':     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}}, 
            'horibaapda350_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E',                                                                                                                                                                      'process_details':'Cannot find Horiba APDA350 instrument, assume it is Horiba APDA350E.',                                                                              'measuring_assumption':True}},
            'horibaapda351e_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA351E'}},
            'horibaapda371_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371'}},
            'horibaapda371_light-scat_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371',                                                                                                                                                                       'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},
            'horibaapna370_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not chemiluminescence (internal molybdenum converter) as reported instrument uses.',                'measuring_assumption':True}},
            'horibafh621r_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba FH621-R'}},
            'horiba_model_apda_350e_tsp_analyser_gravimetric_analysis_beta_ray_attenuation_active':                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E'}},
            'hvs_gravimetric_analysis_active':                                                                                                             {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},               
            'hvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_not-automatic':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_active':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_gravimetric_analysis_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_not-automatic':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'kimoto_168s_beta_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto 168S'}},
            'leckel_-_lvs_3_gravimetric_analysis_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_-_lvs_3_gravimetric_analysis_gravimetric_analysis_active':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_sequential_lvs_lvs_-_automatic_filter_change_2,3_m3_h_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3',           'process_details':'Make assumption sampling instrument is Leckel LVS3.',                              'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'light-scat_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'lvs_gravimetric_analysis_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'lvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'lvs_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'mcz,_pns-lvs_16_seq_gravimetric_analysis_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'mcz,_pns-lvs_16_seq_gravimetric_analysis_gravimetric_analysis_active':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'metone_bam-1020_beta_ray_attenuation':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_beta_ray_attenuation_active':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},  
            'metone_bam-1020_beta_ray_attenuation_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_gravimetric_analysis_beta_ray_attenuation_active':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_ray_attenuation':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_ray_attenuation_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_other_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_teom_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'metone_bam-1020_other_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_teom_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'ml9841b_nephelometry_beta_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                                                                                                                                                                                                                                   'process_details':'Assume method is beta-attenuation - nephelometry as stated and not chemiluminescence (internal molybdenum converter) as reported instrument uses.', 'measuring_assumption':True}},
            'nephelometry_+_beta':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'nephelometry_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'opc-cmc_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'opsis_ar500_open_path_monitor_beta_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not DOAS as reported instrument uses.',                                                             'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                                                 'instrument':'OPSIS AR500'}},
            'opsissm200_beta_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_other_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_gravimetric_analysis_gravimetric_analysis_active':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not gravimetry as stated, but uses beta-attenuation as reported instrument uses.',                                                 'measuring_assumption':True}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation_active':                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation':                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'orion_gravimetric_analysis_beta_ray_attenuation_active':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                                                   'measuring_assumption':True}},
            'orion_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                                                   'measuring_assumption':True}},
            'other_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_beta_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},                                                                                     
            'other_beta_ray_attenuation_automatic':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'other_gravimetric_analysis_beta_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                                                                                                                                                                                         'measuring_assumption':True}},
            'other_gravimetric_analysis_other_automatic':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'other_gravimetric_analysis_teom_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'other_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_opc-cmc_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                                                                                                                                                                                                 'measuring_assumption':True}},                          
            'other_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                                                               'sampling_type':'continuous injection'}},
            'other_light-scat_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'other_opc-cmc_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},                                                                                                                                                                                                                                                                                                                                                                                                                                                 
            'other_other_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_not-automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_teom_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'other_teom-fdms_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_uv-fl_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'palas_model_fidas_200_other_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},                                                                                                                                                          
            'palas_model_fidas_200_teom_automatic':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'palas_model_fidas_200_tapered_element_oscillating_microbalance_(teom)_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'palas_model_fidas_200_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'partisol_gravimetric_analysis_active':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000',  'process_details':'Make assumption sampling instrument is Thermo Partisol 2000.',                     'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'partisol_gravimetric_analysis_gravimetric_analysis_active':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000',  'process_details':'Make assumption sampling instrument is Thermo Partisol 2000.',                     'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'philips_mp_101m_beta_ray_attenuation_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'pm2.5_beta_ray_attenuation_automatic':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}}, 
            'pns_3d_15_gravimetric_analysis_gravimetric_analysis_active':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}},
            'rancon_adam_c_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                                           'measuring_assumption':True}},
            'r&p_partisol_2000_gravimetric_analysis_gravimetric_analysis_active':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2000_gravimetric_analysis_gravimetry_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_gravimetric_analysis_active':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'sentinel_gravimetric_analysis_gravimetric_analysis_active':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Tecora Sentinel',   'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sentinel_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Tecora Sentinel',   'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_gravimetric_analysis_gravimetric_analysis_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_seq47_50_gravimetric_analysis_gravimetric_analysis_active':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry'}},
            'sm200_(adam)_beta_ray_attenuation_active':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm200_(adam)_beta_ray_attenuation_not-automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_beta_ray_attenuation_active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_(air_dust_automatic_monitor)_gravimetric_analysis_beta_ray_attenuation_active':                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},                             
            'sm_200_(air_dust_automatic_monitor)_beta_ray_attenuation_active':                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'spm-613d_gravimetric_analysis_beta_ray_attenuation_active':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto SPM-613D'}},
            'staplex_gravimetric_analysis_gravimetric_analysis_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                        'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_active':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_gravimetric_analysis_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_not-automatic':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_seq47_50_gravimetric_analysis_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_seq47_50_gravimetric_analysis_gravimetric_analysis_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'swam5adual_beta_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam5adual_caps_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a',                                                                                                                                                                          'process_details':'Assume method is not CAPS as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},                                                                                                                                                              
            'swam_5a_dual_channel_monitor_beta_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam_5a_dual_channel_monitor_beta_ray_attenuation':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'tcr_tecora,_model_sentinel_pm_gravimetric_analysis_gravimetric_analysis_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry'}},
            'tcr_tecora,_sentinel_pm_gravimetric_analysis_gravimetric_analysis_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_active':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_not-automatic':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_gravimetric_analysis_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_beta_ray_attenuation_active':                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_optical_particle_counter_+_conversion_to_mass_concentration_active':                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'teom1400_beta_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},    
            'teom_1400a_beta_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400a_beta_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400ab-fdms_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_teom_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_teom-fdms_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_teom-fdms_automatic':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},     
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}}, 
            'teom1400ab_teom_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},                                                                                                    
            'teom1400ab_teom-fdms_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                                                               
            'teom_1400ab_with_fdms_module_(8500)_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_opc-cmc_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_optical_particle_counter_+_conversion_to_mass_concentration_active':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_other_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_active':                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_automatic':                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_teom_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_teom-fdms_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400a_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_active':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400a_teom_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_teom-fdms_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                             
            'teom1400a_teom_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom1400a_teom-fdms_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400fdms-8500b-cb_teom-fdms_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400_fdms_+_8500_b_or_cb_dryer_other_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_fdms_+_8500_b_or_cb_dryer_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':{                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_opc-cmc_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not light scattering photometry as stated but uses TEOM as reported instrument uses.',                                             'measuring_assumption':True}},
            'teom_1400_other_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)_active':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_teom_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1405df-fdms_teom_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_teom-fdms_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405d_teom_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405D'}},
            'teom1405f-fdms_teom_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},
            'teom1405f-fdms_teom-fdms_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                             
            'teom_1405f_with_fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                                                           
            'teom1405_teom_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},
            'teom1405_teom-fdms_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                           
            'teom_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'teom-fdms_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'thermo42i_chemi_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo5014i_beta_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo_5014i_beta_continuous_ambient_particulate_monitor_beta_ray_attenuation_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo5030sharp_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_beta_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_nephelometry_beta_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_nephelometry_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_nephelometry_beta_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_other_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_andersen_esm_fh_62_i-r_beta_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},                                                                                             
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_active':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}}, 
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_gravimetric_analysis_other_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_gravimetric_analysis_beta_ray_attenuation_active':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_nephelometry_beta_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_opc-cmc_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_optical_particle_counter_+_conversion_to_mass_concentration_active':                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_other_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_series_fh_62_c14_continuous_pm10_ambient_particulate_monitor_beta_ray_attenuation_automatic':                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'thermo_esm_andersen,_fh_gravimetric_analysis_gravimetric_analysis_active':                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume instrument is Thermo Andersen FH62 I-R.',                                                                                                    'measuring_assumption':True}},
            'thermo_esm_andersen,_fh_gravimetric_analysis_active':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume instrument is Thermo Andersen FH62 I-R.',                                                                                                    'measuring_assumption':True}},
            'thermoi_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume instrument is Thermo 5014i.',                                                                                                                'measuring_assumption':True}},
            'thermo_i-series_beta_ray_attenuation_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume instrument is Thermo 5014i.',                                                                                                                'measuring_assumption':True}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation_active':                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_gravimetric_analysis_beta_ray_attenuation_active':                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_nephelometry_+_beta_active':                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'too_generic_beta_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'too_generic_beta_ray_attenuation_automatic':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'too_generic_other_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'too_generic_other,_please_specify':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction':                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_teom_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'too_generic_teom-fdms_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'umwelttechnik_mcz_gmbh_micro_pns_lvs_16_gravimetric_analysis_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry'}},
            'undefined_gravimetric_analysis_gravimetry_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'undetermined_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'undetermined_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'undetermined_opc-cmc_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}}, 
            'undetermined_other_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'undetermined_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'undetermined_teom_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'undetermined_teom-fdms_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'unitec_beta_ray_attenuation_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_gravimetric_analysis_beta_ray_attenuation_active':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_lspm10_gravimetric_analysis_beta_ray_attenuation_active':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation as stated but uses Light Scattering Photometry as reported instrument uses.',                                 'measuring_assumption':True}},
            'unitec_lspm10_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation as stated but uses Light Scattering Photometry as reported instrument uses.',                                 'measuring_assumption':True}},
            'unknown_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_active':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_not-automatic':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                           'sampling_type':'injection'}},                                                                               
            'unknown_differential_optical_absorption_spectroscopy_(doas)_active':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},    
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},            
            'unknown_gravimetric_analysis_active':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'unknown_gravimetric_analysis_beta_ray_attenuation_active':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                                                                                                                                                                                         'measuring_assumption':True}},
            'unknown_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                                                                                                                                                                                         'measuring_assumption':True}},
            'unknown_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                                                           'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},  
            'unknown_gravimetric_analysis_gravimetric_analysis_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},                                                                                                                                                                                                                                                                                                                                                                                                                                                               
            'unknown_gravimetric_analysis_gravimetry_active':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_gravimetry_automatic':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_light-scat_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                                                           'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gravimetric_analysis_light_scattering_active':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                                                           'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_gravimetric_analysis_teom_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_gravimetric_analysis_unknown_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'unknown_gravimetric_analysis_unknown_automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                                                                                                                                                                    'measuring_assumption':True}},
            'unknown_gravimetric_analysis_not-automatic':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'unknown_gravimetry_active':                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_ray_attenuation_active':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                                                           'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                                                           'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_light-scat_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering_active':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_nephelometry_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'nephelometry'}},
            'unknown_nephelometry_beta_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_nephelometry_+_beta_active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_opc-cmc_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},                                                                                                                                                                                                                                                                                                                                                                                                                                            
            'unknown_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'unknown_other_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_active':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_not-automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                      'sampling_type':'injection'}},  
            'unknown_teom_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_teom-fdms_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'unknown_unknown_active':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'verewa_f_701_beta_ray_attenuation_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_701_nephelometry_beta_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_703_beta_ray_attenuation_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'verewa_f_703_gravimetric_analysis_beta_ray_attenuation_active':                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'wedding_gravimetric_analysis_beta_ray_attenuation_active':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}}}  
#------------------------------------------- 
        elif run_details['parameter'] in ['Al', 'As', 'BC', 'C', 'C_corrected', 'Ca++', 'Cd', 'Cl-', 'Co', 'Cr', 'Cu', 'EC', 'Fe', 'Hg', 'K+', 'Mg++', 'Mn', 'Na+', 'NH4+', 'Ni', 'NO3-', 'OC', 'OC_corrected', 'Pb', 'Se', 'SO4--', 'SO4--_NSS', 'V', 'Zn']:   
            mapping = {
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_active':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                      'instrument':'GBC 932',        'sampling_type':'continuous injection'}},
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                      'instrument':'GBC 932'}},
            'active':                                                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'ae22-pm2.5_light-abs_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'incremental optical transmission absorption',               'instrument':'Magee AE22'}},
            'ae31_light-abs_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'incremental optical transmission absorption',               'instrument':'Magee AE31'}},
            'ae33_aethalometer_light_absorption':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'incremental optical transmission absorption',               'instrument':'Magee AE33'}},
            'ae33_light-abs_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'incremental optical transmission absorption',               'instrument':'Magee AE33'}}, 
            'atomic_absorption_spectrometry_(aas)':                                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'cold_vapour_atomic_fluorescence_spectrometry_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                             'sampling_type':'continuous injection'}},                                                                   
            'danish_lmp_filterpack_element_collector_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':{'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                      'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_active':                                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                                                                                   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                          'sampling_type':'continuous injection'}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'digitel_dha80_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_active':                                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                  'sampling_type':'continuous injection'}},
            'digitel_dha80_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active': {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                                            'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement_automatic':                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l', 'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'fh62i-n_atomic_absorption_spectrometry_(aas)_active':                                                                                                             {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'Andersen FH62 I-N'},                                                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'fh62i-n_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                      {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'Andersen FH62 I-N'},                                                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},      
            'filter_sampling_atomic_absorption_spectrometry_(aas)_active':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                           'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'filter_sampling_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                           'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'leckel_-_lvs_3_gravimetric_analysis_active':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                             'measurement':{'methodology':'ion chromatography',                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                  'sampling_type':'continuous injection'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'ion_chromatography':                                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection'}},
            'ion_chromatography_passive':                                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                     {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                           'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                              {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                           'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                            {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                           'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'other_light-abs_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'incremental optical transmission absorption'}},
            'other_other_automatic':                                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'sequential_sampler_gravimetry_active':                                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                       'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'}, 'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'sequential_sampler_gravimetry_automatic':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                       'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'}, 'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'sunsettoca_to-ecoc_automatic':                                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l', 'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'sven_leckel,seq_47_50_atomic_absorption_spectrometry_(aas)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'unknown',                                                                                                                            'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel_seq47_50_continous_flow_analysis_(cfa)_continous_flow_analysis_(cfa)_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'continuous flow analysis'}},
            'sven_leckel_seq47_50_ion_chromatography_ion_chromatography_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection'}},
            'termo_esm_andersen,_fh_95_kf_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Andersen FH95-KF'},                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'termo_esm_andersen,_fh_95_kf_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                             {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Andersen FH95-KF'},                                                                                                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'thermo_atomic_absorption_spectrometry_(aas)_active':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'thermo_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                   'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown',                                                                                                                            'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_not-automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown',                                                                                                                            'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                             'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_ion_chromatography_unknown_active':                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                       'sampling_type':'continuous injection'}},                         
            'unknown_other_remote':                                                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif ('PM10_' in run_details['parameter']) or ('PM2.5_' in run_details['parameter']) or ('PM1_' in run_details['parameter']):
            mapping = {
            '':                                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'active':                                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'atmoservice_lvs_3d_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_atomic_absorption_spectrometry_(aas)_active':                               {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                      'sampling_type':'continuous injection'}},
            'atomic_absorption_spectrometry_(aas)':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'atomic_absorption_spectrometry_(aas)_active':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'cold_vapour_atomic_absorption_spectrometry_(cv-aas)_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'cold vapour atomic absorption spectroscopy',                                                           'sampling_type':'continuous injection'}},
            'cold_vapour_atomic_fluorescence_spectrometry':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},
            'cold_vapour_atomic_fluorescence_spectrometry_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                                         'sampling_type':'continuous injection'}},
            'coulometry_active':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'coulometry'}},
            'derenda_continous_flow_analysis_(cfa)_continous_flow_analysis_(cfa)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'continuous flow analysis'}},
            'derenda_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'derenda_ion_chromatography_ion_chromatography_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'derenda_lvs_3.1_ion_chromatography_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'derenda_lvs_3.1_thermal-optical_carbon_analysis_gravimetric_analysis_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'dionex_ion_chromatography_ion_chromatography_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'dionex_2010i_ion_chromatography_ion_chromatography_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                        'instrument':'Dionex 2010i',               'sampling_type':'continuous injection'}},
            'dionex_3000_ion_chromatography_ion_chromatography_active':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                        'instrument':'Thermo ICS-3000',            'sampling_type':'continuous injection'}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement':                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l',                                                       'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement_automatic':                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l',             'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'filter_sampling_ion_chromatography_ion_chromatography_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'filter_3pack_ion_chromatography_ion_chromatography_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'flame_atomic_absorption_spectrometry_(f-aas)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection'}},
            'graphite_furnace_atomic_absorption_spectrometry_(gf-aas)':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                      'sampling_type':'continuous injection'}},  
            'gravimetric_analysis_active':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'}, 'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'hydride_generation_electrothermal_-_atomic_absorption_spectrometry_(hg-et-aas)_active':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'hydride generation atomic absorption spectroscopy',                                                    'sampling_type':'continuous injection'}}, 
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                              'sampling_type':'continuous injection'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'ion_chromatography':                                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'leckel_seq47_50_with_cooler_gravimetric_analysis_thermal-optical_carbon_analysis_active':                                                               {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'leckel_seq47_50_with_cooler_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'low_volume_sampler_leckel_ion_chromatography_ion_chromatography_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'lvs_-_automatic_filter_change_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':{'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'marga_ic_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                        'instrument':'Metrohm MARGA'}},
            'micropns_lvs1_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1'},                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}}, 
            'other_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},    
            'other,_please_specify':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'other,_please_specify_active':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'partisol_2300_other,_please_specify_unknown_active':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection',                                                                                                                                     'measuring_assumption':True}},
            'partisol_2300_other,_please_specify_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                     {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_hydride_generation_electrothermal_-_atomic_absorption_spectr_active':                 {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'partisol_+_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'partisol_+_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'partisol_+_other,_please_specify_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'partisol_+_other,_please_specify_unknown_active':                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection',                                                                                                                                     'measuring_assumption':True}},
            'perkin-elmer_elan_6000_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Perkin Elmer ELAN 6000',     'sampling_type':'continuous injection'}},
            'perkin-elmer_elan_6000_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':{'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Perkin Elmer ELAN 6000'}},
            'spectrometry':                                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                        'measuring_assumption':True}},
            'spectrometry_active':                                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                        'measuring_assumption':True}},
            'spectrophotometry_active':                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},      'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection'}},
            'sunset_laboratory_inc_thermal___optical_carbon_analyser_thermal-optical_carbon_analysis_eusaar_2_method_active':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l',             'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'sunset_laboratory_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l',             'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'sunsettoca_to-ecoc_automatic':                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                  'instrument':'Sunset OCEC 5l',             'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC 5l.',                                                      'measuring_assumption':True}},
            'sven_leckel_seq47_50_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'sven_leckel_seq47_50_ion_chromatography_ion_chromatography_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'sven_leckel_seq47_50_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'teom_accu_sampler_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':     {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'teom_accu_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':            {                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'tapered element oscillating microbalance - gravimetry'},                                                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'thermal-optical_carbon_analysis':                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis'}},
            'thermal-optical_carbon_analysis_active':                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'thermo_icp_iris_intrepid_ii_other_automatic':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',   'instrument':'Thermo IRIS Intrepid II',    'sampling_type':'continuous injection'}},
            'unknown_atomic_absorption_spectrometry_(aas)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_mass_espectrometry_(icp-ms)_active':                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_mass_espectrometry_(icp-ms)_not-automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_spectrophotometry_active':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_unknown_active':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                 'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_cold_vapour_atomic_absorption_spectrometry_(cv-aas)_atomic_absorption_spectrometry_(aas)_active':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'cold vapour atomic absorption spectroscopy',                                                           'sampling_type':'continuous injection'}},
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_mass_espectrometry_(icp-ms)_active':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_mass_espectrometry_(icp-ms)_not-automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_continous_flow_analysis_(cfa)_ion_chromatography_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_mass_espectrometry_(icp-ms)_not-automatic':                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_mass_espectrometry_(icp-ms)_active':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_atomic_absorption_spectrometry_(aas)_active':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                      'sampling_type':'continuous injection'}},
            'unknown_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                      'sampling_type':'continuous injection'}},
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_atomic_absorption_spectrometry_(aas)_active':                        {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_atomic_absorption_spectrometry_active':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_atomic_absorption_spectrometry_(aas)_active':                                   {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_differential_optical_absorption_spectrometry_active':                           {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_ion_chromatography_active':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_unknown_active':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                              'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_chromatography_active':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_high_vol_sampler_active':                                                              {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_hvs_-_automatic_filter_change_30_m3.h-1_active':                                       {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-a_active':                  {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                      {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_lvs_-_automatic_filter_change_2,3_m3_h_active':                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_mass_espectrometry_(icp-ms)_active':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_mass_espectrometry_(icp-ms)_not-automatic':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                         'sampling_type':'continuous injection'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_x-ray_fluorescence_spectroscopy_(xrf)_active':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_ion_chromatography_active':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_ion_chromotagraphy_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_ion_chromatography_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_not-automatic':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'ion chromatography'}},
            'unknown_lvs_-_automatic_filter_change_2,3_m3_h_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                   'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_lvs_-_automatic_filter_change_2,3_m3_h_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_other_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_niosh_active':                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_unknown_active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_spectrophotometry_ion_chromatography_active':                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_spectrophotometry_spectrophotometry_active':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},      'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection'}},
            'unknown_thermal-optical_carbon_analysis_ion_chromatography_active':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_thermal-optical_carbon_analysis_optical_termical_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'thermal-optical analysis',                                                                             'sampling_type':'continuous injection'}},
            'unknown_unknown_active':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'unknown'}},
            'unknown_x-ray_fluorescence_spectroscopy_(xrf)_unknown_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                      'sampling_type':'continuous injection'}},
            'varian_spectre_aa_400_graphit_gta-96_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',           'instrument':'Varian SpectrAA 400 GTA-96', 'sampling_type':'continuous injection'}},
            'xrf_mvs_other_automatic':                                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                     'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                      'sampling_type':'continuous injection'}}} 

#--------------------------------------------------------------
#NCDC ISD
#--------------------------------------------------------------
    elif run_details['network'] == 'NCDC_ISD':
        standard_measurement_information = {'measurement':{'methodology':'unknown'}}

#--------------------------------------------------------------
#SEARCH
#--------------------------------------------------------------
    elif run_details['network'] == 'SEARCH':
        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information  = {'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}

#--------------------------------------------------------------
#US EPA AQS
#--------------------------------------------------------------
    elif run_details['network'] == 'US_EPA_AQS':
        if run_details['parameter'] == 'O3':
            mapping = {
            '3':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Meloy 0A325-2R',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '4':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Meloy 0A350-2R',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '7':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Bendix 8002',                'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '11': {'measurement':{'methodology':'ethylene chemiluminescence',                                                                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}, 'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume ethylene chemiluminescence', 'measuring_assumption':True}},
            '14': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-1',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '15': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-2',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '16': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-3',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}}, 
            '17': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Teledyne ML8410E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '19': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1003-AH',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '20': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Beckman 950A',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '23': {'measurement':{'methodology':'rhodamine B chemiluminescence',                     'instrument':'Philips PW9771',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '36': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'CSI 2000',                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '47': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Thermo 49',                  'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '53': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne ML8810',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '55': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'PCI LC-12',                  'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '56': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1008-AH',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '78': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environics 300',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '87': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne 400',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '91': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne ML9810',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '103':{'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'OPSIS AR500',                'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '105':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O341M',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '112':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA360',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '134':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'DKK-TOA GUX-113E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '137':{'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'Environnement SA SANOA',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '148':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O342M',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '160':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA370',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '161':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Seres OZ 2000G',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '164':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'SIR S-5014',                 'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '165':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Tanabyte 722',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '169':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'DKK-TOA GUX-313E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '187':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Ecotech Serinus10',          'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '190':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 202',                     'reported_lower_limit_of_detection':{'O3':'3.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '199':{'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Teledyne 265E',              'reported_lower_limit_of_detection':{'O3':'0.6'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '214':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne T204-O3',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '215':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 211',                     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '218':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 106-L',                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '222':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Sabio 6030',                 'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '225':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O342e',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '227':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B POM',                     'reported_lower_limit_of_detection':{'O3':'4.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '901':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 202',                     'reported_lower_limit_of_detection':{'O3':'1.5'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2') or (run_details['parameter'] == 'NOx') or (run_details['parameter'] == 'NOy'):
            mapping = {
            '10': {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                              'reported_lower_limit_of_detection':{'NO2':'0.1'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '11': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Lyshkow'},                                                                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry',                                                                                                                     'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'10.0','NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'},                  'sampling_type':'continuous injection'}},
            '12': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'colorimetry',                                                                                                                     'reported_lower_limit_of_detection':{'NO2':'10.0'},                                        'reported_upper_limit_of_detection':{'NO2':'400.0'},                                              'sampling_type':'continuous injection'}}, 
            '13': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'coulometry',                                                                                                                      'reported_lower_limit_of_detection':{'NO2':'10.0','NOx':'10.0'},                           'reported_upper_limit_of_detection':{'NO2':'400.0','NOx':'1200.0'}}},
            '14': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                               'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'},                                                            'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                      'measuring_assumption':True}},
            '16': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman',    'process_details':'Preparation method is stated as Poland Method. Believe this is another name for the Griess-Saltzman method, see paper: https://www.sciencedirect.com/science/article/pii/S0269749197000389', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'10.0'},                                        'reported_upper_limit_of_detection':{'NO2':'400.0'},                                              'sampling_type':'continuous injection',                                                                                                                                                                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            '21': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8440E',       'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '22': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Bendix 8101-C',          'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '25': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'CSI 1600',               'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '31': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Meloy NA530R',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '34': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Beckman 952A',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '35': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 14B/E',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '37': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 14D/E',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '38': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Bendix 8101-B',          'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '40': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Philips PW9762/02',      'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '42': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8840',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '51': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Thermo 17C',             'reported_lower_limit_of_detection':{'NOx':'1.0'},                                         'reported_upper_limit_of_detection':{'NOx':'1200.0'}}},
            '61': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0'},                                         'reported_upper_limit_of_detection':{'NO':'500.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     'measuring_assumption':True}},
            '71': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0'},                                         'reported_upper_limit_of_detection':{'NO':'500.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     'measuring_assumption':True}},
            '74': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42',              'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'1.0', 'NOx':'10.0', 'NOy':'0.5'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '75': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42S',             'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05', 'NOy':'0.05'}, 'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '82': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0', 'NOy':'0.5'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '83': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8841',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '89': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Dasibi 2108',            'reported_lower_limit_of_detection':{'NO':'2.7', 'NO2':'5.0', 'NOx':'2.7'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '90': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML9841',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0', 'NOy':'5.0'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'2.7', 'NOx':'2.7'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measuring_assumption':True}},
            '94': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Sodium Arsenite'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measuring_assumption':True}},
            '95': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TEA'},                                                                                                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measuring_assumption':True}},
            '96': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TGS-ANSA'},                                                                                                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measuring_assumption':True}},
            '97': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TEA'},                                                                                                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measuring_assumption':True}},
            '99': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200E',          'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'2.7', 'NOx':'5.0', 'NOy':'0.5'},   'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOx':'500.0'}}},
            '100':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                               'reported_lower_limit_of_detection':{'NO':'1000.0'},                                       'reported_upper_limit_of_detection':{'NO':'1200.0'},                                                                                        'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                      'measuring_assumption':True}},
            '102':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'differential optical absorption spectroscopy',                                             'instrument':'OPSIS AR500',            'reported_lower_limit_of_detection':{'NO2':'1.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '104':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Environnement SA AC31M', 'reported_lower_limit_of_detection':{'NO2':'5.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '111':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Horiba APNA360',         'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '121':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-114E',       'reported_lower_limit_of_detection':{'NO2':'5.0','NOx':'5.0'},                             'reported_upper_limit_of_detection':{'NO2':'400.0','NOx':'1200.0'}}},
            '139':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'differential optical absorption spectroscopy',                                             'instrument':'Environnement SA SANOA', 'reported_lower_limit_of_detection':{'NO2':'1.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '146':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Environnement SA AC32M', 'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'1.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '152':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'SIR S-5012',             'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'400.0','NO2':'400.0','NOx':'400.0'}}},
            '157':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Horiba APNA370',         'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '163':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Seres 2000G',            'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '171':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-314E',       'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'1200.0','NOx':'1200.0'}}},
            '186':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech Serinus40',      'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'500.0','NOx':'500.0'}}},
            '200':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                                        'instrument':'Teledyne T200UP',        'reported_lower_limit_of_detection':{'NO':'0.1', 'NO2':'0.1', 'NOx':'0.1'},                'reported_upper_limit_of_detection':{'NO':'2000.0','NO2':'2000.0','NOx':'2000.0'}}},
            '210':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                                               'instrument':'Environnement SA AS32M', 'reported_lower_limit_of_detection':{'NO2':'0.04'},                                        'reported_upper_limit_of_detection':{'NO2':'1000.0'}}},
            '212':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                                               'instrument':'Teledyne T500U',         'reported_lower_limit_of_detection':{'NO2':'0.04'},                                        'reported_upper_limit_of_detection':{'NO2':'1000.0'}}},
            '241':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                                        'instrument':'Teledyne T200P',         'reported_lower_limit_of_detection':{'NO':'0.4', 'NO2':'0.4', 'NOx':'0.4'},                'reported_upper_limit_of_detection':{'NO':'4000.0','NO2':'4000.0','NOx':'4000.0'}}},
            '574':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42C-TL',          'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'1000.0','NO2':'1000.0','NOx':'1000.0'}}},
            '590':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech EC9841T',        'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'2000.0','NO2':'2000.0','NOx':'2000.0'}}},
            '591':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Ecotech EC9843',         'reported_lower_limit_of_detection':{'NO':'0.05','NOx':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOx':'2000.0'}}},
            '599':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200EU',         'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'1000.0','NO2':'200.0','NOx':'2000.0'},                                                           'process_details':'In methods_all.csv the instrument is stated as Teledyne 200EU-NOy, but this only measures NO/NOy. It gives an equivalent method of Teledyne 200A/Teledyne 200E, however these have detection limits of 0.4ppbv. In both methods_all.csv and seen in the datafiles, the applicable detection limit is many times 0.05ppbv, this suggests the instrument in fact used is Teledyne 200EU.', 'measuring_assumption':True}},
            '674':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Thermo 42C-NOy',         'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'1000.0','NOy':'1000.0'}}},
            '690':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech EC9841T',        'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOy':'500.0'}}},
            '691':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Ecotech EC9843',         'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOy':'500.0'}}},
            '699':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Teledyne 200EU-NOy',     'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'1000.0','NOy':'520.0'}}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'CO':
            mapping = {
            '8':  {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Bendix 8501-5CA',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '11': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                                                          'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'},                                             'process_details':'Only NDIR given as method. As it is an old AQS code and other methods around it use Non-Dispersive Infrared Absorption (Luft) method, assume this.', 'measuring_assumption':True}},
            '12': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Beckman 866',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '13': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Detection Tube'}, 'measurement':{'methodology':'colorimetry',                                                                                        'reported_lower_limit_of_detection':{'CO':'5000.0'}, 'reported_upper_limit_of_detection':{'CO':'50000.0'},   'sampling_type':'continuous injection'}},
            '14': {                                                                                                                                                                                                         'measurement':{'methodology':'dual isotope fluorescence',                                                                          'reported_lower_limit_of_detection':{'CO':'400.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '18': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'MSA 202S',               'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '21': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'},   'sampling_type':'continuous injection'}}, 
            '33': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Horiba AQM10',           'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '41': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8310',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '48': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Horiba 300E',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '50': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Massachusetts CO1',      'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '51': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Dasibi 3003',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '54': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C',             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '55': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C-TL',          'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '66': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8830',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '67': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Dasibi 3008',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '88': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne ML9830',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '93': {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300',           'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '106':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360',         'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '108':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO11M', 'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '130':{                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'LGR Off-Axis ICOS',      'reported_lower_limit_of_detection':{'CO':'0.05'}}},
            '147':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO12M', 'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '158':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA370',         'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '167':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'DKK-TOA GFC-311E',       'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '172':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'SIR S-5006',             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '174':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Ecotech Serinus30',      'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '554':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48i-TLE',         'reported_lower_limit_of_detection':{'CO':'40.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '562':{                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'LGR Off-Axis ICOS',      'reported_lower_limit_of_detection':{'CO':'0.05'}}},
            '588':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Ecotech EC9830T',        'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '593':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300EU',         'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '594':{                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300EU',         'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '1':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                               'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '5':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'second derivative spectroscopy',                 'instrument':'Ecotech SM1000',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '6':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Meloy SA185-2A',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '9':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43',              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '10': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                     'instrument':'Philips PW9755',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '11': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'colorimetry',                                                                           'reported_lower_limit_of_detection':{'SO2':'10.0'},  'reported_upper_limit_of_detection':{'SO2':'1500.0'},   'sampling_type':'continuous injection'}}, 
            '13': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                                                         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '14': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                                                            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '16': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                                                           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '18': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '19': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                                                           'reported_lower_limit_of_detection':{'SO2':'1.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '20': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '21': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'second derivative spectroscopy',                                                        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '22': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                  'instrument':'Asarco 500',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '23': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '24': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                  'instrument':'Asarco 500',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '29': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Beckman 953',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '30': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Bendix 8303',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '31': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'sulphur chemiluminescence - gas chromatography',                                        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'Chemiluminescence only stated as method. Assume Sulphur Chemiluminescence - Gas Chromatography is method.',                                       'measuring_assumption':True}},
            '32': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Meloy SA285E',           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '33': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                                                         'reported_lower_limit_of_detection':{'SO2':'10.0'},  'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '39': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML8850',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'10000.0'}}},
            '46': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Meloy SA700',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '49': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech AM2020',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'AQS states instrument as using Second Derivative Spectroscopy method, whereas reference suggests Ultraviolet Fluorescence, side with reference.', 'measuring_assumption':True}},
            '50': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '59': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '60': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43A',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'10000.0'}}},
            '61': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Dasibi 4108',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '75': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML8850S',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '77': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100',           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '84': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF21M', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                                                                                                                                                                                                   'measuring_assumption':True}},
            '92': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML9850',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '94': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                                                                                                                                                                                                   'measuring_assumption':True}},
            '95': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'polarography',                                                                          'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '96': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '98': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TGS-ANSA'},                                            'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                                                                                                                                                                                                   'measuring_assumption':True}},
            '100':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100A',          'reported_lower_limit_of_detection':{'SO2':'0.4'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '101':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',   'instrument':'OPSIS AR500',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '114':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA360',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '115':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-32',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '133':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-112E',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '138':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',   'instrument':'Environnement SA SANOA', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '149':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF22M', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '159':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA370',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '166':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'SIR S-5001',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '168':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-312E',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '188':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech Serinus50',      'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '194':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Seres SF2000G',          'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '511':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                     'instrument':'Philips PW9700',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '513':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Teledyne ML8450',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '560':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43C-TLE',         'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '592':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech EC9850T',        'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '600':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100EU',         'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '801':{'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                                                                                                                                                                                                   'measuring_assumption':True}},
            '811':{'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                       'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '850':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                                                                                                                                                                  'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'Only sequential sampler given as method, make assumption concentrations are determined using Ion Chromatography',                                 'measuring_assumption':True}},
            '900':{'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'0.026'}, 'reported_upper_limit_of_detection':{'SO2':'1500.0'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'ISOPRENE':
            mapping = {
            '91': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'},    'sampling_type':'continuous injection',   'process_details':'AQS states purely as FID, but FID cannot solely distinguish Isoprene, therefore assume uses GC beforehand.',         'measuring_assumption':True}},
            '101':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                                                                                      'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',              'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'},                                              'process_details':'AQS states method as multi detector GC, assume Gas Chromatography - Mass Spectrometry - Flame Ionisation Detection', 'measuring_assumption':True}},
            '120':{'primary_sampling':{'sampling_type':'passive',               'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '122':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                                                                                      'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '123':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                                                                                      'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'ISOPRENE':'0.002'}}},
            '125':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.0272'},  'sampling_type':'continuous injection'}},
            '126':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '127':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                           'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '128':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'ISOPRENE':'0.002'},   'sampling_type':'continuous injection'}},
            '129':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                           'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '135':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - direct temperature resolved mass spectrometry',               'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'},    'sampling_type':'continuous injection'}},                            
            '136':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',              'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'},    'sampling_type':'continuous injection'}},
            '137':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'},    'sampling_type':'continuous injection'}},
            '142':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                           'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'ISOPRENE':'0.1141'},  'sampling_type':'continuous injection',   'process_details':'AQS states method as dual column GC, assume Gas Chromatography - Dual Flame Ionisation Detection',                   'measuring_assumption':True}},
            '146':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '147':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.148'}}},
            '148':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - fourier transform infrared spectroscopy - mass spectrometry', 'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '149':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',              'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '150':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                           'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '152':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},
            '175':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                           'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},                                                           
            '177':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',              'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}},                                                          
            '210':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},                                                      'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                           'reported_lower_limit_of_detection':{'ISOPRENE':'0.02'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'NH3':
            mapping = {
            '10': {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'ion chromatography',                                                                                                               'reported_lower_limit_of_detection':{'NH3':'0.1'}}},
            '50': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                    'measurement':{'methodology':'ion chromatography',                                                                                                               'reported_lower_limit_of_detection':{'NH3':'0.03'}}}, 
            '51': {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Thermo 17C',              'reported_lower_limit_of_detection':{'NH3':'1.0'}}},
            '82': {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Teledyne 201E',           'reported_lower_limit_of_detection':{'NH3':'1.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Nessler'},           'measurement':{'methodology':'spectrophotometry',                                                                                                                'reported_lower_limit_of_detection':{'NH3':'7.2'},      'measuring_assumption':True}},
            '92': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Sodium Phenolate'},  'measurement':{'methodology':'colorimetry',                                                                                                                      'reported_lower_limit_of_detection':{'NH3':'7.2'},      'measuring_assumption':True}},
            '93': {                                                                                                                                                                                                            'measurement':{'methodology':'electrochemical membrane diffusion',                                                                                               'reported_lower_limit_of_detection':{'NH3':'1000.0'}}}, 
            '101':{                                                                                                                                                                                                            'measurement':{'methodology':'photoacoustic spectroscopy',                                                               'instrument':'Pranalytica Nitrolux200', 'reported_lower_limit_of_detection':{'NH3':'0.2'}}},
            '121':{                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-114E',        'reported_lower_limit_of_detection':{'NH3':'5.0'}}},
            '400':{                                                                                                                                                                                                            'measurement':{'methodology':'photoacoustic spectroscopy',                                                               'instrument':'Pranalytica Nitrolux200', 'reported_lower_limit_of_detection':{'NH3':'0.2'}}},
            '592':{                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and quartz converters)',                            'instrument':'Ecotech EC9842T',         'reported_lower_limit_of_detection':{'NH3':'0.2'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'PAN':
            mapping = {
            '11':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'measurement':{'methodology':'gas chromatography - electron capture detection', 'reported_lower_limit_of_detection':{'PAN':'0.1'}, 'sampling_type':'continuous injection'}}}

#--------------------------------------------------------------
#US EPA CASTNET
#--------------------------------------------------------------
    elif run_details['network'] == 'US_EPA_CASTNET':
        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}

#--------------------------------------------------------------
#WMO GAW
#--------------------------------------------------------------
    elif run_details['network'] == 'WMO_GAW':
        mapping = {}

#--------------------------------------------------------------
    
    #need to map raw measurement information to standard information? 
    if 'mapping' in locals():
        #check if have mapping information for raw network provided information
        if raw_measurement_info not in list(mapping.keys()):
            #if do not, return 'non-mapped' string
            return 'non-mapped'
        
        #map raw measurement information to standard information
        standard_measurement_information = mapping[raw_measurement_info]

    return standard_measurement_information

###--------------------------------------------------------------------------------------------------###
#DEFINE DICTIONARIES WHICH MAP MEASUREMENT STAGE SPECIFIC MEASUREMENT INFORMATION TO STANDARD METADATA VARIABLES 
###--------------------------------------------------------------------------------------------------###

#syntax --> measurement stage specific variable name : standard metadata name

primary_sampling_metadata_mapping = {'sampling_type':             'primary_sampling_type',
                                     'instrument':                'primary_sampling_instrument_name',
                                     'documented_flow_rate':      'primary_sampling_instrument_documented_flow_rate',
                                     'reported_flow_rate':        'primary_sampling_instrument_reported_flow_rate',
                                     'process_details':           'primary_sampling_process_details',
                                     'instrument_manual_name':    'primary_sampling_instrument_manual_name',
                                     'instrument_further_details':'primary_sampling_further_details'
                                    }

sample_preparation_metadata_mapping = {'sample_preparation_types':          'sample_preparation_types',
                                       'sample_preparation_techniques':     'sample_preparation_techniques',
                                       'sample_preparation_process_details':'sample_preparation_process_details',
                                       'sample_preparation_further_details':'sample_preparation_further_details'
                                      }

measurement_metadata_mapping = {'methodology':                        'measurement_methodology',
                                'instrument':                         'measuring_instrument_name',
                                'sampling_type':                      'measuring_instrument_sampling_type',
                                'documented_flow_rate':               'measuring_instrument_documented_flow_rate',
                                'reported_flow_rate':                 'measuring_instrument_reported_flow_rate',
                                'process_details':                    'measuring_instrument_process_details',
                                'instrument_manual_name':             'measuring_instrument_manual_name',
                                'instrument_further_details':         'measuring_instrument_further_details',
                                'documented_lower_limit_of_detection':'measuring_instrument_documented_lower_limit_of_detection',
                                'reported_lower_limit_of_detection':  'measuring_instrument_reported_lower_limit_of_detection',
                                'documented_upper_limit_of_detection':'measuring_instrument_documented_upper_limit_of_detection',
                                'reported_upper_limit_of_detection':  'measuring_instrument_reported_upper_limit_of_detection',
                                'documented_uncertainty':             'measuring_instrument_documented_uncertainty',
                                'reported_uncertainty':               'measuring_instrument_reported_uncertainty',
                                'documented_accuracy':                'measuring_instrument_documented_accuracy',
                                'reported_accuracy':                  'measuring_instrument_reported_accuracy',
                                'documented_precision':               'measuring_instrument_documented_precision',
                                'reported_precision':                 'measuring_instrument_reported_precision',
                                'documented_zero_drift':              'measuring_instrument_documented_zero_drift',
                                'reported_zero_drift':                'measuring_instrument_reported_zero_drift',
                                'documented_span_drift':              'measuring_instrument_documented_span_drift',
                                'reported_span_drift':                'measuring_instrument_reported_span_drift', 
                                'documented_zonal_drift':             'measuring_instrument_documented_zonal_drift',
                                'reported_zonal_drift':               'measuring_instrument_reported_zonal_drift',
                                'documented_resolution':              'measuring_instrument_documented_measurement_resolution',
                                'reported_resolution':                'measuring_instrument_reported_measurement_resolution',
                                'documented_absorption_cross_section':'measuring_instrument_documented_absorption_cross_section',
                                'reported_absorption_cross_section':  'measuring_instrument_reported_absorption_cross_section'
                               }

###--------------------------------------------------------------------------------------------------###
#DEFINE FUNCTION TO MAP NETWORK PROVIDED MEASUREMENT INFORMATION TO STANDARD MEASUREMENT INFORMATION 
###--------------------------------------------------------------------------------------------------###

def map_network_measurement_information_to_standard_measurement_information(raw_provided_measurement_information, site_metadata, run_details):
    
    '''Function which maps network provided measurement information to standard measurement information.
       This takes information defined in standardised mapping dictionaries to accomplish this.
    '''

    import copy
    from GHOST_standards import standard_sampling_types, standard_sampling_preparation_types, standard_measurement_methodologies
    import glob

    #get list of af all currently unmapped raw network provided measurement information
    unmapped_provided_measurement_information = sorted(glob.glob(run_details['process_root']+'/run/unmapped_provided_measurement_information/*'))
    unmapped_provided_measurement_information = [info.split('unmapped_provided_measurement_information/')[-1] for info in unmapped_provided_measurement_information]
        
    #make sure raw_measurement_info is lower-case, and replace empty single spaces and forward slashes with underscores
    raw_provided_measurement_information = raw_provided_measurement_information.replace(" ", "_").replace("/", "_").lower()
        
    #get standard measurement information specific for network/parameter being processed 
    #on some occasions the measurements are standard by network therefore the information is pre-standardised
    #on other occasions the networks provide non-standard measurement information that need mapping to standard information
    standard_provided_measurement_information = network_measurement_information_to_standard_measurement_information_mapping(raw_provided_measurement_information, site_metadata['meta_update_stamp']['value'][-1], run_details)

    #iterate through standard sampling types/sample preparation types/measurement methodologies dictionaries
    #taking information specific to the standardised reported measurement information, overwriting the stored dictionary information where necessary  

    #initialise dictionary for storing QA flags related to the measurement information (all flags set as False initially)
    measurement_qa_flags = {'Methodology Not Mapped': False, 
                            'Erroneous Primary Sampling': False,
                            'Erroneous Sample Preparation': False,
                            'Erroneous Measurement Methodology': False,
                            'Invalid QA Measurement Methodology': False,
                            'Assumed Primary Sampling' : False,
                            'Assumed Sample Preparation': False,
                            'Assumed Measurement Methodology': False,
                            'Unknown Primary Sampling Instrument': False,
                            'Unknown Measuring Instrument': False
                           }

    #initialise dictionary for storing sample preparation concatenated string information (can be multiple sample preparation information) 
    specific_sample_preparation_strings = {'sample_preparation_types' : '',
                                           'sample_preparation_techniques' : '',
                                           'sample_preparation_process_details' : '',
                                           'sample_preparation_further_details' : ''
                                          }

    #check if the raw measurement information has not been able to be mapped to standard information
    if standard_provided_measurement_information == 'non-mapped':

        #if not, set QA flag informing of this
        measurement_qa_flags['Methodology Not Mapped'] = True
        
        #check if have previously logged that the raw measurement information is unmapped
        if raw_provided_measurement_information not in unmapped_provided_measurement_information:
        
            #if not, create network:species:raw_measurement_info file for writing out a log that measurement information is unmapped 
            unmapped_provided_measurement_information_filename = run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information)
            f= open(unmapped_provided_measurement_information_filename,"w+")
            f.close() 
            #give 770 permissions to file
            os.chmod(unmapped_provided_measurement_information_filename, 0o770)

        #return from function
        return site_metadata, measurement_qa_flags, '?'

    #--------------------------------------------------------------------------------#
    #is there primary sampling information provided?
    specific_primary_sampling_information = {}
    if 'primary_sampling' in list(standard_provided_measurement_information.keys()):
        
        #get standard provided primary sampling information dictionary
        standard_provided_primary_sampling_information = copy.deepcopy(standard_provided_measurement_information['primary_sampling'])
        
        #get information specific to provided primary sampling type in standard sampling types dictionary     
        specific_primary_sampling_information = copy.deepcopy(standard_sampling_types[standard_provided_primary_sampling_information['sampling_type']])
        
        #update specific primary sampling information, updating by "overwrite" variables
        #get list of all variables containing "overwrite" string
        overwrite_vars = [var for var in list(specific_primary_sampling_information.keys()) if 'overwrite' in var]
        for overwrite_var in overwrite_vars:
            original_var = overwrite_var.split('overwrite_')[0]
            specific_primary_sampling_information[original_var] = specific_primary_sampling_information[overwrite_var]

        #if have defined primary sampling instrument name, get dictionary with information specific to that instrument  
        if 'instrument' in list(standard_provided_primary_sampling_information.keys()):
            specific_primary_sampling_instrument_information = standard_sampling_types[standard_provided_primary_sampling_information['sampling_type']]['instruments'][standard_provided_primary_sampling_information['instrument']]   
            #merge specific primary sampling information and specific primary sampling instrument information
            #handle matching variables by keeping instrument specific information by priority
            specific_primary_sampling_information = {**specific_primary_sampling_information, **specific_primary_sampling_instrument_information}

        #merge specific primary sampling information and standard provided primary sampling information
        #handle matching variables by keeping standard provided primary sampling information by priority
        specific_primary_sampling_information  = {**specific_primary_sampling_information, **standard_provided_primary_sampling_information} 
        #get list of variables in generated primary sampling dictionary
        specific_primary_sampling_information_vars = list(specific_primary_sampling_information.keys())        

        #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
        for var in specific_primary_sampling_information_vars:   
            if type(specific_primary_sampling_information[var]) == dict:
                if run_details['parameter'] in list(specific_primary_sampling_information[var].keys()):
                    specific_primary_sampling_information[var] = specific_primary_sampling_information[var][run_details['parameter']]
                else:
                    del specific_primary_sampling_information[var]
                    specific_primary_sampling_information_vars = list(specific_primary_sampling_information.keys()) 

        #if do not have primary sampling type defined (i.e. == unknown), set quality assurance flag informing of this
        if specific_primary_sampling_information['sampling_type'] == 'unknown':
            measurement_qa_flags['Unknown Primary Sampling Type'] = True     

        #if do not have primary sampling instrument defined, set quality assurance flag informing of this
        #this is False for "passive", "remote" and "manual" sampling types as there is no physically real sampling instrument in these cases
        if 'instrument' not in specific_primary_sampling_information_vars:
            if specific_primary_sampling_information['sampling_type'] not in ['manual', 'passive', 'remote']:
                measurement_qa_flags['Unknown Primary Sampling Instrument'] = True     

        #determine if current parameter processing is validly isolated by primary sampling type --> flag if not
        if ('isolated_parameters' in specific_primary_sampling_information_vars) & (specific_primary_sampling_information['sampling_type'] != 'unknown'):
            if (run_details['parameter'] not in specific_primary_sampling_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_primary_sampling_information['isolated_parameters']):
                measurement_qa_flags['Erroneous Primary Sampling'] = True

        #if "primary_sampling_assumption" flag is in specific sample preparation dictionary, update primary sampling assumption flag
        if 'primary_sampling_assumption' in specific_primary_sampling_information_vars:
                measurement_qa_flags['Assumed Primary Sampling'] = True

        #if have 'additional_sample_preparation_type' in specific primary sampling information dictionary, append to sampling_preparation_type variable (if exists)
        if 'additional_sample_preparation_type' in specific_primary_sampling_information_vars:
            if 'sample_preparation_type' in specific_primary_sampling_information_vars:
                specific_primary_sampling_information['sample_preparation_type'] = '{};{}'.format(specific_primary_sampling_information['sample_preparation_type'], specific_primary_sampling_information['additional_sample_preparation_type'])
            else:
                specific_primary_sampling_information['sample_preparation_type'] = specific_primary_sampling_information['additional_sample_preparation_type'] 

        #if have sample preparation information provided for the primary sampling instrument, process it
        if 'sample_preparation_type' in specific_primary_sampling_information_vars:
            #get sample preparation variable
            sample_preparation_type = specific_primary_sampling_information['sample_preparation_type']

            #gather all variable names associated with sample preparation in dictionary (i.e. with 'preparation' prefix)
            sample_preparation_vars = [var for var in specific_primary_sampling_information_vars if ('preparation' in var) & (var != 'sample_preparation_type')]
            #create dictionary of all all sample preparation variables names associated their values (dropping 'preparation' prefix')
            sample_preparation_dict = {}
            for var in sample_preparation_vars:   
                if var not in ['preparation_further_details', 'preparation_manual_name', 'preparation_assumption']:          
                    sample_preparation_dict[var.split('preparation_')[-1]] = specific_primary_sampling_information[var] 
                else:
                    sample_preparation_dict[var] = specific_primary_sampling_information[var] 
    
            #now iterate through all distinct provided sample preparation information
            sample_preparation_type_split = sample_preparation_type.split(';')
            sample_preparation_dict_split = {key: value.split(';') for key, value in list(sample_preparation_dict.items())}            
            for preparation_N, sample_preparation_type_iteration in enumerate(sample_preparation_type_split):

                #create specific distinct sample preparation information dictionary
                standard_provided_sample_preparation_information = {key: value[preparation_N] for key, value in list(sample_preparation_dict.items())}
                standard_provided_sample_preparation_information['sample_preparation_type'] = sample_preparation_type_iteration

                #get information specific to sample preparation type in standard sampling preparation types dictionary
                specific_sample_preparation_information = standard_sampling_preparation_types[sample_preparation_type_iteration]
                
                #if have defined sample preparation technique name, get dictionary with information specific to that technique  
                if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_technique_information = standard_sampling_preparation_types[sample_preparation_type_iteration]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                    #merge specific sample preparation type information and specific sample preparation technique information
                    #handle matching variables by keeping technique specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
                #else if, if have a intermediate measurement defined as the sample preparation type, get dictionary with information specific to that method
                elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                    #if have defined measuring instrument name, get information specific to that instrument
                    if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                        specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                        #merge specific sample preparation methodology information and specific sample preparation instrument information
                        #handle matching variables by keeping instrument specific information by priority
                        specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                    #merge specific sample preparation type information and specific sample preparation methodology information
                    #handle matching variables by keeping methodology specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

                #merge specific sample preparation information and standard provided sampling preparation information
                #handle matching variables by keeping standard provided sampling preparation information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
                #get list of variables in generated sample preparation dictionary
                specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())  

                #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
                for var in specific_sample_preparation_information_vars:   
                    if type(specific_sample_preparation_information[var]) == dict:
                        if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                            specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']]  
                        else:
                            del specific_sample_preparation_information[var]
                            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())

                #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
                if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                    measurement_qa_flags['Unknown Sample Preparation Type'] = True  

                #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
                #this is False for "flask", "preconcentration", "filter" and "denuder" preparation types as there are no specific techniques for these types
                if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                    if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','preconcentration','filter','denuder']:
                            measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

                #determine if current parameter processing is validly isolated by sample preparation
                if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

               #determine if sample preparation measurement method actually measures current parameter processing 
                if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

                #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True                  

                #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
                if 'preparation_assumption' in specific_sample_preparation_information_vars:
                    measurement_qa_flags['Assumed Sample Preparation'] = True

                #concatenate sample preparation strings 
                if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                    specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])
  
    #--------------------------------------------------------------------------------#
    #is there sample preparation information provided?
    #check for all different sample preparation information (i.e. preparation_1, preparation_2, preparation_3..)
    have_preparation = True
    preparation_n = 1
    while have_preparation == True:
        if 'preparation_{}'.format(preparation_n) in list(standard_provided_measurement_information.keys()):
            
            #get standard provided sample preparation information dictionary
            standard_provided_sample_preparation_information = copy.deepcopy(standard_provided_measurement_information['preparation_{}'.format(preparation_n)])

            #get information specific to provided sampling preparation type in standard sampling preparation types dictionary     
            specific_sample_preparation_information = copy.deepcopy(standard_sampling_preparation_types[standard_provided_sample_preparation_information['sample_preparation_type']])
            
            #update specific sample preparation information, updating by "overwrite" variables
            #get list of all variables containing "overwrite" string
            overwrite_vars = [var for var in list(specific_sample_preparation_information.keys()) if 'overwrite' in var]
            for overwrite_var in overwrite_vars:
                original_var = overwrite_var.split('overwrite_')[0]
                specific_sample_preparation_information[original_var] = specific_sample_preparation_information[overwrite_var]

            #if have defined sample preparation technique name, get dictionary with information specific to that technique  
            if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                specific_sample_preparation_technique_information = standard_sampling_preparation_types[standard_provided_sample_preparation_information['sample_preparation_type']]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                #merge specific sample preparation type information and specific sample preparation technique information
                #handle matching variables by keeping technique specific information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
            #else if, if have a intermediate measurement defined as the sample preparation type, get dictionary with information specific to that method
            elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                #if have defined measuring instrument name, get information specific to that instrument
                if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                    #merge specific sample preparation methodology information and specific sample preparation instrument information
                    #handle matching variables by keeping instrument specific information by priority
                    specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                #merge specific sample preparation type information and specific sample preparation methodology information
                #handle matching variables by keeping methodology specific information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

            #merge specific sample preparation information and standard provided sampling preparation information
            #handle matching variables by keeping standard provided sampling preparation information by priority
            specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
            #get list of variables in generated sample preparation dictionary
            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())  

            #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
            for var in specific_sample_preparation_information_vars:   
                if type(specific_sample_preparation_information[var]) == dict:
                    if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                        specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']] 
                    else:
                        del specific_sample_preparation_information[var] 
                        specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys()) 

            #if do not have primary sampling determined yet and sample preparation type is an 'intermediate measurement'
            #make a copy of the specific sampling preparation information dictionary to record the primary sampling information
            if len(specific_primary_sampling_information) == 0:
                specific_primary_sampling_information = copy.deepcopy(specific_sample_preparation_information)

            #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
            if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                measurement_qa_flags['Unknown Sample Preparation Type'] = True  

            #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
            #this is False for "flask", "preconcentration", "filter" and "denuder" preparation types as there are no specific techniques for these types
            if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','preconcentration','filter','denuder']:
                        measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

            #determine if current parameter processing is validly isolated by sample preparation
            if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True

           #determine if sample preparation measurement method actually measures current parameter processing 
            if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True

            #determine if sample preparation measurement method is QA accepted for current parameter processing 
            if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True                 

            #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
            if ('preparation_assumption' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                measurement_qa_flags['Assumed Sample Preparation'] = True

            #concatenate sample preparation strings 
            if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                if 'specific_technique' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                elif 'methodology' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                else:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                if 'process_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                if 'preparation_further_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
            else:
                specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                if 'specific_technique' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                elif 'methodology' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                if 'process_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                if 'preparation_further_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])

            #update preparation_n       
            preparation_n += 1
        else:
            have_preparation = False
        
    #--------------------------------------------------------------------------------#
    #is there measurement information? 
    specific_measurement_method_information = {}
    if 'measurement' in list(standard_provided_measurement_information.keys()):
    
        #get standard provided measurement method information dictionary
        standard_provided_measurement_method_information = copy.deepcopy(standard_provided_measurement_information['measurement'])

        #get information specific to provided measurement method in standard measurement dictionary     
        specific_measurement_method_information = copy.deepcopy(standard_measurement_methodologies[standard_provided_measurement_method_information['methodology']])
        
        #update specific measurement information, updating by "overwrite" variables
        #get list of all variables containing "overwrite" string
        overwrite_vars = [var for var in list(specific_measurement_method_information.keys()) if 'overwrite' in var]
        for overwrite_var in overwrite_vars:
            original_var = overwrite_var.split('overwrite_')[0]
            specific_measurement_method_information[original_var] = specific_measurement_method_information[overwrite_var]

        #if have defined measuring instrument name, get information specific to that instrument
        if 'instrument' in list(standard_provided_measurement_method_information.keys()):
            specific_measurement_instrument_information = standard_measurement_methodologies[standard_provided_measurement_method_information['methodology']]['instruments'][standard_provided_measurement_method_information['instrument']]   
            #merge specific measurement method information and specific measurement instrument information
            #handle matching variables by keeping instrument specific information by priority
            specific_measurement_method_information = {**specific_measurement_method_information, **specific_measurement_instrument_information}

        #merge specific measurement method information and standard provided primary sampling information
        #handle matching variables by keeping standard provided primary sampling information by priority
        specific_measurement_method_information = {**specific_measurement_method_information, **standard_provided_measurement_method_information} 
        #get list of variables in generated primary sampling dictionary
        specific_measurement_method_information_vars = list(specific_measurement_method_information.keys())        
    
        #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
        for var in specific_measurement_method_information_vars:   
            if type(specific_measurement_method_information[var]) == dict:
                if run_details['parameter'] in list(specific_measurement_method_information[var].keys()):
                    specific_measurement_method_information[var] = specific_measurement_method_information[var][run_details['parameter']] 
                else:  
                    del specific_measurement_method_information[var] 
                    specific_measurement_method_information_vars = list(specific_measurement_method_information.keys()) 

        #if do not have measurement method defined (i.e. == unknown), set quality assurance flag informing of this
        if specific_measurement_method_information['methodology'] == 'unknown':
            measurement_qa_flags['Unknown Measurement Method'] = True  

        #if do not have specific measuring instrument defined, set quality assurance flag informing of this
        #if measuring sampling type is "manual" then this should be False, as there is no physically real measuring instrument in this case
        if 'instrument' not in specific_measurement_method_information_vars:
            if specific_measurement_method_information['methodology'] == 'unknown':
                measurement_qa_flags['Unknown Measuring Instrument'] = True
            else:
                if specific_measurement_method_information['sampling_type'] not in ['manual']:
                    measurement_qa_flags['Unknown Measuring Instrument'] = True

       #determine if measurement method actually measures the current parameter processing 
        if ('measured_parameters' in specific_measurement_method_information_vars) & (specific_measurement_method_information['methodology'] != 'unknown'):
            if (run_details['parameter'] not in specific_measurement_method_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_measurement_method_information['measured_parameters']):
                measurement_qa_flags['Erroneous Measurement Methodology'] = True 

        #determine if measurement method is QA accepted for current parameter processing 
        if 'qa_accepted_parameters' in specific_measurement_method_information_vars:
            if (run_details['parameter'] not in specific_measurement_method_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_measurement_method_information['qa_accepted_parameters']):
                measurement_qa_flags['Invalid QA Measurement Methodology'] = True    

        #if "measuring_assumption" flag is in specific measurement method dictionary, update measurement method assumption flag
        if 'measuring_assumption' in specific_measurement_method_information_vars:
                measurement_qa_flags['Assumed Measurement Methodology'] = True     

        #if have 'additional_sample_preparation_type' in specific primary sampling information dictionary, append to sampling_preparation_type variable (if exists)
        if 'additional_sample_preparation_type' in specific_measurement_method_information_vars:
            if 'sample_preparation_type' in specific_measurement_method_information_vars:
                specific_measurement_method_information['sample_preparation_type'] = '{};{}'.format(specific_measurement_method_information['sample_preparation_type'], specific_measurement_method_information['additional_sample_preparation_type'])
            else:
                specific_measurement_method_information['sample_preparation_type'] = specific_measurement_method_information['additional_sample_preparation_type']                 
      
        #if have sample preparation information provided for the measuring instrument, process it
        if 'sample_preparation_type' in specific_measurement_method_information_vars:
            #get sample preparation variable
            sample_preparation_type = specific_measurement_method_information['sample_preparation_type']

            #gather all variable names associated with sample preparation in dictionary (i.e. with 'preparation' prefix)
            sample_preparation_vars = [var for var in specific_measurement_method_information_vars if ('preparation' in var) & (var != 'sample_preparation_type')]
            #create dictionary of all all sample preparation variables names associated their values (dropping 'preparation' prefix'
            sample_preparation_dict = {}
            for var in sample_preparation_vars:   
                if var not in ['preparation_further_details', 'preparation_manual_name', 'preparation_assumption']:          
                    sample_preparation_dict[var.split('preparation_')[-1]] = specific_measurement_method_information[var] 
                else:
                    sample_preparation_dict[var] = specific_measurement_method_information[var] 

            #now iterate through all distinct provided sample preparation information
            sample_preparation_type_split = sample_preparation_type.split(';')
            sample_preparation_dict_split = {key: value.split(';') for key, value in list(sample_preparation_dict.items())}            
            for preparation_N, sample_preparation_type_iteration in enumerate(sample_preparation_type_split):

                #create specific distinct sample preparation information dictionary
                standard_provided_sample_preparation_information = {key: value[preparation_N] for key, value in list(sample_preparation_dict.items())}
                standard_provided_sample_preparation_information['sample_preparation_type'] = sample_preparation_type_iteration

                #get information specific to sample preparation type/specific technique in standard sampling preparation types dictionary
                specific_sample_preparation_information = standard_sampling_preparation_types[sample_preparation_type_iteration]
                if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_technique_information = standard_sampling_preparation_types[sample_preparation_type_iteration]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                    #merge specific sample preparation type information and specific sample preparation technique information
                    #handle matching variables by keeping technique specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
                elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                    if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                        specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                        #merge specific sample preparation methodology information and specific sample preparation instrument information
                        #handle matching variables by keeping instrument specific information by priority
                        specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                    #merge specific sample preparation type information and specific sample preparation methodology information
                    #handle matching variables by keeping methodology specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

                #merge specific sample preparation information and standard provided sampling preparation information
                #handle matching variables by keeping standard provided sampling preparation information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
                #get list of variables in generated sample preparation dictionary
                specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys()) 

                #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
                for var in specific_sample_preparation_information_vars:   
                    if type(specific_sample_preparation_information[var]) == dict:
                        if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                            specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']]  
                        else:
                            del specific_sample_preparation_information[var]
                            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())

                #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
                if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                    measurement_qa_flags['Unknown Sample Preparation Type'] = True  

                #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
                #this is False for "flask", "preconcentration", "filter" and "denuder" preparation types as there are no specific techniques for these types
                if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                    if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','preconcentration','filter','denuder']:
                            measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

                #determine if current parameter processing is validly isolated by sample preparation
                if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

               #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

                #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True                     

                #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
                if ('preparation_assumption' in specific_sample_preparation_information_vars):
                    measurement_qa_flags['Assumed Sample Preparation'] = True

                #concatenate sample preparation strings 
                if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                    specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])

    #--------------------------------------------------------------------------------#

    #if raw provided measurement information has been successfully mapped, but a file is still contained within unmapped_provided_measurement_information directory, remove file from directory
    if raw_provided_measurement_information in unmapped_provided_measurement_information:
        #check if measurement information still exists before removing it (for parallel cases)
        #also put in try/except just in case for parallel cases where file is deleted before trying to remove it (if statement, should catch these cases but have had failures before)
        if os.path.isfile(run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information)) == True:
            try:
                os.remove(run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information))
            except:
                pass
    
    #--------------------------------------------------------------------------------#

    #now add measurement stage specific dictionaries to standard site metadata variables

    #primary sampling metadata
    valid_primary_sampling_vars = list(primary_sampling_metadata_mapping.keys())
    for var in list(specific_primary_sampling_information.keys()):
        if var in valid_primary_sampling_vars:
            if len(site_metadata[primary_sampling_metadata_mapping[var]]['value']) == 0:
                site_metadata[primary_sampling_metadata_mapping[var]]['value'].append(specific_primary_sampling_information[var])
            else:
                site_metadata[primary_sampling_metadata_mapping[var]]['value'] = [specific_primary_sampling_information[var]] * len(site_metadata[primary_sampling_metadata_mapping[var]]['value'])
    #sample preparation metadata
    valid_sample_preparation_vars = list(sample_preparation_metadata_mapping.keys())
    for var in list(specific_sample_preparation_strings.keys()):
        if var in valid_sample_preparation_vars:
            sample_preparation_string = specific_sample_preparation_strings[var]
            if sample_preparation_string != '':
                if len(site_metadata[sample_preparation_metadata_mapping[var]]['value']) == 0: 
                    site_metadata[sample_preparation_metadata_mapping[var]]['value'].append(sample_preparation_string)
                else:
                    site_metadata[sample_preparation_metadata_mapping[var]]['value'] = [sample_preparation_string] * len(site_metadata[sample_preparation_metadata_mapping[var]]['value'])
    #measurement metadata
    valid_measurement_vars = list(measurement_metadata_mapping.keys())
    for var in list(specific_measurement_method_information.keys()):
        if var in valid_measurement_vars:
            if len(site_metadata[measurement_metadata_mapping[var]]['value']) == 0:
                site_metadata[measurement_metadata_mapping[var]]['value'].append(specific_measurement_method_information[var])  
            else:
                site_metadata[measurement_metadata_mapping[var]]['value'] = [specific_measurement_method_information[var]] * len(site_metadata[measurement_metadata_mapping[var]]['value'])
    
    #return from function
    return site_metadata, measurement_qa_flags, specific_measurement_method_information['abbreviation']

###--------------------------------------------------------------------------------------------------###
#DEFINE EXTERNAL METADATA MAPPING INFORMATION
###--------------------------------------------------------------------------------------------------###

#define dictionaries which map external metadata file variable names to standardised metadata variable names
#syntax = 'external_metadata_file_variable_name' : 'standard_variable_name' 

#if the word 'manual' appears in the dictionary variable names, then the external metadata fiel has eben manually compiled

#for mapping variable data contained in 'EBAS/metadata/processed/EBAS_META.npy' to standardised names
EBAS_manual_external_meta_map = {
'other_network_references':'associated_networks',
'setting':                 'area_classification',      
'land_use':                'land_use'
}
    
#for mapping variable data contained in 'EEA_AIRBASE/metadata/processed/EEA_AIRBASE_META.npy' to standardised names
EEA_AIRBASE_external_meta_map = {
'station_european_code':                     'station_reference',                    
'station_latitude_decimal_degrees':          'latitude',
'station_longitude_decimal_degrees':         'longitude', 
'station_altitude':                          'altitude',
'height_sampling_point':                     'sampling_height',                      
'station_name':                              'station_name',                         
'station_type_of_area':                      'area_classification',                  
'type_of_station':                           'station_classification',               
'main_emission_source':                      'main_emission_source',                 
'station_characteristic_of_zone':            'land_use',
'station_subcategory_rural_background':      'area_classification_rural_subdivision',
'station_represented_radius':                'representative_radius',                
'station_city':                              'city',
'station_distance_to_junction':              'distance_to_junction',                 
'station_distance_to_kerb':                  'distance_to_kerb',                     
'station_distance_to_source':                'distance_to_source',                   
'station_street_width':                      'street_width',
'street_type':                               'street_type',
'number_of_vehicles':                        'daily_passing_vehicles',               
'station_traffic_speed':                     'daytime_traffic_speed',                        
'population':                                'population',
'station_description':                       'station_description',
'measurement_equipment':                     'measurement_equipment',
'measurement_technique_principle':           'measurement_method',                   
'measurement_automatic':                     'measurement_automatic',
'calibration_method':                        'calibration_method',
'calibration_frequency':                     'calibration_frequency',
'length_sampling_line':                      'length_sampling_line',
'location_sampling_point':                   'sample_location_miscellaneous',
'EMEP_station':                              'emep_station',                         
'component_code':                            'pollutant_code',                       
'component_name':                            'pollutant',                            
'network_code':                              'local_network_code',                   
'network_short_name':                        'local_network_name',           
'station_local_code':                        'local_reference',               
'network_data_supplier_person_title':        'principal_investigator_title',         
'network_data_supplier_person_first_name':   'principal_investigator_first_name',    
'network_data_supplier_person_last_name':    'principal_investigator_last_name',     
'network_data_supplier_organization_name':   'principal_investigator_institution',     
'network_data_supplier_person_email_address':'principal_investigator_email_address', 
'network_manager_person_title':              'contact_title',                        
'network_manager_person_first_name':         'contact_first_name',                   
'network_manager_person_last_name':          'contact_last_name',                    
'network_manager_organization_name':         'contact_institution',                  
'network_manager_person_email_address':      'contact_email_address'
}       

#for mapping variable data contained in 'EEA_AQ_eReporting/metadata/network_provided/EEA_AQ_eReporting_META.csv' to standardised names
EEA_AQ_eReporting_external_meta_map = {
'AirQualityStationEoICode':'station_reference',     
'Latitude':                'latitude',             
'Longitude':               'longitude',             
'Altitude':                'altitude',
'InletHeight':             'sampling_height',       
'BuildingDistance':        'distance_to_building',  
'KerbDistance':            'distance_to_kerb',      
'AirQualityStationArea':   'area_classification',   
'AirQualityStationType':   'station_classification',
'SamplingPoint':           'sampling_point',        
'SamplingProcess':         'sampling_process',      
'Projection':              'spatial_reference_system',            
'AirQualityNetwork':       'local_network_code',    
'AirQualityStationNatCode':'local_reference',       
'MeasurementEquipment':    'measurement_equipment', 
'MeasurementType':         'measurement_type',
'MeasurementMethod':       'measurement_method',    
'AirPollutantCode':        'pollutant_code',
'AirPollutant':            'pollutant'
}
                                     
#for mapping variable data contained in 'EEA_AQ_eReporting/metadata/network_provided/EEA_AQ_eReporting_AIDE_D.csv' to standardised names
EEA_AQ_eReporting_AIDE_D_external_meta_map = {
'EoICode':                 'station_reference',                 
'Latitude':                'latitude',                          
'Longitude':               'longitude',                         
'Altitude':                'altitude',                          
'StationName':             'station_name',                      
'StationArea':             'area_classification',               
'StationType':             'station_classification',            
'ResponsibleParty':        'principal_investigator_institution',
'SamplingPointId':         'sampling_point',                    
'SamplingProcess':         'sampling_process',                  
'Network':                 'local_network_name',
'MeasurementEquipment':    'measurement_equipment',
'MeasurementMethod':       'measurement_method',
'SamplingEquipment':       'sampling_equipment',
'AnalyticalTechnique':     'analytical_technique',              
'Pollutant':               'pollutant'
}

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###